/*
 *  Lucidor - e-book reader
 *  Copyright © 2008-2010  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("resource://lucidor/modules/utils.jsm");

const XHTML_NS = "http://www.w3.org/1999/xhtml";
const XUL_NS = "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul";

var bcase;
var thumbnail_creator;
var main_control;
var view_mode = "list";
var grid_updated = false;
var selected_iconbox;
var original_view;
var last_predefined = null;	// Last predefined search treeitem
var selectedIndex = -1;	// Currently selected booksource

var treeView = {
    list: [],

    rowCount: 0,

    getCellText: function(row, column) {
	let elem = this.list[row];
	if (!elem)
	    return null;
	if (column.id == "author-col")
	    return elem.author;
	else if (column.id == "title-col")
	    return elem.title;
	else if (column.id == "publisher-col")
	    return elem.publisher;
	else	// id-col
	    return elem.id;
    },

    setTree: function(treebox) {
	this.treebox = treebox;
    },

    cycleHeader: function(col) {
    },

    isContainer: function(row) {
	return false;
    },

    isSeparator: function(row) {
	return false;
    },

    isSorted: function() {
	return false;
    },

    getLevel: function(row) {
	return 0;
    },

    getImageSrc: function(row, col) {
	return null;
    },

    getRowProperties: function(row, props) {
    },

    getCellProperties: function(row, col, props) {
    },

    getColumnProperties: function(colid, col, props) {
    }
};

function init() {
    main_control = Components.classes["@lucidor.org/mainControlService;1"].
	getService().wrappedJSObject;

    let cman = Components.classes["@lucidor.org/collectionManagerService;1"].
	getService().wrappedJSObject;
    cman.addListener(this);

    reread_collections();

    let tree = document.getElementById("book_tree");
    tree.hidden = false;	// FIXME Why is it necessary to have it hidden?
    original_view = tree.view;

    debug("3!");
    // FIXME
    tree = document.getElementById("search_tree");
    tree.view.selection.select(0);

    debug("4!");
}

function reread_collections() {
    debug("reread_collections");

    let cman = Components.classes["@lucidor.org/collectionManagerService;1"].
	getService().wrappedJSObject;
    let collections = cman.getCollectionsJSON2();
    if (!cman.hasCollections())
	cman.saveCollectionsJSON2(collections);

    let start = document.getElementById("noseparator");
    let separator = document.getElementById("separator");
    while (start.nextSibling != separator)
	start.parentNode.removeChild(start.nextSibling);

    for (let coll in loop(collections)) {
	let item = document.createElement("menuitem");
	item.setAttribute("label", coll[0]);
	item.setAttribute("value", coll[1]);
	/*
	if (coll[2]) {
	    item.setAttribute("class", "menuitem-iconic");
	}
	 */
	separator.parentNode.insertBefore(item, separator);
    }

    let src = document.getElementById("book_source");
    if (src.selectedIndex < 0)
	src.selectedIndex = 0;
}

function set_book_source() {
    let src = document.getElementById("book_source");
    if (src.selectedIndex == selectedIndex)
	return;

    debug("set_book_source: " + src.value);
    if (src.value == "manage") {
	src.selectedIndex = selectedIndex;
	// show window
	window.open("chrome://lucidor/content/collections.xul", "manage",
		    "chrome,width=600,height=300");
	return;
    }

    if (bcase && bcase.close)
	bcase.close();

    selectedIndex = src.selectedIndex;

    let tree = document.getElementById("book_tree");
    /*
    if (tree.view)
	tree.view.selection.currentIndex = -1;
    tree.view = original_view;
    */
    set_query("");

    debug("E 1.1");

    try {
    let JSON = Components.classes["@mozilla.org/dom/json;1"].
	createInstance(Components.interfaces.nsIJSON);
    let val = JSON.decode(src.value);
    let cman = Components.classes["@lucidor.org/collectionManagerService;1"].
	getService().wrappedJSObject;
    bcase = cman.getBookcase(val, tree);
    if (!bcase)
	throw "Error";
    } catch (e) {
	debug("E: " + e);
	return;
    }
    //tree.treeBoxObject.invalidate();

    debug("E 2");
    let item = document.getElementById("cache_box");
    item.setAttribute("hidden", !bcase.capabilities.offline_db);

    item = document.getElementById("meta_item");
    item.setAttribute("hidden", !bcase.capabilities.search_meta);
    item = document.getElementById("title_item");
    item.setAttribute("hidden", !bcase.capabilities.search_title);
    item = document.getElementById("author_item");
    item.setAttribute("hidden", !bcase.capabilities.search_author);
    item = document.getElementById("tags_item");
    item.setAttribute("hidden", !bcase.capabilities.search_tags);
    item = document.getElementById("text_item");
    item.setAttribute("hidden", !bcase.capabilities.search_text);
    while (item.parentNode.lastChild != item)
	item.parentNode.removeChild(item.parentNode.lastChild);

    item = document.getElementById("search_text");
    item.disabled = false;
    item = document.getElementById("search_filter");
    item.disabled = false;

    if (item.selectedItem && !item.selectedItem.parentNode)
	item.selectedItem = null;
    if (!item.selectedItem || item.selectedItem.hidden) {
	let hidden = true;
	for (let cld in loop(item.getElementsByTagName("menuitem")))
	    if (!cld.hidden) {
		cld.doCommand();
		hidden = false;
		break;
	    }
	if (hidden) {
	    item = document.getElementById("search_filter");
	    item.disabled = true;
	    item = document.getElementById("search_text");
	    item.disabled = true;
	}
    }

    item = document.getElementById("add_book");
    item.setAttribute("hidden", !bcase.local);
    item = document.getElementById("remove_book");
    item.setAttribute("hidden", !bcase.local);
    //item = document.getElementById("add_to_local");
    //item.setAttribute("hidden", bcase.local);

    /*
    let col = tree.columns.getColumnAt(0);
    if (!bcase.info.author) {
	col.element.setAttribute("ignoreincolumnpicker", "true");
	col.element.setAttribute("hidden", "true");
    } else {
	col.element.removeAttribute("ignoreincolumnpicker");
	col.element.removeAttribute("hidden");
    }
    */

    tree = document.getElementById("search_tree");
    tree.view.selection.currentIndex = -1;
    tree.view.selection.clearSelection();

    let titem = document.getElementById("tree_children");

    // Remove predefined items
    if (last_predefined) {
	debug("D 1 " + titem.firstChild + " " + last_predefined);
	while (titem.firstChild && (titem.firstChild != last_predefined)) {
	    debug("D: " + titem.firstChild.localName);
	    titem.removeChild(titem.firstChild);
	}
	debug("D 4 " + last_predefined.localName);
	titem.removeChild(last_predefined);
	last_predefined = null;
    }

    debug("E 6");

    // Remove children of custom search items
    for (let ite = titem.firstChild; ite; ite = ite.nextSibling)
	for (let ch in loop(ite.childNodes)) {
	    debug("CH: " + ch.localName);
	    if (ch.localName == "treechildren")
		ite.removeChild(ch);
	}

    debug("E 6.5");

    // Add new predefined items

    let rest = titem.firstChild;
    let strbundle = document.getElementById("strings");
    function add_titem(lab, value) {
	debug("TITEM: " + lab + ":" + value);
	let label;
	try {
	    label = strbundle.getString(lab);
	} catch (e) {
	    label = lab;
	}
	let it = create_search_item(label, value);
	if (rest)
	    titem.insertBefore(it, rest);
	else
	    titem.appendChild(it);
	last_predefined = it;
    }

    if (bcase.listers)
	for (let br in loop(bcase.listers))
	    add_titem("show_" + br, br);

    if (bcase.browsers)
	for (let br in loop(bcase.browsers))
	    add_titem("browse_" + br, "browse=" + br);

    current_searcher = null;
    opensearchers = {};

    grid_updated = false;
    if (view_mode == "icons")
	update_grid();
    update_book_info();
}

function create_search_item(label, value) {
    let it = document.createElement("treeitem");
    let row = document.createElement("treerow");
    let cell = document.createElement("treecell");
    cell.setAttribute("label", label);
    row.appendChild(cell);
    cell = document.createElement("treecell");
    cell.setAttribute("label", value);
    row.appendChild(cell);
    it.appendChild(row);
    return it;
}

function collapse_sidebar() {
    let splitter = document.getElementById("splitter");
    splitter.setAttribute("state", "collapsed");
    //splitter.setAttribute("hidden", "true");
    //let sidebar = document.getElementById("mysidebar");
    //sidebar.setAttribute("hidden", "true");
}

function show_warning(text_msg, more_msg, autohide) {
    main_control.show_warning(window, [ text_msg ], more_msg, autohide);
}

function on_search_keypress(event) {
    if (event.keyCode == 13)
	search();
}

function search() {
    debug("SEARCH");
    let text = document.getElementById("search_text").value;
    if (!text)
	return;

    let filter = document.getElementById("search_filter");
    if (filter.value.substring(0, 11) == "opensearch:") {
	debug(filter.value);
	let searcher = opensearchers[filter.value.substring(11)].searcher;
	let uri = searcher.get_search_uri(text);
	search_add("browse=uri:" + uri, text, "meta", null);
    } else {
	let label = null;
	if (filter.value.indexOf("=") > 0)
	    label = filter.label;
	let value = filter.value + "=" + text;
	search_add(value, text, filter.value, label);
    }
}

/*
 * @value: string used for searching
 * @text: search phrase - only used for display
 * @filter: filter type - only used for display
 * @label: custom label - only used for display (instead of filter)
 */
function search_add(value, text, filter, label) {
    debug("search_add: " + text + " : " + filter + " : " + value + " : " + label);
    let tree = document.getElementById("search_tree");

    let tch = document.getElementById("tree_children");
    let elems = tch.getElementsByTagName("treeitem");	// NOTE only children necessary
    let titem;

    // Check if there is an old item with the same value
    for (let el in loop(elems)) {
	let cell = el.getElementsByTagName("treecell").item(1);
	if (cell.getAttribute("label") == value) {
	    titem = el;
	    tch.removeChild(titem);
	    break;
	}
    }

    // Otherwise create a new item
    if (!titem) {
	let strbundle = document.getElementById("strings");
	if (!label)
	    try {
		label = strbundle.getString("filter_" + filter);
	    } catch (e) {
		label = filter;
	    }
	label += ": " + text;
	titem = create_search_item(label, value);
    }

    if (tch.lastChild == last_predefined)
	tch.appendChild(titem);
    else
	tch.insertBefore(titem, last_predefined.nextSibling);

    let col = tree.columns.getLastColumn();
    for (let i = 0; i < tree.view.rowCount; i++) {
	if (tree.view.getCellText(i, col) == value) {
	    tree.treeBoxObject.ensureRowIsVisible(i);
	    tree.view.selection.select(i);
	    break;
	}
    }

    // Remove oldest item if there are too many
    let count = 0;
    if (last_predefined) {
	for (let ch = tch.firstChild; ch && ch != last_predefined; ch = ch.nextSibling)
	    count++;
	count++;
    }
    if (tch.childNodes.length - count > 8)
	tch.removeChild(tch.lastChild);

    debug("search_add finished");
}

function treeitem_select(event) {
    let bar = document.getElementById("opensearch_bar");
    bar.hidden = true;
    let tree = document.getElementById("search_tree");
    if (tree.currentIndex < 0) {
	set_query("");
	let query_tree = document.getElementById("book_tree");
	query_tree.builder.rebuild();
	return;
    }

    let col = tree.columns.getLastColumn();
    let value = tree.view.getCellText(tree.currentIndex, col);
    debug("ASDF: " + value);
    let i = value.indexOf("=");
    if (i == -1) {
	//value == "all" || value == "feeds" || value == "untagged" ||
	// value == "top" || value == "recent"
	do_search(value, "", null);
    } else {
	let filter = value.substring(0, i);
	let text = value.substring(i + 1);
	debug("FILTER: " + filter + " TEXT: " + text);
	let node = tree.view.getItemAtIndex(tree.currentIndex);
	if (filter == "browse") {
	    do_browse(text, node);
	} else {
	    do_search(filter, text, null, node);
	}
    }
}

function do_search(filter, text, text2, node) {
    debug("do_search:" + filter + ":" + text + ":" + text2 + ":");

    if (current_searcher) {
	let searchTerms = "";
	let prefix = "";
	if (filter == "author")
	    prefix = "author:";
	else if (filter == "title")
	    prefix = "title:";
	if (current_searcher.syntax["compound"]) {
	    for (let phrase in loop(text.split(" ")))
		searchTerms += prefix + phrase + " ";
	    searchTerms = searchTerms.substring(0, searchTerms.length - 1);
	} else {
	    searchTerms = prefix + text;
	}
	let uri = current_searcher.get_search_uri(searchTerms);
	do_browse("uri:" + uri, node);
	return;
    }

    try {
	if (bcase.search) {
	    bcase.search(filter, text, text2);
	} else if (bcase.search_cb) {
	    function searcher(gen, sid) {
		debug("do search_cb");
		if (sid == current_sid) {
		    if (gen == null) {
			show_searchbar("failed");
		    } else {
			do_browse_gen(gen, node);
			grid_updated = false;
			if (view_mode == "icons")
			    update_grid();
			//update_book_info();
		    }
		}
	    }
	    show_searchbar("searching");
	    bcase.search_cb(filter, text, text2, searcher, new_sid());
	    return;
	} else if (bcase.search_gen) {
	    debug("do search_gen");
	    let gen = bcase.search_gen(filter, text, text2);
	    do_browse_gen(gen, null);
	} else if (bcase.capabilities.standard_search) {
	    search_standard(filter, text, text2);
	} else {
	    throw "incorrect type";
	}
    } catch (e) {
	debug("ER: " + bcase.capabilities);
	if (e == "error") {
	    show_searchbar("failed");
	} else if (e == "incorrect type") {
	    show_searchbar("failed unknown");
	} else
	    show_searchbar("failed unknown");
    }

    grid_updated = false;
    if (view_mode == "icons")
	update_grid();
    //update_book_info();
}

function show_searchbar(msg) {
    let text = "";
    if (msg == "searching") {
	text = "Searching...";
    } else if (msg == "failed") {
	text = "Search failed.";
    } else if (msg == "failed unknown") {
	text = "Search failed.";
    } else if (msg == "zero") {
	text = "Search returned no results.";
    }
    let bar = document.getElementById("opensearch_bar");
    bar.hidden = false;
    let label = document.getElementById("opensearch_label");
    label.value = text;

    let button = document.getElementById("prev_results");
    button.disabled = true;
    button = document.getElementById("next_results");
    button.disabled = true;
}

function search_standard(filter, text, text2) {
    debug("search_standard: " + filter + " " + text + " " + text2);
    let stat = "SELECT author, title, publisher, id FROM books";
    if (filter == "all") {
    } else if (filter == "feeds") {
	stat += " WHERE type='feed'";
    } else if (filter == "untagged") {
	stat += " WHERE tags ISNULL";
    } else if (filter == "range") {
	stat += " WHERE id BETWEEN " + text + " AND " + text2;
    } else if (filter == "author" || filter == "title" ||
	       filter == "meta") {
	if (text) {
	    if (filter == "author" || filter == "title") {
		stat += " WHERE LOWER(" + filter + ") LIKE LOWER(?1)";
	    } else {	// meta
		stat += " WHERE LOWER(title) LIKE LOWER(?1) OR LOWER(author) LIKE LOWER(?1)";
	    }
	    //text = text.replace("*", "%", "g");
	    // FIXME escape text
	    text = "%" + text + "%";
	    text = "'" + text + "'";
	    while (stat.indexOf("?1") >= 0)
		stat = stat.replace("?1", text, "g");
	}
    } else {
	throw "incorrect type";
    }

    set_query(stat);
    let query_tree = document.getElementById("book_tree");
    query_tree.builder.rebuild();
    return true;
}

function set_query(stat) {
    debug("set_query: " + stat);
    let query_tree = document.getElementById("book_tree");
    if (query_tree.view) {
	query_tree.view.selection.currentIndex = -1;
	query_tree.view.selection.clearSelection();
    }
    query_tree.view = original_view;

    let query = query_tree.ownerDocument.getElementById("query");
    while (query.hasChildNodes())
	query.removeChild(query.lastChild);
    let node = query_tree.ownerDocument.createTextNode(stat);
    query.appendChild(node);
}

var current_sid = 0;
function new_sid() {
    return ++current_sid;
}

function do_browse(text, node) {
    debug("do_browse: " + text + " : " + node.localName);

    if (bcase.browse_cb) {
	show_searchbar("searching");
	function browsed(gen, sid) {
	    debug("browsed");
	    if (sid == current_sid) {
		if (gen == null) {
		    show_searchbar("failed");
		} else {
		    do_browse_gen(gen, node);
		}
	    }
	}
	bcase.browse_cb(text, browsed, new_sid());
    } else if (bcase.browse_gen) {
	let gen = bcase.browse_gen(text);
	do_browse_gen(gen, node);
    } else {
	throw "Error: cannot browse";
    }
}

function do_browse_gen(gen, node) {
    debug("do_browse_gen");

    if (node.lastChild && node.lastChild.localName == "treechildren")
	while (node.lastChild.hasChildNodes())
	    node.lastChild.removeChild(node.lastChild.firstChild);

    try {
	let tree = document.getElementById("book_tree");
	tree.view = treeView;
	tree.view.selection.currentIndex = -1;
	tree.view.selection.clearSelection();
	treeView.rowCount = 0;
	treeView.list = [];

	let label = document.getElementById("opensearch_label");
	label.value = "";

	let jump = false;
	for (let link in gen) {
	    if (link.thisnode) {
		if (link.links) {
		    for (let li in loop(link.links)) {
			if (li.type == "opensearch") {
			    add_opensearch(li.href);
			} else if (li.type == "search") {
			    add_stanza_search(li.href, li.title);
			}
		    }
		}
		if (link.firstpath) {
		    node = add_first_link(node, link);
		    jump = true;
		    for (let link in gen)
			;
		}
		if (link.nextpath) {
		    if (!node.hasAttribute("is_first") &&
			!node.hasAttribute("is_next")) {
			node = add_first_link(node, link);
			jump = true;
			add_next_link(node, link);
			for (let link in gen)
			    ;
		    } else if (!node.hasAttribute("has_next")) {
			link.title = "↳ " + link.title;
			add_next_link(node, link);
		    }
		}
		if (link.pageinfo) {
		    // FIXME localize
		    let first = link.pageinfo.startIndex;
		    let total = link.pageinfo.totalResults;
		    let last = -1;
		    if (link.pageinfo.startIndex != -1 &&
			link.pageinfo.itemsPerPage != -1) {
			last = link.pageinfo.startIndex +
			    link.pageinfo.itemsPerPage - 1;
		    }
		    if (last > link.pageinfo.totalResults)
			last = link.pageinfo.totalResults;
		    let text = "";
		    if (first > -1 && last > -1 && total > -1) {
			text = "Showing results " + first +
			    " to " + last + " of " + total;
		    } else if (first > -1 && last > -1) {
			text = "Showing results " + first + " to " + last;
		    }
		    //let label = document.getElementById("opensearch_label");
		    label.value = text;
		}

		let bar = document.getElementById("opensearch_bar");
		if (node.hasAttribute("is_first") ||
		    node.hasAttribute("is_next") ||
		    node.hasAttribute("has_next")) {
		    let button = document.getElementById("prev_results");
		    if (node.hasAttribute("is_next"))
			button.disabled = false;
		    else
			button.disabled = true;
		    button = document.getElementById("next_results");
		    if (node.hasAttribute("has_next"))
			button.disabled = false;
		    else
			button.disabled = true;
		    bar.hidden = false;
		}
	    } else
		add_link(node, link);
	}

	tree.treeBoxObject.beginUpdateBatch();
	//tree.treeBoxObject.rowCountChanged(0, treeView.list.length);
	treeView.rowCount = treeView.list.length;
	tree.treeBoxObject.endUpdateBatch();

	grid_updated = false;
	if (view_mode == "icons")
	    update_grid();
	//update_book_info();
	if (jump)
	    child_results();
    } catch (e) {
	debug("--- " + e);
    }
}

function previous_results() {
    let stree = document.getElementById("search_tree");
    let current = stree.view.selection.currentIndex;
    if (current < 1)
	return;
    let view = stree.view.QueryInterface(Components.interfaces.nsITreeContentView);
    let item = view.getItemAtIndex(current);
    if (!item.previousSibling)
	return;
    let index = view.getIndexOfItem(item.previousSibling);
    // This should not be needed
    //if (!stree.view.isContainerOpen(current))
    //	stree.view.toggleOpenState(current);
    stree.view.selection.select(index);
    scroll_to_row(stree.view.selection.tree, index);
}

function next_results() {
    let stree = document.getElementById("search_tree");
    let current = stree.view.selection.currentIndex;
    debug("CURRENT: " + current);
    if (current < 0)
	return;
    let view = stree.view.QueryInterface(Components.interfaces.nsITreeContentView);
    let item = view.getItemAtIndex(current);
    if (!item.nextSibling)
	return;
    let index = view.getIndexOfItem(item.nextSibling);
    // This should not be needed
    //if (!stree.view.isContainerOpen(current))
    //	stree.view.toggleOpenState(current);
    stree.view.selection.select(index);
    scroll_to_row(stree.view.selection.tree, index);
}

function scroll_to_row(tree, index) {
    // FIXME this is not exactly what I want
    //tree.scrollToRow(index);
}

function child_results() {
    let stree = document.getElementById("search_tree");
    let current = stree.view.selection.currentIndex;
    if (current < 0)
	return;
    if (!stree.view.isContainerOpen(current))
	stree.view.toggleOpenState(current);
    stree.view.selection.select(current + 1);
}

function add_first_link(node, link) {
    debug("add_first_link " + link.title + " " + link.type);

    //if (node.hasAttribute("is_next"))
    //	return;

    let ndoc = node.ownerDocument;
    let chld = node.lastChild;
    if (chld.localName != "treechildren") {
	chld = ndoc.createElement("treechildren");
	node.appendChild(chld);
	node.setAttribute("container", "true");
	node.setAttribute("open", "true");
    }

    // ☞»⁝…›↓→↪↳‣
    let cit = create_catalog_item(ndoc, "↳ " + link.title, link.path);
    chld.appendChild(cit);

    cit.setAttribute("is_first", "true");
    //cit.setAttribute("has_next", "true");
    return cit;
}

function add_next_link(node, link) {
    debug("add_next_link " + link.nexttitle + " " + link.type);

    if (node.hasAttribute("has_next"))
	return;

    let ndoc = node.ownerDocument;
    let cit = create_catalog_item(ndoc, "↳ " + link.nexttitle, link.nextpath);
    if (node.nextSibling)
	node.parentNode.insertBefore(cit, node.nextSibling);
    else
	node.parentNode.appendChild(cit);
    cit.setAttribute("is_next", "true");
    node.setAttribute("has_next", "true");
}

function add_link(node, link) {
    debug("add_link " + link.title + " " + link.type);
    if (node && link.type == "folder") {
	debug("CATALOG " + node);
	let ndoc = node.ownerDocument;
	let chld = node.lastChild;
	if (chld.localName != "treechildren") {
	    chld = ndoc.createElement("treechildren");
	    node.appendChild(chld);
	    node.setAttribute("container", "true");
	    node.setAttribute("open", "true");
	}
	let cit = create_catalog_item(ndoc, link.title, link.path);
	chld.appendChild(cit);
    } else if (link.type == "epub" || link.type == "link") {
	debug("ITEM");
	if (node)
	    node.my_has_items = true;

	//let id, updated;
	/*
	  if (link.entry.id)
	  id = link.entry.id;
	  if (link.entry.updated)
	  updated = link.entry.updated;
	*/
	let bi = {};
	bi.id = link.id;	// FIXME
	//bi.uid = id;	// FIXME
	bi.title = link.title;
	bi.author = link.author;
	bi.publisher = link.publisher;
	bi.location = link.location;

	treeView.list.push(bi);
    }
}

function create_catalog_item(doc, title, path) {
    let it = doc.createElement("treeitem");
    let row = doc.createElement("treerow");
    let cell = doc.createElement("treecell");
    cell.setAttribute("label", title);	// FIXME begränsa längd
    row.appendChild(cell);
    cell = doc.createElement("treecell");
    cell.setAttribute("label", "browse=" + path);
    row.appendChild(cell);
    it.appendChild(row);
    return it;
}

var current_searcher = null;
var opensearchers = {};

function set_current_searcher(new_searcher) {
    if (new_searcher != current_searcher) {
	current_searcher = new_searcher;

	let item = document.getElementById("meta_item");
	if (current_searcher)
	    item.setAttribute("hidden", false);
	else
	    item.setAttribute("hidden", true);

	item = document.getElementById("title_item");
	if (current_searcher && current_searcher.syntax["author"])
	    item.setAttribute("hidden", false);
	else
	    item.setAttribute("hidden", true);

	item = document.getElementById("author_item");
	if (current_searcher && current_searcher.syntax["title"])
	    item.setAttribute("hidden", false);
	else
	    item.setAttribute("hidden", true);

	item = document.getElementById("search_text");
	if (current_searcher)
	    item.disabled = false;
	else
	    item.disabled = true;

	item = document.getElementById("search_filter");
	if (current_searcher)
	    item.disabled = false;
	else
	    item.disabled = true;
    }
}

function new_opensearch(se, uri) {
    debug("new_opensearch: " + se + " : " + uri);
    if (se && opensearchers[uri]) {
	opensearchers[uri].searcher = se;
	let title = "Search";
	if (se.shortname)
	    title = se.shortname;
	if (!current_searcher)
	    show_warning("Search option found", null, true);
	set_current_searcher(se);
    }
}

function add_opensearch(uri) {
    debug("add_opensearch: " + uri);
    if (!(uri in opensearchers)) {
	opensearchers[uri] = { node: null };
	bcase.get_opensearch(uri, new_opensearch);
    }
}

function add_custom_opensearch(uri, title) {
    let top = document.getElementById("text_item");
    let parent = top.parentNode;
    let value = "opensearch:" + uri;
    let item = document.createElement("menuitem");
    item.setAttribute("label", title);
    item.setAttribute("value", value);
    parent.appendChild(item);

    let aitem = document.getElementById("search_filter");
    aitem.selectedItem = item;
    aitem.disabled = false;
    aitem = document.getElementById("search_text");
    aitem.disabled = false;
}

function stanzasearch(href, title) {
    this.type = "stanza";
    this.shortname = title;
    this.syntax = {};
    this.template = href;
    this.wrappedJSObject = this;
}

stanzasearch.prototype = {
    get_search_uri: function(phrase) {
	return this.template.replace("{searchTerms}", escape(phrase));
    }
};

function add_stanza_search(href, title) {
    if (current_searcher)
	return false;	// Do not override OpenSearch
    let se = new stanzasearch(href, title);
    //let title = "Search";
    //if (se.shortname)
    //	title = se.shortname;
    set_current_searcher(se);
    //show_warning("New search option found", null, true);
    return true;
}

function sort_click() {
    grid_updated = false;
    /*
    let tree = document.getElementById("book_tree");
    tree.view.selection.clearSelection();
    */
    /* FIXME horrible hack */
    setTimeout(reselect, 250);
}

function reselect() {
    update_book_info();
}

function update_book_info() {
    let bookinfo = null;
    let tree = document.getElementById("book_tree");

    if (tree.view && tree.view.selection.currentIndex >= 0) {
	let col = tree.columns.getLastColumn();
	let id;
	try {
	    // FIXME sometimes this will fail for an unknown reason
	    id = tree.view.getCellText(tree.view.selection.currentIndex, col);
	    bookinfo = bcase.get_book_info(id);
	    if (!bookinfo) {
		// This happens if the book has been removed in another
		// instance of the bookcase
		// FIXME localize
		show_warning("Error: Could not find book", null, true);
	    }
	} catch (e) {
	}
    }
    update_book_info_2(bookinfo);
    if (bookinfo) {
	old_book_id = bookinfo.id;
	if (bcase.get_full_entry &&
	    bcase.get_full_entry(bookinfo.id, update_full_info)) {
	    // FIXME start visual feedback that entry is loading
	    // must make sure bcase.get_full_entry returns before
	    // it calls callback
	}
    }
}

var old_book_id = null;

function update_full_info(bookinfo) {
    debug("update_full_info");
    // FIXME stop visual feedback that entry is loading
    if (bookinfo && (bookinfo.id == old_book_id))
	update_book_info_2(bookinfo);
}

// FIXME
function xmlescape(uri) {
    return uri.replace("'", "\'", "g");
}

function update_book_info_2(bookinfo) {
    let book_box = document.getElementById("book_info_box");
    while (book_box.lastChild)
    	book_box.removeChild(book_box.lastChild);

    if (!bookinfo) {
	let button = document.getElementById("open_book");
	button.setAttribute("disabled", "true");
	button = document.getElementById("open_book_in_tab");
	button.setAttribute("disabled", "true");
	button = document.getElementById("show_properties");
	button.setAttribute("disabled", "true");
	button = document.getElementById("remove_book");
	button.setAttribute("disabled", "true");
	return;
    }

    let bool = "false";
    if (bookinfo.type == "link")
	bool = "true";
    let button = document.getElementById("open_book");
    button.setAttribute("disabled", bool);
    button = document.getElementById("open_book_in_tab");
    button.setAttribute("disabled", bool);

    bool = "true";
    if (bookinfo.location.substring(0, 5) == "file:")
	bool = "false";
    button = document.getElementById("show_properties");
    button.setAttribute("disabled", bool);

    button = document.getElementById("remove_book");
    button.setAttribute("disabled", "false");

    // Cover preview
    let imgbox = document.createElementNS(XUL_NS, "vbox");
    imgbox.setAttribute("class", "info_cover");
    let img = get_book_preview(bookinfo);
    if (img) {
	imgbox.appendChild(img);
	book_box.appendChild(imgbox);
    }

    let item = document.createElementNS(XUL_NS, "spacer");
    item.setAttribute("flex", 1);
    imgbox.appendChild(item);

    // Info
    let infobox = document.createElementNS(XUL_NS, "vbox");
    infobox.setAttribute("flex", "1");
    book_box.appendChild(infobox);

    let info_frame = document.createElementNS(XUL_NS, "vbox");
    info_frame.setAttribute("class", "infoframe");
    infobox.appendChild(info_frame);

    item = document.createElementNS(XHTML_NS, "div");
    //item = document.createElementNS(XUL_NS, "description");
    let text = document.createTextNode(bookinfo.title);
    item.appendChild(text);
    info_frame.appendChild(item);

    if (bookinfo.author) {
	item = document.createElementNS(XHTML_NS, "div");
	//item = document.createElementNS(XUL_NS, "description");
	item.setAttribute("class", "author");
	text = document.createTextNode(bookinfo.author);
	item.appendChild(text);
	info_frame.appendChild(item);
    }

    if (bookinfo.publisher) {
	item = document.createElementNS(XHTML_NS, "div");
	//item = document.createElementNS(XUL_NS, "description");
	text = document.createTextNode(bookinfo.publisher);
	item.appendChild(text);
	info_frame.appendChild(item);
    }

    if (bookinfo.type != "link") {
	// FIXME Show bookinfo.location, but with max-width and full in tooltiptext
	item = document.createElementNS(XHTML_NS, "div");
	//item = document.createElementNS(XUL_NS, "description");
	//item.setAttribute("crop", "end");
	let loc = bookinfo.location;
	if (loc.length > 70) {
	    loc = loc.substring(0, 67) + "...";
	    item.setAttribute("tooltiptext", bookinfo.location);
	}
	text = document.createTextNode(loc);
	item.appendChild(text);
	//item.setAttribute("value", loc);
	info_frame.appendChild(item);
    }

    // Links
    if (bookinfo.links) {
	debug("LINKS!!");
	for (link in loop(bookinfo.links)) {
	    item = document.createElementNS(XHTML_NS, "div");
	    item.setAttribute("class", "makeroom");
	    let item2 = document.createElementNS(XHTML_NS, "a");
	    item2.setAttribute("class", "link");
	    item2.setAttribute("href", "#");
	    let title = link.title;
	    if (!title)
		title = link.href;
	    if (link.type == "application/atom+xml")
		item2.setAttribute("onclick", "open_browse('" +
				   xmlescape(link.href) + "', '" +
				   xmlescape(title) + "');");
	    else {
		item2.setAttribute("onclick", "open_weblink('" +
				   xmlescape(link.href) + "');");
		item2.setAttribute("tooltiptext", link.href);
	    }
	    text = document.createTextNode(title);
	    item2.appendChild(text);
	    item.appendChild(item2);
	    info_frame.appendChild(item);
	}
    }

    item = document.createElementNS(XUL_NS, "spacer");
    item.setAttribute("flex", 100);
    book_box.appendChild(item);
}

function show_properties() {
    let tree = document.getElementById("book_tree");

    if (tree.view && tree.view.selection.currentIndex >= 0) {
	try {
	    let col = tree.columns.getLastColumn();
	    let id = tree.view.getCellText(tree.view.selection.currentIndex,
					   col);
	    let bookinfo = bcase.get_book_info(id);
	    if (bookinfo && bookinfo.type)
		window.open("chrome://lucidor/content/epubproperties.xul?uri=" +
			    escape(bookinfo.location), null,
			    "chrome,titlebar,centerscreen");
	} catch (e) {
	}
    }
}

function open_browse(url, text) {
    debug("open_browse: " + url);
    search_add("browse=uri:" + url, text, "browse", null);
}

function open_weblink(url) {
    debug("open_weblink: " + url);
    if (!bcase.warn_about_drm)
	main_control.add_web_tab(url);
    else {
	let texts = [ "If you want to purchase ebooks, " +
		      "please make sure that the books are not encumbered " +
		      "with DRM (Digital Rights Management).",
		      //"DRM limits your freedom to read your books on your " +
		      //"own conditions, " +
		      "DRM puts artificial restrictions on your freedom " +
		      "to read your books on your own conditions. " +
		      "It is not possible to read books " +
		      "tainted with DRM using Lucidor." ];
	//let checkbox = "I understand, and do not need to be reminded again.";
	//let checkbox = "I agree that DRM is bad, and do not need to be reminded again.";
	main_control.add_web_tab_with_warning(url, texts, true);
    }
}

function add_book_from_location() {
    // FIXME
    if (!bcase.add_book)
	return -1;
    let fp = Components.classes["@mozilla.org/filepicker;1"].
	createInstance(Components.interfaces.nsIFilePicker);
    let strbundle = document.getElementById("strings");
    let text = strbundle.getString("add_book");
    fp.init(window, text,
	    Components.interfaces.nsIFilePicker.modeOpenMultiple);
    let sharedbundle = document.getElementById("shared_strings");
    text = sharedbundle.getString("books");
    fp.appendFilter(text, "*.epub ; *.EPUB");
    //fp.appendFilter(text, "*.epub ; *.EPUB ; *.pdf ; *.PDF");
    text = sharedbundle.getString("all_files");
    fp.appendFilter(text, "*");
    let res = fp.show();
    if (res == fp.returnOK) {
	let first = -1;
	let en = fp.files;
	let id;
	while (en.hasMoreElements()) {
	    let file = en.getNext().QueryInterface(Components.interfaces.nsIFile);
	    id = bcase.add_book(file);
	    if (id == -1) {
		text = strbundle.getString("open_error");
		alert(text);
	    } else if (first == -1)
		first = id;
	}
	if (first != -1) {
	    let tree = document.getElementById("search_tree");
	    tree.view.selection.currentIndex = -1;
	    tree.view.selection.clearSelection();
	    // FIXME
	    do_search("range", first, id);
	    tree = document.getElementById("book_tree");
	    tree.view.selection.select(0);
	}
    }
}

function open_book_from_bookcase(newtab) {
    let item = document.getElementById("book_tree");
    if (item.currentIndex < 0)
	return;
    let col = item.columns.getLastColumn();
    let id = item.view.getCellText(item.currentIndex, col);

    let bookinfo = bcase.get_book_info(id);
    if (!bookinfo || !bookinfo.type) {
	// Detta händer om boken tagits bort i en annan instans av bookcase
	let strbundle = document.getElementById("strings");
	let text = strbundle.getString("book_unavailable");
	show_warning(text, null, true);
	return;
    }
    debug("open_book_from_bookcase: " + bookinfo.location);
    debug(bookinfo.type);
    if (bookinfo.type == "feed") {
	open_feed(bookinfo, newtab);
	return;
    }

    let location;

    let type = bookinfo.type.split(";");
    if (type[0] == "text" || type[0] == "text/plain") {
	debug("Text document");
	location = get_text_uri(bookinfo);
    } else {	// epub
	location = bookinfo.location;
    }

    if (newtab)
	main_control.add_epub_tab(location);
    else
	main_control.open_epub_location(location);
}

function get_text_uri(bookinfo) {
    let type = bookinfo.type.split(";");
    let location = "x-lucidor-convert://text/?uri=" +
	encodeURI(bookinfo.location);
    /*
    for (typ in type) {
	if (typ.substring(0, 8) == "charset=")
	    location += "?charset=" + typ.substring(8);
    }
    */
    return location;
}

function open_feed(bookinfo, newtab) {
    let location = "x-lucidor-convert://feed/?";

    let JSON = Components.classes["@mozilla.org/dom/json;1"].
	createInstance(Components.interfaces.nsIJSON);
    let feed = JSON.decode(bookinfo.location);
    for (let item in loop(feed.items))
	location += "uri=" + escape(item.uri) + "&";

    if (bookinfo.title)
	location += "title=" + escape(bookinfo.title);

    if (newtab)
	main_control.add_epub_tab(location);
    else
	main_control.open_epub_location(location);
}

function remove_book() {
    if (!bcase.remove_book)
	return -1;
    let tree = document.getElementById("book_tree");
    if (tree.currentIndex < 0)
	return;
    let col = tree.columns.getLastColumn();
    let id = tree.view.getCellText(tree.currentIndex, col);

    let bookinfo = bcase.get_book_info(id);
    if (!bookinfo) {
	// FIXME show warning?
	// Detta händer om boken tagits bort i en annan instans av bookcase
	treeitem_select();
	update_book_info();
	return;
    }

    let location = bookinfo.location;

    let prompts = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);
    let chk = {value: false};
    let ok;

    let strbundle = document.getElementById("strings");
    let title = strbundle.getString("remove_book");
    let text = strbundle.getString("confirm_remove");
    if (bookinfo.type != "feed" && location.indexOf("file://") == 0) {
	// FIXME set to true if location is contained within bookcase folder?
	//chk = {value: true};
	let del_text = strbundle.getString("confirm_delete");
	ok = prompts.confirmCheck(window, title, text, del_text, chk);
    } else {
	ok = prompts.confirm(window, title, text);
    }

    if (ok) {
	bcase.remove_book(id);

	if (chk.value) {
	    let file = Components.classes["@mozilla.org/network/protocol;1?name=file"].
		getService(Components.interfaces.nsIFileProtocolHandler).
		getFileFromURLSpec(location);
	    try {
		file.remove(false);
	    } catch (e) {
		// FIXME Should this be reported to the user?
	    }
	}

	treeitem_select();
	update_book_info();
    }
}

function set_view_mode(mode) {
    view_mode = mode;
    if (mode == "list")
	document.getElementById("deck").selectedIndex = 0;
    else {
	update_grid();
	document.getElementById("deck").selectedIndex = 1;
    }
}

function select_iconbox(event, id) {
    if (event.detail >= 2)
	open_book_from_bookcase(id, false);

    if (selected_iconbox && selected_iconbox != event.currentTarget)
	selected_iconbox.setAttribute("class", "iconbox");
    selected_iconbox = event.currentTarget;
    selected_iconbox.setAttribute("class", "iconbox-selected");

    let tree = document.getElementById("book_tree");
    let col = tree.columns.getLastColumn();
    for (let i = 0; i < tree.view.rowCount; i++) {
	let r_id = tree.view.getCellText(i, col);
	if (id == r_id) {
	    tree.view.selection.select(i);
	    break;
	}
    }
}

function set_selected_iconbox() {
    let tree = document.getElementById("book_tree");
    let grid = document.getElementById("grid");
    let iconbox = grid.childNodes.item(tree.currentIndex);
    if (selected_iconbox && (selected_iconbox != iconbox))
	selected_iconbox.setAttribute("class", "iconbox");
    if (iconbox) {
	selected_iconbox = iconbox;
	selected_iconbox.setAttribute("class", "iconbox-selected");
    }
}

function get_book_preview(bookinfo, empty) {
    let preview = null;
    if (bcase.get_preview)
	preview = bcase.get_preview(bookinfo.id);
    if (preview) {
	debug("PREVIEW 1");
	let item = document.createElementNS(XUL_NS, "image");
	item.setAttribute("src", preview);
	return item;
    } else if (bcase.get_preview_url) {
	preview = bcase.get_preview_url(bookinfo.id);
	debug("PREVIEW 1.5: " + preview);
	let item = document.createElementNS(XUL_NS, "image");
	item.setAttribute("src", preview);
	return item;
    } else if (bcase.generate_preview) {
	debug("PREVIEW 2");
	let item = document.createElementNS(XHTML_NS, "canvas");
	item.setAttribute("width", 100);
	item.setAttribute("height", 100);
	bcase.generate_preview(bookinfo, item, true);
	return item;
    } else if (bookinfo.location.substring(0, 7) == "file://") {
	debug("PREVIEW 3");
	let item = document.createElementNS(XHTML_NS, "canvas");
	item.setAttribute("width", 100);
	item.setAttribute("height", 100);

	try {
	    if (bookinfo.type == "epub") {
		debug("YEAH");
		let previews = Components.classes["@lucidor.org/previewCreatorService;1"].
		    getService().wrappedJSObject;
		previews.generate_preview(bookinfo.location, item,
					  bcase, bookinfo.id);
		return item;
	    }
	} catch (e) {
	    debug("Error: " + e);
	}
	debug("PREVIEW 3 done");
    }

    if (empty) {
	let item = document.createElementNS(XHTML_NS, "canvas");
	item.setAttribute("width", 100);
	item.setAttribute("height", 100);
	return item;
    }
    return null;
}

function update_grid() {
    if (grid_updated) {
	set_selected_iconbox();
	return;
    }
    grid_updated = true;

    let grid = document.getElementById("grid");
    while (grid.hasChildNodes())
	grid.removeChild(grid.firstChild);

    let tree = document.getElementById("book_tree");
    let rows = tree.view.rowCount;
    if (rows == 0)
	return;

    let id_col = tree.columns.getLastColumn();
    let author_col = tree.columns.getColumnAt(0);
    let title_col = tree.columns.getColumnAt(1);
    for (let i = 0; i < rows; i++) {
	let id = tree.view.getCellText(i, id_col);
	let author = tree.view.getCellText(i, author_col);
	let title = tree.view.getCellText(i, title_col);

	let box = document.createElementNS(XUL_NS, "vbox");
	box.setAttribute("onclick", "select_iconbox(event, " + id + ");");
	box.setAttribute("class", "iconbox");
	grid.appendChild(box);

	let imgbox = document.createElementNS(XUL_NS, "hbox");
	box.appendChild(imgbox);

	let item = document.createElementNS(XUL_NS, "spacer");
	item.setAttribute("flex", "1");
	imgbox.appendChild(item);

	let bookinfo = bcase.get_book_info(id);
	if (bookinfo) {
	    let img = get_book_preview(bookinfo, true);
	    if (img) {
		img.setAttribute("class", "iconimg");
		imgbox.appendChild(img);
	    }
	} else {
	}

	debug("AA");
	item = document.createElementNS(XUL_NS, "spacer");
	item.setAttribute("flex", "1");
	imgbox.appendChild(item);

	debug("AA1");

	item = document.createElementNS(XHTML_NS, "div");
	item.setAttribute("class", "icontitle");
	let text;
	if (title.length > 35) {
	    text = document.createTextNode(title.substring(0, 33) + "...");
	    item.setAttribute("tooltiptext", title);
	} else
	    text = document.createTextNode(title);
	item.appendChild(text);
	box.appendChild(item);

	debug("AA2");

	if (author) {
	    item = document.createElementNS(XHTML_NS, "div");
	    item.setAttribute("class", "iconauthor");
	    if (author.length > 35) {
		text = document.createTextNode(author.substring(0, 33) + "...");
		item.setAttribute("tooltiptext", author);
	    } else
		text = document.createTextNode(author);
	    item.appendChild(text);
	    box.appendChild(item);
	}
	debug("AAAA");
    }

    debug("END");

    set_selected_iconbox();
}

function update_cache() {
    debug("update_cache");

    if (bcase.update_message) {
	window.openDialog("chrome://lucidor/content/sync_dialog.xul",
			  "sd", "chrome", bcase, update_finish);
	debug("OPEN + CLOSED");
	return;
    }

    bcase.sync_cache();

    let tree = document.getElementById("search_tree");
    tree.view.selection.clearSelection();
    debug("update_cache finished");
}

function update_finish() {
    debug("update_finish");
    let tree = document.getElementById("search_tree");
    tree.view.selection.clearSelection();
    debug("update_cache finished");
}
