/*
 *  Lucidor - e-book reader
 *  Copyright © 2008-2010  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("resource://lucidor/modules/utils.jsm");

const XUL_NS = "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul";
const XHTML_NS = "http://www.w3.org/1999/xhtml";

function init() {
    let ioService = Components.classes["@mozilla.org/network/io-service;1"].
	getService(Components.interfaces.nsIIOService);
    let docuri = ioService.newURI(document.documentURI, "UTF-8", null).
	QueryInterface(Components.interfaces.nsIURL);
    let use_title = true;
    let location = unescape(docuri.query);
    if (!location) {
	use_title = false;
	location = "resource://lucidor/samples/";
    }
    let fph = Components.classes["@mozilla.org/network/protocol;1?name=file"].
	getService(Components.interfaces.nsIFileProtocolHandler);

    let dir;
    try {
	let uri = ioService.newURI(location, "UTF-8", null);
	if (uri.schemeIs("resource"))
	    location = Components.classes["@mozilla.org/network/protocol;1?name=resource"].
		getService(Components.interfaces.nsIResProtocolHandler).
		resolveURI(uri);
	dir = fph.getFileFromURLSpec(location);
    } catch (e) {
	return;
    }

    let strbundle = document.getElementById("strings");

    if (use_title) {
	let text = strbundle.getFormattedString("books_in_folder", [dir.path]);
	let header_text = document.createTextNode(text);
	let header = document.getElementById("header");
	for (let child in loop(header.childNodes))
	    if (child.nodeType == child.TEXT_NODE)
		header.removeChild(child);
	header.appendChild(header_text);
    }

    if (!dir.exists() || !dir.isDirectory()) {
	return;
    }

    let language_mapper = Components.classes["@lucidor.org/languageMapperService;1"].getService().wrappedJSObject;
    language_mapper.init();

    let sharedbundle = document.getElementById("shared_strings");

    function add_line(frame, key, value) {
	if (!value)
	    return;
	let label = sharedbundle.getString(key);
	return add_line2(frame, label, value, null);
    }

    function add_line2(frame, label, value, xclass) {
	let div = document.createElementNS(XHTML_NS, "div");
	let cls = "infoitem";
	if (xclass)
	    cls += " " + xclass;
	div.setAttribute("class", cls);
	frame.appendChild(div);

	if (label) {
	    let span = document.createElementNS(XHTML_NS, "span");
	    span.setAttribute("class", "infoheader");
	    div.appendChild(span);
	    let node = document.createTextNode(label + ": ");
	    span.appendChild(node);
	}

	let node;
	if (value.length >= 100)
	    node = document.createTextNode(value.substring(0, 90) + "...");
	else
	    node = document.createTextNode(value);
	div.appendChild(node);
	if (value.length >= 100)
	    div.setAttribute("tooltiptext", value);
    }

    let box = document.getElementById("list");
    let entries = dir.directoryEntries;
    while (entries.hasMoreElements()) {
	let entry = entries.getNext().QueryInterface(Components.interfaces.nsIFile);
	let loc = fph.getURLSpecFromFile(entry);
	if (loc.substring(loc.length - 5) != ".epub")
	    continue;

	let book = Components.classes["@lucidor.org/epubDocument;1"].
	    createInstance().wrappedJSObject;
	try {
	    book.init(entry);
	} catch(e) {
	    continue;
	}

	let hbox = document.createElementNS(XUL_NS, "hbox");
	box.appendChild(hbox);

	let main_frame = document.createElementNS(XUL_NS, "hbox");
	main_frame.setAttribute("class", "frame");
	main_frame.setAttribute("flex", "1");
	hbox.appendChild(main_frame);

	let imgbox = document.createElementNS(XUL_NS, "vbox");
	main_frame.appendChild(imgbox);

	let img = document.createElementNS(XHTML_NS, "canvas");
	img.setAttribute("class", "info_cover");
	imgbox.appendChild(img);
	let previews = Components.classes["@lucidor.org/previewCreatorService;1"].
	    getService().wrappedJSObject;
	previews.generate_preview(loc, img, null, -1);

	item = document.createElementNS(XUL_NS, "spacer");
	item.setAttribute("flex", "1");
	imgbox.appendChild(item);

	let infobox = document.createElementNS(XUL_NS, "vbox");
	infobox.setAttribute("flex", "1");
	main_frame.appendChild(infobox);

	let info_frame = document.createElementNS(XUL_NS, "vbox");
	infobox.appendChild(info_frame);

	add_line2(info_frame, null, book.opf.metadata.title, null);

	for (let creator in loop(book.opf.metadata.creator))
	    if (!creator.role || creator.role == "aut")
		add_line2(info_frame, null, creator.name, "author");

	if (book.opf.metadata.publisher != undefined)
	    add_line(info_frame, "publisher",
		     book.opf.metadata.publisher);

	if (book.opf.metadata.language != undefined) {
	    let lang = language_mapper.translate(book.opf.metadata.language);
	    add_line(info_frame, "language", lang);
	}

	let div = document.createElementNS(XHTML_NS, "div");
	div.setAttribute("align", "right");
	let item = document.createElementNS(XHTML_NS, "a");
	item.setAttribute("href", loc);
	let label = strbundle.getString("view_this_book");
	let text = document.createTextNode(label);
	item.appendChild(text);
	div.appendChild(item);
	info_frame.appendChild(div);

	item = document.createElementNS(XUL_NS, "spacer");
	item.setAttribute("flex", "1");
	infobox.appendChild(item);

	item = document.createElementNS(XUL_NS, "spacer");
	item.setAttribute("flex", 3);
	hbox.appendChild(item);
    }
}
