/*
 *  Lucidor - e-book reader
 *  Copyright © 2009-2010  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("resource://lucidor/modules/utils.jsm");

var current_collection;

function init() {
    let cman = Components.classes["@lucidor.org/collectionManagerService;1"].
	getService().wrappedJSObject;
    let collections = cman.getCollectionsJSON2();
    update_collection_list(collections);
}

function update_collection_list(collections) {
    debug("A");

    let list = document.getElementById("list");
    let count = list.getRowCount();
    while (count > 0) {
	list.removeItemAt(0);
	count--;
    }
    for (let coll in loop(collections))
	list.appendItem(coll[0], coll[1]);

    update_fields();
    debug("B");
}

function select_collection() {
    let list = document.getElementById("list");
    if (list.selectedIndex < 0)
	return;

    list.ensureIndexIsVisible(list.selectedIndex);
    let JSON = Components.classes["@mozilla.org/dom/json;1"].
	createInstance(Components.interfaces.nsIJSON);
    debug(": " + list.selectedIndex);
    debug(": " + list.selectedItem);
    debug(": " + list.selectedItem.label);
    debug("JSON.decode: " + list.selectedItem.value);
    current_collection = JSON.decode(list.selectedItem.value);

    let item = document.getElementById("nametext");
    item.value = current_collection.title;

    item = document.getElementById("typelist");
    item.value = current_collection.type;

    item = document.getElementById("urltext");
    if (current_collection.url)
	item.value = current_collection.url;
    else
	item.value = "";

    item = document.getElementById("path_text");
    if (current_collection.folder) {
	let fph = Components.classes["@mozilla.org/network/protocol;1?name=file"].
	    getService(Components.interfaces.nsIFileProtocolHandler);
	let dir;
	try {
	    dir = fph.getFileFromURLSpec(current_collection.folder);
	} catch (e) {
	}
	if (dir)
	    item.value = dir.path;
	else
	    item.value = current_collection.folder;
    } else
	item.value = "";

    item = document.getElementById("promptcheck");
    debug("CHECKED: " + current_collection.prompt);
    if (current_collection.prompt == undefined)
	item.checked = true;
    else if (current_collection.prompt)
	item.checked = current_collection.prompt;
    else
	item.checked = false;

    item = document.getElementById("usertext");
    if (current_collection.user)
	item.value = current_collection.user;
    else
	item.value = "";

    item = document.getElementById("passtext");
    if (current_collection.pass)
	item.value = current_collection.pass;
    else
	item.value = "";

    update_fields();
}

function select_type() {
    let typelist = document.getElementById("typelist");
    let cman = Components.classes["@lucidor.org/collectionManagerService;1"].
	getService().wrappedJSObject;
    let conf = cman.getConfig(typelist.value);
    if (!conf)
	conf = { url: false, urls: false, folder: false, auth: false };

    let elem = document.getElementById("urlbox");
    if (conf.url)
	elem.removeAttribute("hidden");
    else
	elem.setAttribute("hidden", "true");
    elem = document.getElementById("folderbox");
    if (conf.folder)
	elem.removeAttribute("hidden");
    else
	elem.setAttribute("hidden", "true");
    elem = document.getElementById("authbox");
    if (conf.auth) {
	elem.removeAttribute("hidden");
	set_prompt();
    } else
	elem.setAttribute("hidden", "true");

    update_current("type");
}

function add_collection() {
    let coll = { title: "New Collection",
		 type: "opds" };
    let JSON = Components.classes["@mozilla.org/dom/json;1"].
	createInstance(Components.interfaces.nsIJSON);
    let list = document.getElementById("list");
    let item = list.appendItem(coll.title, JSON.encode(coll));
    list.ensureIndexIsVisible(list.getRowCount());
    list.selectItem(item);
}

function remove_collection() {
    let list = document.getElementById("list");
    list.removeItemAt(list.selectedIndex);
    update_fields();
}

function default_collections() {
    let cman = Components.classes["@lucidor.org/collectionManagerService;1"].
	getService().wrappedJSObject;
    let collections = cman.getDefaultCollectionsJSON2();
    update_collection_list(collections);
}

function move_up() {
    let list = document.getElementById("list");
    let ind = list.selectedIndex;
    if (ind == 0)
	return;
    let it = list.getItemAtIndex(ind - 1);
    let lab = it.label;
    let val = it.value;
    list.ensureIndexIsVisible(ind - 1);
    list.removeItemAt(ind - 1);
    list.insertItemAt(ind, lab, val);
    list.ensureIndexIsVisible(ind - 1);
    update_buttons();
}

function move_down() {
    let list = document.getElementById("list");
    let ind = list.selectedIndex;
    if (ind == list.getRowCount())
	return;
    let it = list.getItemAtIndex(ind + 1);
    let lab = it.label;
    let val = it.value;
    list.ensureIndexIsVisible(ind + 1);
    list.removeItemAt(ind + 1);
    list.insertItemAt(ind, lab, val);
    list.ensureIndexIsVisible(ind + 1);
    update_buttons();
}

function update_buttons() {
    let list = document.getElementById("list");
    if (list.selectedIndex < 0) {
	let item = document.getElementById("removebutton");
	item.setAttribute("disabled", "true");
	item = document.getElementById("upbutton");
	item.setAttribute("disabled", "true");
	item = document.getElementById("downbutton");
	item.setAttribute("disabled", "true");
    } else {
	let item = document.getElementById("removebutton");
	item.removeAttribute("disabled");
	if (list.selectedIndex == 0) {
	    item = document.getElementById("upbutton");
	    item.setAttribute("disabled", "true");
	} else {
	    item = document.getElementById("upbutton");
	    item.removeAttribute("disabled");
	}
	if (list.selectedIndex == list.getRowCount() - 1) {
	    item = document.getElementById("downbutton");
	    item.setAttribute("disabled", "true");
	} else {
	    item = document.getElementById("downbutton");
	    item.removeAttribute("disabled");
	}
    }
}

function update_fields() {
    let list = document.getElementById("list");
    if (list.selectedIndex < 0) {
	debug("DISABLE");
	for (let id in loop(["nametext", "typelist", "urltext", "promptcheck",
			     "usertext", "passtext"])) {
	    let item = document.getElementById(id);
	    item.value = "";
	    item.setAttribute("disabled", "true");
	}
    } else {
	debug("ENABLE");
	/*
	let item = document.getElementById("nametext");
	if (!item.hasAttribute("disabled"))
	return;
	*/
	for (let id in loop(["nametext", "typelist", "urltext", "promptcheck",
			     "usertext", "passtext"])) {
	    let item = document.getElementById(id);
	    item.removeAttribute("disabled");
	}
	update_prompt();
    }
    update_buttons();
}

function on_prompt() {
    update_current("prompt");
    update_prompt();
}

function update_prompt() {
    let item = document.getElementById("promptcheck");
    if (item.checked) {
	item = document.getElementById("usertext");
	item.setAttribute("disabled", "true");
	item = document.getElementById("passtext");    
	item.setAttribute("disabled", "true");
    } else {
	item = document.getElementById("usertext");
	item.removeAttribute("disabled");
	item = document.getElementById("passtext");    
	item.removeAttribute("disabled");
    }
}

function update_current(hint) {
    let list = document.getElementById("list");
    if (list.selectedIndex < 0)
	return;

    let item;
    if (hint == "name") {
	item = document.getElementById("nametext");
	current_collection.title = item.value;
	list.selectedItem.label = current_collection.title;
    } else if (hint == "type") {
	item = document.getElementById("typelist");
	current_collection.type = item.value;
    } else if (hint == "url") {
	item = document.getElementById("urltext");
	current_collection.url = item.value;
    } else if (hint == "user") {
	item = document.getElementById("usertext");
	current_collection.user = item.value;
    } else if (hint == "pass") {
	item = document.getElementById("passtext");
	current_collection.pass = item.value;
    } else if (hint == "prompt") {
	item = document.getElementById("promptcheck");
	current_collection.prompt = item.checked;
    }

    let JSON = Components.classes["@mozilla.org/dom/json;1"].
	createInstance(Components.interfaces.nsIJSON);
    list.selectedItem.value = JSON.encode(current_collection);
}


function select_directory() {
    let fp = Components.classes["@mozilla.org/filepicker;1"].
	createInstance(Components.interfaces.nsIFilePicker);
    //let strbundle = document.getElementById("strings");
    //let title = strbundle.getString("select_path.title");
    let title = "Select Folder";
    fp.init(window, title, Components.interfaces.nsIFilePicker.modeGetFolder);

    let fph = Components.classes["@mozilla.org/network/protocol;1?name=file"].
	getService(Components.interfaces.nsIFileProtocolHandler);
    let dir;
    try {
	dir = fph.getFileFromURLSpec(current_collection.folder);
    } catch (e) {
    }
    if (dir)
	fp.displayDirectory = dir;

    let res = fp.show();
    if (res == fp.returnOK) {
	let tp = document.getElementById("path_text");
	tp.value = fp.file.path;

	current_collection.folder = fp.fileURL.spec;

	let JSON = Components.classes["@mozilla.org/dom/json;1"].
	    createInstance(Components.interfaces.nsIJSON);
	let list = document.getElementById("list");
	list.selectedItem.value = JSON.encode(current_collection);
    }
}

function save_changes() {
    debug("save_changes");
    let list = document.getElementById("list");
    let num = list.getRowCount();
    let collections = [];
    for (let i = 0; i < num; i++) {
	let item = list.getItemAtIndex(i);
	collections.push(item.value);
    }
    let cman = Components.classes["@lucidor.org/collectionManagerService;1"].
	getService().wrappedJSObject;
    cman.saveCollectionsJSON(collections);

    window.close();
}

function cancel() {
    window.close();
}
