/*
 *  Lucidor - e-book reader
 *  Copyright © 2008-2009  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("resource://lucidor/modules/utils.jsm");

const XHTML_NS = "http://www.w3.org/1999/xhtml";
const XUL_NS = "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul";

function init () {
    let ioService = Components.classes["@mozilla.org/network/io-service;1"].
	getService(Components.interfaces.nsIIOService);
    let docuri = ioService.newURI(document.documentURI, "UTF-8", null);
    docuri.QueryInterface(Components.interfaces.nsIURL);

    let urispec;
    let params = docuri.query.split("&");
    for (let par in loop(params)) {
	let pair = par.split("=");
	if (pair[0] == "uri")
	    urispec = unescape(pair[1]);
    }

    let nsuri = ioService.newURI(urispec, "UTF-8", null);
    if (nsuri.schemeIs("file")) {
	let file = Components.classes["@mozilla.org/network/protocol;1?name=file"].
	    getService(Components.interfaces.nsIFileProtocolHandler).
	    getFileFromURLSpec(urispec);
	try {
	    return init2(file, nsuri.spec);
	} catch (e) {
	    debug("init2 failed: " + e);
	}
    }
}

function init2(file, uri) {
    let book = Components.classes["@lucidor.org/epubDocument;1"].
	createInstance().wrappedJSObject;
    try {
	book.init(file);
    } catch (e) {
	debug("Error opening: " + e);
	return -1;
    }

    let box = document.getElementById("ve_info");
    let strbundle = document.getElementById("strings");
    let sharedbundle = document.getElementById("shared_strings");

    function add_line(entity_id, value) {
	add_line1(entity_id, value, null);
    }

    function add_line1(entity_id, value, extra) {
	if (!value)
	    return;
	let label = sharedbundle.getString(entity_id);
	if (extra)
	    label += " (" + extra + ")";
	add_line2(label, value);
    }

    let last_header = null;

    function add_line2(label, value) {
	if (!value)
	    return;
	if (label == last_header && last_header)
	    label = null;
	else if (label)
	    last_header = label;
	add_line3(label, value);
    }

    function add_line3(label, value) {
	if (label) {
	    let div = document.createElementNS(XHTML_NS, "div");
	    div.setAttribute("class", "infoheader");
	    box.appendChild(div);
	    let node = document.createTextNode(label);
	    div.appendChild(node);
	}
	if (value) {
	    let div = document.createElementNS(XHTML_NS, "div");
	    div.setAttribute("class", "infoitem");
	    box.appendChild(div);
	    let node;
	    if (value.length >= 100)
		node = document.createTextNode(value.substring(0, 90) + "...");
	    else
		node = document.createTextNode(value);
	    div.appendChild(node);
	    if (value.length >= 100)
		div.setAttribute("tooltiptext", value);
	}
    }

    function add_header(entity_id) {
	let label = sharedbundle.getString(entity_id);
	add_line3(label, null);
    }

    function add_role(person, default_role) {
	let label;
	try {
	    label = strbundle.getString("role_" + person.role);
	} catch (e) {
	    label = strbundle.getString(default_role);
	    if (person.role)
		label += " (" + person.role + ")";
	}
	add_line2(label, person.name);
    }

    function add_date(date) {
	let label;
	try {
	    label = strbundle.getString("date_" + date.event);
	} catch (e) {
	    label = sharedbundle.getString("date");
	    if (date.event)
		label += " (" + date.event + ")";
	}
	add_line2(label, date.date);
    }

    add_line("title", book.opf.metadata.title);
    if ((book.opf.metadata.creator.length == 1) &&
	!book.opf.metadata.creator[0].role)
	add_line("author", book.opf.metadata.creator[0].name);
    else
	for (let creator in loop(book.opf.metadata.creator))
	    add_role(creator, "creator");
    if (book.opf.metadata.contributor != undefined)
	for (let contrib in loop(book.opf.metadata.contributor))
	    add_role(contrib, "contributor");
    if (book.opf.metadata.publisher != undefined)
	add_line("publisher", book.opf.metadata.publisher);
    if (book.opf.metadata.rights != undefined)
	add_line("rights", book.opf.metadata.rights);
    if (book.opf.metadata.date != undefined)
	for (let date in loop(book.opf.metadata.date))
	    add_date(date);
    if (book.opf.metadata.languages != undefined) {
	let language_mapper = Components.classes["@lucidor.org/languageMapperService;1"].getService().wrappedJSObject;
	language_mapper.init();
	for (let language in loop(book.opf.metadata.languages)) {
	    let lang = language_mapper.translate(language);
	    add_line("language", lang);
	}
    }
    if (book.opf.metadata.description != undefined)
	add_line("description", book.opf.metadata.description);
    if (book.opf.metadata.subject != undefined)
	for (let subject in loop(book.opf.metadata.subject))
	    add_line("subject", subject);
    if (book.opf.metadata.coverage != undefined)
	for (let coverage in loop(book.opf.metadata.coverage))
	    add_line("coverage", coverage);
    if (book.opf.metadata.type != undefined)
	for (let type in loop(book.opf.metadata.type))
	    add_line("type", type);
    if (book.opf.metadata.format != undefined)
	for (let format in loop(book.opf.metadata.format))
	    add_line("format", format);
    if (book.opf.metadata.source != undefined)
	for (let source in loop(book.opf.metadata.source))
	    add_line("source", source);
    if (book.opf.metadata.relation != undefined)
	for (let relation in loop(book.opf.metadata.relation))
	    add_line("relation", relation);
    if (book.opf.uid != undefined)
	add_line("uid", book.opf.uid);
    for (let id in loop(book.opf.metadata.identifier))
	if (!id.uid)
	    add_line1("identifier", id.id, id.scheme);
    add_line("location", book.original_location);
}

function unload() {
}
