/*
 *  Lucidor - e-book reader
 *  Copyright © 2008-2011  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("resource://lucidor/modules/utils.jsm");

const XUL_NS = "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul";

var progressListener = {
    onLocationChange: function(webProgress, request, location) {
    },

    onProgressChange: function(webProgress, request, curSelfProgress,
			       maxSelfProgress, curTotalProgress,
			       maxTotalProgress) {
    },

    onSecurityChange: function(webProgress, request, state) {
    },

    onStateChange: function(webProgress, request, stateFlags, status) {
	if ((stateFlags & 16) && (stateFlags & 131072))
	    update_tab_title();
    },

    onStatusChange: function(webProgress, request, status, message) {
    },

    QueryInterface: function(aIID) {
	if (aIID.equals(Components.interfaces.nsIWebProgressListener) ||
	    aIID.equals(Components.interfaces.nsISupportsWeakReference))
	    return this;
	throw Components.results.NS_NOINTERFACE;
    }
};

var historyListener = {
    OnHistoryGoBack: function(backURI) {
	update_bar();
	return true;
    },

    OnHistoryGoForward: function(forwardURI) {
	update_bar();
	return true;
    },

    OnHistoryGotoIndex: function(index, gotoURI) {
	update_bar();
	return true;
    },

    OnHistoryNewEntry: function(newURI) {
	update_bar();
	return true;
    },

    OnHistoryPurge: function(numEntries) {
	update_bar();
	return true;
    },

    OnHistoryReload: function(reloadURI, reloadFlags) {
	update_bar();
	return true;
    },

    QueryInterface: function(aIID) {
	if (aIID.equals(Components.interfaces.nsISHistoryListener) ||
	    aIID.equals(Components.interfaces.nsISupportsWeakReference) ||
	    aIID.equals(Components.interfaces.nsISupports))
	    return this;
	throw Components.results.NS_NOINTERFACE;
    }
};

function init() {
    /*
    let chromeRegistry = Components.classes["@mozilla.org/chrome/chrome-registry;1"].getService(Components.interfaces.nsIXULChromeRegistry);
    let currentLocale = chromeRegistry.getSelectedLocale("lucidor");
    debug("Locale: " + currentLocale);
    */

    // Must keep reference to contentListener when using this
    //let uriLoader = Components.classes["@mozilla.org/uriloader;1"].
    //	getService(Components.interfaces.nsIURILoader);
    //uriLoader.registerContentListener(contentListener);

    // Hide Mac OS X application menu from other OSes
    let appinfo = Components.classes["@mozilla.org/xre/app-info;1"].
	getService(Components.interfaces.nsIXULRuntime);
    if (appinfo.OS != "Darwin" && appinfo.OS != "MacOSX") {
	let mb = document.getElementById("osx-app-menu");
	mb.setAttribute("hidden", "true");
	mb = document.getElementById("full_screen_button");
	mb.setAttribute("hidden", "false");
    }

    let control = Components.classes["@lucidor.org/mainControlService;1"].
	getService().wrappedJSObject;
    control.init(window);

    // Register default stylesheets
    let sss = Components.classes["@mozilla.org/content/style-sheet-service;1"].
	getService(Components.interfaces.nsIStyleSheetService);
    let ios = Components.classes["@mozilla.org/network/io-service;1"].
	getService(Components.interfaces.nsIIOService);
    let uri = ios.newURI("chrome://lucidor/content/dtbook.css", null, null);
    if (!sss.sheetRegistered(uri, sss.AGENT_SHEET))
	sss.loadAndRegisterSheet(uri, sss.AGENT_SHEET);
    uri = ios.newURI("chrome://lucidor/content/xhtml.css", null, null);
    if (!sss.sheetRegistered(uri, sss.AGENT_SHEET))
	sss.loadAndRegisterSheet(uri, sss.AGENT_SHEET);
    uri = ios.newURI("chrome://lucidor/content/ordbrand.css", null, null);
    if (!sss.sheetRegistered(uri, sss.AGENT_SHEET))
	sss.loadAndRegisterSheet(uri, sss.AGENT_SHEET);

    // Initialize previewCreator service
    let previews = Components.classes["@lucidor.org/previewCreatorService;1"].
	getService().wrappedJSObject;
    let pbr = document.getElementById("preview_browser");
    pbr.docShell.allowAuth = false;
    pbr.docShell.allowJavascript = false;
    pbr.docShell.allowMetaRedirects = false;
    pbr.docShell.allowPlugins = false;
    previews.init(pbr);

    // Initialize languageMapper service
    //let language_mapper = Components.classes["@lucidor.org/languageMapperService;1"].getService().wrappedJSObject;
    //language_mapper.init();
    //language_mapper.debug();

    // Check and maybe update preferences
    let prefs = Components.classes["@mozilla.org/preferences-service;1"].
	getService(Components.interfaces.nsIPrefService).
	getBranch("lucidor.");
    let major = 0;
    let minor = 0;
    let extra = 0;
    try {
	let ver = prefs.getCharPref("lastVersion");
	let parts = ver.split(".");
	major = Number(parts[0]);
	if (parts.length > 1)
	    minor = Number(parts[1]);
	if (parts.length > 2)
	    extra = Number(parts[2]);
    } catch (e) {
    }
    if (prefs.prefHasUserValue("collection.order")) {
	if (major == 0 && minor < 8) {
	    let cman = Components.classes["@lucidor.org/collectionManagerService;1"].
		getService().wrappedJSObject;
	    cman.addCollection({ title: "Google Books",
				 type: "googlebooks" });
	}
	if (major == 0 && minor < 7) {
	    let cman = Components.classes["@lucidor.org/collectionManagerService;1"].
		getService().wrappedJSObject;
	    cman.addCollection({ title: "epubcatalog.org",
				 type: "epubcatalog_atom" });
	    cman.addCollection({ title: "Meta Catalog", type: "opds",
				 url: "http://lucidor.org/opds/meta.atom" });
	}
	if (major == 0 && minor < 9) {
	    let cman = Components.classes["@lucidor.org/collectionManagerService;1"].
		getService().wrappedJSObject;
	    cman.addCollection({ title: "Smashwords",
				 type: "opds",
				 url: "http://www.smashwords.com/atom" });
	}
    }
    prefs.setCharPref("lastVersion", "0.9.7");

    // Make sure lucidor.bookcase.folder is set
    let bookcase = Components.classes["@lucidor.org/bookcaseSqlService;1"].
	getService().wrappedJSObject;
    bookcase.get_books_directory();

    // Handle command line arguments
    let clh = Components.classes["@lucidor.org/cmdlineHandler;1"].
	getService().wrappedJSObject;
    if (clh.defaultURIS.length > 0)
	for (let duri in loop(clh.defaultURIS))
	    add_epub_tab(duri.spec);
    else {
	let url = "chrome://lucidor/content/homepage.xul?startup";
	if (major + minor == 0)
	    url += ";newuser";
	add_tab(url);
    }
}

function toggle_full_screen() {
    let mb = document.getElementById("full_screen_button");
    if (mb.getAttribute("checked"))
	setTimeout('window.fullScreen = true;', 1);
    else
	setTimeout('window.fullScreen = false;', 1);
}

function open_preferences() {
    window.openDialog("chrome://lucidor/content/preferences.xul",
		      "Preferences",
		      "chrome,titlebar,toolbar,centerscreen,dialog=yes");
}

function open_addons() {
    //window.open("chrome://mozapps/content/extensions/extensions.xul?type=extensions",
    //		"Add-ons", "chrome,titlebar,toolbar,centerscreen");
    return add_tab("chrome://mozapps/content/extensions/extensions.xul?type=extensions");
}

function add_epub_tab(location) {
    return add_tab("chrome://lucidor/content/viewepub.xul?uri=" +
		   escape(location));
}

function add_web_tab(location) {
    return add_tab("chrome://lucidor/content/viewweb.xul?uri=" +
		   escape(location));
}

function add_web_tab_post(location, post) {
    return add_tab("chrome://lucidor/content/viewweb.xul?uri=" +
		   escape(location) + "&post=" + escape(post));
}

function add_tab(location) {
    let tabs = document.getElementById("tabs");
    let tab = document.createElementNS(XUL_NS, "tab");
    tab.setAttribute("label", " ");
    //tab.setAttribute("style", "overflow: hidden; crop: start; width: 8em;");
    tabs.appendChild(tab);

    let tabpanels = document.getElementById("tabpanels");
    let tabpanel = document.createElementNS(XUL_NS, "tabpanel");
    //tabpanel.setAttribute("flex", "1");
    let topbox = document.createElementNS(XUL_NS, "vbox");
    topbox.setAttribute("flex", "1");
    let vbox = document.createElementNS(XUL_NS, "vbox");
    topbox.appendChild(vbox);
    let browser = document.createElementNS(XUL_NS, "browser");
    browser.setAttribute("src", location);
    browser.setAttribute("disablehistory", "true");
    //browser.setAttribute("type", "content");
    browser.setAttribute("flex", "1");
    //tabpanel.appendChild(browser);
    topbox.appendChild(browser);
    tabpanel.appendChild(topbox);
    tabpanels.appendChild(tabpanel);
    tabs.selectedItem = tab;

    browser.addProgressListener(progressListener);
    //browser.sessionHistory.addSHistoryListener(historyListener);

    /*
    let cl = Components.classes["@lucidor.org/myURIContentListener;1"].createInstance(Components.interfaces.nsIURIContentListener);
    //let browser = document.getElementById("book_view");
    let wnd = browser.docShell.QueryInterface(Components.interfaces.nsIInterfaceRequestor).getInterface(Components.interfaces.nsIURIContentListener);
    wnd.parentContentListener = cl;
    */

    let button = document.getElementById("close-tab-button");
    button.setAttribute("disabled", "false");

    let box = document.getElementById("scrollbox");
    let sb = box.boxObject.QueryInterface(Components.interfaces.nsIScrollBoxObject);
    let x = tab.boxObject.x - tabs.childNodes.item(0).boxObject.x;	// FIXME
    sb.scrollTo(x, 0);
    return vbox;
}

function close_tab() {
    let tabs = document.getElementById("tabs");
    let index = tabs.selectedIndex;
    if (index == -1)
	return;

    // FIXME there must be a better way to do this
    let browser = get_browser();
    if (browser.contentWindow && browser.contentWindow.warn_unload)
	browser.contentWindow.warn_unload();

    tabs.removeItemAt(index);
    let tabpanels = document.getElementById("tabpanels");
    tabpanels.removeChild(tabpanels.selectedPanel);

    if (index == tabs.childNodes.length)
	index--;
    tabs.selectedIndex = index;
    if (index == -1) {
	let button = document.getElementById("close-tab-button");
	button.setAttribute("disabled", "true");
	update_bar();
    }

    let box = document.getElementById("scrollbox");
    let sb = box.boxObject.QueryInterface(Components.interfaces.nsIScrollBoxObject);
    let xp = {};
    sb.getPosition(xp, {});
    sb.scrollTo(xp.value, 0);	// FIXME
}

function overflow() {
    //debug("overflow");
    let box = document.getElementById("scrollbox");
    let sb = box.boxObject.QueryInterface(Components.interfaces.nsIScrollBoxObject);
    let wp = {};
    sb.getScrolledSize(wp, {});
    //debug("W: " + wp.value + " : " + box.boxObject.width);
    // NOTE Necessary because with some themes there are false overflow events
    if (wp.value <= box.boxObject.width)
	return underflow();

    let button = document.getElementById("tab_left");
    button.setAttribute("hidden", "false");
    button = document.getElementById("tab_right");
    button.setAttribute("hidden", "false");
}

function underflow() {
    //debug("underflow");
    let button = document.getElementById("tab_left");
    button.setAttribute("hidden", "true");
    button = document.getElementById("tab_right");
    button.setAttribute("hidden", "true");

    let box = document.getElementById("scrollbox");
    let sb = box.boxObject.QueryInterface(Components.interfaces.nsIScrollBoxObject);
    sb.scrollTo(0, 0);
}

function scroll_left() {
    let tabs = document.getElementById("tabs");
    let box = document.getElementById("scrollbox");
    let sb = box.boxObject.QueryInterface(Components.interfaces.nsIScrollBoxObject);
    let xp = {};
    let yp = {};
    sb.getPosition(xp, yp);
    let off = tabs.childNodes.item(0).boxObject.x;
    let j = 0;
    for (; j < tabs.childNodes.length; j++) {
	let p = tabs.childNodes.item(j).boxObject.x - off;
	//debug(";; " + tabs.childNodes.item(j).boxObject.x + " " + p + " " + xp.value);
	if (xp.value < p + 5)
	    break;
    }
    j = Math.max(j - 1, 0);
    let x = tabs.childNodes.item(j).boxObject.x - off;
    sb.scrollTo(x, yp.value);

    if (j == 0 || x <= 0) {
	let button = document.getElementById("tab_left");
	button.setAttribute("disabled", "true");
    }
    let button = document.getElementById("tab_right");
    button.setAttribute("disabled", "false");
}

function scroll_right() {
    let tabs = document.getElementById("tabs");
    let box = document.getElementById("scrollbox");
    let sb = box.boxObject.QueryInterface(Components.interfaces.nsIScrollBoxObject);
    let xp = {};
    sb.getPosition(xp, {});
    let off = tabs.childNodes.item(0).boxObject.x;
    let j = 0;
    for (; j < tabs.childNodes.length; j++) {
	let p = tabs.childNodes.item(j).boxObject.x - off;
	//debug(";; " + tabs.childNodes.item(j).boxObject.x + " " + p + " " + xp.value);
	if (xp.value < p + 5)
	    break;
    }
    j = Math.min(j + 1, tabs.childNodes.length - 1);
    let x = tabs.childNodes.item(j).boxObject.x - off;
    sb.scrollTo(x, 0);

    let last = tabs.childNodes.item(tabs.childNodes.length - 1);
    // FIXME
    if (x >= (last.boxObject.x + last.boxObject.width - box.boxObject.width)) {
	let button = document.getElementById("tab_right");
	button.setAttribute("disabled", "true");
    }
    let button = document.getElementById("tab_left");
    button.setAttribute("disabled", "false");
}

var dragObserver = {
    onDragOver: function(event, flavour, session) {
    },

    onDrop: function(event, transferData, session) {
	debug("DROP: " + transferData.data + " " + transferData.flavour.contentType);
	if (transferData.flavour.contentType == "application/x-moz-file") {
	    let uri = Components.classes["@mozilla.org/network/protocol;1?name=file"].getService(Components.interfaces.nsIFileProtocolHandler).
		getURLSpecFromFile(transferData.data);
	    add_epub_tab(uri);
	} else if (transferData.flavour.contentType == "text/x-moz-url") {
	    let ss = transferData.data.split("\n");
	    add_web_tab(ss[0]);
	}
    },

    getSupportedFlavours: function() {
	let flavours = new FlavourSet();
	flavours.appendFlavour("application/x-moz-file","nsIFile");
	flavours.appendFlavour("text/x-moz-url");
	return flavours;
    }
}

function open_epub_location(location) {
    return open_location("chrome://lucidor/content/viewepub.xul?uri=" +
			 escape(location));
}

function open_location(uri) {
    debug("open_location: " + uri);
    let browser = get_browser();
    if (browser.loadURI)
	browser.loadURI(uri, null, null);
    else
	add_tab(uri);
}

function open_book_from_location() {
    let fp = Components.classes["@mozilla.org/filepicker;1"].
	createInstance(Components.interfaces.nsIFilePicker);
    let strbundle = document.getElementById("strings");
    let text = strbundle.getString("open_file");
    fp.init(window, text, Components.interfaces.nsIFilePicker.modeOpen);
    let sharedbundle = document.getElementById("shared_strings");
    text = sharedbundle.getString("books");
    fp.appendFilter(text, "*.epub ; *.EPUB");
    //fp.appendFilter(text, "*.epub ; *.EPUB ; *.pdf ; *.PDF");
    text = sharedbundle.getString("all_files");
    fp.appendFilter(text, "*");
    let res = fp.show();
    if (res == fp.returnOK)
	open_epub_location(fp.fileURL.spec);
}

function save_file() {
    let browser = get_browser();
    let file;
    if (browser.contentWindow.get_save_file)
	file = browser.contentWindow.get_save_file();

    let fp = Components.classes["@mozilla.org/filepicker;1"].
	createInstance(Components.interfaces.nsIFilePicker);
    let strbundle = document.getElementById("strings");
    //let text = strbundle.getString("open_file");
    let text = "Save File";
    fp.init(window, text, Components.interfaces.nsIFilePicker.modeSave);

    let sharedbundle = document.getElementById("shared_strings");
    text = sharedbundle.getString("books");
    fp.appendFilter(text, "*.epub ; *.EPUB");
    //fp.appendFilter(text, "*.epub ; *.EPUB ; *.pdf ; *.PDF");
    text = sharedbundle.getString("all_files");
    fp.appendFilter(text, "*");

    if (file)
	fp.defaultString = file.leafName;
    else
	fp.defaultString = browser.contentWindow.get_save_name();
    fp.defaultExtension = "epub";
    let res = fp.show();

    function my_save_file(targetFile, replace) {
	debug("my_save_file");
	if (!replace && targetFile.exists())
	    return;
	// FIXME honour "replace", it is now OS specific, ok on GNU/Linux
	try {
	    file.copyTo(targetFile.parent, targetFile.leafName);
	} catch (e) {
	    let main_control = Components.classes["@lucidor.org/mainControlService;1"].getService().wrappedJSObject;
	    // FIXME localize
	    let text = "An error occurred when trying to save the file.";
	    main_control.show_warning(browser.contentWindow, [ text ],
				      null, false);
	}
    }
    let save_it = my_save_file;

    if (!file)
	save_it = browser.contentWindow.save_file;

    if (res == fp.returnOK)
	save_it(fp.file, false);
    else if (res == fp.returnReplace)
	save_it(fp.file, true);
}

function get_panel(win) {
    let tabpanels = document.getElementById("tabpanels");
    if (!win)
	return tabpanels.selectedPanel;
    let i = 0;
    for (let panel in loop(tabpanels.childNodes)) {
	if (panel.firstChild.firstChild.nextSibling.contentWindow == win)
	    return [panel, i];
	i++;
    }
    throw "Error";
}

function get_browser(panel) {
    if (panel)
	return panel.firstChild.firstChild.nextSibling;
    let tabpanels = document.getElementById("tabpanels");
    return tabpanels.selectedPanel.firstChild.firstChild.nextSibling;
}

function go_back() {
    let main_browser = get_browser();
    if (main_browser.contentWindow.has_browser) {
	let browser = main_browser.contentWindow.get_browser();
	if (browser.canGoBack) {
	    browser.goBack();
	    return;
	}
    }
    /*
    if (main_browser.canGoBack)
	main_browser.goBack();
    */
}

function go_forward() {
    let main_browser = get_browser();
    if (main_browser.contentWindow.has_browser) {
	let browser = main_browser.contentWindow.get_browser();
	if (browser.canGoForward) {
	    browser.goForward();
	    return;
	}
    }
    /*
    if (main_browser.canGoForward)
	main_browser.goForward();
    */
}

function zoom_in_text() {
    let browser = get_browser();
    if (browser.contentWindow.zoomable) {
	let mdv = browser.contentWindow.getMarkupDocumentViewer();
	mdv.textZoom += 0.05;
    }
}

function zoom_out_text() {
    let browser = get_browser();
    if (browser.contentWindow.zoomable) {
	let mdv = browser.contentWindow.getMarkupDocumentViewer();
	mdv.textZoom -= 0.05;
    }
}

function zoom_in() {
    let browser = get_browser();
    if (browser.contentWindow.zoomable) {
	let mdv = browser.contentWindow.getMarkupDocumentViewer();
	mdv.fullZoom += 0.05;
    }
}

function zoom_out() {
    let browser = get_browser();
    if (browser.contentWindow.zoomable) {
	let mdv = browser.contentWindow.getMarkupDocumentViewer();
	mdv.fullZoom -= 0.05;
    }
}

function update_bar() {
    //debug("update_bar");
    let main_browser = get_browser();
    let browser = { canGoBack: false, canGoForward: false };
    if (main_browser.contentWindow && main_browser.contentWindow.has_browser)
	browser = main_browser.contentWindow.get_browser();

    // For now
    main_browser = { canGoBack: false, canGoForward: false,
		     contentWindow: main_browser.contentWindow };

    let value = "true";
    if (main_browser.contentWindow && main_browser.contentWindow.zoomable)
	value = "false";
    let button = document.getElementById("zoom_in_text_button");
    button.setAttribute("disabled", value);
    button = document.getElementById("zoom_out_text_button");
    button.setAttribute("disabled", value);
    button = document.getElementById("zoom_in_button");
    button.setAttribute("disabled", value);
    button = document.getElementById("zoom_out_button");
    button.setAttribute("disabled", value);

    let item = document.getElementById("save_file_as_item");
    item.setAttribute("disabled", !(main_browser.contentWindow &&
				    main_browser.contentWindow.save_file));

    button = document.getElementById("forward-button");
    if (browser.canGoForward || main_browser.canGoForward)
	button.setAttribute("disabled", "false");
    else
	button.setAttribute("disabled", "true");

    button = document.getElementById("back-button");
    if (browser.canGoBack || main_browser.canGoBack)
	button.setAttribute("disabled", "false");
    else
	button.setAttribute("disabled", "true");

    //debug("update_bar OK");
}

function update_tab_title(win) {
    //debug("update_tab_title");
    let tabs = document.getElementById("tabs");
    let tab = tabs.selectedItem;
    let index = tabs.selectedIndex;
    let panel = null;
    if (win) {
	let pan = get_panel(win);
	panel = pan[0];
	index = pan[1];
	tab = tabs.childNodes.item(index);
    }
    if (tab) {
	let browser = get_browser(panel);
	let title = "[Untitled]";	// FIXME
	if (browser.contentDocument && browser.contentDocument.title)
	    title = browser.contentDocument.title;
	let shorttitle = title.substring(0, 25);
	if (title.length > 25)
	    shorttitle += "...";
	tab.setAttribute("label", shorttitle);
	tab.setAttribute("tooltiptext", title);
    }

    // Update tab buttons
    let box = document.getElementById("scrollbox");
    let sb = box.boxObject.QueryInterface(Components.interfaces.nsIScrollBoxObject);
    let xp = {};
    sb.getPosition(xp, {});
    let x = xp.value;
    let off = tabs.childNodes.item(0).boxObject.x;
    let last = tabs.childNodes.item(tabs.childNodes.length - 1);

    let button = document.getElementById("tab_right");
    if (x >= (last.boxObject.x + last.boxObject.width - box.boxObject.width))
	button.setAttribute("disabled", "true");
    else
	button.setAttribute("disabled", "false");

    button = document.getElementById("tab_left");
    if (x > 0)
	button.setAttribute("disabled", "false");
    else
	button.setAttribute("disabled", "true");
}

function tab_onselect() {
    update_bar();
}

function toggle_tabbar() {
    let tabbar = document.getElementById("tabbar");
    let val = !tabbar.hidden;
    tabbar.hidden = val;
    return val;
}

function show_findbar() {
    let bar = document.getElementById("findbar");
    bar.setAttribute("hidden", "false");
    let box = document.getElementById("find-box");
    box.focus();
}

function hide_findbar() {
    let bar = document.getElementById("findbar");
    bar.setAttribute("hidden", "true");
}

function on_find_keypress(event) {
    if (event.keyCode == 13)
	find(false);
}

var search_start;
var search_string;
var finder;

function find(backwards) {
    let browser = get_browser();
    if (browser.contentWindow.has_browser)
	browser = browser.contentWindow.get_browser();
    let doc = browser.contentDocument;
    let root = doc.documentElement;
    let fbox = document.getElementById("find-box");

    let word = fbox.value;
    if (!word)
	return;

    let range = doc.createRange();
    range.selectNodeContents(root);

    let start;
    // FIXME should search continue from the same place even if search string has changed?
    if (search_start && (search_start.startContainer.ownerDocument == doc) &&
	(search_string == word)) {
	start = search_start;
	if (backwards)
	    start.setEnd(start.startContainer, start.startOffset);
	else
	    start.setStart(start.endContainer, start.endOffset);
    } else {
	start = doc.createRange();
	if (backwards) {
	    start.setStart(range.endContainer, range.endOffset);
	    start.setEnd(range.endContainer, range.endOffset);
	} else {
	    start.setStart(range.startContainer, range.startOffset);
	    start.setEnd(range.startContainer, range.startOffset);
	}
    }

    let end = doc.createRange();
    if (backwards) {
	end.setStart(range.startContainer, range.startOffset);
	end.setEnd(range.startContainer, range.startOffset);
    } else {
	end.setStart(range.endContainer, range.endOffset);
	end.setEnd(range.endContainer, range.endOffset);
    }

    if (!finder)
	finder = Components.classes["@mozilla.org/embedcomp/rangefind;1"].
	    createInstance(Components.interfaces.nsIFind);
    let match_case = document.getElementById("match-case");
    finder.caseSensitive = match_case.checked;
    finder.findBackwards = backwards;

    let res = finder.Find(word, range, start, end);
    if (res) {
	debug("FIND " + res + " " + res.startOffset);
	let sel = browser.contentWindow.getSelection();
	sel.removeAllRanges();
	sel.addRange(res);
	// FIXME must refresh view so that the old selection disappears
	let sel2 = sel.QueryInterface(Components.interfaces.nsISelection2);
	sel2.scrollIntoView(0, false, -1, -1);
    }

    search_start = res;
    search_string = word;
}

function find3(backwards) {
    let browser = get_browser();
    if (browser.contentWindow.has_browser)
	browser = browser.contentWindow.get_browser();
    debug("title: " + browser.contentDocument.title);
    let wbfind = browser.webBrowserFind;
    let fbox = document.getElementById("find-box");
    wbfind.searchString = fbox.value;
    wbfind.findBackwards = backwards;
    wbfind.searchFrames = true;
    wbfind.entireWord = false;
    wbfind.matchCase = false;
    let i = wbfind.findNext();
    debug("FIND '" + wbfind.searchString + "' " + i);
}

function onclose() {
    let tabpanels = document.getElementById("tabpanels");
    for (let i = 0; i < tabpanels.childNodes.length; i++) {
	let tabpanel = tabpanels.childNodes.item(i);
	let browser = tabpanel.firstChild.firstChild.nextSibling;
	if (browser.contentWindow && browser.contentWindow.warn_unload)
	    browser.contentWindow.warn_unload();
    }
}

function onunload() {
    let main_control = Components.classes["@lucidor.org/mainControlService;1"].
	getService().wrappedJSObject;
    main_control.remove_tempfiles();
}

function quit() {
    window.close();
    let appStartup = Components.classes["@mozilla.org/toolkit/app-startup;1"].
        getService(Components.interfaces.nsIAppStartup);
    try {
	appStartup.quit(appStartup.eAttemptQuit);
    } catch (e) {
	appStartup.quit(appStartup.eForceQuit);
    }
}
