/*
 *  Lucidor - e-book reader
 *  Copyright © 2009-2010  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("resource://lucidor/modules/utils.jsm");

const XHTML_NS = "http://www.w3.org/1999/xhtml";
const XUL_NS = "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul";

var bcase;
var is_extension = false;

function init() {
    let appinfo = Components.classes["@mozilla.org/xre/app-info;1"].
	getService(Components.interfaces.nsIXULAppInfo);
    if (appinfo.ID == "Lucidor@lucidor.org")
	is_extension = false;
    else
	is_extension = true;
    if (is_extension) {
	get_livemarks();
	return update_sense();
    }

    let tree = document.getElementById("book_tree");
    tree.hidden = false;	// FIXME Why is it necessary to have it hidden?

    bcase = Components.classes["@lucidor.org/localCase;1"].
	createInstance().wrappedJSObject;
    bcase.init(null, tree);
    search_standard_feed();

    update_sense();
}

function get_livemarks() {
    let showLivemarks = false;
    let treechildren = document.getElementById("livemarkchildren");

    function add_item(title, uri) {
	let ti = document.createElementNS(XUL_NS, "treeitem");
	//ti.setAttribute("value", uri);
	let tr = document.createElementNS(XUL_NS, "treerow");
	ti.appendChild(tr);
	let tc = document.createElementNS(XUL_NS, "treecell");
	tc.setAttribute("label", title);
	tr.appendChild(tc);
	tc = document.createElementNS(XUL_NS, "treecell");
	tc.setAttribute("label", uri);
	tr.appendChild(tc);
	treechildren.appendChild(ti);
	showLivemarks = true;
    }

    let lmservice = Components.classes["@mozilla.org/browser/livemark-service;2"].getService(Components.interfaces.nsILivemarkService);
    let historyService = Components.classes["@mozilla.org/browser/nav-history-service;1"].getService(Components.interfaces.nsINavHistoryService);
    let options = historyService.getNewQueryOptions();
    let query = historyService.getNewQuery();
    //query.onlyBookmarked = true;

    let bookmarksService = Components.classes["@mozilla.org/browser/nav-bookmarks-service;1"].
        getService(Components.interfaces.nsINavBookmarksService);
    let folders = [bookmarksService.bookmarksMenuFolder,
    		   bookmarksService.toolbarFolder,
    		   bookmarksService.unfiledBookmarksFolder];

    for (let folder in loop(folders)) {
	query.setFolders([folder], 1);
	let res = historyService.executeQuery(query, options);
	let node = res.root;
	let done = false;
	let i = 0;
	while (node) {
	    i += 1;
	    if (i > 10000)
		break;
	    let is_folder = node.type == node.RESULT_TYPE_FOLDER;
	    if (!done && is_folder) {
		if (node.itemId >= 0 && lmservice.isLivemark(node.itemId)) {
		    add_item(node.title, lmservice.getFeedURI(node.itemId).spec);
		} else {
		    if (is_folder) {
			node.QueryInterface(Components.interfaces.nsINavHistoryContainerResultNode);
			node.containerOpen = true;
			if (node.childCount > 0) {
			    node = node.getChild(0);
			    continue;
			}
		    }
		}
	    }
	    if (is_folder) {
		node.QueryInterface(Components.interfaces.nsINavHistoryContainerResultNode);
		node.containerOpen = false;
	    }
	    if (node == res.root)
		break;
	    let par = node.parent.QueryInterface(Components.interfaces.nsINavHistoryContainerResultNode);
	    if (node.bookmarkIndex < (par.childCount - 1)) {
		node = par.getChild(node.bookmarkIndex + 1)
		done = false;
	    } else {
		node = par;
		done = true;
	    }
	}
    }

    //let tree = document.getElementById("livemarktree");
    //if (tree.view.rowCount > 0) {
    if (showLivemarks) {
	let box = document.getElementById("livebox");
	box.hidden = false;
    }
}

function search_standard_feed() {
    let stat = "SELECT title, location, id FROM books";
    stat += " WHERE type='feed'";

    let query_tree = document.getElementById("book_tree");
    if (query_tree.view)
	query_tree.view.selection.currentIndex = -1;

    let query = query_tree.ownerDocument.getElementById("query");
    while (query.hasChildNodes())
	query.removeChild(query.lastChild);

    let node = query_tree.ownerDocument.createTextNode(stat);
    query.appendChild(node);

    query_tree.builder.rebuild();
}

function update_sense() {
    if (is_extension)
	return update_sense_x();

    let list = document.getElementById("book_tree");
    let button = document.getElementById("remove_book");
    if (list.currentIndex >= 0)
	button.removeAttribute("disabled");
    else
	button.setAttribute("disabled", "true");

    button = document.getElementById("open_book");
    if (list.currentIndex >= 0)
	button.removeAttribute("disabled");
    else
	button.setAttribute("disabled", "true");

    list = document.getElementById("urllist");
    button = document.getElementById("remove_feed");
    if (list.selectedIndex >= 0)
	button.removeAttribute("disabled");
    else
	button.setAttribute("disabled", "true");

    button = document.getElementById("retrieve_title");
    if (list.getRowCount() > 0)
	button.removeAttribute("disabled");
    else
	button.setAttribute("disabled", "true");

    button = document.getElementById("add_book");
    if (list.getRowCount() > 0)
	button.removeAttribute("disabled");
    else
	button.setAttribute("disabled", "true");

    button = document.getElementById("open_new_book");
    if (list.getRowCount() > 0)
	button.removeAttribute("disabled");
    else
	button.setAttribute("disabled", "true");
}

function update_sense_x() {
    let list = document.getElementById("urllist");
    let button = document.getElementById("remove_feed");
    if (list.selectedIndex >= 0)
	button.removeAttribute("disabled");
    else
	button.setAttribute("disabled", "true");

    button = document.getElementById("retrieve_title");
    if (list.getRowCount() > 0)
	button.removeAttribute("disabled");
    else
	button.setAttribute("disabled", "true");

    button = document.getElementById("open_new_book");
    if (list.getRowCount() > 0)
	button.removeAttribute("disabled");
    else
	button.setAttribute("disabled", "true");
}

function select_book() {
    let tree = document.getElementById("book_tree");
    if (tree.currentIndex < 0)
	return;
    let col = tree.columns.getNamedColumn("location-col");
    let location = tree.view.getCellText(tree.currentIndex, col);

    let tb = document.getElementById("title_textbox");
    col = tree.columns.getNamedColumn("title-col");
    let title = tree.view.getCellText(tree.currentIndex, col);
    tb.value = title;

    debug("location: " + location);
    try {
    let list = document.getElementById("urllist");
    while (list.getRowCount() > 0)
	list.removeItemAt(0);

    let JSON = Components.classes["@mozilla.org/dom/json;1"].
	createInstance(Components.interfaces.nsIJSON);
    let feed = JSON.decode(location);

    for (let item in loop(feed.items))
	if (item.title)
	    list.appendItem(item.title, item.uri);
	else
	    list.appendItem(item.uri, item.uri);

    //tb = document.getElementById("feed_url_textbox");
    //tb.value = location;

    update_sense();
    } catch (e) {
	debug("ER: " + e);
    }
}

function remove_book() {
    // FIXME show prompt?
    let tree = document.getElementById("book_tree");
    if (tree.currentIndex < 0)
	return;
    let col = tree.columns.getLastColumn();
    let id = tree.view.getCellText(tree.currentIndex, col);

    bcase.remove_book(id);
    search_standard_feed();

    update_sense();
}

function retrieve_feed_info() {
    // FIXME
}

function add_url() {
    let tb = document.getElementById("feed_url_textbox");
    let location = tb.value;

    let list = document.getElementById("urllist");
    let ioService = Components.classes["@mozilla.org/network/io-service;1"].
	getService(Components.interfaces.nsIIOService);
    try {
	let nsuri = ioService.newURI(location, "UTF-8", null).
	    QueryInterface(Components.interfaces.nsIURL);
    } catch (e) {
	//show_error();
	debug("Incorrect URL");
	return;
    }
    list.appendItem(location, location);

    tb.value = "";
    update_sense();
}

function add_livemark() {
    let list = document.getElementById("urllist");
    let ioService = Components.classes["@mozilla.org/network/io-service;1"].
	getService(Components.interfaces.nsIIOService);

    let tree = document.getElementById("livemarktree");
    let start = new Object();
    let end = new Object();
    let numRanges = tree.view.selection.getRangeCount();
    let col = tree.columns.getLastColumn();

    for (let t = 0; t < numRanges; t++) {
	tree.view.selection.getRangeAt(t, start, end);
	for (let v = start.value; v <= end.value; v++) {
	    let location = tree.view.getCellText(v, col);
	    try {
		let nsuri = ioService.newURI(location, "UTF-8", null).
		    QueryInterface(Components.interfaces.nsIURL);
	    } catch (e) {
		//show_error();
		debug("Incorrect URL");
		continue;
	    }
	    list.appendItem(location, location);
	}
    }

    update_sense();
}

function remove_url() {
    let list = document.getElementById("urllist");
    if (list.selectedIndex >= 0)
	list.removeItemAt(list.selectedIndex);

    update_sense();
}

function add_book() {
    try {
    let list = document.getElementById("urllist");
    let rows = list.getRowCount();
    let location;
    if (rows == 0) {
	// FIXME show error
	return;
    } else {
	let ls = new Array(rows);
	for (let i = 0; i < rows; i++)
	    ls[i] = { uri: list.getItemAtIndex(i).value };
	let JSON = Components.classes["@mozilla.org/dom/json;1"].
	    createInstance(Components.interfaces.nsIJSON);
	let feed = { items: ls, version: 1 };
	location = JSON.encode(feed);
    }

    let tb = document.getElementById("title_textbox");
    let data = {
	title: tb.value,
	publisher: "Lucidor Generator",
	location: location,
	type: "feed"
    };
    bcase.add_book_data(data);
    search_standard_feed();

    clear_fields();

    update_sense();
    } catch (e) {
	debug("ERR: " + e);
    }
}

function clear_fields() {
    let list = document.getElementById("urllist");
    while (list.getRowCount() > 0)
	list.removeItemAt(0);

    let tb = document.getElementById("title_textbox");
    tb.value = "";
    tb = document.getElementById("feed_url_textbox");
    tb.value = "";
}

function generate_book() {
    debug("generate_book");

    let tree = document.getElementById("book_tree");
    if (tree.currentIndex < 0)
	return;
    let col = tree.columns.getNamedColumn("title-col");
    let title = tree.view.getCellText(tree.currentIndex, col);
    col = tree.columns.getNamedColumn("location-col");
    let loc = tree.view.getCellText(tree.currentIndex, col);

    let JSON = Components.classes["@mozilla.org/dom/json;1"].
	createInstance(Components.interfaces.nsIJSON);
    let feed = JSON.decode(loc);

    generate2(feed, title);
}

function generate_new_book() {
    debug("generate_new_book");
    let list = document.getElementById("urllist");
    let rows = list.getRowCount();
    let ls = new Array(rows);
    for (let i = 0; i < rows; i++)
	ls[i] = { uri: list.getItemAtIndex(i).value };
    let title;
    let tb = document.getElementById("title_textbox");
    if (tb.value)
	title = tb.value;
    let feed = { items: ls, version: 1 };
    generate2(feed, title);
}

function generate2(feed, title) {
    debug("generate2: " + feed.items + " : " + title);
    let location = "x-lucidor-convert://feed/?";

    for (let item in loop(feed.items))
	location += "uri=" + escape(item.uri) + "&";

    if (title)
	location += "title=" + escape(title);

    let uri = "chrome://lucidor/content/viewepub.xul?uri=" +
	escape(location);

    debug("location: " + location);
    if (is_extension) {
	let mainWindow = window.QueryInterface(Components.interfaces.nsIInterfaceRequestor)
            .getInterface(Components.interfaces.nsIWebNavigation)
            .QueryInterface(Components.interfaces.nsIDocShellTreeItem)
            .rootTreeItem
            .QueryInterface(Components.interfaces.nsIInterfaceRequestor)
            .getInterface(Components.interfaces.nsIDOMWindow);
	mainWindow.gBrowser.addTab(uri);
    } else {
	let main_control = Components.classes["@lucidor.org/mainControlService;1"].
	    getService().wrappedJSObject;
	main_control.add_tab(uri);
    }
}
