/*
 *  Lucidor - e-book reader
 *  Copyright © 2009-2010  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("resource://lucidor/modules/utils.jsm");

const XHTML_NS = "http://www.w3.org/1999/xhtml";
const XUL_NS = "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul";

var opdsparser = null;
var simple = true;
var current_feed = null;
var current_book = null;
var urihistory = [];
var opensearchers = {};
var current_feed_box = null;
var bookmark_box;

function init() {
    let frame = document.getElementById("frame");
    let body = frame.contentDocument.getElementById("body");
    body.setAttribute("class", "emptygrey");

    window.addEventListener("message", receiveMessage, false);
    opdsparser = Components.classes["@lucidor.org/opdsParser;1"].
        createInstance().wrappedJSObject;
    init_bookmarks();
    show_bookmarks();
}

function switch_search_deck() {
    let deck = document.getElementById("search-deck");
    let button = document.getElementById("search-deck-button");
    if (button.checked)
	deck.selectedIndex = 1;
    else
	deck.selectedIndex = 0;
}

function go_personal() {
    browse_feed("my:");
}

function go_bookmark(event) {
    let location = event.target.value;
    browse_feed(location);
}

function on_go_keypress(event) {
    if (event.keyCode == 13)
	go();
}

function go() {
    let elem = document.getElementById("location_text");
    let location = elem.value;
    let ioService = Components.classes["@mozilla.org/network/io-service;1"].
	getService(Components.interfaces.nsIIOService);
    let uri;
    try {
	try {
	    uri = ioService.newURI(location, "UTF-8", null);
	} catch (e) {
	    location = "http://" + location;
	    uri = ioService.newURI(location, "UTF-8", null);
	}
    } catch (e) {
	show_searchbar("invalid");
	return;
    }

    browse_feed(location);
}

function on_search_keypress(event) {
    if (event.keyCode == 13)
	search();
}

function search() {
    debug("Search");
    let text = document.getElementById("search_text").value;
    if (!text)
	return;

    if (!(current_feed.opensearch in opensearchers)) {
	//show_warning("Failed");
	return;
    }

    //let bar = document.getElementById("opensearch_bar");
    //bar.hidden = true;

    let current_searcher = opensearchers[current_feed.opensearch];

    let searchTerms = "";
    let prefix = "";
    /*
    let sfilter = document.getElementById("search_filter");
    if (filter == "author")
	prefix = "author:";
    else if (filter == "title")
	prefix = "title:";
    */

    if (current_searcher.syntax["compound"]) {
	for (let phrase in loop(text.split(" ")))
	    searchTerms += prefix + phrase + " ";
	searchTerms = searchTerms.substring(0, searchTerms.length - 1);
    } else {
	searchTerms = prefix + text;
    }
    let uri = current_searcher.get_search_uri(searchTerms);
    browse_feed(uri);
}

function show_searchbar(msg) {
    let text = "";
    if (msg == "searching") {
	text = "Searching...";
    } else if (msg == "failed") {
	text = "Search failed.";
    } else if (msg == "failed unknown") {
	text = "Search failed.";
    } else if (msg == "zero") {
	text = "Search returned no results.";
    } else if (msg == "invalid") {
	text = "Invalid location.";
    }
    let bar = document.getElementById("opensearch_bar");
    bar.hidden = false;
    let label = document.getElementById("opensearch_label");
    label.value = text;

    let button = document.getElementById("prev_results");
    button.disabled = true;
    button = document.getElementById("next_results");
    button.disabled = true;
}

var current_sid = 0;
function new_sid() {
    return ++current_sid;
}

function browse_feed(uri) {
    debug("browse_feed: " + uri);

    unzoom_book();
    show_searchbar("searching");

    if (uri.substring(0, 3) == "my:") {
	let bookcase = Components.classes["@lucidor.org/myBooksService;1"].
            getService().wrappedJSObject;
	try {
	    bookcase.open();
	} catch (e) {
	    throw e;
	}
	bookcase.get_feed(uri, browse_result, new_sid());
    } else
	opdsparser.get_feed(uri, browse_result, new_sid());
}

function browse_result(result, sid) {
    debug("browse_result: " + sid);
    if (sid != current_sid)
	return;

    if (!result) {
	show_searchbar("failed");
	return;
    }

    current_feed = result;
    update_location(result.uri);
    update_opensearch();
    if (result.opensearch)
	add_opensearch(result.opensearch);

    let label = document.getElementById("opensearch_label");
    label.value = "";
    if (result.totalResults != -1) {
	let first = result.startIndex;
	let total = result.totalResults;
	let last = -1;
	if (first == -1)	// FIXME necessary for feedbooks.com
	    first = 1;
	if (first != -1 && result.itemsPerPage != -1)
	    last = first + result.itemsPerPage - 1;
	if (last > total)
	    last = total;
	let text = "";
	if (first > -1 && last > -1 && total > -1) {
	    if (first == last)
		text = "" + first + " (" + total + ")";
	    else
		text = "" + first + ".." + last + " (" + total + ")";
	} else if (first > -1 && last > -1) {
	    text = "" + first + ".." + last;
	}
	//let label = document.getElementById("opensearch_label");
	label.value = text;
    }

    let button = document.getElementById("start_button");
    if (result.start)
	button.disabled = false;
    else
	button.disabled = true;
    button = document.getElementById("prev_results");
    if (result.previous)
	button.disabled = false;
    else
	button.disabled = true;
    button = document.getElementById("next_results");
    if (result.next)
	button.disabled = false;
    else
	button.disabled = true;

    let mode = 0;
    if (!simple)
	mode = 1;

    let newbox = document.createElementNS(XUL_NS, "vbox");
    newbox.setAttribute("flex", "1");
    current_feed_box = newbox;

    // Content box
    let box = document.createElementNS(XUL_NS, "vbox");
    box.setAttribute("flex", "1");
    newbox.appendChild(box);

    for (entry in loop(result.entries))
	add_entry(box, entry, mode);

    // Link box
    box = document.createElementNS(XUL_NS, "vbox");
    newbox.appendChild(box);

    for (let link in loop(result.links))
	if (link.type == "toplink")
	    add_top_link(box, link);

    unzoom_book();
}

function zoom_book(idx) {
    current_book = idx;
    debug("IDX: " + idx + " ENTS: " + current_feed.entries + " L: " + current_feed.entries.length);

    let newbox = document.createElementNS(XUL_NS, "vbox");
    newbox.setAttribute("flex", "1");
    add_entry(newbox, current_feed.entries[idx], 2);

    let button = document.getElementById("prev_book");
    if (idx > 0)
	button.disabled = false;
    else
	button.disabled = true;
    button = document.getElementById("next_book");
    if (current_feed.entries.length > (idx + 1))
	button.disabled = false;
    else
	button.disabled = true;

    let label = document.getElementById("book_label");
    let first = 1;
    let text;
    if (current_feed.startIndex >= 1)
	first = current_feed.startIndex
    if (current_feed.entries.length == 1)
	text = "" + (first + idx) + " (" + first + ")";
    else
	text = "" + (first + idx) + " (" + first + ".." + (first + current_feed.entries.length - 1) + ")";
    label.value = text;

    switch_content_deck(1, newbox);
}

function switch_content_deck(i, box) {
    debug("switch_content_deck: " + i + " " + box);
    let deck = document.getElementById("content-deck");
    deck.selectedIndex = i;

    let frame = document.getElementById("frame");
    let body = frame.contentDocument.getElementById("body");
    while (body.hasChildNodes())
	body.removeChild(body.firstChild);
    if (box)
	body.appendChild(box);
    frame.contentWindow.scrollTo(0, 0);
}

function unzoom_book() {
    switch_content_deck(0, current_feed_box);
}

function init_bookmarks() {
    let bookmarks = [{ title: "Ebooks libres et gratuits",
		       href: "http://www.ebooksgratuits.com/opds/",
		       icon: null },
		     { title: "Feedbooks",
		       href: "http://feedbooks.com/catalog.atom",
		       icon: "http://www.feedbooks.com/favicon.ico" },
		     { title: "Internet Archive Catalog",
		       href: "http://bookserver.archive.org/catalog/",
		       icon: null },
		     /* { title: "Internet Archive Aggregator",
		       href: "http://bookserver.archive.org/aggregator/",
		       icon: null }, */
		     { title: "Project Gutenberg",
		       href: "http://lucidor.org/opds/pg.php",
		       icon: "http://lucidor.org/favicon.ico" },
		     /* { title: "Lucidor",
		       href: "http://lucidor.org/opds/lucidor.php",
		       icon: "http://lucidor.org/favicon.ico" } */
		     /* { title: "Lucicat",
		       href: "http://localhost/opds/lucicat.php",
		       icon: null }, */
		    ];

    bookmark_box = document.createElementNS(XUL_NS, "vbox");
    bookmark_box.setAttribute("style", "margin-bottom: 5mm;");

    for (let bookmark in loop(bookmarks))
	create_bookmark_item(bookmark_box, bookmark.title,
			     bookmark.href, bookmark.icon);
}

function show_bookmarks() {
    let newbox = document.createElementNS(XUL_NS, "vbox");
    newbox.setAttribute("flex", "1");

    newbox.appendChild(bookmark_box);

    let box = document.createElementNS(XUL_NS, "vbox");
    newbox.appendChild(box);

    for (let i = 0; i < 10; i++) {
	let j = urihistory.length - i - 1;
	if (j < 0)
	    break;
	let ent = urihistory[j];
	create_bookmark_item(box, ent, ent, null);
    }

    switch_content_deck(2, newbox);
}

// FIXME
function xmlescape(uri) {
    return uri.replace("'", "\'", "g");
}

function create_bookmark_item(box, title, uri, icon) {
    let frame = document.createElementNS(XUL_NS, "vbox");
    frame.setAttribute("class", "opdsframe");
    box.appendChild(frame);

    let item = document.createElementNS(XHTML_NS, "div");
    let item2 = document.createElementNS(XHTML_NS, "a");
    item2.setAttribute("class", "link");
    item2.setAttribute("href", "#");
    item2.setAttribute("onclick", "browse_f('" + xmlescape(uri) + "');");
    item2.setAttribute("tooltiptext", uri);

    if (icon) {
	let img = document.createElementNS(XHTML_NS, "object");
	img.setAttribute("data", icon);
	item2.appendChild(img);
    }

    let text = document.createTextNode(title);
    item2.appendChild(text);
    item.appendChild(item2);
    frame.appendChild(item);
}

function previous_book() {
    zoom_book(current_book - 1);
}

function next_book() {
    zoom_book(current_book + 1);
}

function previous_results() {
    browse_feed(current_feed.previous);
}

function next_results() {
    browse_feed(current_feed.next);
}

function add_entry(box, link, mode) {
    debug("add_entry " + link.title + " " + link.type + " mode=" + mode);
    for (a in link)
	debug("A:" + a + " " + link[a]);

    if (link.type == "folder") {
	create_folder_box(link, box);
    } else if (link.type == "epub" || link.type == "link") {
	create_book_box(link, box, mode);
    }
    debug("Finished");
}

function create_folder_box(bookinfo, book_box) {
    let infobox = document.createElementNS(XUL_NS, "vbox");
    book_box.appendChild(infobox);

    let info_frame = document.createElementNS(XUL_NS, "vbox");
    info_frame.setAttribute("class", "opdsframe");
    infobox.appendChild(info_frame);

    let item = document.createElementNS(XHTML_NS, "div");
    let item2 = document.createElementNS(XHTML_NS, "a");
    item2.setAttribute("class", "link");
    item2.setAttribute("href", "#");
    item2.setAttribute("onclick", "browse_f('" +
		       xmlescape(bookinfo.location) + "');");
    item2.setAttribute("tooltiptext", bookinfo.location);
    let text = document.createTextNode(bookinfo.title);
    item2.appendChild(text);
    item.appendChild(item2);
    info_frame.appendChild(item);
}

function create_book_box(bookinfo, top_box, mode) {
    function add_line(frame, label, value) {
	let div = document.createElementNS(XHTML_NS, "div");
	//div = document.createElementNS(XUL_NS, "description");
	div.setAttribute("class", "opdsitem");
	if (label) {
	    let span = document.createElementNS(XHTML_NS, "span");
	    span.setAttribute("class", "opdsheader");
	    div.appendChild(span);
	    let text = document.createTextNode(label + ": ");
	    span.appendChild(text);
	}
	let text = document.createTextNode(value);
	div.appendChild(text);
	frame.appendChild(div);
    }

    let book_box = document.createElementNS(XUL_NS, "hbox");
    top_box.appendChild(book_box);

    // Cover preview
    if ((mode == 1 || mode == 2) && bookinfo.thumbnail_link) {
	let imgbox = document.createElementNS(XUL_NS, "vbox");
	imgbox.setAttribute("class", "opdscover");
	//let img = get_book_preview(bookinfo);
	let img = document.createElementNS(XUL_NS, "image");
	img.setAttribute("src", bookinfo.thumbnail_link);
	if (img) {
	    imgbox.appendChild(img);
	    book_box.appendChild(imgbox);
	}
    }

    let item;

    // Info
    let infobox = document.createElementNS(XUL_NS, "vbox");
    infobox.setAttribute("flex", "1");
    book_box.appendChild(infobox);

    let info_frame = document.createElementNS(XUL_NS, "vbox");
    info_frame.setAttribute("class", "opdsframe");
    infobox.appendChild(info_frame);

    if (mode == 0 || mode == 1) {
	item = document.createElementNS(XHTML_NS, "div");
	item.setAttribute("class", "opdsitem");
	let item2 = document.createElementNS(XHTML_NS, "a");
	item2.setAttribute("class", "link");
	item2.setAttribute("href", "#");
	item2.setAttribute("onclick", "zoom_b(" + bookinfo.idx + ");");
	let text = document.createTextNode(bookinfo.title);
	item2.appendChild(text);
	item.appendChild(item2);
	info_frame.appendChild(item);
    } else {
	add_line(info_frame, null, bookinfo.title);
    }

    if (bookinfo.author) {
	item = document.createElementNS(XHTML_NS, "div");
	//item = document.createElementNS(XUL_NS, "description");
	item.setAttribute("class", "author");
	let text = document.createTextNode(bookinfo.author);
	item.appendChild(text);
	info_frame.appendChild(item);
    }

    if ((mode == 1 || mode == 2) && bookinfo.publisher)
	add_line(info_frame, "Publisher", bookinfo.publisher);

    if ((mode == 1 || mode == 2) && bookinfo.language)
	add_line(info_frame, "Language", bookinfo.language);

    if ((mode == 1 || mode == 2) && bookinfo.type != "link") {
	item = document.createElementNS(XHTML_NS, "div");
	item.setAttribute("class", "makeroom");
	let item2 = document.createElementNS(XHTML_NS, "a");
	item2.setAttribute("class", "link");
	item2.setAttribute("href", "#");
	item2.setAttribute("onclick", "open_b('" +
			   xmlescape(bookinfo.location) + "');");
	item2.setAttribute("tooltiptext", bookinfo.location);
	let text = document.createTextNode("Open book");
	item2.appendChild(text);
	item.appendChild(item2);
	info_frame.appendChild(item);
    }

    // Links
    if ((mode == 1 || mode == 2) && bookinfo.links) {
	debug("LINKS!!");
	for (link in loop(bookinfo.links)) {
	    item = document.createElementNS(XHTML_NS, "div");
	    item.setAttribute("class", "makeroom");
	    let item2 = document.createElementNS(XHTML_NS, "a");
	    item2.setAttribute("class", "link");
	    item2.setAttribute("href", "#");
	    let title = link.title;
	    if (!title)
		title = link.href;
	    if (link.type == "application/atom+xml") {
		item2.setAttribute("onclick", "browse_f('" +
				   xmlescape(link.href) + "');");
	    } else if (link.type == "application/epub+zip") {
		item2.setAttribute("onclick", "open_b('" +
				   xmlescape(link.href) + "');");
		item2.setAttribute("tooltiptext", link.href);
	    } else {
		item2.setAttribute("onclick", "open_w('" +
				   xmlescape(link.href) + "');");
		item2.setAttribute("tooltiptext", link.href);
	    }
	    let text = document.createTextNode(title);
	    item2.appendChild(text);
	    item.appendChild(item2);
	    info_frame.appendChild(item);
	}
    }

    if (mode == 2 && bookinfo.mybooksid != undefined) {
	// Add remove link
	item = document.createElementNS(XHTML_NS, "div");
	item.setAttribute("class", "bigroom");
	let item2 = document.createElementNS(XHTML_NS, "a");
	item2.setAttribute("class", "removelink");
	item2.setAttribute("href", "#");
	item2.setAttribute("onclick", "remove_b(" +
			   bookinfo.mybooksid + ");");
	let text = document.createTextNode("Remove book");
	item2.appendChild(text);
	item.appendChild(item2);
	info_frame.appendChild(item);
    }

    /*
    item = document.createElementNS(XUL_NS, "spacer");
    item.setAttribute("flex", 100);
    book_box.appendChild(item);
    */
}

function add_top_link(area, link) {
    let info_frame = document.createElementNS(XUL_NS, "vbox");
    info_frame.setAttribute("class", "toplinkframe");
    info_frame.setAttribute("flex", "1");
    area.appendChild(info_frame);

    let box = document.createElementNS(XHTML_NS, "div");
    info_frame.appendChild(box);

    create_link_box(link, box);
}

function create_link_box(link, top_box) {
    let item = document.createElementNS(XHTML_NS, "a");
    item.setAttribute("class", "link toplink");
    item.setAttribute("href", "#");
    item.setAttribute("onclick", "browse_f('" +
		      xmlescape(link.href) + "');");
    let text = document.createTextNode(link.title);
    item.appendChild(text);
    top_box.appendChild(item);
}

function update_opensearch() {
    let item = document.getElementById("search_text");
    if (current_feed.opensearch in opensearchers) {
	//if (!item.disabled)
	//    show_warning("Search option found", null, true);
	item.disabled = false;
    } else {
	item.disabled = true;
    }
}

function add_opensearch(uri) {
    debug("add_opensearch: " + uri);
    if (!(uri in opensearchers)) {
	if (uri == "my:os") {
	    let bookcase = Components.classes["@lucidor.org/myBooksService;1"].
		getService().wrappedJSObject;
	    try {
		bookcase.open();
		bookcase.get_opensearch(uri, new_opensearch);
	    } catch (e) {
	    }
	} else {
	    opdsparser.get_opensearch(uri, new_opensearch);
	}
    }
}

function new_opensearch(se, uri) {
    debug("new_opensearch: " + se + " : " + uri);
    if (!se)
	return;

    opensearchers[uri] = se;
    let title = "Search";
    if (se.shortname)
	title = se.shortname;
    update_opensearch();
}

function open_book(uri) {
    open_weblink(uri);
}

function open_weblink(uri) {
    let mainWindow = window.QueryInterface(Components.interfaces.nsIInterfaceRequestor)
        .getInterface(Components.interfaces.nsIWebNavigation)
        .QueryInterface(Components.interfaces.nsIDocShellTreeItem)
        .rootTreeItem
        .QueryInterface(Components.interfaces.nsIInterfaceRequestor)
        .getInterface(Components.interfaces.nsIDOMWindow);
    mainWindow.Browser.addTab(uri, true);
}

function update_location(uri) {
    let elem = document.getElementById("location_text");
    elem.value = uri;

    if (uri != urihistory[urihistory.length - 1]) {
	if (urihistory.length > 30)
	    urihistory.shift();
	urihistory.push(uri);
	update_back_button();
    }
}

function update_back_button() {
    let button = document.getElementById("back_button");
    if (urihistory.length > 1)
	button.disabled = false;
    else
	button.disabled = true;
}

function go_start() {
    if (current_feed.start)
	browse_feed(current_feed.start);
}

function go_back() {
    urihistory.pop();
    let uri = urihistory.pop();
    let elem = document.getElementById("location_text");
    elem.value = uri;
    update_back_button();
    browse_feed(uri);
}

function remove_book(bookid) {
    let bookinfo;
    let bookcase = Components.classes["@lucidor.org/myBooksService;1"].
        getService().wrappedJSObject;
    try {
	bookcase.open();
	bookinfo = bookcase.get_book_info(bookid);
	bookcase.remove_book(bookid);
    } catch (e) {
	throw e;
    }

    if (bookinfo.location.indexOf("file://") == 0) {
	let file = Components.classes["@mozilla.org/network/protocol;1?name=file"].
            getService(Components.interfaces.nsIFileProtocolHandler).
            getFileFromURLSpec(bookinfo.location);
        try {
            file.remove(false);
        } catch (e) {
            // FIXME Should this be reported to the user?
        }
    }

    browse_feed(current_feed.uri);
}

function receiveMessage(event) {
    if (event.origin !== "chrome://lucidor")
	return;
    let i = event.data.indexOf(":");
    let cmd = event.data.substring(0, i);
    let arg = event.data.substring(i + 1);
    if (cmd == "browse_f")
	browse_feed(arg);
    else if (cmd == "zoom_b")
	zoom_book(Number(arg));
    else if (cmd == "open_b")
	open_book(arg);
    else if (cmd == "open_w")
	open_web(arg);
    else if (cmd == "remove_b")
	remove_book(Number(arg));
}
