/*
 *  Lucidor - e-book reader
 *  Copyright © 2008-2011  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("resource://lucidor/modules/utils.jsm");

const XHTML_NS = "http://www.w3.org/1999/xhtml";
const XUL_NS = "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul";
const DTB_NS = "http://www.daisy.org/z3986/2005/dtbook/";
const OPS_NS = "http://www.idpf.org/2007/ops";

var is_extension = false;
var zoomable = true;
var has_browser = true;
var browser;
var book = null;
var bookid;
var bookcase_id = -1;
var spine;
var partinfo;
var old_part = -1;
var view_mode = "fixed";
var view_border = false;
var column_adjustment = "center";
var ppdm;	// Pixels Per Decimeter (x-axis)
var shown_errors = false;
var shown_encryption_warning = false;
var scroll_ratio = 0;
var scroll_to_search = null;
var extra_stylesheet = null;
var scroll_time = 0;
var scroll_count = 0;
var language_mapper;
var main_control;
var xpointer_service;

function fixup_dom(doc) {
    /* Set root elements; glue is necessary for columns */
    let oldroot = doc.documentElement;
    if (oldroot.localName != "x-ordbrand-root") {
	/* should not use XHTML_NS here, but needed to get .style */
	let root = doc.createElementNS(XHTML_NS, "x-ordbrand-root");
	let glue = doc.createElementNS(XHTML_NS, "x-ordbrand-glue");
	doc.replaceChild(root, oldroot);
	root.appendChild(glue);
	glue.appendChild(oldroot);

	if (extra_stylesheet) {
	    let pi = doc.createProcessingInstruction("xml-stylesheet",
						     extra_stylesheet);
	    doc.insertBefore(pi, doc.firstChild);
	    extra_stylesheet = null;
	} else {
	    /* If <body> has background-color set, then <html>/<dtbook> should
	       inherit the same value unless they set a different value */
	    /* FIXME does not yet work correctly if <html>/<dtbook> has
	       background-color set to a different value (rare case) */
	    let doc_elem = doc.documentElement;
	    let du = Components.classes["@mozilla.org/inspector/dom-utils;1"].
		createInstance(Components.interfaces.inIDOMUtils);
	    let background;
	    let body;
	    if (oldroot.localName == "dtbook")
		body = doc_elem.getElementsByTagNameNS(DTB_NS, "book").item(0);
	    else
		body = doc_elem.getElementsByTagNameNS(XHTML_NS, "body").item(0);
	    if (!body)
		throw "Error FIXME";
	    let rules;
	    try {
		rules = du.getCSSStyleRules(body);
	    } catch (e) {
		debug("error: " + e);
		throw e;
	    }
	    for (let i = 0; i < rules.Count(); i += 1) {
		let el = rules.GetElementAt(i);
		for (let j = 0; j < el.style.length; j += 1) {
		    let item = el.style.item(j);
		    if (item == "background-color") {
			let val = el.style.getPropertyValue("background-color");
			if (val != "rgb(255, 255, 255)") {
			    background = val;
			    break;
			}
		    }
		}
		if (background)
		    break;
	    }
	    if (background)
		body.parentNode.style.backgroundColor = background;
	}

	// FIXME handle opf:switch element
	let elems = doc.getElementsByTagNameNS(OPS_NS, "switch");
	for (let elem in loop(elems)) {
	    for (let child in loop(elem.childNodes)) {
		if (child.nodeType == doc.ELEMENT_TYPE &&
		    child.namespaceURI == OPS_NS) {
		    if (child.localName == "case")
			;
		    else if (child.localName == "default")
			;
		}
	    }
	}

	// add event handlers for links
	elems = doc.getElementsByTagNameNS(XHTML_NS, "a");
	for (let elem in loop(elems)) {
	    elem.addEventListener("mouseover", link_handler, false);
	    elem.addEventListener("mouseout", link_out_handler, false);
	}

	/*
	elems = doc.getElementsByTagNameNS(DTB_NS, "a");
	for (let elem in loop(elems))
	    elem.addEventListener("mouseover", link_handler, false);
	*/
	/*
	elems = doc.getElementsByTagNameNS(DTB_NS, "noteref");
	for (let elem in loop(elems))
	    elem.addEventListener("mouseover", noteref_handler, false);
	*/
    }
}

var progressListener = {
    onLocationChange: function(webProgress, request, location) {
	debug("onLocationChange: " + location.spec);

	let part = find_part(location.spec);
	if (part < 0) {
	    // FIXME is this possible with standards conforming documents?
	    debug("FIXME Error: unknown spine location");
	    return;
	}
	debug("PART: " + part);

	if (part != old_part) {
	    update_buttons(part);
	    update_navigation_bar();

	    for (let i = 0; i < 2 && i < partinfo[part].length; i++)
		partinfo[part][i].setAttribute("class", "first-current-navitem");
	    for (let i = 2; i < partinfo[part].length; i++)
		partinfo[part][i].setAttribute("class", "current-navitem");

	    if (old_part >= 0)
		for (let link in loop(partinfo[old_part]))
		    link.removeAttribute("class");
	    old_part = part;
	}
    },

    onProgressChange: function(webProgress, request, curSelfProgress,
			       maxSelfProgress, curTotalProgress,
			       maxTotalProgress) {
    },

    onSecurityChange: function(webProgress, request, state) {
    },

    onStateChange: function(webProgress, request, stateFlags, status) {
	let doc = webProgress.DOMWindow.document;
	if ((stateFlags & 16) && (stateFlags & 131072)) {
	    debug("on_STATE_change: " + request.name);

	    // This tests whether the loaded document is a top-level document
	    // or included as an object (or subframe?)
	    /*
	    if (webProgress.DOMWindow != webProgress.DOMWindow.parent)
		return;
	    */
	    if (webProgress.DOMWindow != browser.contentWindow)
		return;

	    if (webProgress.DOMWindow.document.documentElement.tagName ==
		"parsererror") {
		book.add_error("ill-formed document");
		// FIXME göra något mer?
	    } else {
		fixup_dom(doc);

		set_view_mode(null);
		if (view_mode == "flow") {
		    let button = document.getElementById("flow_button");
		    button.checked = true;
		} else if (view_mode == "columns") {
		    let button = document.getElementById("columns_button");
		    button.checked = true;
		}

		debug("scroll_ratio: " + scroll_ratio);
		if (scroll_ratio > 0) {
		    let wind = webProgress.DOMWindow;
		    if (view_mode == "columns")
			wind.scrollTo(wind.scrollMaxX * scroll_ratio,
				      wind.scrollY);
		    else
			wind.scrollTo(wind.scrollX,
				      wind.scrollMaxY * scroll_ratio);
		    scroll_ratio = 0;
		}

		if (scroll_to_search) {
		    scroll_to_xpointer(scroll_to_search, doc,
				       webProgress.DOMWindow);
		    scroll_to_search = null;
		}

		update_last_date_and_position();
		update_bar();	// FIXME should not be necessary
	    }
	}

	if (book && book.uses_encryption && !shown_encryption_warning) {
	    let strbundle = document.getElementById("strings");
	    let text = strbundle.getString("uses_encryption");
	    show_warning(text, null, false);
	    shown_encryption_warning = true;
	}

	if (book && book.has_errors && !shown_errors) {
	    let strbundle = document.getElementById("strings");
	    let text = strbundle.getString("has_errors");
	    let details = strbundle.getString("show_details");

	    if (book.errors.length > 0)
		show_warning(text, details, true);
	    else
		show_warning(text, null, true);
	    shown_errors = true;

	    let item = document.getElementById("details_item");
	    item.setAttribute("disabled", "false");
	}
    },

    onStatusChange: function(webProgress, request, status, message) {
    },

    QueryInterface: function(aIID) {
	if (aIID.equals(Components.interfaces.nsIWebProgressListener) ||
	    aIID.equals(Components.interfaces.nsISupportsWeakReference))
	    return this;
	throw Components.results.NS_NOINTERFACE;
    }
};

function scroll_to_xpointer(xptr, doc, window) {
    debug("scroll_to_xpointer: '" + xptr + "'");

    /*
    if (xptr.indexOf("string-range(/") >= 0) {
	xptr = xptr.replace("string-range(/",
			"string-range(/x-ordbrand-root[1]/x-ordbrand-glue[1]/");
    }
    */

    /*
      let xpeval = Components.classes["@mozilla.org/xmlextras/xpointerevaluator;1"].createInstance(Components.interfaces.nsIXPointerEvaluator);
      let xpres = xpeval.evaluate(doc, xptr);
    */

    let dr;
    try {
	dr = xpointer_service.parseXPointerToRange(xptr, doc);
	debug(dr + ":" + dr.startOffset);
    } catch (e) {
	debug("Couldn't resolve: " + e);
    }

    let sel = window.getSelection();
    sel.removeAllRanges();
    sel.addRange(dr);
    let sel2 = sel.QueryInterface(Components.interfaces.nsISelection2);
    sel2.scrollIntoView(0, false, -1, -1);
}

function update_bar() {
    main_control.update_bar();
}

var historyListener = {
    timeout: 200,	// FIXME hack

    OnHistoryGoBack: function(backURI) {
	setTimeout(update_bar, this.timeout);
	return true;
    },

    OnHistoryGoForward: function(forwardURI) {
	setTimeout(update_bar, this.timeout);
	return true;
    },

    OnHistoryGotoIndex: function(index, gotoURI) {
	setTimeout(update_bar, this.timeout);
	return true;
    },

    OnHistoryNewEntry: function(newURI) {
	setTimeout(update_bar, this.timeout);
	return true;
    },

    OnHistoryPurge: function(numEntries) {
	setTimeout(update_bar, this.timeout);
	return true;
    },

    OnHistoryReload: function(reloadURI, reloadFlags) {
	setTimeout(update_bar, this.timeout);
	return true;
    },

    QueryInterface: function(aIID) {
	if (aIID.equals(Components.interfaces.nsISHistoryListener) ||
	    aIID.equals(Components.interfaces.nsISupportsWeakReference) ||
	    aIID.equals(Components.interfaces.nsISupports))
	    return this;
	throw Components.results.NS_NOINTERFACE;
    }
};

var uriContentListener = {
    loadCookie: null,
    parentContentListener: null,

    QueryInterface: function(aIID) {
	if (aIID.equals(Components.interfaces.nsIURIContentListener) ||
	    aIID.equals(Components.interfaces.nsISupportsWeakReference) ||
	    aIID.equals(Components.interfaces.nsISupports))
	    return this;
	throw Components.results.NS_NOINTERFACE;
    },
 
    canHandleContent: function(aContentType, aIsContentPreferred) {
	if (this.parentContentListener)
	    return this.parentContentListener.canHandleContent(aContentType,
							       aIsContentPreferred);
	return false;
    },

    doContent: function(aContentType, aIsContentPreferred, aRequest, aContentHandler) {
	if (this.parentContentListener)
	    return this.parentContentListener.doContent(aContentType,
							aIsContentPreferred,
							aRequest,
							aContentHandler);
	throw Components.results.NS_ERROR_NOT_IMPLEMENTED;	// FIXME
    },

    isPreferred: function(aContentType, desiredContentType) {
	if (this.parentContentListener)
	    return this.parentContentListener.isPreferred(aContentType,
							  desiredContentType);
	return false;
    },

    onStartURIOpen: function(aURI) {
	debug("onStartURIOpen: " + aURI.scheme);
	if (aURI.scheme == "x-lucidor")
	    return false;
	debug("ZZ");
	// if BUSY_FLAGS_PAGE_LOADING
	if (browser.docShell.busyFlags & 4)
	    return true;

	let ok = true;
	let prefs = Components.classes["@mozilla.org/preferences-service;1"].
	    getService(Components.interfaces.nsIPrefService).
	    getBranch("lucidor.");
	let prompt_on = prefs.getBoolPref("viewer.promptOnExternalLink");
	if (prompt_on) {
	    let strbundle = document.getElementById("strings");
	    let title = strbundle.getString("external_link");
	    let text = strbundle.getString("external_link_question");
	    text += "\n(" + aURI.spec + ")";
	    let prompts = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);
	    ok = prompts.confirm(window, title, text);
	}
	if (ok)
	    main_control.add_web_tab(aURI.spec);
	return true;
    }
};

var noExternalURIContentListener = {
    loadCookie: null,
    parentContentListener: null,

    QueryInterface: function(aIID) {
	if (aIID.equals(Components.interfaces.nsIURIContentListener) ||
	    aIID.equals(Components.interfaces.nsISupportsWeakReference) ||
	    aIID.equals(Components.interfaces.nsISupports))
	    return this;
	throw Components.results.NS_NOINTERFACE;
    },
 
    canHandleContent: function(aContentType, aIsContentPreferred) {
	if (this.parentContentListener)
	    return this.parentContentListener.canHandleContent(aContentType,
							       aIsContentPreferred);
	return false;
    },

    doContent: function(aContentType, aIsContentPreferred, aRequest, aContentHandler) {
	if (this.parentContentListener)
	    return this.parentContentListener.doContent(aContentType,
							aIsContentPreferred,
							aRequest,
							aContentHandler);
	throw Components.results.NS_ERROR_NOT_IMPLEMENTED;	// FIXME
    },

    isPreferred: function(aContentType, desiredContentType) {
	if (this.parentContentListener)
	    return this.parentContentListener.isPreferred(aContentType,
							  desiredContentType);
	return false;
    },

    onStartURIOpen: function(aURI) {
	debug("ONSTARTURIOPEN: " + aURI.scheme);
	if (aURI.scheme == "x-lucidor")
	    return false;
	return true;
    }
};

var contextHandler = {
    handleEvent: function(event) {
	if (event.type == "contextmenu") {
	    debug("E: " + event.type);
	    let popup = document.getElementById("contextmenu");
	    popup.showPopup(null, event.screenX + 2, event.screenY + 2,
			    "context");
	    event.stopPropagation();
	}
    }
};

var sidebarContextHandler = {
    handleEvent: function(event) {
	if (event.type == "contextmenu") {
	    debug("E: " + event.type);
	    /*
	    let popup = document.getElementById("contextmenu");
	    popup.showPopup(null, event.screenX + 2, event.screenY + 2,
			    "context");
	     */
	    event.stopPropagation();
	}
    }
};

function init() {
    debug("viewepub init: " + document.documentURI);

    let prefs = Components.classes["@mozilla.org/preferences-service;1"].
	getService(Components.interfaces.nsIPrefService).
	getBranch("lucidor.");

    let appinfo = Components.classes["@mozilla.org/xre/app-info;1"].
	getService(Components.interfaces.nsIXULAppInfo);
    if (appinfo.ID == "Lucidor@lucidor.org")
	is_extension = false;
    else
	is_extension = true;

    if (appinfo.platformVersion.substring(0, 2) != "1.") {
	// For some strange reason nsIFind does not work in XULRunner >= 2.0
	let button = document.getElementById("search_button");
	button.setAttribute("hidden", "true");
    }

    browser = document.getElementById("epub_browser");
    browser.addProgressListener(progressListener);
    browser.docShell.allowAuth = false;
    browser.docShell.allowJavascript = prefs.getBoolPref("viewer.javascript");
    browser.docShell.allowMetaRedirects = false;
    //browser.docShell.allowPlugins = false;
    browser.docShell.useErrorPages = true;
    let wnd = browser.docShell.QueryInterface(Components.interfaces.nsIInterfaceRequestor).getInterface(Components.interfaces.nsIURIContentListener);
    uriContentListener.parentContentListener = wnd.parentContentListener;
    wnd.parentContentListener = uriContentListener;

    if (!is_extension) {
	// FIXME Why is it necessary to manually create a sessionHistory?
	// It is not necessary for viewweb.
	try {
	    browser.webNavigation.sessionHistory = Components.classes["@mozilla.org/browser/shistory;1"].createInstance(Components.interfaces.nsISHistory);
	    browser.sessionHistory.addSHistoryListener(historyListener);
	} catch (e) {
	}
    }

    if (is_extension) {
	browser.contentWindow.addEventListener("contextmenu", contextHandler,
					       true);
	let el = document.getElementById("ve_deck");
	el.addEventListener("contextmenu", sidebarContextHandler, true);

	// Register default stylesheets
	let sss = Components.classes["@mozilla.org/content/style-sheet-service;1"].
	    getService(Components.interfaces.nsIStyleSheetService);
	let ios = Components.classes["@mozilla.org/network/io-service;1"].
	    getService(Components.interfaces.nsIIOService);
	let uri = ios.newURI("chrome://lucidor/content/dtbook.css", null, null);
	if (!sss.sheetRegistered(uri, sss.AGENT_SHEET))
	    sss.loadAndRegisterSheet(uri, sss.AGENT_SHEET);
	uri = ios.newURI("chrome://lucidor/content/xhtml.css", null, null);
	if (!sss.sheetRegistered(uri, sss.AGENT_SHEET))
	    sss.loadAndRegisterSheet(uri, sss.AGENT_SHEET);
	uri = ios.newURI("chrome://lucidor/content/ordbrand.css", null, null);
	if (!sss.sheetRegistered(uri, sss.AGENT_SHEET))
	    sss.loadAndRegisterSheet(uri, sss.AGENT_SHEET);
    }

    let hbrowser = document.getElementById("hidden_browser_tt");
    hbrowser.docShell.allowAuth = false;
    hbrowser.docShell.allowImages = false;
    hbrowser.docShell.allowJavascript = false;
    hbrowser.docShell.allowMetaRedirects = false;
    hbrowser.docShell.allowPlugins = false;
    hbrowser.docShell.useErrorPages = true;
    hbrowser.addProgressListener(tooltipProgressListener);
    wnd = hbrowser.docShell.QueryInterface(Components.interfaces.nsIInterfaceRequestor).getInterface(Components.interfaces.nsIURIContentListener);
    noExternalURIContentListener.parentContentListener = wnd.parentContentListener;
    wnd.parentContentListener = noExternalURIContentListener;

    hbrowser = document.getElementById("hidden_browser_s");
    hbrowser.docShell.allowAuth = false;
    hbrowser.docShell.allowImages = false;
    hbrowser.docShell.allowJavascript = false;
    hbrowser.docShell.allowMetaRedirects = false;
    hbrowser.docShell.allowPlugins = false;
    hbrowser.docShell.useErrorPages = true;
    hbrowser.addProgressListener(searchProgressListener);
    wnd = hbrowser.docShell.QueryInterface(Components.interfaces.nsIInterfaceRequestor).getInterface(Components.interfaces.nsIURIContentListener);
    wnd.parentContentListener = noExternalURIContentListener;

    let box = document.getElementById("hidden_box");
    ppdm = box.boxObject.width;

    language_mapper = Components.classes["@lucidor.org/languageMapperService;1"].getService().wrappedJSObject;
    language_mapper.init();

    if (is_extension) {
	fakeControlService.init(null);
	main_control = fakeControlService;
    } else {
	main_control = Components.classes["@lucidor.org/mainControlService;1"].
	    getService().wrappedJSObject;
    }
    view_border = prefs.getBoolPref("viewer.border");
    view_mode = prefs.getCharPref("viewer.viewMode");
    column_adjustment = prefs.getCharPref("viewer.columnAdjustment");

    try {
	let textbox = document.getElementById("w-box");
	textbox.value = prefs.getCharPref("viewer.columnWidth");
	textbox = document.getElementById("h-box");
	textbox.value = prefs.getCharPref("viewer.columnHeight");
    } catch (e) {
    }
	
    myPrefObserver.register();

    document.addEventListener("LucidorEvent", lucidor_event, false, true);

    let ioService = Components.classes["@mozilla.org/network/io-service;1"].
	getService(Components.interfaces.nsIIOService);
    let docuri = ioService.newURI(document.documentURI, "UTF-8", null);
    docuri.QueryInterface(Components.interfaces.nsIURL);

    let urispec;
    let params = docuri.query.split("&");
    for (let par in loop(params)) {
	let pair = par.split("=");
	if (pair[0] == "uri")
	    urispec = unescape(pair[1]);
	else if (pair[0] == "id")
	    bookcase_id = unescape(pair[1]);
    }

    debug("URI: " + urispec);

    let nsuri;
    try {
	nsuri = ioService.newURI(urispec, "UTF-8", null);
    } catch (e) {
	debug("strange uri: " + e);
	error_opening();
	return;
    }
    if (nsuri.schemeIs("file")) {
	let file = Components.classes["@mozilla.org/network/protocol;1?name=file"].
	    getService(Components.interfaces.nsIFileProtocolHandler).
	    getFileFromURLSpec(urispec);
	try {
	    return init2(file, nsuri.spec);
	} catch (e) {
	    debug("init2 failed: " + e);
	    error_opening();
	    return;
	}
    }

    let file = main_control.get_tempdir();
    file.append("lucidor.epub");
    file.createUnique(Components.interfaces.nsIFile.NORMAL_FILE_TYPE, 0600);

    let name = file.path;

    let wbp = Components.classes["@mozilla.org/embedding/browser/nsWebBrowserPersist;1"].createInstance(Components.interfaces.nsIWebBrowserPersist);
    wbp.persistFlags = wbp.PERSIST_FLAGS_SERIALIZE_OUTPUT;

    wbp.progressListener = {
	onLocationChange: function(webProgress, request, location) {
	},

	onProgressChange: function(webProgress, request, curSelfProgress,
				   maxSelfProgress, curTotalProgress,
				   maxTotalProgress) {
	},

	onSecurityChange: function(webProgress, request, state) {
	},

	onStateChange: function(webProgress, request, stateFlags, status) {
	    debug("onStateChange");
	    if (wbp.currentState == wbp.PERSIST_STATE_FINISHED) {
		debug("Persister has finished saving data.");
		init2(file, nsuri.spec);
	    }
	},

	onStatusChange: function(webProgress, request, status, message) {
	},

	QueryInterface: function(aIID) {
	    if (aIID.equals(Components.interfaces.nsIWebProgressListener))
		return this;
	    throw Components.results.NS_NOINTERFACE;
	}
    };

    try {
	wbp.saveURI(nsuri, null, null, null, null, file);
    } catch(e) {
	error_opening();
	return;
    }

    function fail() {
	if (wbp.currentState != wbp.PERSIST_STATE_FINISHED) {
	    wbp.progressListener = null;
	    debug("Fail");
	    error_opening();
	}
    };
}

function unload() {
    let bookman = Components.classes["@lucidor.org/bookmanagerService;1"].
	getService().wrappedJSObject;
    bookman.removeBook(bookid);
    myPrefObserver.unregister();
    if (is_extension && main_control)
	main_control.remove_tempfiles();
}

function warn_unload() {
    update_last_date_and_position();
}

function update_date_read() {
    if (is_extension)
	return;

    let bookcase = Components.classes["@lucidor.org/bookcaseSqlService;1"].
	getService().wrappedJSObject;
    bookcase.open();
    let id = bookcase.get_id_from_sha1(book.sha1);
    if (id >= 0)
	bookcase.update_date_read(id);
}

function update_last_date_and_position() {
    if (!book)
	return;

    if (!is_extension) {
	let bookcase = Components.classes["@lucidor.org/bookcaseSqlService;1"].
	    getService().wrappedJSObject;
	bookcase.open();
	let id = bookcase.get_id_from_sha1(book.sha1);
	if (id >= 0) {
	    bookcase.update_date_read(id);
	    bookcase.update_last_position(id, get_current_position());
	    return;
	}
    }

    /*
    let prefs = Components.classes["@mozilla.org/preferences-service;1"].
	getService(Components.interfaces.nsIPrefService).
	getBranch("lucidor.");
    let bookmark_lite = prefs.getBoolPref("bookcase.bookmarkLite");
    if (bookmark_lite) {
	let bookcase = Components.classes["@lucidor.org/bookmarkLiteService;1"].
	    getService().wrappedJSObject;
	bookcase.open();
	let id = bookcase.get_or_add_from_sha1(book.sha1);
	bookcase.update_date_read(id);
	bookcase.update_last_position(id, get_current_position());
    }
    */
}

var book_file;

function init2(file, uri) {
    debug("init2");
    book = Components.classes["@lucidor.org/epubDocument;1"].
	createInstance().wrappedJSObject;
    try {
	book.init(file);
    } catch (e) {
	debug("Error opening: " + e);
	error_opening();
	return -1;	// FIXME
    }

    let main = document.getElementById("open_box");
    main.setAttribute("hidden", "true");
    main = document.getElementById("main_box");
    main.setAttribute("hidden", "false");

    book.original_location = uri;	// FIXME

    if (!is_extension && book.original_location !=
	"chrome://lucidor/content/lucidor-manual.epub") {
	let bookcase = Components.classes["@lucidor.org/bookcaseSqlService;1"].
	    getService().wrappedJSObject;
	bookcase.open();
	if (bookcase.get_id_from_sha1(book.sha1) < 0) {
	    let prefs = Components.classes["@mozilla.org/preferences-service;1"].
		getService(Components.interfaces.nsIPrefService).
		getBranch("lucidor.");
	    let prompt_before = prefs.getBoolPref("bookcase.promptBeforeAdd");
	    if (prompt_before) {
		book_file = file;	// FIXME
		let box = document.getElementById("question_box");
		box.setAttribute("hidden", "false");
	    } else {
		try_copy_and_add_book(bookcase, book, file);
	    }
	}
    }

    let bookman = Components.classes["@lucidor.org/bookmanagerService;1"].
	getService().wrappedJSObject;
    bookid = bookman.addBook("epub", book);

    document.title = book.opf.metadata.title;
    main_control.update_tab_title(window);

    partinfo = new Array(book.opf.spine.length);

    // Info sidebar
    let box = document.getElementById("ve_info");
    let strbundle = document.getElementById("strings");
    let sharedbundle = document.getElementById("shared_strings");

    function add_line(entity_id, value) {
	add_line1(entity_id, value, null);
    }

    function add_line1(entity_id, value, extra) {
	if (!value)
	    return;
	let label = sharedbundle.getString(entity_id);
	if (extra)
	    label += " (" + extra + ")";
	add_line2(label, value);
    }

    let last_header = null;

    function add_line2(label, value) {
	if (!value)
	    return;
	if (label == last_header && last_header)
	    label = null;
	else if (label)
	    last_header = label;
	add_line3(label, value);
    }

    function add_line3(label, value) {
	if (label) {
	    let div = document.createElementNS(XHTML_NS, "div");
	    div.setAttribute("class", "infoheader");
	    box.appendChild(div);
	    let node = document.createTextNode(label);
	    div.appendChild(node);
	}
	if (value) {
	    let div = document.createElementNS(XHTML_NS, "div");
	    div.setAttribute("class", "infoitem");
	    box.appendChild(div);
	    let node;
	    if (value.length >= 100)
		node = document.createTextNode(value.substring(0, 90) + "...");
	    else
		node = document.createTextNode(value);
	    div.appendChild(node);
	    if (value.length >= 100)
		div.setAttribute("tooltiptext", value);
	}
    }

    function add_header(entity_id) {
	let label = sharedbundle.getString(entity_id);
	add_line3(label, null);
    }

    function add_role(person, default_role) {
	let label;
	try {
	    label = strbundle.getString("role_" + person.role);
	} catch (e) {
	    label = strbundle.getString(default_role);
	    if (person.role)
		label += " (" + person.role + ")";
	}
	add_line2(label, person.name);
    }

    function add_date(date) {
	let label;
	try {
	    label = strbundle.getString("date_" + date.event);
	} catch (e) {
	    label = sharedbundle.getString("date");
	    if (date.event)
		label += " (" + date.event + ")";
	}
	add_line2(label, date.date);
    }

    add_line("title", book.opf.metadata.title);
    if ((book.opf.metadata.creator.length == 1) &&
	!book.opf.metadata.creator[0].role)
	add_line("author", book.opf.metadata.creator[0].name);
    else
	for (let creator in loop(book.opf.metadata.creator))
	    add_role(creator, "creator");
    if (book.opf.metadata.contributor != undefined)
	for (let contrib in loop(book.opf.metadata.contributor))
	    add_role(contrib, "contributor");
    if (book.opf.metadata.publisher != undefined)
	add_line("publisher", book.opf.metadata.publisher);
    if (book.opf.metadata.rights != undefined)
	add_line("rights", book.opf.metadata.rights);
    if (book.opf.metadata.date != undefined)
	for (let date in loop(book.opf.metadata.date))
	    add_date(date);
    if (book.opf.metadata.languages != undefined) {
	for (let language in loop(book.opf.metadata.languages)) {
	    let lang = language_mapper.translate(language);
	    add_line("language", lang);
	}
    }
    if (book.opf.metadata.description != undefined)
	add_line("description", book.opf.metadata.description);
    if (book.opf.metadata.subject != undefined)
	for (let subject in loop(book.opf.metadata.subject))
	    add_line("subject", subject);
    if (book.opf.metadata.coverage != undefined)
	for (let coverage in loop(book.opf.metadata.coverage))
	    add_line("coverage", coverage);
    if (book.opf.metadata.type != undefined)
	for (let type in loop(book.opf.metadata.type))
	    add_line("type", type);
    if (book.opf.metadata.format != undefined)
	for (let format in loop(book.opf.metadata.format))
	    add_line("format", format);
    if (book.opf.metadata.source != undefined)
	for (let source in loop(book.opf.metadata.source))
	    add_line("source", source);
    if (book.opf.metadata.relation != undefined)
	for (let relation in loop(book.opf.metadata.relation))
	    add_line("relation", relation);
    if (book.opf.uid != undefined)
	add_line("uid", book.opf.uid);
    for (let id in loop(book.opf.metadata.identifier))
	if (!id.uid)
	    add_line1("identifier", id.id, id.scheme);
    add_line("location", book.original_location);

    // Spine sidebar
    spine = book.getSpine();
    box = document.getElementById("ve_spine");
    for (let i = 0; i < spine.length; i++) {
	let item = spine[i];
	let div = document.createElementNS(XHTML_NS, "div");
	div.setAttribute("class", "navitem");
	box.appendChild(div);
	let link = document.createElementNS(XHTML_NS, "a");
	div.appendChild(link);

	if (book.is_encrypted(item.href)) {
	    link.setAttribute("style", "text-decoration: line-through");
	} else {
	    link.setAttribute("href", "#");
	    //link.setAttribute("onclick", "view_uri('" +
	    //		  bookid + "/" + item.href + "');");
	    link.setAttribute("onclick", "view_part(" + i + ", false);");
	}

	let span = document.createElementNS(XHTML_NS, "span");
	span.setAttribute("class", "infoheader");
	span.appendChild(document.createTextNode(item.idref + ":"));
	link.appendChild(span);
	let text = " " + item.href;
	if (!item.linear) {
	    let nl_text = strbundle.getString("auxiliary");
	    text += " (" + nl_text + ")";
	}
	link.appendChild(document.createTextNode(text));

	if (!partinfo[i])
	    partinfo[i] = [];
	partinfo[i].push(link);
    }

    // Navigation sidebar

	function traverse_items(node, level, container) {
	    debug("*: " + node + ";" + container);
	    if (node.href) {
		debug("*: " + node.href);
		let div = document.createElementNS(XHTML_NS, "div");
		div.setAttribute("class", "navitem");
		div.setAttribute("level", level);
		container.appendChild(div);

		let link = document.createElementNS(XHTML_NS, "a");
		div.appendChild(link);
		if (node.broken || book.is_encrypted(node.href)) {
		    link.setAttribute("style", "text-decoration: line-through");
		} else {
		    link.setAttribute("href", "#");
		    link.setAttribute("onclick", "view_uri('" +
				      bookid + "/" + node.href + "');");
		}
		let text = node.text;
		link.appendChild(document.createTextNode(text));
		let part = find_part(node.href);
		if (part >= 0) {
		    if (!partinfo[part])
			partinfo[part] = [];
		    partinfo[part].push(link);
		}
	    } else if (node.list) {
		let spa = document.createElementNS(XHTML_NS, "span");
		div.appendChild(spa);
		let handle = document.createElementNS(XHTML_NS, "a");
		spa.appendChild(handle);
		handle.setAttribute("href", "#");
		handle.setAttribute("onclick", "toggle_it(event);");
		// ⊞⊟﹢﹣▸▾▹▿
		handle.appendChild(document.createTextNode("▹"));
		spa.appendChild(document.createTextNode(" "));	// NBSP

		let text = node.head;
		handle.appendChild(document.createTextNode(text));

		let child_div = document.createElementNS(XHTML_NS, "div");
		child_div.setAttribute("class", "navchild");
		child_div.setAttribute("style", "display: none");
		div.appendChild(child_div);
		for (let n in loop(node.list))
		    traverse_items(n, level + 1, child_div);
	    }
	}

    let toc_nav = null;
    let landmarks_nav = null;
    let pagelist_nav = null;
    for (let nav in loop(book.navs))
	if (nav.type == "toc") {
	    toc_nav = nav;
	} else if (nav.type == "landmarks") {
	    landmarks_nav = nav;
	} else if (nav.type == "page-list") {
	    pagelist_nav = nav;
	}

    if (toc_nav) {
	debug("TOC Navigation sidebar");
	let nav = toc_nav;
	let button = document.getElementById("content_button");

 	button.setAttribute("hidden", "false");
	let box = document.getElementById("ve_toc");

	if (nav.head) {
	    let div = document.createElementNS(XHTML_NS, "div");
	    div.setAttribute("class", "navheader");
	    div.appendChild(document.createTextNode(nav.head));
	    box.appendChild(div);
	}

	for (let n in loop(nav.list))
	    traverse_items(n, 0, box);

	// FIXME Add preference for this?
	if (false) {
	    expand_all(true);
	} else if (false) {
	    collapse_all(true);
	} else {
	    show_all(true);
	}
    }

    if (pagelist_nav) {
	debug("TOC Navigation sidebar");
	let nav = pagelist_nav;
	let button = document.getElementById("pages_button");

 	button.setAttribute("hidden", "false");
	let box = document.getElementById("ve_pages");

	if (nav.head) {
	    let div = document.createElementNS(XHTML_NS, "div");
	    div.setAttribute("class", "navheader");
	    div.appendChild(document.createTextNode(nav.head));
	    box.appendChild(div);
	}

	for (let n in loop(nav.list))
	    traverse_items(n, 0, box);

	// FIXME Add preference for this?
	if (false) {
	    expand_all(true);
	} else if (false) {
	    collapse_all(true);
	} else {
	    show_all(true);
	}
    }

    if (landmarks_nav) {
	debug("Landmarks Navigation sidebar");
	let nav = landmarks_nav;
	let button = document.getElementById("guide_button");

 	button.setAttribute("hidden", "false");
	let box = document.getElementById("ve_guide");

	if (nav.head) {
	    let div = document.createElementNS(XHTML_NS, "div");
	    div.setAttribute("class", "navheader");
	    div.appendChild(document.createTextNode(nav.head));
	    box.appendChild(div);
	}

	for (let n in loop(nav.list))
	    traverse_items(n, 0, box);

	// FIXME Add preference for this?
	if (false) {
	    expand_all(true);
	} else if (false) {
	    collapse_all(true);
	} else {
	    show_all(true);
	}
    }

    // Contents sidebar
    if (book.ncx && book.ncx.navmap) {
	let button = document.getElementById("content_button");
 	button.setAttribute("hidden", "false");
	let navmap = book.ncx.navmap;
	let box = document.getElementById("ve_toc");

	if (navmap.label) {
	    let div = document.createElementNS(XHTML_NS, "div");
	    div.setAttribute("class", "navheader");
	    div.appendChild(document.createTextNode(navmap.label));
	    box.appendChild(div);
	}

	function traverse_points(node, level, container) {
	    debug("*: " + node + ";" + container);
	    if (node.label && node.content) {
		debug("*: " + node.label + ";" + node.content);
		let div = document.createElementNS(XHTML_NS, "div");
		div.setAttribute("class", "navitem");
		div.setAttribute("level", level);
		container.appendChild(div);
		if (node.children) {
		    let spa = document.createElementNS(XHTML_NS, "span");
		    div.appendChild(spa);
		    let handle = document.createElementNS(XHTML_NS, "a");
		    spa.appendChild(handle);
		    handle.setAttribute("href", "#");
		    handle.setAttribute("onclick", "toggle_it(event);");
		    // ⊞⊟﹢﹣▸▾▹▿
		    handle.appendChild(document.createTextNode("▹"));
		    spa.appendChild(document.createTextNode(" "));	// NBSP
		}
		let link = document.createElementNS(XHTML_NS, "a");
		div.appendChild(link);
		if (node.broken || book.is_encrypted(node.content)) {
		    link.setAttribute("style", "text-decoration: line-through");
		} else {
		    link.setAttribute("href", "#");
		    link.setAttribute("onclick", "view_uri('" +
				      bookid + "/" + node.content + "');");
		}
		let text = node.label;
		link.appendChild(document.createTextNode(text));
		let part = find_part(node.content);
		if (part >= 0) {
		    if (!partinfo[part])
			partinfo[part] = [];
		    partinfo[part].push(link);
		}

		if (node.children) {
		    let child_div = document.createElementNS(XHTML_NS, "div");
		    child_div.setAttribute("class", "navchild");
		    child_div.setAttribute("style", "display: none");
		    div.appendChild(child_div);
		    for (let n in loop(node.children))
			traverse_points(n, level + 1, child_div);
		}
	    } else {
		if (level > 0)
		    debug("Error: No label");
		for (let n in loop(node.children))
		    traverse_points(n, level, container);
	    }
	}
	traverse_points(navmap, 0, box);

	// FIXME Add preference for this?
	if (false) {
	    expand_all(true);
	} else if (false) {
	    collapse_all(true);
	} else {
	    show_all(true);
	}
    }

    // Contents sidebar FIXME
    if (false && book.ncx && book.ncx.navmap) {
	let button = document.getElementById("content_button");
 	button.setAttribute("hidden", "false");
	let navmap = book.ncx.navmap;
	let box = document.getElementById("ve_toc");

	if (navmap.label) {
	    let div = document.createElementNS(XHTML_NS, "div");
	    div.setAttribute("class", "navheader");
	    div.appendChild(document.createTextNode(navmap.label));
	    box.appendChild(div);
	}

	let tree = document.createElementNS(XUL_NS, "tree");
	tree.setAttribute("seltype", "single");
	//tree.setAttribute("onselect", "FIXME");
	tree.setAttribute("flex", "1");
	box.appendChild(tree);
	let treecols = document.createElementNS(XUL_NS, "treecols");
	//treecols.setAttribute("hidden", "true");
	//treecols.setAttribute("collapsed", "true");
	//treecols.setAttribute("style", "display: none");
	tree.appendChild(treecols);
	let treecol = document.createElementNS(XUL_NS, "treecol");
	treecol.setAttribute("primary", "true");
	treecol.setAttribute("flex", "1");
	treecols.appendChild(treecol);

	function traverse_points(node, treeparent) {
	    let treeitem;
	    debug("*: " + node.label + ";" + node.content);
	    if (treeparent) {
		treeitem = document.createElementNS(XUL_NS, "treeitem");
		if (node.label && node.content) {
		    let row = document.createElementNS(XUL_NS, "treerow");
		    treeitem.appendChild(row);
		    let cell = document.createElementNS(XUL_NS, "treecell");
		    cell.setAttribute("label", node.label);
		    row.appendChild(cell);
		}
		treeparent.appendChild(treeitem);
	    }
	    let treechildren;
	    if (node.children) {
		treechildren = document.createElementNS(XUL_NS, "treechildren");
		if (treeitem) {
		    treeitem.setAttribute("container", "true");
		    treeitem.setAttribute("open", "true");
		    treeitem.appendChild(treechildren);
		}
		for (let n in loop(node.children))
		    traverse_points(n, treechildren);
	    }
	    return treechildren;
	}

	let childs = traverse_points(navmap, null);
	tree.appendChild(childs);
    }

    // Secondary content sidebar
    if (book.ncx && book.ncx.navlists) {
	let button = document.getElementById("list_button");
 	button.setAttribute("hidden", "false");
	box = document.getElementById("ve_list");
	for (let navlist in loop(book.ncx.navlists)) {
	    if (navlist.label) {
		let div = document.createElementNS(XHTML_NS, "div");
		div.setAttribute("class", "navheader");
		div.appendChild(document.createTextNode(navlist.label));
		box.appendChild(div);
	    }
	    for (let node in loop(navlist.list)) {
		if (node.label && node.content) {
		    debug("$: " + node.label + ";" + node.content);
		    let div = document.createElementNS(XHTML_NS, "div");
		    div.setAttribute("class", "navitem");
		    box.appendChild(div);
		    let link = document.createElementNS(XHTML_NS, "a");
		    div.appendChild(link);
		    link.setAttribute("href", "#");
		    link.setAttribute("onclick", "view_uri('" +
				      bookid + "/" + node.content + "');");
		    let text = node.label;
		    link.appendChild(document.createTextNode(text));
		}
	    }
	}
    }

    // Pages sidebar
    if (book.ncx && book.ncx.pagelist) {
	let button = document.getElementById("pages_button");
 	button.setAttribute("hidden", "false");
	box = document.getElementById("ve_pages");

	if (book.ncx.pagelist.label) {
	    let div = document.createElementNS(XHTML_NS, "div");
	    div.setAttribute("class", "navheader");
	    div.appendChild(document.createTextNode(book.ncx.pagelist.label));
	    box.appendChild(div);
	}
	for (let node in loop(book.ncx.pagelist.list)) {
	    if (node.label && node.content) {
		debug("$: " + node.label + ";" + node.content);
		let div = document.createElementNS(XHTML_NS, "div");
		div.setAttribute("class", "navitem");
		box.appendChild(div);
		let link = document.createElementNS(XHTML_NS, "a");
		div.appendChild(link);
		link.setAttribute("href", "#");
		link.setAttribute("onclick", "view_uri('" +
				  bookid + "/" + node.content + "');");
		let text = node.label;
		link.appendChild(document.createTextNode(text));
	    }
	}
    }

    // Guide sidebar
    if (book.opf.guide) {
	let button = document.getElementById("guide_button");
	button.setAttribute("hidden", "false");
	box = document.getElementById("ve_guide");
	for (let item in loop(book.opf.guide)) {
	    let div = document.createElementNS(XHTML_NS, "div");
	    div.setAttribute("class", "navitem");
	    box.appendChild(div);
	    let link = document.createElementNS(XHTML_NS, "a");
	    div.appendChild(link);
	    if (item.broken || book.is_encrypted(item.href)) {
		link.setAttribute("style", "text-decoration: line-through");
	    } else {
		link.setAttribute("href", "#");
		// FIXME encode item.href?
		link.setAttribute("onclick", "view_uri('" +
				  bookid + "/" + item.href + "');");
	    }
	    let text = item.title;

	    let type_text;
	    try {
		type_text = strbundle.getString("type_" + item.type);
	    } catch (e) {
		if (item.type.substring(0, 6) == "other.")
		    type_text = strbundle.getString("type_other");
		else
		    book.add_error("invalid guide item type: " + item.type);
	    }
	    //if (type_text)
	    //	link.setAttribute("tooltiptext", type_text);
	    if (!text) {
		if (type_text)
		    text = type_text;
		else if (item.type)
		    text = item.type;
		else
		    text = "";
	    } else if (type_text &&
		       (type_text.toLowerCase() != text.toLowerCase())) {
		text += " (" + type_text + ")";
	    }
	    link.appendChild(document.createTextNode(text));
	}
    }

    // Tools sidebar
    let show_tools = false;

    debug("Original location: " + book.original_location)
    let uri = null;
    try {
	let ioService = Components.classes["@mozilla.org/network/io-service;1"].
	    getService(Components.interfaces.nsIIOService);
	uri = ioService.newURI(book.original_location, "UTF-8", null);
    } catch (e) {
    }

    /*
    if (uri && (uri.schemeIs("http") || uri.schemeIs("https"))) {
	let button = document.getElementById("epubcat");
	button.setAttribute("hidden", "false");
	show_tools = true;
    }
    */
    if (book.opf.metadata.autopublication) {
	let button = document.getElementById("autopub");
	button.setAttribute("hidden", "false");
	show_tools = true;

	if (uri && !uri.schemeIs("x-lucidor-convert")) {
	    let box = document.getElementById("update_box");
	    box.setAttribute("hidden", "false");
	}
    }
    if (show_tools) {
	let button = document.getElementById("tools_button");
	button.setAttribute("hidden", "false");
    }

    let mlist = document.getElementById("sidelist");
    for (let cld in loop(mlist.getElementsByTagName("menuitem")))
	if (!cld.hidden) {
	    cld.doCommand();
	    break;
	}

    if (!show_last_position()) {
	let part = book.get_first_part();
	if (part >= 0)
	    view_uri_x(bookid + "/" + spine[part].href, spine[part].stylesheet,
		       0, null);
	else
	    book.add_error("has no content");
    }
    setTimeout(maybe_hide_toolbar, 15000);

    update_date_read();
}

function toggle_it(event) {
    let target = event.currentTarget;
    let child_div = target.parentNode.parentNode.lastChild;
    if (child_div.style.display == "none") {
	$(child_div).show("fast");
	$(target).text("▿");
    } else {
	$(child_div).hide("fast");
	$(target).text("▹");
    }
    // ⊞⊟﹢﹣▸▾▹▿
}

// Default Contents state: Expanded, Collapsed, Show all
function expand_all(imm) {
    let box = document.getElementById("ve_toc");
    $("div:has(div.navchild) > span:first-child > a:first-child", box).text("▿");
    $("div:has(div.navchild) > span:first-child", box).show();
    if (imm)
	$(".navchild", box).show();
    else
	$(".navchild", box).show("slow");
}

function collapse_all(imm) {
    let box = document.getElementById("ve_toc");
    $("div:has(div.navchild) > span:first-child > a:first-child", box).text("▹");
    $("div:has(div.navchild) > span:first-child", box).show();
    if (imm)
	$(".navchild", box).hide();
    else
	$(".navchild", box).hide("slow");
}

function show_all(imm) {
    let box = document.getElementById("ve_toc");
    $("div:has(div.navchild) > span:first-child", box).hide();
    if (imm)
	$(".navchild", box).show();
    else
	$(".navchild", box).show("slow");
}

function show_last_position() {
    if (is_extension)
	return false;

    let bookcase = Components.classes["@lucidor.org/bookcaseSqlService;1"].
	getService().wrappedJSObject;
    bookcase.open();
    let id = bookcase.get_id_from_sha1(book.sha1);
    let bm_text = bookcase.get_last_position(id);
    if (bm_text) {
	let JSON = Components.classes["@mozilla.org/dom/json;1"].
	    createInstance(Components.interfaces.nsIJSON);
	let bm = JSON.decode(bm_text);
	if (bm.version == 1) {
	    let part = find_part(bm.href);
	    if (part >= 0) {
		view_uri_x(bookid + "/" + spine[part].href,
			   spine[part].stylesheet, bm.scroll, null);
		return true;
	    }
	}
    }
    return false;
}

function get_current_position() {
    let JSON = Components.classes["@mozilla.org/dom/json;1"].
	createInstance(Components.interfaces.nsIJSON);
    let scroll;
    let wind = browser.contentWindow;
    if (view_mode == "columns")
	scroll = (1.0 * wind.scrollX) / wind.scrollMaxX;
    else
	scroll = (1.0 * wind.scrollY) / wind.scrollMaxY;
    let href = href_to_local_href(browser.currentURI.spec);
    let bm = { version: 1, href: href, scroll: scroll };
    return JSON.encode(bm);
}

function error_opening() {
    document.getElementById("main_box").hidden = true;
    document.getElementById("open_box").hidden = true;
    document.getElementById("error_box").hidden = false;
}

function answer_yes() {
    let box = document.getElementById("question_box");
    box.setAttribute("hidden", "true");

    try_copy_and_add_book(null, book, book_file);
    book_file = null;

    update_last_date_and_position();
}

function answer_no() {
    let box = document.getElementById("question_box");
    box.setAttribute("hidden", "true");
}

function update_yes() {
    let box = document.getElementById("update_box");
    box.setAttribute("hidden", "true");
    update_autopub();
}

function update_no() {
    let box = document.getElementById("update_box");
    box.setAttribute("hidden", "true");
}

function show_warning(text_msg, more_msg, autohide) {
    main_control.show_warning(window, [ text_msg ], more_msg, autohide);
}

function close_warning(box_id) {
    let box = document.getElementById(box_id);
    if (box)
	box.parentNode.removeChild(box);
}

function show_details() {
    let text = "";
    for (let w in loop(book.errors))
	text += "• " + w + "\n";
    let prompts = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);
    let strbundle = document.getElementById("strings");
    let title = strbundle.getString("errors");
    prompts.alert(window, title, text);
}

function show_source() {
    main_control.add_web_tab("view-source:" + browser.currentURI.spec);
}

function lookup_word() {
    let prefs = Components.classes["@mozilla.org/preferences-service;1"].
	getService(Components.interfaces.nsIPrefService).
	getBranch("lucidor.");
    let uri = prefs.getCharPref("viewer.lookup");

    let sel = browser.contentWindow.getSelection();
    let word = sel.toString();
    word = escape(word);
    uri = uri.replace("%%", word);

    main_control.add_web_tab(uri);
}

function contextmenu_onpopup() {
    let sel = browser.contentWindow.getSelection();
    let item = document.getElementById("lookup_item");
    if (sel.toString().length < 1)
	item.disabled = true;
    else
	item.disabled = false;
}

function update_autopub() {
    let location = "x-lucidor-convert://feed/?";
    for (let source in loop(book.opf.metadata.source))
	location += "uri=" + escape(source) + "&";
    location += "title=" + escape(book.opf.metadata.title);
    debug("location: " + location);
    main_control.add_epub_tab(location);
}

function send_to_epubcatalog() {
    let url = "http://www.epubcatalog.org/epubcat/addcat";
    let dataString = "url=" + book.original_location;
    main_control.add_web_tab_post(url, dataString);
}

function try_copy_and_add_book(bookcase, book, file) {
    if (is_extension)
	return;

    if (!bookcase) {
	bookcase = Components.classes["@lucidor.org/bookcaseSqlService;1"].
	    getService().wrappedJSObject;
	bookcase.open();
    }

    try {
	bookcase.add_open_book(book, file);
    } catch (e) {
	// FIXME make more detailed and localized
	show_warning("There was an error when trying to add the book",
		     null, false);
    }
}

function get_browser() {
    return browser;
    //return document.getElementById("epub_browser");
}

function getMarkupDocumentViewer() {
    return browser.markupDocumentViewer;
    //return document.getElementById("epub_browser").markupDocumentViewer;
}

function get_save_file() {
    return book.zipreader.file;
}

function save_file() {
};

function switch_deck(i) {
    document.getElementById("ve_deck").selectedIndex = i;
}

function view_part(part, endscroll) {
    let es = 0;
    if (endscroll)
	es = 1;
    view_uri_x(bookid + "/" + spine[part].href, spine[part].stylesheet,
	       es, null);
}

function view_uri(location) {
    view_uri_x(location, null, 0, null);
}

function view_uri_x(location, stylesheet, endscroll, searchscroll) {
    debug("view_uri_x: " + location);

    if (location.indexOf("x-lucidor://") == 0)
	location = location.substring(12);

    let nu = Components.classes["@mozilla.org/network/util;1"].
	createInstance(Components.interfaces.nsINetUtil);
    let loca = location.split("#");
    if (loca.length == 2)
	location = nu.escapeString(loca[0], nu.ESCAPE_URL_PATH)
	    + "#" +  loca[1];
    else
	location = nu.escapeString(location, nu.ESCAPE_URL_PATH);
    if (stylesheet)
	stylesheet = nu.escapeString(stylesheet, nu.ESCAPE_URL_PATH);

    let tmploc = location.substring(bookid.length + 1);
    let item = book.opf.manifest_href[tmploc];
    if (item &&	item.encryption) {
	let strbundle = document.getElementById("strings");
	let text = strbundle.getString("encrypted_error");
	show_warning(text, null, false);
	update_buttons();
	update_col_buttons();
	return;
    }

    location = "x-lucidor://" + location;

    debug("(" + browser.currentURI.spec + ")");

    scroll_ratio = endscroll;
    scroll_to_search = searchscroll;
    extra_stylesheet = stylesheet;

    try {
	debug("Location: " + location);
	browser.loadURI(location, null, null);
    } catch (e) {
	let strbundle = document.getElementById("strings");
	let text = strbundle.getString("load_error");
	show_warning(text, null, false);
	update_buttons();
	update_col_buttons();
    }
}

function href_to_local_href(href) {
    if (href.indexOf("x-lucidor://") == 0)
	href = href.substring(12);
    // FIXME there is a rare bug here: a name of a folder may be equal to bookid
    if (href.indexOf(bookid + "/") == 0)
	href = href.substring(bookid.length + 1);
    let i = href.indexOf("#");
    if (i >= 0)
	href = href.substring(0, i);
    return href;
}

function find_part(href) {
    let nu = Components.classes["@mozilla.org/network/util;1"].
	createInstance(Components.interfaces.nsINetUtil);
    let nh = nu.unescapeString(href, nu.ESCAPE_ALL);
    return book.hrefToPart(href_to_local_href(nh));
}

function show_next_part() {
    let part = find_part(browser.currentURI.spec);
    let next = book.get_next_part(part);
    if (next < 0)
	return;
    view_part(next, false);
}

function show_previous_part() {
    let part = find_part(browser.currentURI.spec);
    let prev = book.get_previous_part(part);
    if (prev < 0)
	return;
    view_part(prev, true);
}

function go_back() {
    browser.goBack();
}

function go_forward() {
    browser.goForward();
}

function update_buttons(part) {
    let linear;
    if (spine[part])
	linear = spine[part].linear;
    else
	linear = false;	// Hack to disable buttons

    let button = document.getElementById("prev_button");
    if (linear && (book.get_previous_part(part) >= 0))
	button.setAttribute("disabled", "false");
    else
	button.setAttribute("disabled", "true");

    button = document.getElementById("next_button");
    if (linear && (book.get_next_part(part) >= 0))
	button.setAttribute("disabled", "false");
    else
	button.setAttribute("disabled", "true");
}

function update_col_buttons() {
    let val = "true";
    if (view_mode == "columns")
	val = "false";
    let button = document.getElementById("prev_col_button");
    button.setAttribute("hidden", val);
    button = document.getElementById("next_col_button");
    button.setAttribute("hidden", val);
}

var show_bar = false;
function update_navigation_bar() {
    if (is_extension)
	return;

    let buttonbar = document.getElementById("nav_buttons");
    let item = document.getElementById("navbar_item");
    let margin = 20;
    if (show_bar)
	margin += 20;
    if (item.getAttribute("checked") &&
	((browser.contentWindow.scrollY < margin) ||
	(browser.contentWindow.scrollY >
	 (browser.contentWindow.scrollMaxY - margin)))) {
	show_bar = true;
	buttonbar.hidden = false;
    } else {
	show_bar = false;
	buttonbar.hidden = true;
    }
}

function toggle_navigation_bar() {
    let buttonbar = document.getElementById("nav_buttons");
    let item = document.getElementById("navbar_item");
    if (item.getAttribute("checked")) {
	show_bar = true;
	buttonbar.hidden = false;
    } else {
	show_bar = false;
	buttonbar.hidden = true;
    }
}

function update_contextmenu() {
    let bar = document.getElementById("mysidebar");
    let item = document.getElementById("sidebar_item");
    item.checked = !bar.hidden;
    bar = document.getElementById("toolbar");
    item = document.getElementById("toolbar_item");
    item.checked = !bar.hidden;
}

function toggle_sidebar() {
    let sidebar = document.getElementById("mysidebar");
    let splitter = document.getElementById("epub_splitter");
    let val = !sidebar.hidden;
    sidebar.hidden = val;
    splitter.hidden = val;

    let item = document.getElementById("sidebar_item");
    if (val)
	item.setAttribute("checked", "false");
    else
	item.setAttribute("checked", "true");
}

function toggle_toolbar() {
    hide_toolbar_ok = false;
    let toolbar = document.getElementById("toolbar");
    let val = !toolbar.hidden;
    toolbar.hidden = val;

    let item = document.getElementById("toolbar_item");
    if (val)
	item.setAttribute("checked", "false");
    else
	item.setAttribute("checked", "true");
}

var hide_toolbar_ok = true;

function maybe_hide_toolbar() {
    if (hide_toolbar_ok) {
	let toolbar = document.getElementById("toolbar");
	if (!toolbar.hidden)
	    toggle_toolbar();
    }
}

function toggle_tabbar() {
    let val = main_control.toggle_tabbar();
    let item = document.getElementById("tabbar_item");
    if (val)
	item.setAttribute("checked", "false");
    else
	item.setAttribute("checked", "true");
}

function set_view_mode(mode) {
    if (mode)
	view_mode = mode;
    let root = browser.contentDocument.documentElement;
    let glue = root.firstChild;
    let wbox = document.getElementById("w-box");
    let hbox = document.getElementById("h-box");
    let autosize = document.getElementById("auto-button");

    let defclass = "";
    if (view_border)
	defclass = "border ";

    if (view_mode == "columns") {
	root.setAttribute("class", defclass + "columns");
	glue.setAttribute("class", defclass + "columns");
	wbox.disabled = false;
	hbox.disabled = false;
	autosize.disabled = false;
    } else if (view_mode == "flow") {
	root.setAttribute("class", defclass + "flow");
	root.style.width = "";
	glue.setAttribute("class", defclass + "flow");
	glue.style.height = "";
	glue.style.MozColumnWidth = "";
	wbox.disabled = true;
	hbox.disabled = true;
	autosize.disabled = true;
    } else {	// view_mode == "fixed"
	root.setAttribute("class", defclass + "fixed");
	glue.setAttribute("class", defclass + "fixed");
	glue.style.height = "";
	glue.style.MozColumnWidth = "";
	wbox.disabled = false;
	hbox.disabled = true;
	autosize.disabled = false;
    }
    update_col_buttons();
    size_change();
}

function w_scroll(event) {
    let textbox = document.getElementById("w-box");
    if (event.detail < 0)
	textbox.increase();
    else
	textbox.decrease();
    size_change();
}

function h_scroll(event) {
    let textbox = document.getElementById("h-box");
    if (event.detail < 0)
	textbox.increase();
    else
	textbox.decrease();
    size_change();
}

function auto_size() {
    let val = browser.boxObject.width * 10 / ppdm - 1;
    // FIXME make two columns visible at once if enough space is available?
    if (view_mode == "columns" && val > 30)
	val = val / 2 - 1;
    //val = Math.min(val, 30);
    let textbox = document.getElementById("w-box");
    textbox.value = val;

    val = browser.boxObject.height * 10 / ppdm - 2;
    textbox = document.getElementById("h-box");
    textbox.value = val;

    size_change();
}

function size_change() {
    if (view_mode == "flow")
	return;

    let prefs = Components.classes["@mozilla.org/preferences-service;1"].
	getService(Components.interfaces.nsIPrefService).
	getBranch("lucidor.");

    let textbox = document.getElementById("w-box");
    let width = textbox.value + "cm";
    let root = browser.contentDocument.documentElement;
    root.style.width = width;
    prefs.setCharPref("viewer.columnWidth", textbox.value);

    textbox = document.getElementById("h-box");
    prefs.setCharPref("viewer.columnHeight", textbox.value);
    if (view_mode == "columns") {
	let height = textbox.value + "cm";
	let glue = root.firstChild;
	glue.style.MozColumnWidth = width;
	glue.style.height = height;
    }
}

function mouse_scroll(event) {
    //debug("::: " + event.detail);
    if (event.timeStamp > scroll_time + 600) {
	scroll_time = event.timeStamp;
	scroll_count = 0;
	return;
    }

    /*
    if (scroll_count == -1) {
	debug("DDD");
	event.preventDefault();
	return;
    }
    */

    if (event.detail > 0 && browser.contentWindow.scrollY >
	(browser.contentWindow.scrollMaxY - 2)) {
	scroll_count++;
	scroll_time = event.timeStamp;
	if (scroll_count > 8) {
	    scroll_count = -1;
	    //event.preventDefault();
	    if (view_mode != "columns" || browser.contentWindow.scrollX >
		(browser.contentWindow.scrollMaxX - 2)) {
		show_next_part();
		// This is to work around the problem that you cannot continue scrolling without moving mouse after changing part
		let utils = window.QueryInterface(Components.interfaces.nsIInterfaceRequestor).getInterface(Components.interfaces.nsIDOMWindowUtils);
		utils.sendMouseEvent("mousemove", 1, 0, 0, 0, 0);
		utils.sendMouseEvent("mousemove", -1, 0, 0, 0, 0);
	    } else	// view_mode == columns && not at right end
		next_column();
	    // FIXME must wait until part is loaded before scrolling
	    //browser.contentWindow.scrollY = 30;
	}
    } else if (browser.contentWindow.scrollY == 0 && event.detail < 0) {
	scroll_count++;
	scroll_time = event.timeStamp;
	if (scroll_count > 8) {
	    scroll_count = -1;
	    //event.preventDefault();
	    if (view_mode != "columns" || browser.contentWindow.scrollX == 0) {
		show_previous_part();
		// This is to work around the problem that you cannot continue scrolling without moving mouse after changing part
		let utils = window.QueryInterface(Components.interfaces.nsIInterfaceRequestor).getInterface(Components.interfaces.nsIDOMWindowUtils);
		utils.sendMouseEvent("mousemove", 1, 0, 0, 0, 0);
		utils.sendMouseEvent("mousemove", -1, 0, 0, 0, 0);
	    } else	// view_mode == columns && not at left end
		previous_column();
	}
    } else {
	scroll_count = 0;
    }
}

function next_column() {
    change_column(1);
}

function previous_column() {
    change_column(-1);
}

function change_column(dir) {
    if (view_mode != "columns")
	return;

    let textbox = document.getElementById("w-box");
    let w = browser.contentWindow.innerWidth + browser.contentWindow.scrollMaxX;
    let cw = textbox.value * ppdm * 0.1;
    let margin = 0.4 * ppdm * 0.1;	// FIXME not exact
    cw += margin;
    let cols = Math.floor(w / cw);
    let rcw = w / cols;
    debug(w + " " + cw + " " + cols + " " + rcw + " " + margin + " " + w / cw);

    if (column_adjustment == "none") {
	browser.contentWindow.scrollBy(dir * rcw,
				       dir * -browser.contentWindow.scrollMaxY);
	return;
    }

    let cx = browser.contentWindow.scrollX;
    let offset = 0;
    if (column_adjustment == "center" &&
	browser.contentWindow.innerWidth > rcw) {
	offset = (browser.contentWindow.innerWidth - rcw) / 2 - margin / 2;
	cx += offset;
	//debug("OFFSET: " + offset);
    }

    let current_col = Math.floor(cx / rcw);
    if (current_col < 0)	// This should not happen, but it does
	current_col = 0;
    if (dir == 1) {
	if ((cx % rcw) / rcw > 0.9)
	    current_col += 1;
    } else {
	if ((cx % rcw) / rcw > 0.1)
	    current_col += 1;
    }

    let next = (current_col + dir) * rcw - offset;
    debug(current_col + " " + next);

    if (dir == 1)
	browser.contentWindow.scrollTo(next, 0);
    else
	browser.contentWindow.scrollTo(next, browser.contentWindow.scrollMaxY);
}

function add_note() {
    debug("add_note");
    // FIXME
}

function on_search_keypress(event) {
    if (event.keyCode == 13)
	find_in_document();
}

var finder;
var spart;
var search_string;
var search_stylesheet;
var searches;

function find_in_document() {
    debug("find_in_document");
    finder = Components.classes["@mozilla.org/embedcomp/rangefind;1"].
	createInstance(Components.interfaces.nsIFind);
    finder.caseSensitive = false;
    finder.findBackwards = false;

    spart = 0;
    searches = new Array();

    let fbox = document.getElementById("search-box");
    search_string = fbox.value;
    if (search_string != "") {
	let frame = document.getElementById("search_results");
	while (frame.childNodes.length > 0)
	    frame.removeChild(frame.childNodes.item(0));

	let div = document.createElementNS(XHTML_NS, "div");
	div.setAttribute("class", "navitem");
	let strbundle = document.getElementById("strings");
	let text = strbundle.getString("searching");
	div.appendChild(document.createTextNode(text));
	frame.appendChild(div);

	start_search();
    }
}

var searchProgressListener = {
    onLocationChange: function(webProgress, request, location) {
    },

    onProgressChange: function(webProgress, request, curSelfProgress,
			       maxSelfProgress, curTotalProgress,
			       maxTotalProgress) {
    },

    onSecurityChange: function(webProgress, request, state) {
    },

    onStateChange: function(webProgress, request, stateFlags, status) {
	if ((stateFlags & 16) && (stateFlags & 131072)) {
	    let hbrowser = document.getElementById("hidden_browser_s");
	    if (webProgress.DOMWindow != hbrowser.contentWindow)
		return;

	    let doc = webProgress.DOMWindow.document;
	    let oldroot = doc.documentElement;
	    let root = doc.createElementNS(XHTML_NS, "x-ordbrand-root");
	    let glue = doc.createElementNS(XHTML_NS, "x-ordbrand-glue");
	    doc.replaceChild(root, oldroot);
	    root.appendChild(glue);
	    glue.appendChild(oldroot);
	    if (search_stylesheet) {
		let pi = doc.createProcessingInstruction("xml-stylesheet",
							 search_stylesheet);
		doc.insertBefore(pi, doc.firstChild);
	    }
	    find_in_part(doc);
	    return start_search();
	}
    },

    onStatusChange: function(webProgress, request, status, message) {
    },

    QueryInterface: function(aIID) {
	if (aIID.equals(Components.interfaces.nsIWebProgressListener) ||
	    aIID.equals(Components.interfaces.nsISupportsWeakReference))
	    return this;
	throw Components.results.NS_NOINTERFACE;
    }
};

function start_search() {
    debug("start_search");
    if (spart >= spine.length) {
	debug("Search ended");

	let iframe = document.getElementById("search_results");
	let div = document.createElementNS(XHTML_NS, "div");
	div.setAttribute("class", "navitem");
	let strbundle = document.getElementById("strings");
	let text = strbundle.getFormattedString("matches",
						['"' + search_string + '"',
						 searches.length.toString()]);
	div.appendChild(document.createTextNode(text));

	let oldfirst = iframe.firstChild;
	iframe.insertBefore(div, oldfirst);
	iframe.removeChild(oldfirst);

	return;
    }

    debug("Part " + spart);
    let location = "x-lucidor://" + bookid + "/" + spine[spart].href;
    search_stylesheet = "x-lucidor://" + bookid + "/" + spine[spart].stylesheet;
    spart++;

    try {
	let hbrowser = document.getElementById("hidden_browser_s");
	hbrowser.loadURI(location, null, null);
    } catch (e) {
	// FIXME show_warning
	debug("Error: opening page for search failed: " + location);
	return start_search();
    }
}

function find_in_part(doc) {
    let rootNode = doc.documentElement;

    let searchRange = doc.createRange();
    searchRange.selectNodeContents(rootNode);

    let startPoint = doc.createRange();
    startPoint.setStart(searchRange.startContainer, searchRange.startOffset);
    startPoint.setEnd(searchRange.startContainer, searchRange.startOffset);

    let endPoint = doc.createRange();
    endPoint.setStart(searchRange.endContainer, searchRange.endOffset);
    endPoint.setEnd(searchRange.endContainer, searchRange.endOffset);

    debug("ROOT: " + rootNode.localName + " " + searchRange.startOffset + " " +
	 searchRange.endOffset);

    while (true) {
	let res = finder.Find(search_string, searchRange, startPoint, endPoint);
	debug("RES: " + res);
	if (!res)
	    return;
	debug(res.startOffset + " " + res.endOffset);
	try {
	let iframe = document.getElementById("search_results");
	let div = document.createElementNS(XHTML_NS, "div");
	div.setAttribute("class", "navitem");
	iframe.appendChild(div);
	let link = document.createElementNS(XHTML_NS, "a");
	div.appendChild(link);
	link.setAttribute("href", "#");
	link.setAttribute("onclick", "view_search(" + searches.length + ");");

	let span = document.createElementNS(XHTML_NS, "span");
	span.setAttribute("class", "infoheader");
	span.appendChild(document.createTextNode(searches.length + 1));
	link.appendChild(span);
	let context = get_context(res);
	link.appendChild(document.createTextNode(" " + context.pre));
	span = document.createElementNS(XHTML_NS, "span");
	span.setAttribute("class", "match");
	span.appendChild(document.createTextNode(res.toString()));
	link.appendChild(span);
	link.appendChild(document.createTextNode(context.post));

	let item = {};
	item.location = doc.location.toString();
	//item.location = spine[spart - 1].href;
	item.stylesheet = spine[spart - 1].stylesheet;
	if (!xpointer_service) {
	    xpointer_service = Components.classes["@mozilla.org/xpointer-service;1"].getService();
	    xpointer_service.QueryInterface(Components.interfaces.nsIXPointerService);
	}
	let xp = xpointer_service.createXPointerFromRange(res, doc);
	debug("XPointer " + xp);
	item.xpointer = xp;	// FIXME
	searches.push(item);

	startPoint.setStart(res.endContainer, res.endOffset);
	startPoint.setEnd(res.endContainer, res.endOffset);
	} catch (e) {
	    debug("EEE: " + e);
	}
    }
}

function node_is_block(node) {
    return node.ownerDocument.defaultView.
	getComputedStyle(node, "").display == "block";
}

function node_is_blocklike(node) {
    let d = node.ownerDocument.defaultView.getComputedStyle(node, "").display;
    return (d == "block") || (d == "list-item");
}

// Get the first enclosing node (or self) which is blocklike
function get_first_block(node) {
    for (; node.parentNode; node = node.parentNode)
	if ((node.nodeType == node.ELEMENT_NODE) && node_is_blocklike(node))
	    return node;
    return node;
}

// Get the first enclosing node (or self) which is blocklike, and which
// has no nonwhitespace text before self
function get_first_starting_block(node) {
    let top = node;
    for (; node.parentNode; node = node.parentNode) {
	if ((node.nodeType == node.ELEMENT_NODE) && node_is_blocklike(node))
	    return node;
	for (let n = node.previousSibling; n; n = n.previousSibling)
	    if (n.nodeType == node.TEXT_NODE)
		for (let ch in loop(n.nodeValue))
		    if (!is_ws[ch])
			return top;
    }
    return top;
}

// FIXME should collapse whitespace
function get_context(drange) {
    const CLEN = 22;
    let doc = drange.startContainer.ownerDocument;

    // Pre-context
    let pre = doc.createRange();
    pre.setStart(drange.startContainer, Math.max(0, drange.startOffset - CLEN));
    pre.setEnd(drange.startContainer, drange.startOffset);

    // Expand range to previous nodes if necessary
    let node = pre.startContainer;
    let block = get_first_block(node);
    let walker = doc.createTreeWalker(block, 0x04, null, false);
    walker.currentNode = node;
    //debug("LEN: " + pre.toString().length + " " + pre.toString());
    while (pre.toString().length < CLEN) {
	node = walker.previousNode();
	if (!node)
	    break;
	pre.setStart(node, Math.max(0, node.textContent.length -
				    (CLEN - pre.toString().length)));
	//debug("len: " + pre.toString().length + " " + pre.toString());
    }

    // Start of a block?
    let prebreak = false;
    if ((pre.startOffset == 0) && !walker.previousNode())
	prebreak = true;

    // Post-context
    let post = doc.createRange();
    post.setStart(drange.endContainer, drange.endOffset);
    post.setEnd(drange.endContainer,
		Math.min(drange.endContainer.textContent.length,
			 drange.endOffset + CLEN));

    // Expand range to following nodes if necessary
    node = post.endContainer;
    block = get_first_block(node);
    walker = doc.createTreeWalker(block, 0x04, null, false);
    walker.currentNode = node;
    //debug("lEN: " + post.toString().length + " " + post.toString());
    while (post.toString().length < CLEN) {
	node = walker.nextNode();
	if (!node)
	    break;
	post.setEnd(node, Math.min(node.textContent.length,
				   (CLEN - post.toString().length)));
	//debug("len: " + post.toString().length + " " + post.toString());
    }

    // End of the block?
    let postbreak = false;
    if ((post.endOffset == post.endContainer.textContent.length) &&
	!walker.nextNode())
	postbreak = true;

    let pretext = pre.toString();
    let posttext = post.toString();

    // Align with word-breaks if not too much context is lost
    if (!prebreak) {
	let i = pretext.indexOf(" ");
	if (i < 5)
	    pretext = pretext.substring(i);

	pretext = "..." + pretext;
    }

    if (!postbreak) {
	let i = posttext.lastIndexOf(" ");
	if (i >= (posttext.length - 5))
	    posttext = posttext.substring(0, i + 1);

	posttext += "...";
    }

    return {pre: pretext, post: posttext};
}

function view_search(num) {
    debug("view_search: " + searches[num].xpointer);
    if (browser.currentURI.spec == searches[num].location) {
	scroll_to_xpointer(searches[num].xpointer, browser.contentDocument,
			   browser.contentWindow);
    } else {
	view_uri_x(searches[num].location, searches[num].stylesheet,
		   0, searches[num].xpointer);
    }
}

var is_ws = [];
is_ws[" "] = true;
is_ws["\n"] = true;
is_ws["\t"] = true;
is_ws["\r"] = true;
is_ws[String.fromCharCode(160)] = true;	// &nbsp;

// FIXME This can be improved
function get_nodetext(node) {
    let text = "";
    let cut = false;
    let nls = 0;
    let walker = node.ownerDocument.createTreeWalker(node, 0x01 | 0x04,
						     null, false);
    for (let n = walker.currentNode; n; n = walker.nextNode()) {
	if (n.nodeType == n.TEXT_NODE) {
	    let aa = "";
	    let ws = (text.length > 0) && is_ws[text[text.length - 1]];
	    for (let ch in loop(n.nodeValue)) {
		if (is_ws[ch]) {
		    if (!ws) {
			aa += " ";
			ws = true;
		    }
		} else {
		    aa += ch;
		    ws = false;
		}
	    }
	    text += aa;
	} else if (text.length > 0) {
	    if (node_is_block(n)) {
		if (is_ws[text[text.length - 1]]) {
		    if ((text.length > 1) && !is_ws[text[text.length - 2]]) {
			text = text.substring(0, text.length - 1) + "\n\n";
			nls++;
		    }
		} else {
		    text += "\n\n";
		    nls += 2;
		}
	    } else if (n.localName == "br") {
		text += "\n";
		nls++;
	    }
	}
	if (text.length > 600 || nls > 5) {
	    cut = true;
	    break;
	}
    }

    // Strip whitespace from the beginning and end
    let i;
    for (i = 0; i < text.length; i++)
	if (!is_ws[text[i]])
	    break;
    let j;
    if (cut) {
	for (j = Math.min(600, text.length - 1); j > i; j--)
	    if (!is_ws[text[j]])
		break;
	return text.substring(i, j + 1) + "[...]";
    } else {
	for (j = text.length - 1; j > i; j--)
	    if (!is_ws[text[j]])
		break;
	return text.substring(i, j + 1);
    }
}

function get_linktext(link) {
    let href = link.getAttribute("href");
    if (!href)
	return;

    let ioService = Components.classes["@mozilla.org/network/io-service;1"].
	getService(Components.interfaces.nsIIOService);
    let baseuri = ioService.newURI(link.ownerDocument.baseURI, "UTF-8", null).
	QueryInterface(Components.interfaces.nsIURL);
    let linkuri = ioService.newURI(href, "UTF-8", baseuri);

    if (linkuri.schemeIs("http") || linkuri.schemeIs("https") ||
	linkuri.schemeIs("mailto")) {
	if (is_extension)
	    return;
	link.tip_text = href;
	link.tip_color = "#ffe7e7";
	return show_tooltip(link);
    } else if (!linkuri.schemeIs("x-lucidor")) {
	return;
    }

    linkuri.QueryInterface(Components.interfaces.nsIURL);
    if (linkuri.filePath == baseuri.filePath)
	return get_and_show_tooltip(link, link.ownerDocument, linkuri.ref, 2);
    else
	return load_tooltip(link, linkuri.spec, linkuri.ref);
}

function get_and_show_tooltip(link, doc, idref, mmm) {
    let node;
    if (idref) {
	node = doc.getElementById(idref);
	if (!node) {
	    book.add_error("fragment identifier missing: " + idref + " " + mmm);
	    for (let n in loop(doc.getElementsByTagNameNS(XHTML_NS, "a")))
		if (n.hasAttribute("name") && n.getAttribute("name") == idref) {
		    node = n;
		    break;
		}
	}
	if (!node) {
	    link.tip_text = null;
	    return;
	}

	node = get_first_starting_block(node);
    } else {
	node = doc.documentElement;
	for (let n in loop(node.childNodes))
	    if (n.nodeType == node.ELEMENT_NODE &&
		((n.localName == "body" && n.namespaceURI == XHTML_NS) ||
		 (n.localName == "book" && n.namespaceURI == DTB_NS))) {
		node = n;
		break;
	    }
    }
    link.tip_text = get_nodetext(node);
    link.tip_color = "";
    if (link.tip_text)
	show_tooltip(link);
}

var tt_target = null;

function show_tooltip(target) {
    tt_target = target;
    setTimeout(do_show_tooltip, 250);
}

function do_show_tooltip() {
    if (!tt_target)
	return;
    let target = tt_target;
    tt_target = null;
    let tooltip = document.getElementById("tooltip");
    let parent = tooltip.parentNode;
    parent.removeChild(tooltip);
    parent.appendChild(tooltip);
    tooltip.label = target.tip_text;
    if (target.tip_color)
	tooltip.setAttribute("style",
			     "background-color: " + target.tip_color + ";");
    else
	tooltip.removeAttribute("style");
    tooltip.openPopup(target, "after_pointer", -1 , -1, false, false);
}

var link_handler = {
    handleEvent: function(event) {
	let target = event.currentTarget;
	do_remove = false;
	if (target.tip_text) {
	    show_tooltip(target);
	    return;
	}
	if (target.tip_text == undefined)
	    get_linktext(target);
	//if (target.parentNode.localName == "noteref")	// FIXME
    }
};

var do_remove = false;

function remove_tooltip() {
    if (do_remove) {
	let tooltip = document.getElementById("tooltip");
	let parent = tooltip.parentNode;
	parent.removeChild(tooltip);
	parent.appendChild(tooltip);
    }
}

var link_out_handler = {
    handleEvent: function(event) {
	do_remove = true;
	tt_target = null;
	setTimeout(remove_tooltip, 25);
    }
};

function tt_over() {
    do_remove = false;
}

function tt_out() {
    do_remove = true;
    tt_target = null;
    setTimeout(remove_tooltip, 25);
}

function load_tooltip(link, location, fragment) {
    tooltipProgressListener.ok = false;
    let hbrowser = document.getElementById("hidden_browser_tt");
    hbrowser.stop();

    if (hbrowser.currentURI.spec == location)
	return get_and_show_tooltip(link, hbrowser.contentDocument, fragment, 1);

    tooltipProgressListener.link = link;
    tooltipProgressListener.idref = fragment;
    //tooltipProgressListener.location = location;
    try {
	hbrowser.loadURI(location, null, null);
    } catch (e) {
    }
}

var tooltipProgressListener = {
    onLocationChange: function(webProgress, request, location) {
    },

    onProgressChange: function(webProgress, request, curSelfProgress,
			       maxSelfProgress, curTotalProgress,
			       maxTotalProgress) {
    },

    onSecurityChange: function(webProgress, request, state) {
    },

    onStateChange: function(webProgress, request, stateFlags, status) {
	if ((stateFlags & 1) && (stateFlags & 131072)) {
	    let hbrowser = document.getElementById("hidden_browser_tt");
	    if (webProgress.DOMWindow != hbrowser.contentWindow)
		return;
	    this.ok = true;
	} else if ((stateFlags & 16) && (stateFlags & 131072)) {
	    if (!this.ok)
		return;

	    let hbrowser = document.getElementById("hidden_browser_tt");
	    if (webProgress.DOMWindow != hbrowser.contentWindow)
		return;

	    /*
	    request.QueryInterface(Components.interfaces.nsIChannel);
	    if (request.URI.spec != this.location) {
		debug("QERROR");
		return;
	    }
	     */

	    let doc = webProgress.DOMWindow.document;
	    get_and_show_tooltip(this.link, doc, this.idref, 3);
	    this.link = null;
	    this.idref = null;
	    //this.location = null;
	}
    },

    onStatusChange: function(webProgress, request, status, message) {
    },

    QueryInterface: function(aIID) {
	if (aIID.equals(Components.interfaces.nsIWebProgressListener) ||
	    aIID.equals(Components.interfaces.nsISupportsWeakReference))
	    return this;
	throw Components.results.NS_NOINTERFACE;
    }
};

var myPrefObserver = {
  register: function() {
      let prefService = Components.classes["@mozilla.org/preferences-service;1"].
          getService(Components.interfaces.nsIPrefService);
      this._branch = prefService.getBranch("lucidor.");
      this._branch.QueryInterface(Components.interfaces.nsIPrefBranch2);
      this._branch.addObserver("", this, false);
  },

  unregister: function() {
      if (!this._branch)
	  return;
      this._branch.removeObserver("", this);
  },

  observe: function(aSubject, aTopic, aData) {
      if (aTopic != "nsPref:changed")
	  return;

      let prefs = Components.classes["@mozilla.org/preferences-service;1"].
	  getService(Components.interfaces.nsIPrefService).
	  getBranch("lucidor.");
      if (aData == "viewer.border") {
	  view_border = prefs.getBoolPref("viewer.border");
	  set_view_mode(null);
      } else if (aData == "viewer.columnAdjustment") {
	  column_adjustment = prefs.getCharPref("viewer.columnAdjustment");
      }
  }
};

var fakeControlService = {
    init: function(win) {
	let tmp = Components.classes["@mozilla.org/file/directory_service;1"].
	    getService(Components.interfaces.nsIProperties).
	    get("TmpD", Components.interfaces.nsIFile);
	tmp.append("lucifox");
	tmp.createUnique(Components.interfaces.nsIFile.DIRECTORY_TYPE, 0700);
	this.tempdir = tmp;
    },

    get_tempdir: function() {
	return this.tempdir;
    },

    remove_tempfiles: function() {
	debug("remove_tempfiles");
	try {
	    if (this.tempdir && this.tempdir.isDirectory())
		this.tempdir.remove(true);
	} catch (e) {
	}
    },

    online: function() {
    },

    add_tab: function(location) {
	debug("ADD_TAB: " + location);
	let mainWindow = window.QueryInterface(Components.interfaces.nsIInterfaceRequestor)
            .getInterface(Components.interfaces.nsIWebNavigation)
            .QueryInterface(Components.interfaces.nsIDocShellTreeItem)
            .rootTreeItem
            .QueryInterface(Components.interfaces.nsIInterfaceRequestor)
            .getInterface(Components.interfaces.nsIDOMWindow);
	mainWindow.gBrowser.addTab(location);
    },

    add_epub_tab: function(location) {
	return this.add_tab("chrome://lucidor/content/viewepub.xul?uri=" +
			    escape(location));
    },

    add_web_tab: function(location) {
	return this.add_tab(location);
    },

    add_web_tab_post: function(location, post) {
	// FIXME
	let stringStream = Components.classes["@mozilla.org/io/string-input-stream;1"].
            createInstance(Components.interfaces.nsIStringInputStream);
	stringStream.data = post;
	postData = Components.classes["@mozilla.org/network/mime-input-stream;1"].
            createInstance(Components.interfaces.nsIMIMEInputStream);
	postData.addHeader("Content-Type", "application/x-www-form-urlencoded");
	postData.addContentLength = true;
	postData.setData(stringStream);

	let mainWindow = window.QueryInterface(Components.interfaces.nsIInterfaceRequestor)
            .getInterface(Components.interfaces.nsIWebNavigation)
            .QueryInterface(Components.interfaces.nsIDocShellTreeItem)
            .rootTreeItem
            .QueryInterface(Components.interfaces.nsIInterfaceRequestor)
            .getInterface(Components.interfaces.nsIDOMWindow);
	mainWindow.gBrowser.webNavigation.loadURI(location, null, null,
						  postData, null);
    },

    add_web_tab_with_warning: function(location, texts, autohide) {
    },

    open_epub_location: function(location) {
    },

    open_location: function(location) {
    },

    update_bar: function() {
	return;

	let button = document.getElementById("forward-button");
	if (browser.canGoForward)
	    button.setAttribute("disabled", "false");
	else
	    button.setAttribute("disabled", "true");

	button = document.getElementById("back-button");
	if (browser.canGoBack)
	    button.setAttribute("disabled", "false");
	else
	    button.setAttribute("disabled", "true");
    },

    update_tab_title: function(win) {
    },

    toggle_tabbar: function() {
    },

    show_warning: function(win, texts, more_msg, autohide) {
	let msg_box = document.getElementById("message_area");
	this.show_warning_vbox(win, msg_box, texts, more_msg, autohide);
    },

    show_warning_vbox: function(win, msg_box, texts, more_msg, autohide, type) {
	let doc = msg_box.ownerDocument;

	let new_box = doc.createElementNS(XUL_NS, "notification");
	if (type)	// "info", "warning", "critical"
	    new_box.setAttribute("type", type);
	let label = "";
	for (let text_msg in loop(texts)) {
	    if (label)
		label += "\n\n";
	    label += text_msg;
	}
	new_box.setAttribute("label", label);

	if (more_msg) {
	    let button = doc.createElementNS(XUL_NS, "button");
	    button.setAttribute("label", more_msg);
	    button.setAttribute("oncommand", "show_details();");
	    //button.setAttribute("oncommand", "main_control.show_more(event);");
	    //button.setUserData("more", win, null);
	    new_box.appendChild(button);
	}

	msg_box.appendChild(new_box);

	if (autohide) {
	    let this_proxy = this;
	    var tim = {
		notify: function(timer) {
		    try {
			new_box.close();
		    } catch (e) {
		    }
		    tim.timer = null;	// break circular reference
		}
	    };

	    let timer = Components.classes["@mozilla.org/timer;1"].
		createInstance(Components.interfaces.nsITimer);
	    tim.timer = timer;	// NOTE circular reference
	    timer.initWithCallback(tim, 15000, timer.TYPE_ONE_SHOT);
	}
    },

    show_more: function(event) {
	event.target.getUserData("more").show_details();
    },
};

function lucifox_save_file() {
    let file = get_save_file();

    let fp = Components.classes["@mozilla.org/filepicker;1"].
	createInstance(Components.interfaces.nsIFilePicker);
    let strbundle = document.getElementById("strings");
    //let text = strbundle.getString("save_file");
    let text = "Save File";	// FIXME
    fp.init(window, text, Components.interfaces.nsIFilePicker.modeSave);

    let sharedbundle = document.getElementById("shared_strings");
    text = sharedbundle.getString("books");
    fp.appendFilter(text, "*.epub ; *.EPUB");
    text = sharedbundle.getString("all_files");
    fp.appendFilter(text, "*");

    let defaultString = file.leafName;
    try {
	let ios = Components.classes["@mozilla.org/network/io-service;1"].
	    getService(Components.interfaces.nsIIOService);
	let uri = ios.newURI(book.original_location, "UTF-8", null);
	uri.QueryInterface(Components.interfaces.nsIURL);
	defaultString = uri.fileBaseName;
    } catch (e) {
    }
    if (!defaultString)
	defaultString = "book.epub";

    if ((defaultString.length < 5) ||
	(defaultString.substring(defaultString.length - 5).toLowerCase()
	 != ".epub")) {
	defaultString += ".epub";
    }

    fp.defaultString = defaultString;
    fp.defaultExtension = "epub";

    let prefs = Components.classes["@mozilla.org/preferences-service;1"].
	getService(Components.interfaces.nsIPrefService).
	getBranch("lucidor.");
    try {
	let dir = prefs.getComplexValue("lucifox.folder",
					Components.interfaces.nsILocalFile);
	fp.displayDirectory = dir;
    } catch (e) {
    }

    let res = fp.show();

    function my_save_file(targetFile, replace) {
	debug("my_save_file");
	if (!replace && targetFile.exists())
	    return;
	// FIXME honour "replace", it is now OS specific, ok on GNU/Linux
	try {
	    file.copyTo(targetFile.parent, targetFile.leafName);
	} catch (e) {
	    // FIXME localize
	    let text = "An error occurred when trying to save the file.";
	    main_control.show_warning(browser.contentWindow, [ text ],
				      null, false);
	}
    }
    let save_it = my_save_file;

    if (!file)
	save_it = browser.contentWindow.save_file;

    if (res == fp.returnOK)
	save_it(fp.file, false);
    else if (res == fp.returnReplace)
	save_it(fp.file, true);
}

function lucidor_event(evt) {
    let cmd = evt.target.getAttribute("command");
    if (cmd == "previous_part") {
	show_previous_part();
    } else if  (cmd == "next_part") {
	show_next_part();
    } else if  (cmd == "previous_column") {
	previous_column();
    } else if  (cmd == "next_column") {
	next_column();
    }
    document.documentElement.removeChild(evt.target);
}
