/*
 *  Lucidor - e-book reader
 *  Copyright © 2009-2010  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("resource://lucidor/modules/utils.jsm");

const XHTML_NS = "http://www.w3.org/1999/xhtml";
const XUL_NS = "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul";

var is_extension = false;
var bcase = null;

function init() {
    let appinfo = Components.classes["@mozilla.org/xre/app-info;1"].
	getService(Components.interfaces.nsIXULAppInfo);
    if (appinfo.ID == "Lucidor@lucidor.org")
	is_extension = false;
    else
	is_extension = true;

    bcase = Components.classes["@lucidor.org/opdsCase;1"].
        createInstance().wrappedJSObject;
    bcase.set_in_memory(true);
    bcase.init({url: '', db: true}, null);
    current_searcher = true;
    set_current_searcher(null);
    let bookmarks = [/*{ title: "Arxiv",
		       href: "http://arxiv-opds.heroku.com/catalog.atom",
		       icon: "http://arxiv-opds.heroku.com/favicon.ico" },*/
		     { title: "Ebooks libres et gratuits",
		       href: "http://www.ebooksgratuits.com/opds/",
		       icon: null },
		     { title: "Feedbooks",
		       href: "http://feedbooks.com/catalog.atom",
		       icon: "http://www.feedbooks.com/favicon.ico" },
		     { title: "Internet Archive Catalog",
		       href: "http://bookserver.archive.org/catalog/",
		       icon: "http://www.archive.org/favicon.ico" },
		     /*{ title: "Internet Archive Aggregator",
		       href: "http://bookserver.archive.org/aggregator/",
		       icon: "http://www.archive.org/favicon.ico" },*/
		     /*{ title: "PragPub Magazine",
		       href: "http://pragprog.com/magazines.opds",
		       icon: "http://pragprog.com/favicon.ico" },*/
		     { title: "Project Gutenberg",
		       href: "http://lucidor.org/opds/pg.php",
		       icon: "http://lucidor.org/favicon.ico" },
		     /*{ title: "Revues.org",
		       href: "http://bookserver.revues.org/",
		       icon: "http://www.revues.org/favicon.ico" },*/
		     { title: "Lucidor",
		       href: "http://lucidor.org/opds/lucidor.php",
		       icon: "http://lucidor.org/favicon.ico" },
		    ];
    let menu = document.getElementById("bookmarks");
    let sep = document.getElementById("separator1");
    for (let bookmark in loop(bookmarks)) {
	let item = document.createElementNS(XUL_NS, "menuitem");
	item.setAttribute("label", bookmark.title);
	item.setAttribute("value", bookmark.href);
	if (bookmark.icon)
	    item.setAttribute("image", bookmark.icon);
	item.setAttribute("class", "menuitem-iconic");
	item.setAttribute("oncommand", "go_bookmark(event);");
	menu.insertBefore(item, sep);
    }
}

function switch_deck() {
    let deck = document.getElementById("deck");
    let button = document.getElementById("deck_button");
    if (button.checked)
	deck.selectedIndex = 1;
    else
	deck.selectedIndex = 0;
}

function collapse_sidebar() {
    let splitter = document.getElementById("splitter");
    splitter.setAttribute("state", "collapsed");
    //splitter.setAttribute("hidden", "true");
    //let sidebar = document.getElementById("mysidebar");
    //sidebar.setAttribute("hidden", "true");
}

function go_personal() {
    let file;
    let prefs = Components.classes["@mozilla.org/preferences-service;1"].
	getService(Components.interfaces.nsIPrefService).
	getBranch("lucidor.");
    try {
	if (is_extension) {
	    file = prefs.getComplexValue("lucifox.folder",
					 Components.interfaces.nsILocalFile);
	} else {
	    file = prefs.getComplexValue("bookcase.folder",
					 Components.interfaces.nsILocalFile);
	}
    } catch (e) {
	let props = Components.classes["@mozilla.org/file/directory_service;1"].
	    getService(Components.interfaces.nsIProperties);
	if (props.has("Home"))
	    file = props.get("Home", Components.interfaces.nsIFile);
	else
	    return;
    }
    let fph = Components.classes["@mozilla.org/network/protocol;1?name=file"].getService(Components.interfaces.nsIFileProtocolHandler);
    let uri = fph.getURLSpecFromFile(file);
    let location = "x-lucidor-opds://folder/?uri=" + escape(uri);
    open_browse(location, null);
}

function go_bookmark(event) {
    let location = event.target.value;
    open_browse(location, null);
}

function on_go_keypress(event) {
    if (event.keyCode == 13)
	go();
}

function go() {
    let elem = document.getElementById("location_text");
    let location = elem.value;
    let ioService = Components.classes["@mozilla.org/network/io-service;1"].
	getService(Components.interfaces.nsIIOService);
    let uri;
    try {
	try {
	    uri = ioService.newURI(location, "UTF-8", null);
	} catch (e) {
	    location = "http://" + location;
	    uri = ioService.newURI(location, "UTF-8", null);
	}
	if (uri.schemeIs("file")) {
	    let fph = Components.classes["@mozilla.org/network/protocol;1?name=file"].
		getService(Components.interfaces.nsIFileProtocolHandler);
	    dir = fph.getFileFromURLSpec(location);
	    if (dir.isDirectory()) {
		location = "x-lucidor-opds://folder/?uri=" + escape(location);
	    }
	}
    } catch (e) {
	show_searchbar("invalid");
	return;
    }

    open_browse(location, null);
}

function create_search_item(label, value) {
    let it = document.createElement("treeitem");
    let row = document.createElement("treerow");
    let cell = document.createElement("treecell");
    cell.setAttribute("label", label);
    row.appendChild(cell);
    cell = document.createElement("treecell");
    cell.setAttribute("label", value);
    row.appendChild(cell);
    it.appendChild(row);
    return it;
}

function collapse_sidebar() {
    let splitter = document.getElementById("splitter");
    splitter.setAttribute("state", "collapsed");
    //splitter.setAttribute("hidden", "true");
    //let sidebar = document.getElementById("mysidebar");
    //sidebar.setAttribute("hidden", "true");
}

function on_search_keypress(event) {
    if (event.keyCode == 13)
	search();
}

function search() {
    debug("SEARCH");
    let text = document.getElementById("search_text").value;
    if (!text)
	return;

    let filter = document.getElementById("search_filter");
    let searcher = current_searcher;
    if (filter.value.substring(0, 11) == "opensearch:") {
	debug(filter.value);
	searcher = opensearchers[filter.value.substring(11)].searcher;
    }
    let uri = searcher.get_search_uri(text);
    search_add("browse=uri:" + uri, text, "meta", null);
}

/*
 * @value: string used for searching
 * @text: search phrase - only used for display
 * @filter: filter type - only used for display
 * @label: custom label - only used for display (instead of filter)
 */
function search_add(value, text, filter, label) {
    debug("search_add: " + text + " : " + filter + " : " + value + " : " + label);
    let tree = document.getElementById("search_tree");

    let tch = document.getElementById("tree_children");
    let elems = tch.getElementsByTagName("treeitem");	// NOTE only children necessary
    let titem;

    // Check if there is an old item with the same value
    for (let el in loop(elems)) {
	let cell = el.getElementsByTagName("treecell").item(1);
	if (cell.getAttribute("label") == value) {
	    titem = el;
	    tch.removeChild(titem);
	    break;
	}
    }

    // Otherwise create a new item
    if (!titem) {
	if (filter == "browse") {
	    label = text;
	} else {
	    let strbundle = document.getElementById("strings");
	    if (!label)
		try {
		    label = strbundle.getString("filter_" + filter);
		} catch (e) {
		    label = filter;
		}
	    label += ": " + text;
	}
	titem = create_search_item(label, value);
    }

    if (tch.firstChild)
	tch.insertBefore(titem, tch.firstChild);
    else
	tch.appendChild(titem);

    tree.treeBoxObject.ensureRowIsVisible(0);
    tree.view.selection.select(0);

    // Remove oldest item if there are too many
    if (tch.childNodes.length > 8)
	tch.removeChild(tch.lastChild);

    debug("search_add finished");
}

function treeitem_select(event) {
    let bar = document.getElementById("opensearch_bar");
    bar.hidden = true;
    let tree = document.getElementById("search_tree");
    if (tree.currentIndex < 0) {
	// FIXME
	return;
    }

    let col = tree.columns.getLastColumn();
    let value = tree.view.getCellText(tree.currentIndex, col);
    debug("ASDF: " + value);
    let i = value.indexOf("=");
    if (i == -1) {
	//value == "all" || value == "feeds" || value == "untagged" ||
	// value == "top" || value == "recent"
	do_search(value, "", null);
    } else {
	let filter = value.substring(0, i);
	let text = value.substring(i + 1);
	debug("FILTER: " + filter + " TEXT: " + text);
	let node = tree.view.getItemAtIndex(tree.currentIndex);
	if (filter == "browse") {
	    do_browse(text, node);
	} else {
	    do_search(filter, text, null, node);
	}
    }
}

function do_search(filter, text, text2, node) {
    debug("do_search:" + filter + ":" + text + ":" + text2 + ":");

    if (current_searcher) {
	let searchTerms = "";
	let prefix = "";
	if (filter == "author")
	    prefix = "author:";
	else if (filter == "title")
	    prefix = "title:";
	if (current_searcher.syntax["compound"]) {
	    for (let phrase in loop(text.split(" ")))
		searchTerms += prefix + phrase + " ";
	    searchTerms = searchTerms.substring(0, searchTerms.length - 1);
	} else {
	    searchTerms = prefix + text;
	}
	let uri = current_searcher.get_search_uri(searchTerms);
	do_browse("uri:" + uri, node);
	return;
    }

    grid_updated = false;
    if (view_mode == "icons")
	update_grid();
    //update_book_info();
}

function show_searchbar(msg) {
    let text = "";
    if (msg == "searching") {
	text = "Searching...";
    } else if (msg == "failed") {
	text = "Search failed.";
    } else if (msg == "failed unknown") {
	text = "Search failed.";
    } else if (msg == "zero") {
	text = "Search returned no results.";
    } else if (msg == "invalid") {
	text = "Invalid location.";
    }
    let bar = document.getElementById("opensearch_bar");
    bar.hidden = false;
    let label = document.getElementById("opensearch_label");
    label.value = text;

    let button = document.getElementById("prev_results");
    button.disabled = true;
    button = document.getElementById("next_results");
    button.disabled = true;
}

var current_sid = 0;
function new_sid() {
    return ++current_sid;
}

function do_browse(text, node) {
    debug("do_browse: " + text + " : " + node.localName);

    if (bcase.browse_cb) {
	show_searchbar("searching");
	function browsed(gen, sid) {
	    debug("browsed");
	    if (sid == current_sid) {
		if (gen == null) {
		    show_searchbar("failed");
		} else {
		    do_browse_gen(gen, node);
		}
	    }
	}
	bcase.browse_cb(text, browsed, new_sid());
    } else if (bcase.browse_gen) {
	let gen = bcase.browse_gen(text);
	do_browse_gen(gen, node);
    } else {
	throw "Error: cannot browse";
    }
}

function do_browse_gen(gen, node) {
    debug("do_browse_gen");

    if (node.lastChild && node.lastChild.localName == "treechildren")
	while (node.lastChild.hasChildNodes())
	    node.lastChild.removeChild(node.lastChild.firstChild);

    try {
	let box = document.getElementById("content");
	while (box.hasChildNodes())
	    box.removeChild(box.firstChild);

	box = document.getElementById("link_area");
	while (box.hasChildNodes())
	    box.removeChild(box.firstChild);

	let label = document.getElementById("opensearch_label");
	label.value = "";

	let jump = false;
	for (let link in gen) {
	    if (link.thisnode) {
		update_location(link.location);
		if (link.links) {
		    for (let li in loop(link.links)) {
			if (li.type == "opensearch") {
			    add_opensearch(li.href);
			} else if (li.type == "search") {
			    add_stanza_search(li.href, li.title);
			} else if (li.type == "toplink") {
			    add_top_link(li);
			}
		    }
		}
		if (link.firstpath) {
		    node = add_first_link(node, link);
		    jump = true;
		    for (let link in gen)
			;
		}
		if (link.nextpath) {
		    if (!node.hasAttribute("is_first") &&
			!node.hasAttribute("is_next")) {
			node = add_first_link(node, link);
			jump = true;
			add_next_link(node, link);
			for (let link in gen)
			    ;
		    } else if (!node.hasAttribute("has_next")) {
			link.title = "↳ " + link.title;
			add_next_link(node, link);
		    }
		}
		if (link.pageinfo) {
		    // FIXME localize
		    let first = link.pageinfo.startIndex;
		    let total = link.pageinfo.totalResults;
		    let last = -1;
		    if (link.pageinfo.startIndex != -1 &&
			link.pageinfo.itemsPerPage != -1) {
			last = link.pageinfo.startIndex +
			    link.pageinfo.itemsPerPage - 1;
		    }
		    if (last > link.pageinfo.totalResults)
			last = link.pageinfo.totalResults;
		    let text = "";
		    if (first > -1 && last > -1 && total > -1) {
			text = "Showing results " + first +
			    " to " + last + " of " + total;
		    } else if (first > -1 && last > -1) {
			text = "Showing results " + first + " to " + last;
		    }
		    //let label = document.getElementById("opensearch_label");
		    label.value = text;
		}

		let bar = document.getElementById("opensearch_bar");
		if (node.hasAttribute("is_first") ||
		    node.hasAttribute("is_next") ||
		    node.hasAttribute("has_next")) {
		    let button = document.getElementById("prev_results");
		    if (node.hasAttribute("is_next"))
			button.disabled = false;
		    else
			button.disabled = true;
		    button = document.getElementById("next_results");
		    if (node.hasAttribute("has_next"))
			button.disabled = false;
		    else
			button.disabled = true;
		    bar.hidden = false;
		}
	    } else
		add_link(node, link);
	}

	if (jump)
	    child_results();
    } catch (e) {
	debug("--- " + e);
    }
}

function previous_results() {
    let stree = document.getElementById("search_tree");
    let current = stree.view.selection.currentIndex;
    if (current < 1)
	return;
    let view = stree.view.QueryInterface(Components.interfaces.nsITreeContentView);
    let item = view.getItemAtIndex(current);
    if (!item.previousSibling)
	return;
    let index = view.getIndexOfItem(item.previousSibling);
    // This should not be needed
    //if (!stree.view.isContainerOpen(current))
    //	stree.view.toggleOpenState(current);
    stree.view.selection.select(index);
    scroll_to_row(stree.view.selection.tree, index);
}

function next_results() {
    let stree = document.getElementById("search_tree");
    let current = stree.view.selection.currentIndex;
    debug("CURRENT: " + current);
    if (current < 0)
	return;
    let view = stree.view.QueryInterface(Components.interfaces.nsITreeContentView);
    let item = view.getItemAtIndex(current);
    if (!item.nextSibling)
	return;
    let index = view.getIndexOfItem(item.nextSibling);
    // This should not be needed
    //if (!stree.view.isContainerOpen(current))
    //	stree.view.toggleOpenState(current);
    stree.view.selection.select(index);
    scroll_to_row(stree.view.selection.tree, index);
}

function scroll_to_row(tree, index) {
    // FIXME this is not exactly what I want
    //tree.scrollToRow(index);
}

function child_results() {
    let stree = document.getElementById("search_tree");
    let current = stree.view.selection.currentIndex;
    if (current < 0)
	return;
    if (!stree.view.isContainerOpen(current))
	stree.view.toggleOpenState(current);
    stree.view.selection.select(current + 1);
}

function add_first_link(node, link) {
    debug("add_first_link " + link.title + " " + link.type);

    //if (node.hasAttribute("is_next"))
    //	return;

    let ndoc = node.ownerDocument;
    let chld = node.lastChild;
    if (chld.localName != "treechildren") {
	chld = ndoc.createElement("treechildren");
	node.appendChild(chld);
	node.setAttribute("container", "true");
	node.setAttribute("open", "true");
    }

    // ☞»⁝…›↓→↪↳‣
    let cit = create_catalog_item(ndoc, "↳ " + link.title, link.path);
    chld.appendChild(cit);

    cit.setAttribute("is_first", "true");
    //cit.setAttribute("has_next", "true");
    return cit;
}

function add_next_link(node, link) {
    debug("add_next_link " + link.nexttitle + " " + link.type);

    if (node.hasAttribute("has_next"))
	return;

    let ndoc = node.ownerDocument;
    let cit = create_catalog_item(ndoc, "↳ " + link.nexttitle, link.nextpath);
    if (node.nextSibling)
	node.parentNode.insertBefore(cit, node.nextSibling);
    else
	node.parentNode.appendChild(cit);
    cit.setAttribute("is_next", "true");
    node.setAttribute("has_next", "true");
}

function add_link(node, link) {
    debug("add_link " + link.title + " " + link.type);
    for (a in link)
	debug("A:" + a + " " + link[a]);

    if (node && link.type == "folder") {
	debug("CATALOG " + node);
	let ndoc = node.ownerDocument;
	let chld = node.lastChild;
	if (chld.localName != "treechildren") {
	    chld = ndoc.createElement("treechildren");
	    node.appendChild(chld);
	    node.setAttribute("container", "true");
	    node.setAttribute("open", "true");
	}
	let cit = create_catalog_item(ndoc, link.title, link.path);
	chld.appendChild(cit);

	let area = document.getElementById("content");
	let box = document.createElement("vbox");
	area.appendChild(box);
	create_folder_box(link, box, cit);
    } else if (link.type == "epub" || link.type == "link") {
	debug("ITEM");
	if (node)
	    node.my_has_items = true;

	let area = document.getElementById("content");
	let box = document.createElement("vbox");
	area.appendChild(box);
	create_book_box(link, box);
    }
}

function get_book_preview(bi) {
    return null;
}

function open_folder(event, location) {
    debug("open_folder:" + event + " " + location);
    let stree = document.getElementById("search_tree");
    let view = stree.view.QueryInterface(Components.interfaces.nsITreeContentView);
    let node = event.target.getUserData("lucifox");
    debug("node: " + node);
    let index = view.getIndexOfItem(node);
    // This should not be needed
    //if (!stree.view.isContainerOpen(current))
    //	stree.view.toggleOpenState(current);
    stree.view.selection.select(index);
    scroll_to_row(stree.view.selection.tree, index);
}

// FIXME
function xmlescape(uri) {
    return uri.replace("'", "\'", "g");
}

function create_folder_box(bookinfo, book_box, node) {
    // Info
    let infobox = document.createElementNS(XUL_NS, "vbox");
    infobox.setAttribute("flex", "1");
    book_box.appendChild(infobox);

    let info_frame = document.createElementNS(XUL_NS, "vbox");
    info_frame.setAttribute("class", "infoframe");
    infobox.appendChild(info_frame);

    let item = document.createElementNS(XHTML_NS, "div");
    let item2 = document.createElementNS(XHTML_NS, "a");
    item2.setAttribute("class", "link");
    item2.setAttribute("href", "#");
    item2.setAttribute("onclick", "open_folder(event, '" +
		       xmlescape(bookinfo.location) + "');");
    item2.setUserData("lucifox", node, null);
    item2.setAttribute("tooltiptext", bookinfo.location);
    let text = document.createTextNode(bookinfo.title);
    item2.appendChild(text);
    item.appendChild(item2);
    info_frame.appendChild(item);
}

function create_book_box(bookinfo, top_box) {
    function add_line(frame, label, value) {
	let div = document.createElementNS(XHTML_NS, "div");
	//div = document.createElementNS(XUL_NS, "description");
	div.setAttribute("class", "infoitem");
	if (label) {
	    let span = document.createElementNS(XHTML_NS, "span");
	    span.setAttribute("class", "infoheader");
	    div.appendChild(span);
	    let text = document.createTextNode(label + ": ");
	    span.appendChild(text);
	}
	let text = document.createTextNode(value);
	div.appendChild(text);
	frame.appendChild(div);
    }

    let book_box = document.createElementNS(XUL_NS, "hbox");
    top_box.appendChild(book_box);

    // Cover preview
    if (bookinfo.preview) {
	let imgbox = document.createElementNS(XUL_NS, "vbox");
	imgbox.setAttribute("class", "info_cover");
	//let img = get_book_preview(bookinfo);
	let img = document.createElementNS(XUL_NS, "image");
	img.setAttribute("src", bookinfo.preview);
	if (img) {
	    imgbox.appendChild(img);
	    book_box.appendChild(imgbox);
	}
    }

    let item;

    // Info
    let infobox = document.createElementNS(XUL_NS, "vbox");
    infobox.setAttribute("flex", "1");
    book_box.appendChild(infobox);

    let info_frame = document.createElementNS(XUL_NS, "vbox");
    info_frame.setAttribute("class", "infoframe");
    infobox.appendChild(info_frame);

    add_line(info_frame, null, bookinfo.title);

    if (bookinfo.author) {
	item = document.createElementNS(XHTML_NS, "div");
	//item = document.createElementNS(XUL_NS, "description");
	item.setAttribute("class", "author");
	let text = document.createTextNode(bookinfo.author);
	item.appendChild(text);
	info_frame.appendChild(item);
    }

    if (bookinfo.publisher)
	add_line(info_frame, "Publisher", bookinfo.publisher);

    if (bookinfo.language)
	add_line(info_frame, "Language", bookinfo.language);

    if (bookinfo.type == "epub") {
	// FIXME Show bookinfo.location, but with max-width and full in tooltiptext
	// item = document.createElementNS(XHTML_NS, "div");
	// //item = document.createElementNS(XUL_NS, "description");
	// //item.setAttribute("crop", "end");
	// let loc = bookinfo.location;
	// if (loc.length > 70) {
	//     loc = loc.substring(0, 67) + "...";
	//     item.setAttribute("tooltiptext", bookinfo.location);
	// }
	// text = document.createTextNode(loc);
	// item.appendChild(text);
	// //item.setAttribute("value", loc);
	// info_frame.appendChild(item);

	item = document.createElementNS(XHTML_NS, "div");
	item.setAttribute("class", "makeroom");
	let item2 = document.createElementNS(XHTML_NS, "a");
	item2.setAttribute("class", "link");
	item2.setAttribute("href", "#");
	item2.setAttribute("onclick", "open_book('" +
			   xmlescape(bookinfo.location) + "');");
	item2.setAttribute("tooltiptext", bookinfo.location);
	let text = document.createTextNode("Open book");
	item2.appendChild(text);
	item.appendChild(item2);
	info_frame.appendChild(item);
    }

    // Links
    if (bookinfo.links) {
	debug("LINKS!!");
	for (link in loop(bookinfo.links)) {
	    item = document.createElementNS(XHTML_NS, "div");
	    item.setAttribute("class", "makeroom");
	    let item2 = document.createElementNS(XHTML_NS, "a");
	    item2.setAttribute("class", "link");
	    item2.setAttribute("href", "#");
	    let title = link.title;
	    if (!title)
		title = link.href;
	    if (link.type == "application/atom+xml") {
		item2.setAttribute("onclick", "open_browse('" +
				   xmlescape(link.href) + "', '" +
				   xmlescape(title) + "');");
	    } else if (link.type == "application/epub+zip") {
		item2.setAttribute("onclick", "open_book('" +
				   xmlescape(link.href) + "');");
		item2.setAttribute("tooltiptext", link.href);
	    } else {
		item2.setAttribute("onclick", "open_weblink('" +
				   xmlescape(link.href) + "');");
		item2.setAttribute("tooltiptext", link.href);
	    }
	    text = document.createTextNode(title);
	    item2.appendChild(text);
	    item.appendChild(item2);
	    info_frame.appendChild(item);
	}
    }

    /*
    item = document.createElementNS(XUL_NS, "spacer");
    item.setAttribute("flex", 100);
    book_box.appendChild(item);
    */
}

function add_top_link(link) {
    /*if (node)
      node.my_has_items = true;
    */
    let box = document.getElementById("top_links");
    if (!box) {
	let area = document.getElementById("link_area");

	let info_frame = document.createElementNS(XUL_NS, "vbox");
	info_frame.setAttribute("class", "toplinkframe");
	info_frame.setAttribute("flex", "1");
	area.appendChild(info_frame);

	box = document.createElementNS(XHTML_NS, "div");
	box.setAttribute("id", "top_links");
	info_frame.appendChild(box);
    }

    create_link_box(link, box);
}

function create_link_box(link, top_box) {
    let item = document.createElementNS(XHTML_NS, "a");
    item.setAttribute("class", "link toplink");
    item.setAttribute("href", "#");
    item.setAttribute("onclick", "open_browse('" +
		      xmlescape(link.href) + "', '" +
		      xmlescape(link.title) + "');");
    let text = document.createTextNode(link.title);
    item.appendChild(text);
    top_box.appendChild(item);
}

function create_catalog_item(doc, title, path) {
    let it = doc.createElement("treeitem");
    let row = doc.createElement("treerow");
    let cell = doc.createElement("treecell");
    cell.setAttribute("label", title);	// FIXME begränsa längd
    row.appendChild(cell);
    cell = doc.createElement("treecell");
    cell.setAttribute("label", "browse=" + path);
    row.appendChild(cell);
    it.appendChild(row);
    return it;
}

var current_searcher = null;
var opensearchers = {};

function set_current_searcher(new_searcher) {
    if (new_searcher != current_searcher) {
	current_searcher = new_searcher;

	let item = document.getElementById("meta_item");
	if (current_searcher)
	    item.setAttribute("hidden", false);
	else
	    item.setAttribute("hidden", true);

	item = document.getElementById("title_item");
	if (current_searcher && current_searcher.syntax["author"])
	    item.setAttribute("hidden", false);
	else
	    item.setAttribute("hidden", true);

	item = document.getElementById("author_item");
	if (current_searcher && current_searcher.syntax["title"])
	    item.setAttribute("hidden", false);
	else
	    item.setAttribute("hidden", true);

	item = document.getElementById("tags_item");
	if (current_searcher && current_searcher.syntax["tags"])
	    item.setAttribute("hidden", false);
	else
	    item.setAttribute("hidden", true);

	item = document.getElementById("text_item");
	if (current_searcher && current_searcher.syntax["text"])
	    item.setAttribute("hidden", false);
	else
	    item.setAttribute("hidden", true);

	item = document.getElementById("search_text");
	if (current_searcher)
	    item.disabled = false;
	else
	    item.disabled = true;

	item = document.getElementById("search_filter");
	if (current_searcher)
	    item.disabled = false;
	else
	    item.disabled = true;
    }
}

function new_opensearch(se, uri) {
    debug("new_opensearch: " + se + " : " + uri);
    if (se && opensearchers[uri]) {
	debug("OK");
	opensearchers[uri].searcher = se;
	let title = "Search";
	if (se.shortname)
	    title = se.shortname;
	//if (!current_searcher)
	//    show_warning("Search option found", null, true);
	set_current_searcher(se);
    }
}

function add_opensearch(uri) {
    debug("add_opensearch: " + uri);
    if (!(uri in opensearchers)) {
	opensearchers[uri] = { node: null };
	bcase.get_opensearch(uri, new_opensearch);
    }
}

function add_custom_opensearch(uri, title) {
    let top = document.getElementById("text_item");
    let parent = top.parentNode;
    let value = "opensearch:" + uri;
    let item = document.createElement("menuitem");
    item.setAttribute("label", title);
    item.setAttribute("value", value);
    parent.appendChild(item);

    let aitem = document.getElementById("search_filter");
    aitem.selectedItem = item;
    aitem.disabled = false;
    aitem = document.getElementById("search_text");
    aitem.disabled = false;
}

function open_browse(url, text) {
    debug("open_browse: " + url);
    if (!text)
	text = url;
    search_add("browse=uri:" + url, text, "browse", null);
}

function open_book(url) {
    if (is_extension) {
	open_weblink(url);
    } else {
	let main_control = Components.classes["@lucidor.org/mainControlService;1"].
	    getService().wrappedJSObject;
	main_control.add_epub_tab(url);
    }
}

function open_weblink(url) {
    if (is_extension) {
	let mainWindow = window.QueryInterface(Components.interfaces.nsIInterfaceRequestor)
            .getInterface(Components.interfaces.nsIWebNavigation)
            .QueryInterface(Components.interfaces.nsIDocShellTreeItem)
            .rootTreeItem
            .QueryInterface(Components.interfaces.nsIInterfaceRequestor)
            .getInterface(Components.interfaces.nsIDOMWindow);
	mainWindow.gBrowser.addTab(url);
    } else {
	main_control = Components.classes["@lucidor.org/mainControlService;1"].
	    getService().wrappedJSObject;
	main_control.add_web_tab(url);
    }
}

function update_location(url) {
    let elem = document.getElementById("location_text");
    elem.value = url;
}
