/*
 *  Lucidor - e-book reader
 *  Copyright © 2008-2010  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("resource://lucidor/modules/utils.jsm");

const XUL_NS = "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul";

var zoomable = true;
var has_browser = true;
var main_control;

var progressListener = {
    onLocationChange: function(webProgress, request, location) {
	debug("onLocationChange: " + location.spec);
	let browser = document.getElementById("browser");
	if (webProgress.DOMWindow.document == browser.contentDocument) {
	    // eller: if (webProgress.DOMWindow == browser.contentWindow) {
	    let lbox = document.getElementById("location-box");
	    //lbox.value = request.name;
	    lbox.value = location.spec;
	}
    },

    onProgressChange: function(webProgress, request, curSelfProgress,
			       maxSelfProgress, curTotalProgress,
			       maxTotalProgress) {
    },

    onSecurityChange: function(webProgress, request, state) {
    },

    onStateChange: function(webProgress, request, stateFlags, status) {
	/* 
	// This must be done in epubContentSniffer instead
	try {
	    let chan = request.QueryInterface(Components.interfaces.nsIHttpChannel);
	    let head = chan.getResponseHeader("Content-Disposition");
	    debug("HEADER: " + head);
	    chan.setResponseHeader("Content-Disposition", "", false);
	} catch (e) {
	}
	*/
	if ((stateFlags & 16) && (stateFlags & 131072)) {
	    let browser = document.getElementById("browser");
	    if (webProgress.DOMWindow == browser.contentWindow) {
		if (browser.contentTitle)
		    document.title = browser.contentTitle;
		else {
		    let strbundle = document.getElementById("strings");
		    let text = strbundle.getString("untitled");
		    document.title = "[" + text + "]";
		}
		main_control.update_tab_title();
		main_control.update_bar();	// FIXME should not be necessary
	    }
	}
    },

    onStatusChange: function(webProgress, request, status, message) {
    },

    QueryInterface: function(aIID) {
	if (aIID.equals(Components.interfaces.nsIWebProgressListener) ||
	    aIID.equals(Components.interfaces.nsISupportsWeakReference))
	    return this;
	throw Components.results.NS_NOINTERFACE;
    }
};

function update_bar() {
    main_control.update_bar();
}

var historyListener = {
    timeout: 200,	// FIXME hack

    OnHistoryGoBack: function(backURI) {
	setTimeout(update_bar, this.timeout);
	return true;
    },

    OnHistoryGoForward: function(forwardURI) {
	setTimeout(update_bar, this.timeout);
	return true;
    },

    OnHistoryGotoIndex: function(index, gotoURI) {
	setTimeout(update_bar, this.timeout);
	return true;
    },

    OnHistoryNewEntry: function(newURI) {
	setTimeout(update_bar, this.timeout);
	return true;
    },

    OnHistoryPurge: function(numEntries) {
	setTimeout(update_bar, this.timeout);
	return true;
    },

    OnHistoryReload: function(reloadURI, reloadFlags) {
	setTimeout(update_bar, this.timeout);
	return true;
    },

    QueryInterface: function(aIID) {
	if (aIID.equals(Components.interfaces.nsISHistoryListener) ||
	    aIID.equals(Components.interfaces.nsISupportsWeakReference) ||
	    aIID.equals(Components.interfaces.nsISupports))
	    return this;
	throw Components.results.NS_NOINTERFACE;
    }
};

var contentListener;

function init() {
    main_control = Components.classes["@lucidor.org/mainControlService;1"].
	getService().wrappedJSObject;

    contentListener = Components.classes["@lucidor.org/epubURIContentListener;1"].createInstance(Components.interfaces.nsIURIContentListener);
    /*
    let oldcontentListener = browser.docShell.QueryInterface(Components.interfaces.nsIInterfaceRequestor).getInterface(Components.interfaces.nsIURIContentListener);
    oldcontentListener.parentContentListener = contentListener;
    */
    let uriLoader = Components.classes["@mozilla.org/uriloader;1"].
	getService(Components.interfaces.nsIURILoader);
    uriLoader.registerContentListener(contentListener);

    let fph = Components.classes["@mozilla.org/network/protocol;1?name=file"].
	getService(Components.interfaces.nsIFileProtocolHandler);

    // Get Bookcase Folder
    let bookcaseuri = "";
    let prefs = Components.classes["@mozilla.org/preferences-service;1"].
	getService(Components.interfaces.nsIPrefService).
	getBranch("lucidor.");
    try {
	let file = prefs.getComplexValue("bookcase.folder",
					 Components.interfaces.nsILocalFile);
	bookcaseuri = "chrome://lucidor/content/booklist.xul?" +
	    fph.getURLSpecFromFile(file);
    } catch(e) {
    }

    // Get Home Folder
    let props = Components.classes["@mozilla.org/file/directory_service;1"].
	getService(Components.interfaces.nsIProperties);
    let home = props.get("Home", Components.interfaces.nsIFile);
    let homeuri = fph.getURLSpecFromFile(home);

    let menupopup = document.getElementById("bookmarks");
    let a = ["Local Sample Books", "chrome://lucidor/content/booklist.xul",
	     "Local Bookcase", bookcaseuri,
	     "Home Folder", homeuri,
	     "-", "-",
	     "lucidor.org", "http://lucidor.org/"
	     //"FeedBooks", "http://feedbooks.com/"
	     ];
    for (let i = 0; i < a.length; i += 2) {
	if (!a[i + 1])
	    continue;
	if (a[i] == "-") {
	    let menuitem = document.createElementNS(XUL_NS, "menuseparator");
	    menupopup.appendChild(menuitem);
	} else {
	    let menuitem = document.createElementNS(XUL_NS, "menuitem");
	    menuitem.setAttribute("label", a[i]);
	    menuitem.setAttribute("tooltiptext", a[i + 1]);
	    menuitem.setAttribute("oncommand", "go_to('" + a[i + 1] + "');");
	    menupopup.appendChild(menuitem);
	}
    }

    let ioService = Components.classes["@mozilla.org/network/io-service;1"].
	getService(Components.interfaces.nsIIOService);
    let docuri = ioService.newURI(document.documentURI, "UTF-8", null);
    docuri.QueryInterface(Components.interfaces.nsIURL);

    let uri = null;
    let postparams = null;
    let params = docuri.query.split("&");
    for (let par in loop(params)) {
	let pair = par.split("=");
	if (pair[0] == "uri")
	    uri = unescape(pair[1]);
	else if (pair[0] == "post")
	    postparams = unescape(pair[1]);
    }

    debug("webview: " + uri);

    let lbox = document.getElementById("location-box");
    lbox.value = uri;

    let browser = document.getElementById("browser");
    browser.addProgressListener(progressListener);
    //browser.docShell.allowAuth = false;
    //browser.docShell.allowJavascript = false;
    //browser.docShell.allowMetaRedirects = false;
    //browser.docShell.allowPlugins = false;
    browser.docShell.useErrorPages = true;
    browser.sessionHistory.addSHistoryListener(historyListener);
    let postData = null;
    if (postparams) {
	let stringStream = Components.classes["@mozilla.org/io/string-input-stream;1"].
            createInstance(Components.interfaces.nsIStringInputStream);
	stringStream.data = postparams;
	postData = Components.classes["@mozilla.org/network/mime-input-stream;1"].
            createInstance(Components.interfaces.nsIMIMEInputStream);
	postData.addHeader("Content-Type", "application/x-www-form-urlencoded");
	postData.addContentLength = true;
	postData.setData(stringStream);
    }

    if (uri != null)
	//browser.loadURI(uri, null, null);
	//browser.loadURIWithFlags(uri, null, null, null, postData);
	// ¿eller?
	browser.webNavigation.loadURI(uri, null, null, postData, null);
}

function get_browser() {
    return document.getElementById("browser");
}

function getMarkupDocumentViewer() {
    return document.getElementById("browser").markupDocumentViewer;
}

function on_location_keypress(event) {
    if (event.keyCode == 13)
	go();
}

function go() {
    let lbox = document.getElementById("location-box");
    let browser = document.getElementById("browser");
    browser.loadURI(lbox.value, null, null);
}

function go_to(location) {
    let lbox = document.getElementById("location-box");
    lbox.value = location;
    let browser = document.getElementById("browser");
    browser.loadURI(lbox.value, null, null);
}

function stop() {
    let browser = document.getElementById("browser");
    browser.stop();
}

function toggle_tabbar() {
    let val = window.parent.toggle_tabbar();
    let item = document.getElementById("tabbar_item");
    if (val)
	item.setAttribute("checked", "false");
    else
	item.setAttribute("checked", "true");
}

function toggle_locationbar() {
    let locationbar = document.getElementById("locationbar");
    let val = !locationbar.hidden;
    locationbar.hidden = val;

    let item = document.getElementById("locationbar_item");
    if (val)
	item.setAttribute("checked", "false");
    else
	item.setAttribute("checked", "true");
}

function popupShowing() {
    let element = document.popupNode;

    // FIXME a med img (eller annat som är tillåtet i a) inuti?

    let hide = true;
    if (element.localName.toLowerCase() == "a")
	hide = false;
    for (let name in loop(["open_link_item", "save_link_item"])) {
	let item = document.getElementById(name);
	item.setAttribute("hidden", hide);
    }

    /*
    item = document.getElementById("save_image_item");
    hide = true;
    if (element.localName.toLowerCase() == "img")
	hide = false;
    item.setAttribute("hidden", hide);
    */
}

function open_link_tab() {
    let element = document.popupNode;
    let href = element.getAttribute("href");

    let ios = Components.classes['@mozilla.org/network/io-service;1'].
	getService(Components.interfaces.nsIIOService);
    let baseuri = null;
    try {
	baseuri = ios.newURI(element.ownerDocument.baseURI, null, null);
    } catch (e) {
    }
    let location = href;
    try {
	let uri = ios.newURI(href, null, baseuri);
	location = uri.spec;
    } catch (e) {
    }

    main_control.add_web_tab(location);
}

var proglisten =  {
    onStateChange: function(aWebProgress, aRequest, aStateFlags, aStatus) {
	if (aStateFlags & 0x00000010) {	// STATE_STOP
	    let ok;
	    let name = "";
	    try {
		name = aRequest.name;
		aRequest.QueryInterface(Components.interfaces.nsIHttpChannel);
		if (aRequest.requestSucceeded)
		    ok = 0;
		else
		    ok = 1;
	    } catch (e) {
		ok = aStatus;
	    }
	    debug("STATUS: " + ok + ":" + aStatus);

	    if (ok == 0) {
		let text = "Download finished: " + name;
		main_control.show_warning(window, [text], null, true);
	    } else {
		let text = "Download failed: " + name;
		main_control.show_warning(window, [text], null, false);
	    }
	}
    },

    onProgressChange: function(aWebProgress, aRequest, aCurSelfProgress,
			       aMaxSelfProgress, aCurTotalProgress,
			       aMaxTotalProgress) {
    },

    onLocationChange: function(aWebProgress, aRequest, aLocation) {
    },

    onStatusChange: function(aWebProgress, aRequest, aStatus, aMessage) {
    },

    onSecurityChange: function(aWebProgress, aRequest, aState) {
    }
};

function prompt_file(fn) {
    let fp = Components.classes["@mozilla.org/filepicker;1"].
	createInstance(Components.interfaces.nsIFilePicker);
    //let strbundle = document.getElementById("strings");
    //let title = strbundle.getString("select_path.title");
    let title = "Save link";
    fp.init(window, title, Components.interfaces.nsIFilePicker.modeSave);
    if (fn)
	fp.defaultString = fn;
    let res = fp.show();
    if (res == fp.returnOK || res == fp.returnReplace)
	return fp.fileURL;
    return null;
}

function save_page() {
    let browser = document.getElementById("browser");
    let ios = Components.classes['@mozilla.org/network/io-service;1'].
	getService(Components.interfaces.nsIIOService);
    let source_uri = null;
    let fn = null;
    try {
	let uri = browser.currentURI;
	if (uri.schemeIs("view-source")) {
	    uri = ios.newURI(uri.path, null, null);
	    source_uri = uri;
	}
	uri.QueryInterface(Components.interfaces.nsIURL);
	fn = unescape(uri.fileName);
    } catch (e) {
    }

    let fileURL = prompt_file(fn);
    if (!fileURL)
	return;

    let wbp = Components.classes['@mozilla.org/embedding/browser/nsWebBrowserPersist;1'].createInstance(Components.interfaces.nsIWebBrowserPersist);
    //wbp.progressListener = proglisten;
    if (source_uri)
	try {
	    wbp.saveURI(source_uri, null, null, null, null, fileURL);
	} catch (e) {
	    let text = "Save failed: " + source_uri.spec;
	    main_control.show_warning(window, [text], null, false);
	}
    else
	wbp.saveDocument(browser.contentDocument, fileURL,
			 null, null, null, null);
}

function save_link() {
    let element = document.popupNode;
    let href = element.getAttribute("href");

    let ios = Components.classes['@mozilla.org/network/io-service;1'].
	getService(Components.interfaces.nsIIOService);
    let baseuri = null;
    try {
	baseuri = ios.newURI(element.ownerDocument.baseURI, null, null);
    } catch (e) {
    }
    let uri = ios.newURI(href, null, baseuri);

    let fn = null;
    try {
	uri.QueryInterface(Components.interfaces.nsIURL);
	fn = unescape(uri.fileName);
    } catch (e) {
    }

    let fileURL = prompt_file(fn);
    if (!fileURL)
	return;

    let wbp = Components.classes['@mozilla.org/embedding/browser/nsWebBrowserPersist;1'].createInstance(Components.interfaces.nsIWebBrowserPersist);
    /*
    let dm = Components.classes["@mozilla.org/download-manager;1"].
	getService(Components.interfaces.nsIDownloadManager);
    let dn = dm.addDownload(0, uri, fileURL, "....", null,
			    Date().UTC() * 1000, null, wbp);
    wbp.progressListener = dn;
    */

    wbp.progressListener = proglisten;

    try {
	wbp.saveURI(uri, null, null, null, null, fileURL);
    } catch (e) {
	let text = "Download failed: " + uri.spec;
	main_control.show_warning(window, [text], null, false);
    }
}
