/*
 *  Lucidor - e-book reader
 *  Copyright © 2009  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

var EXPORTED_SYMBOLS = [ "init_cache", "get_from_cache", "put_in_cache" ];

function init_cache(this_proxy, name) {
    let ok = false;
    try {
	let prefs = Components.classes["@mozilla.org/preferences-service;1"].
	    getService(Components.interfaces.nsIPrefService).
	    getBranch("lucidor.");
	let value = prefs.getBoolPref("debug.cache");
	ok = value;
    } catch (e) {
    }

    if (!ok)
	return false;

    if (!this_proxy.debug_cache) {
	let file = Components.classes["@mozilla.org/file/directory_service;1"].
	    getService(Components.interfaces.nsIProperties).
	    get("ProfD", Components.interfaces.nsIFile);
	let fn = "debug_cache";
	if (name)
	    fn = name;
	file.append(fn + ".sqlite");
	let ss = Components.classes["@mozilla.org/storage/service;1"].
	    getService(Components.interfaces.mozIStorageService);
	let db = ss.openDatabase(file);
	if (!db.tableExists("cache")) {
	    db.createTable("cache", "id INTEGER NOT NULL PRIMARY KEY, uri TEXT, doc TEXT");
	}
	this_proxy.debug_cache = db;
    }
    return true;
}

function get_from_cache(this_proxy, url) {
    if (!this_proxy.debug_cache)
	return null;

    let doc = null;
    let statement = this_proxy.debug_cache.createStatement("SELECT doc FROM cache WHERE uri=?1");
    try {
	statement.bindUTF8StringParameter(0, url);
	if (statement.executeStep())
	    doc = statement.getUTF8String(0);
    } finally {
	statement.reset();
    }
    return doc;
}

function put_in_cache(this_proxy, url, doc) {
    if (!this_proxy.debug_cache)
	return;

    let statement = this_proxy.debug_cache.createStatement("INSERT INTO cache (id, uri, doc) VALUES (NULL, ?1, ?2)");
    statement.bindUTF8StringParameter(0, url);
    statement.bindUTF8StringParameter(1, doc);
    statement.execute();
}
