/*
 *  Lucidor - e-book reader
 *  Copyright © 2009-2010  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

function debug(msg) {
    if (false) {
	try {
	    let cons = Components.classes["@mozilla.org/consoleservice;1"].
		getService(Components.interfaces.nsIConsoleService);
	    cons.logStringMessage(msg);
	} catch (e) {
	}
    }
}

function loop(thing) {
    if (!thing)
	return;
    for (let i = 0; i < thing.length; i++)
	yield thing[i];
}

const ATOM_NS = "http://www.w3.org/2005/Atom";
const DC_NS = "http://purl.org/dc/terms";
const DCE_NS = "http://purl.org/dc/elements/1.1/";
const GBS_NS = "http://schemas.google.com/books/2008";
const OS_NS = "http://a9.com/-/spec/opensearch/1.1/";
const LUCI_NS = "http://lucidor.org/-/x-opds/";

var is_ws = [];
is_ws[" "] = true;
is_ws["\n"] = true;
is_ws["\t"] = true;
is_ws["\r"] = true;

function get_text(node) {
    let text = node.textContent;

    // Strip beginning and ending whitespace
    let i;
    for (i = 0; i < text.length; i++)
	if (!is_ws[text[i]])
	    break;
    let j;
    for (j = text.length - 1; j > i; j--)
	if (!is_ws[text[j]])
	    break;
    return text.substring(i, j + 1);
}

function atomProcessor() {
    this.wrappedJSObject = this;
    this.listener = null;
}

atomProcessor.prototype = {
    classDescription: "atomProcessor",
    classID: Components.ID("{099af2a8-ae0c-4a0c-b726-e85f481bf6e5}"),
    contractID: "@lucidor.org/atomProcessor;1",
    QueryInterface: XPCOMUtils.generateQI([Components.interfaces.nsISupports]),

    parseFromString: function(text, uri) {
	let parser = Components.classes["@mozilla.org/xmlextras/domparser;1"].
	    getService(Components.interfaces.nsIDOMParser);
	let doc = parser.parseFromString(text, "application/xml");
	let doc_elem = doc.documentElement;
	if ((doc_elem.localName != "entry") || (doc_elem.namespaceURI != ATOM_NS)) {
	    listener.handleResult(null);
	    return;
	}

	let entry = {};
	entry.fields = Components.classes["@mozilla.org/hash-property-bag;1"].
	    createInstance(Components.interfaces.nsIWritablePropertyBag2);
	entry.authors = Components.classes["@mozilla.org/array;1"].
	    createInstance(Components.interfaces.nsIMutableArray);
	let links = Components.classes["@mozilla.org/array;1"].
	    createInstance(Components.interfaces.nsIMutableArray);
	entry.fields.setPropertyAsInterface("links", links);
	entry.baseURI = uri;	// FIXME
	for (let node in loop(doc_elem.childNodes)) {
	    if (!(node.nodeType == doc_elem.ELEMENT_NODE))
		continue;
	    if (node.namespaceURI == ATOM_NS) {
		if (node.localName == "id") {
		    entry.id = node.textContent;
		} else if (node.localName == "summary") {
		    // FIXME
		} else if (node.localName == "updated") {
		    // FIXME
		} else if (node.localName == "category") {
		    // FIXME <category term="girlfriend"/>
		} else if (node.localName == "author") {
		    let person = Components.classes["@mozilla.org/feed-person;1"].
			createInstance(Components.interfaces.nsIFeedPerson);
		    for (let ch in loop(node.childNodes)) {
			if (ch.localName == "name")
			    person.name = ch.textContent;
		    }
		    entry.authors.appendElement(person, false);
		} else if (node.localName == "title") {
		    if (!entry.title)
			entry.title = { plainText: function() {
					    return node.textContent;
					} };
		} else if (node.localName == "link") {
		    let link = Components.classes["@mozilla.org/hash-property-bag;1"].
			createInstance(Components.interfaces.nsIWritablePropertyBag);
		    for (let att in loop(node.attributes)) {
			att.QueryInterface(Components.interfaces.nsIDOMAttr);
			link.setProperty(att.name, att.value);
		    }
		    links.appendElement(link, false);
		}
	    } else if (node.namespaceURI == GBS_NS) {
		if (node.localName == "embeddability") {
		} else if (node.localName == "openAccess") {
		} else if (node.localName == "viewability") {
		}
	    } else if (node.namespaceURI == DC_NS || node.namespaceURI == DCE_NS) {
		if (node.localName == "creator") {
		    if (!entry.fields.hasKey("nullcreator"))
			entry.fields.setPropertyAsAUTF8String("nullcreator", node.textContent);
		} else if (node.localName == "date") {
		    if (!entry.fields.hasKey("nulldate"))
			entry.fields.setPropertyAsAUTF8String("nulldate", node.textContent);
		} else if (node.localName == "format") {
		    if (!entry.fields.hasKey("nullformat"))
			entry.fields.setPropertyAsAUTF8String("nullformat", node.textContent);
		} else if (node.localName == "identifier") {
		    if (!entry.fields.hasKey("nullidentifier"))
			entry.fields.setPropertyAsAUTF8String("nullidentifier", node.textContent);
		} else if (node.localName == "issued") {
		    if (!entry.fields.hasKey("nullissued"))
			entry.fields.setPropertyAsAUTF8String("nullissued", node.textContent);
		} else if (node.localName == "language") {
		    if (!entry.fields.hasKey("nulllanguage"))
			entry.fields.setPropertyAsAUTF8String("nulllanguage", node.textContent);
		} else if (node.localName == "publisher") {
		    if (!entry.fields.hasKey("nullpublisher"))
			entry.fields.setPropertyAsAUTF8String("nullpublisher", node.textContent);
		} else if (node.localName == "title") {
		    if (!entry.fields.hasKey("nulltitle"))
			entry.fields.setPropertyAsAString("nulltitle", node.textContent);
		}
	    }
	}

	if (this.listener)
	    this.listener.handleResult({ doc: entry });
    }
};

var components = [atomProcessor];

if (XPCOMUtils.generateNSGetFactory)
    var NSGetFactory = XPCOMUtils.generateNSGetFactory(components);
else
    var NSGetModule = XPCOMUtils.generateNSGetModule(components);
