/*
 *  Lucidor - e-book reader
 *  Copyright © 2008-2010  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

function debug(msg) {
    if (false) {
	try {
	    let cons = Components.classes["@mozilla.org/consoleservice;1"].
		getService(Components.interfaces.nsIConsoleService);
	    cons.logStringMessage(msg);
	} catch (e) {
	}
    }
}

function loop(thing) {
    if (!thing)
	return;
    for (let i = 0; i < thing.length; i++)
	yield thing[i];
}

function bookcase() {
    this.wrappedJSObject = this;
}

bookcase.prototype = {
    classDescription: "bookcaseLite",
    classID: Components.ID("{f122caeb-d59a-4e42-9b6d-389f72c0c924}"),
    contractID: "@lucidor.org/bookcaseLiteService;1",
    _xpcom_factory: bookcaseFactory,
    QueryInterface: XPCOMUtils.generateQI([Components.interfaces.nsISupports]),

    db: null,
    file_uri: null,

    get_db_file: function() {
	let file = Components.classes["@mozilla.org/file/directory_service;1"].
		getService(Components.interfaces.nsIProperties).
		get("ProfD", Components.interfaces.nsIFile);
	file.append("lucifox_bookcase.sqlite");

	let fph = Components.classes["@mozilla.org/network/protocol;1?name=file"].getService(Components.interfaces.nsIFileProtocolHandler);
	this.file_uri = fph.getURLSpecFromFile(file);
	debug("DB_FILE: " + this.file_uri);
	return file;
    },

    open: function() {
	if (this.db)
	    return;
	let file = this.get_db_file();
	let ss = Components.classes["@mozilla.org/storage/service;1"].
		getService(Components.interfaces.mozIStorageService);
	this.db = ss.openDatabase(file);

	if (!this.db.tableExists("books"))
	    this.create();
    },

    create: function() {
	this.db.createTable("books", "id INTEGER PRIMARY KEY, sha1 TEXT, date_read TEXT, bookmark TEXT");
	this.db.schemaVersion = 1;
    },

    add_open_book: function(book, file) {
	debug("YY " + book.original_location + " " + book.location);

	return this.internal_add_open_book(book);
    },

    add_book_data: function(book) {
	let statement = this.db.createStatement("INSERT INTO books (id, sha1, date_read) VALUES (NULL, ?1, datetime('now'))");
	statement.bindUTF8StringParameter(0, book.sha1);
	statement.execute();
	return this.db.lastInsertRowID;
    },

    update_date_read: function(id) {
	let statement = this.db.createStatement("UPDATE books SET date_read=datetime('now') WHERE id=?1");
	statement.bindUTF8StringParameter(0, id);
	statement.execute();
    },

    update_last_position: function(id, position) {
	let statement = this.db.createStatement("UPDATE books SET bookmark=?1 WHERE id=?2");
	statement.bindUTF8StringParameter(0, position);
	statement.bindUTF8StringParameter(1, id);
	statement.execute();
    },

    remove_book: function(id) {
	let statement = this.db.createStatement("DELETE FROM books WHERE id=?1");
	statement.bindUTF8StringParameter(0, id);
	statement.execute();
    },

    get_id_from_sha1: function(sha1) {
	let statement = this.db.createStatement("SELECT id FROM books WHERE sha1=?1");
	let id = -1;
	try {
	    statement.bindUTF8StringParameter(0, sha1);
	    if (statement.executeStep())
		id = statement.getInt32(0);
	} finally {
	    statement.reset();
	}
	return id;
    },

    get_book_info: function(id) {
	debug("bookcase.get_book_info: " + id);

	let book;
	let statement = this.db.createStatement("SELECT id, uid, title, author, publisher, location, type FROM books WHERE id=?1");
	try {
	    statement.bindInt32Parameter(0, id);
	    if (statement.executeStep()) {
		book = {};
		book.id = statement.getInt32(0);
		//book.uid = statement.getUTF8String(1);
		book.title = statement.getUTF8String(2);
		book.author = statement.getUTF8String(3);
		book.publisher = statement.getUTF8String(4);
		book.location = statement.getUTF8String(5);
		book.type = statement.getUTF8String(6);
	    }
	} finally {
	    statement.reset();
	}
	return book;
    },

    get_last_position: function(id) {
	debug("bookcase.get_last_position: " + id);

	let position;
	let statement = this.db.createStatement("SELECT bookmark FROM books WHERE id=?1");
	try {
	    statement.bindInt32Parameter(0, id);
	    if (statement.executeStep()) {
		position = statement.getUTF8String(0);
	    }
	} finally {
	    statement.reset();
	}
	return position;
    }
};

var bookcaseService = new bookcase();

var bookcaseFactory = {
    createInstance: function(outer, aIID) {
	if (outer != null)
	    throw Components.results.NS_ERROR_NO_AGGREGATION;
	return bookcaseService.QueryInterface(aIID);
    }
};

var components = [bookcase];

if (XPCOMUtils.generateNSGetFactory)
    var NSGetFactory = XPCOMUtils.generateNSGetFactory(components);
else
    var NSGetModule = XPCOMUtils.generateNSGetModule(components);
