/*
 *  Lucidor - e-book reader
 *  Copyright © 2008-2010  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

function debug(msg) {
    if (false) {
	try {
	    let cons = Components.classes["@mozilla.org/consoleservice;1"].
		getService(Components.interfaces.nsIConsoleService);
	    cons.logStringMessage(msg);
	} catch (e) {
	}
    }
}

function loop(thing) {
    if (!thing)
	return;
    for (let i = 0; i < thing.length; i++)
	yield thing[i];
}

function bookcase() {
    this.wrappedJSObject = this;
}

bookcase.prototype = {
    classDescription: "bookcaseSql",
    classID: Components.ID("{e20c9bef-4e16-4f7b-989a-08e40ce4298a}"),
    contractID: "@lucidor.org/bookcaseSqlService;1",
    _xpcom_factory: bookcaseFactory,
    QueryInterface: XPCOMUtils.generateQI([Components.interfaces.nsISupports]),

    db: null,
    file_uri: null,

    get_books_directory: function() {
	let file;
	let prefs = Components.classes["@mozilla.org/preferences-service;1"].
		getService(Components.interfaces.nsIPrefService).
		getBranch("lucidor.");
	try {
	    file = prefs.getComplexValue("bookcase.folder",
					 Components.interfaces.nsILocalFile);
	} catch (e) {
	    let props = Components.classes["@mozilla.org/file/directory_service;1"].
		getService(Components.interfaces.nsIProperties);

	    // None?
	    if (!file && props.has("Docs"))
		file = props.get("Docs", Components.interfaces.nsIFile);
	    // Windows
	    if (!file && props.has("Pers"))
		file = props.get("Pers", Components.interfaces.nsIFile);
	    // MacOSX, Darwin
	    if (!file && props.has("UsrDocs"))
		file = props.get("UsrDocs", Components.interfaces.nsIFile);
	    //if (!file && props.has("LocDocs"))
	    //	file = props.get("LocDocs", Components.interfaces.nsIFile);
	    // Unices
	    if (!file && props.has("XDGDocs"))
		file = props.get("XDGDocs", Components.interfaces.nsIFile);
	    // Fallback
	    if (!file) {
		debug("No OS specific docs directory found");
		if (props.has("Home"))
		    file = props.get("Home", Components.interfaces.nsIFile);
		else
		    file = props.get("ProfD", Components.interfaces.nsIFile);
		file.append("Documents");
	    }

	    file.append("Lucidor");
	    prefs.setComplexValue("bookcase.folder",
				  Components.interfaces.nsILocalFile, file);
	}
	debug("Bookcase directory: " + file.path);
	return file;
    },

    get_db_file: function() {
	let file = Components.classes["@mozilla.org/file/directory_service;1"].
		getService(Components.interfaces.nsIProperties).
		get("ProfD", Components.interfaces.nsIFile);
	file.append("bookcase.sqlite");

	let fph = Components.classes["@mozilla.org/network/protocol;1?name=file"].getService(Components.interfaces.nsIFileProtocolHandler);
	this.file_uri = fph.getURLSpecFromFile(file);
	debug("DB_FILE: " + this.file_uri);
	return file;
    },

    open: function() {
	if (this.db)
	    return;
	let file = this.get_db_file();
	let ss = Components.classes["@mozilla.org/storage/service;1"].
		getService(Components.interfaces.mozIStorageService);
	this.db = ss.openDatabase(file);

	if (!this.db.tableExists("books"))
	    this.create();
	if (this.db.schemaVersion != 2)
	    this.update();
    },

    create: function() {
	this.db.createTable("books", "id INTEGER PRIMARY KEY, uid TEXT, title TEXT, author TEXT, publisher TEXT, language TEXT, location TEXT, sha1 TEXT, type TEXT, tags TEXT, preview TEXT, date_added TEXT, date_read TEXT, bookmark TEXT");
	this.db.schemaVersion = 2;
    },

    update: function() {
	// schemaVersion = 1 (Lucidor 0.5):
	// "books", "id INTEGER PRIMARY KEY, uid TEXT, title TEXT, author TEXT, publisher TEXT, language TEXT, location TEXT, sha1 TEXT, type TEXT, tags TEXT, preview TEXT, date_added TEXT, date_read TEXT"
	//
	// schemaVersion = 2 (Lucidor 0.6, 0.7, 0.8):
	// "books", "id INTEGER PRIMARY KEY, uid TEXT, title TEXT, author TEXT, publisher TEXT, language TEXT, location TEXT, sha1 TEXT, type TEXT, tags TEXT, preview TEXT, date_added TEXT, date_read TEXT, bookmark TEXT"

	if (this.db.schemaVersion > 2) {
	    //FIXME
	    return;
	}

	if (this.db.schemaVersion < 2) {
	    let statement = this.db.createStatement("ALTER TABLE books ADD COLUMN bookmark TEXT");
	    statement.execute();
	    this.db.schemaVersion = 2;
	}
    },

    add_book: function(file) {
	debug("bookcase.add_book: " + file);
	let book = Components.classes["@lucidor.org/epubDocument;1"].
		getService().wrappedJSObject;
	try {
	    book.init(file);
	} catch (e) {
	    return -1;
	}
	return this.add_open_book(book, file);
    },

    add_open_book: function(book, file) {
	debug("YY " + book.original_location + " " + book.location);
	let prefs = Components.classes["@mozilla.org/preferences-service;1"].
		getService(Components.interfaces.nsIPrefService).
		getBranch("lucidor.");
	let copy_local = prefs.getBoolPref("bookcase.copyLocal");
	let copy_remote = prefs.getBoolPref("bookcase.copyRemote");

	let leaf = "unnamed.epub";
	let fph = Components.classes["@mozilla.org/network/protocol;1?name=file"].getService(Components.interfaces.nsIFileProtocolHandler);
	let ori_uri = null;
	try {
	    let ioService = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);
	    ori_uri = ioService.newURI(book.original_location, "UTF-8", null).
		QueryInterface(Components.interfaces.nsIURL);
	    leaf = ori_uri.fileName;
	} catch (e) {
	}

	// Note that "chrome:" uris are treated as remote
	if (ori_uri && ori_uri.schemeIs("file") && !copy_local) {
	    // In this case "file" is always at book.original_location
	    //let ori_file = fph.getFileFromURLSpec(book.original_location);
	    let tmp = Components.classes["@mozilla.org/file/directory_service;1"].
		getService(Components.interfaces.nsIProperties).
		get("TmpD", Components.interfaces.nsIFile);
	    // Always copy file if it is in a temporary directory
	    let path = file.path.toLowerCase();
	    if (!(tmp.contains(file, true) ||
		  path.indexOf("/tmp/") == 0 ||	// Unix
		  path.indexOf("/temp/") == 0 ||	// FIXME Mac OS X?
		  path.indexOf("c:\\windows\\temp\\") == 0)) {	// FIXME Windows
		return this.internal_add_open_book(book);
	    }
	} else if (!copy_remote)
	    return this.internal_add_open_book(book);

	let dir = this.get_books_directory();
	if (!dir.exists()) {
	    try {
		dir.create(dir.DIRECTORY_TYPE, 0700);
	    } catch (e) {
		throw e;
	    }
	}

	debug("DIR: " + dir.path + " : " + leaf);
	let ok = false;

	try {
	    let target = dir.clone();
	    target.append(leaf);
	    if (!target.exists()) {
		file.copyTo(dir, leaf);
		ok = true;
	    }
	} catch (e) {
	}

	if (!ok) {
	    let leaf_name = leaf;
	    let i = leaf.lastIndexOf(".");
	    if (i > 0)
		leaf_name = leaf.substring(0, i);
	    let rand = Math.floor(Math.random() * 1000000);
	    leaf = leaf_name + "-" + rand.toString() + ".epub";
	    try {
		let target = dir.clone();
		target.append(leaf);
		if (target.exists())
		    throw "Exists";
		file.copyTo(dir, leaf);
	    } catch (e) {
		throw e;
	    }
	}

	dir.append(leaf);
	book.original_location = fph.getURLSpecFromFile(dir);

	return this.internal_add_open_book(book);
    },

    internal_add_open_book: function(book) {
	let statement = this.db.createStatement("INSERT INTO books (id, uid, title, author, publisher, location, sha1, type, date_added) VALUES (NULL, ?1, ?2, ?3, ?4, ?5, ?6, ?7, datetime('now'))");
	statement.bindUTF8StringParameter(0, book.opf.uid);
	statement.bindUTF8StringParameter(1, book.opf.metadata.title);
	if (book.opf.metadata.author.name)
	    statement.bindUTF8StringParameter(2, book.opf.metadata.author.name);
	if (book.opf.metadata.publisher)
	    statement.bindUTF8StringParameter(3, book.opf.metadata.publisher);
	statement.bindUTF8StringParameter(4, book.original_location);
	statement.bindUTF8StringParameter(5, book.sha1);
	statement.bindUTF8StringParameter(6, book.type);
	statement.execute();

	//this.update_date_read(this.db.lastInsertRowID);
	return this.db.lastInsertRowID;
    },

    add_book_data: function(book) {
	let statement = this.db.createStatement("INSERT INTO books (id, uid, title, author, publisher, location, sha1, type, date_added) VALUES (NULL, ?1, ?2, ?3, ?4, ?5, ?6, ?7, datetime('now'))");
	if (book.uid)
	    statement.bindUTF8StringParameter(0, book.uid);
	statement.bindUTF8StringParameter(1, book.title);
	if (book.author)
	    statement.bindUTF8StringParameter(2, book.author);
	if (book.publisher)
	    statement.bindUTF8StringParameter(3, book.publisher);
	statement.bindUTF8StringParameter(4, book.location);
	if (book.sha1)
	    statement.bindUTF8StringParameter(5, book.sha1);
	statement.bindUTF8StringParameter(6, book.type);
	statement.execute();

	return this.db.lastInsertRowID;
    },

    set_preview: function(id, url) {
	let statement = this.db.createStatement("UPDATE books SET preview=?1 WHERE id=?2");
	statement.bindUTF8StringParameter(0, url);
	statement.bindUTF8StringParameter(1, id);
	statement.execute();
    },

    update_date_read: function(id) {
	let statement = this.db.createStatement("UPDATE books SET date_read=datetime('now') WHERE id=?1");
	statement.bindUTF8StringParameter(0, id);
	statement.execute();
    },

    update_last_position: function(id, position) {
	let statement = this.db.createStatement("UPDATE books SET bookmark=?1 WHERE id=?2");
	statement.bindUTF8StringParameter(0, position);
	statement.bindUTF8StringParameter(1, id);
	statement.execute();
    },

    remove_book: function(id) {
	let statement = this.db.createStatement("DELETE FROM books WHERE id=?1");
	statement.bindUTF8StringParameter(0, id);
	statement.execute();
    },

    get_id: function(uid) {
	let statement = this.db.createStatement("SELECT id FROM books WHERE uid=?1");
	let id = -1;
	try {
	    statement.bindUTF8StringParameter(0, uid);
	    if (statement.executeStep())
		id = statement.getInt32(0);
	} finally {
	    statement.reset();
	}
	return id;
    },

    get_id_from_sha1: function(sha1) {
	let statement = this.db.createStatement("SELECT id FROM books WHERE sha1=?1");
	let id = -1;
	try {
	    statement.bindUTF8StringParameter(0, sha1);
	    if (statement.executeStep())
		id = statement.getInt32(0);
	} finally {
	    statement.reset();
	}
	return id;
    },

    get_id_from_location: function(uri) {
	let statement = this.db.createStatement("SELECT id FROM books WHERE location=?1");
	let id = -1;
	try {
	    statement.bindUTF8StringParameter(0, uri);
	    if (statement.executeStep())
		id = statement.getInt32(0);
	} finally {
	    statement.reset();
	}
	return id;
    },

    get_book_info: function(id) {
	debug("bookcase.get_book_info: " + id);

	let book;
	let statement = this.db.createStatement("SELECT id, uid, title, author, publisher, location, type FROM books WHERE id=?1");
	try {
	    statement.bindInt32Parameter(0, id);
	    if (statement.executeStep()) {
		book = {};
		book.id = statement.getInt32(0);
		//book.uid = statement.getUTF8String(1);
		book.title = statement.getUTF8String(2);
		book.author = statement.getUTF8String(3);
		book.publisher = statement.getUTF8String(4);
		book.location = statement.getUTF8String(5);
		book.type = statement.getUTF8String(6);
	    }
	} finally {
	    statement.reset();
	}
	return book;
    },

    get_preview: function(id) {
	let url = null;
	let statement = this.db.createStatement("SELECT preview FROM books WHERE id=?1");
	try {
	    statement.bindInt32Parameter(0, id);
	    if (statement.executeStep())
		url = statement.getUTF8String(0);
	} finally {
	    statement.reset();
	}
	return url;
    },

    get_last_position: function(id) {
	debug("bookcase.get_last_position: " + id);

	let position;
	let statement = this.db.createStatement("SELECT bookmark FROM books WHERE id=?1");
	try {
	    statement.bindInt32Parameter(0, id);
	    if (statement.executeStep()) {
		position = statement.getUTF8String(0);
	    }
	} finally {
	    statement.reset();
	}
	return position;
    },

    search: function(text, filter) {
	debug("Filter :'" + filter + "' Text: '" + text + "'");
	let stat = "SELECT id, author, title, location, date_read FROM books";
	let statement;
	if (filter == "where") {
	    stat += " " + text;
	    statement = this.db.createStatement(stat);
	} else if (filter == "untagged") {
	    stat += " WHERE tags = ''";	// FIXME
	    statement = this.db.createStatement(stat);
	} else if (text) {
	    if (filter == "author" || filter == "title") {
		stat += " WHERE LOWER(" + filter + ") LIKE LOWER(?1)";
		statement = this.db.createStatement(stat);
	    } else {	// meta
		stat += " WHERE LOWER(title) LIKE LOWER(?1) OR LOWER(author) LIKE LOWER(?1)";
		statement = this.db.createStatement(stat);
	    }
	    //text = text.replace("*", "%");
	    text = "%" + text + "%";
	    //text = statement.escapeStringForLIKE(text, "%");
	    statement.bindUTF8StringParameter(0, text);
	} else {
	    statement = this.db.createStatement(stat);
	}
	try {
	    while (statement.executeStep()) {
		let book = {};
		book.id = statement.getInt32(0);
		book.author = statement.getUTF8String(1);
		book.title = statement.getUTF8String(2);
		book.location = statement.getUTF8String(3);
		yield book;
	    }
	} finally {
	    statement.reset();
	}
    }
};

var bookcaseService = new bookcase();

var bookcaseFactory = {
    createInstance: function(outer, aIID) {
	if (outer != null)
	    throw Components.results.NS_ERROR_NO_AGGREGATION;
	return bookcaseService.QueryInterface(aIID);
    }
};

var components = [bookcase];

if (XPCOMUtils.generateNSGetFactory)
    var NSGetFactory = XPCOMUtils.generateNSGetFactory(components);
else
    var NSGetModule = XPCOMUtils.generateNSGetModule(components);
