/*
 *  Lucidor - e-book reader
 *  Copyright © 2009-2010  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

function debug(msg) {
    if (false) {
	try {
	    let cons = Components.classes["@mozilla.org/consoleservice;1"].
		getService(Components.interfaces.nsIConsoleService);
	    cons.logStringMessage(msg);
	} catch (e) {
	}
    }
}

function loop(thing) {
    if (!thing)
	return;
    for (let i = 0; i < thing.length; i++)
	yield thing[i];
}

function bookdb() {
    this.wrappedJSObject = this;
}

bookdb.prototype = {
    classDescription: "bookdb",
    classID: Components.ID("{fe14e5ba-2c5e-497b-b175-7b1f93823167}"),
    contractID: "@lucidor.org/bookdb;1",
    QueryInterface: XPCOMUtils.generateQI([Components.interfaces.nsISupports]),

    db: null,
    root_node: 1,
    file_uri: null,

    getFilename: function() {
	if (!this.db.databaseFile)
	    return null;
	let str = this.db.databaseFile.leafName;
	return str.substring(0, str.length - 7);
    },

    open: function(name) {
	debug("bookdb.open");
	if (this.db)
	    return;
	let file = Components.classes["@mozilla.org/file/directory_service;1"].
		getService(Components.interfaces.nsIProperties).
		get("ProfD", Components.interfaces.nsIFile);
	file.append(name + ".sqlite");

	let fph = Components.classes["@mozilla.org/network/protocol;1?name=file"].getService(Components.interfaces.nsIFileProtocolHandler);
	this.file_uri = fph.getURLSpecFromFile(file);

	let ss = Components.classes["@mozilla.org/storage/service;1"].
		getService(Components.interfaces.mozIStorageService);
	this.db = ss.openDatabase(file);

	if (!this.db.tableExists("books"))
	    this.create_tables();

	if (this.db.schemaVersion < 2)
	    this.update();
    },

    create_memory: function() {
	let ss = Components.classes["@mozilla.org/storage/service;1"].
	    getService(Components.interfaces.mozIStorageService);
	this.db = ss.openSpecialDatabase("memory");

	this.create_tables();
    },

    create: function(name) {
	debug("bookdb.create");
	if (this.db)
	    return;

	let file = Components.classes["@mozilla.org/file/directory_service;1"].
		getService(Components.interfaces.nsIProperties).
		get("ProfD", Components.interfaces.nsIFile);
	file.append(name + ".sqlite");
	file.createUnique(Components.interfaces.nsIFile.NORMAL_FILE_TYPE, 0600);

	let fph = Components.classes["@mozilla.org/network/protocol;1?name=file"].getService(Components.interfaces.nsIFileProtocolHandler);
	this.file_uri = fph.getURLSpecFromFile(file);

	let ss = Components.classes["@mozilla.org/storage/service;1"].
		getService(Components.interfaces.mozIStorageService);
	this.db = ss.openDatabase(file);

	this.create_tables();
    },

    create_tables: function() {
	this.db.createTable("books", "id INTEGER NOT NULL PRIMARY KEY, xid TEXT, fullid TEXT, uid TEXT, title TEXT, author TEXT, publisher TEXT, language TEXT, content TEXT, sha1 TEXT, location TEXT, links TEXT, type TEXT, preview TEXT, cover TEXT, date_added TEXT");
	this.db.createTable("nodes", "id INTEGER NOT NULL PRIMARY KEY, xid TEXT, path TEXT, title TEXT, content TEXT, links TEXT, next TEXT, pageinfo TEXT, visited TEXT");
	this.db.createTable("tree", "id INTEGER NOT NULL PRIMARY KEY, parent INTEGER, node INTEGER, book INTEGER");
	this.db.schemaVersion = 2;
    },

    update: function() {
	// schemaVersion = 1 (Lucidor 0.6, 0.7):
	// "books", "id INTEGER NOT NULL PRIMARY KEY, xid TEXT, uid TEXT, title TEXT, author TEXT, publisher TEXT, language TEXT, sha1 TEXT, location TEXT, links TEXT, type TEXT, preview TEXT, date_added TEXT"
	// "nodes", "id INTEGER NOT NULL PRIMARY KEY, xid TEXT, path TEXT, title TEXT, links TEXT, next TEXT, visited TEXT"
	// "tree", "id INTEGER NOT NULL PRIMARY KEY, parent INTEGER, node INTEGER, book INTEGER"
	//
	// schemaVersion = 2 (Lucidor 0.8):
	// "books", "id INTEGER NOT NULL PRIMARY KEY, xid TEXT, fullid TEXT, uid TEXT, title TEXT, author TEXT, publisher TEXT, language TEXT, content TEXT, sha1 TEXT, location TEXT, links TEXT, type TEXT, preview TEXT, cover TEXT, date_added TEXT"
	// "nodes", "id INTEGER NOT NULL PRIMARY KEY, xid TEXT, path TEXT, title TEXT, content TEXT, links TEXT, next TEXT, pageinfo TEXT, visited TEXT"

	if (this.db.schemaVersion == 1) {
	    let statement = this.db.createStatement("ALTER TABLE books ADD COLUMN fullid TEXT");
	    statement.execute();
	    statement = this.db.createStatement("ALTER TABLE books ADD COLUMN content TEXT");
	    statement.execute();
	    statement = this.db.createStatement("ALTER TABLE books ADD COLUMN cover TEXT");
	    statement.execute();

	    statement = this.db.createStatement("ALTER TABLE nodes ADD COLUMN content TEXT");
	    statement.execute();
	    statement = this.db.createStatement("ALTER TABLE nodes ADD COLUMN pageinfo TEXT");
	    statement.execute();

	    this.db.schemaVersion = 2;
	}
    },

    close: function() {
	try {
	    this.db.close();
	} catch (e) {
	}
	this.db = null;
    },

    delete: function() {
	let file = this.db.databaseFile;
	try {
	    this.db.close();
	} catch (e) {
	}
	try {
	    file.remove(false);
	} catch (e) {
	}
	this.db = null;
    },

    add_or_get_book: function(book) {
	let id = this.get_id_from_location(book.location);
	if (id >= 0)
	    return id;
	else
	    return this.add_book_info(book);
    },

    add_or_update_book: function(book) {
	let id = this.get_id_from_xid(book.xid);
	if (id >= 0)
	    return this.update_book(id, book);
	else
	    return this.add_book_info(book);
    },

    add_book_info: function(book) {
	let statement = this.db.createStatement("INSERT INTO books (id, uid, title, author, publisher, location, sha1, type, xid, links, preview, cover, fullid, content, date_added) VALUES (NULL, ?1, ?2, ?3, ?4, ?5, ?6, ?7, ?8, ?9, ?10, ?11, ?12, ?13, datetime('now'))");
	// FIXME language
	if (book.uid)
	    statement.bindUTF8StringParameter(0, book.uid);
	if (book.title)
	    statement.bindUTF8StringParameter(1, book.title);
	if (book.author)
	    statement.bindUTF8StringParameter(2, book.author);
	if (book.publisher)
	    statement.bindUTF8StringParameter(3, book.publisher);
	if (book.location)
	    statement.bindUTF8StringParameter(4, book.location);
	if (book.sha1)
	    statement.bindUTF8StringParameter(5, book.sha1);
	if (book.type)
	    statement.bindUTF8StringParameter(6, book.type);
	if (book.xid)
	    statement.bindUTF8StringParameter(7, book.xid);
	if (book.links) {
	    let JSON = Components.classes["@mozilla.org/dom/json;1"].
		createInstance(Components.interfaces.nsIJSON);
	    statement.bindUTF8StringParameter(8, JSON.encode(book.links));
	}
	if (book.thumbnail_link)
	    statement.bindUTF8StringParameter(9, book.thumbnail_link);
	if (book.cover_link)
	    statement.bindUTF8StringParameter(10, book.cover_link);
	if (book.fullid)
	    statement.bindUTF8StringParameter(11, book.fullid);
	if (book.content)
	    statement.bindUTF8StringParameter(12, book.content);
	statement.execute();

	return this.db.lastInsertRowID;
    },

    add_book_file: function(fil) {
	let book = Components.classes["@lucidor.org/epubDocument;1"].
		getService().wrappedJSObject;
	try {
	    book.init(fil);
	} catch (e) {
	    return -1;
	}

	let statement = this.db.createStatement("INSERT INTO books (id, uid, title, author, publisher, location, sha1, type, date_added) VALUES (NULL, ?1, ?2, ?3, ?4, ?5, ?6, ?7, datetime('now'))");
	statement.bindUTF8StringParameter(0, book.opf.uid);
	statement.bindUTF8StringParameter(1, book.opf.metadata.title);
	if (book.opf.metadata.author.name)
	    statement.bindUTF8StringParameter(2, book.opf.metadata.author.name);
	if (book.opf.metadata.publisher)
	    statement.bindUTF8StringParameter(3, book.opf.metadata.publisher);
	statement.bindUTF8StringParameter(4, book.location);
	statement.bindUTF8StringParameter(5, book.sha1);
	statement.bindUTF8StringParameter(6, book.type);
	statement.execute();

	return this.db.lastInsertRowID;
    },

    update_book: function(id, book) {
	let statement = this.db.createStatement("UPDATE books SET uid=?1, title=?2, author=?3, publisher=?4, location=?5, sha1=?6, type=?7, xid=?8, links=?9, preview=?10, cover=?11, fullid=?12, content=?13, date_added=datetime('now') WHERE id=?14");
	// FIXME language
	if (book.uid)
	    statement.bindUTF8StringParameter(0, book.uid);
	if (book.title)
	    statement.bindUTF8StringParameter(1, book.title);
	if (book.author)
	    statement.bindUTF8StringParameter(2, book.author);
	if (book.publisher)
	    statement.bindUTF8StringParameter(3, book.publisher);
	if (book.location)
	    statement.bindUTF8StringParameter(4, book.location);
	if (book.sha1)
	    statement.bindUTF8StringParameter(5, book.sha1);
	if (book.type)
	    statement.bindUTF8StringParameter(6, book.type);
	if (book.xid)
	    statement.bindUTF8StringParameter(7, book.xid);
	if (book.links) {
	    let JSON = Components.classes["@mozilla.org/dom/json;1"].
		createInstance(Components.interfaces.nsIJSON);
	    statement.bindUTF8StringParameter(8, JSON.encode(book.links));
	}
	if (book.thumbnail_link)
	    statement.bindUTF8StringParameter(9, book.thumbnail_link);
	if (book.cover_link)
	    statement.bindUTF8StringParameter(10, book.cover_link);
	if (book.fullid)
	    statement.bindUTF8StringParameter(11, book.fullid);
	if (book.content)
	    statement.bindUTF8StringParameter(12, book.content);
	statement.bindUTF8StringParameter(13, id);
	statement.execute();

	return id;
    },

    add_or_get_node: function(parent, path, title) {
	let id = this.get_node_by_path(path);
	if (id >= 0)
	    return id;
	return this.add_node(parent, path, title);
    },

    add_node: function(parent, path, title, content) {
	debug("PARENT: " + parent + " PATH: " + path + " NAME: " + title);
	let statement = this.db.createStatement("INSERT INTO nodes (id, path, title, content) VALUES (NULL, ?1, ?2, ?3)");
	statement.bindUTF8StringParameter(0, path);
	statement.bindUTF8StringParameter(1, title);
	if (content)
	    statement.bindUTF8StringParameter(2, content);
	statement.execute();
	let nodeid = this.db.lastInsertRowID;

	if (parent == null)
	    this.root_node = nodeid;
	else {
	    statement = this.db.createStatement("INSERT INTO tree (id, parent, node) VALUES (NULL, ?1, ?2)");
	    statement.bindInt32Parameter(0, parent);
	    statement.bindInt32Parameter(1, nodeid);
	    statement.execute();
	}

	return nodeid;
    },

    set_node_links: function(id, links) {
	let statement = this.db.createStatement("UPDATE nodes SET links=?1 WHERE id=?2");
	let JSON = Components.classes["@mozilla.org/dom/json;1"].
	    createInstance(Components.interfaces.nsIJSON);
	statement.bindUTF8StringParameter(0, JSON.encode(links));
	statement.bindInt32Parameter(1, id);
	statement.execute();
    },

    set_node_pageinfo: function(id, totalResults, startIndex, itemsPerPage) {
	let statement = this.db.createStatement("UPDATE nodes SET pageinfo=?1 WHERE id=?2");
	let JSON = Components.classes["@mozilla.org/dom/json;1"].
	    createInstance(Components.interfaces.nsIJSON);
	let pi = { totalResults: totalResults,
		   startIndex: startIndex,
		   itemsPerPage: itemsPerPage };
	statement.bindUTF8StringParameter(0, JSON.encode(pi));
	statement.bindInt32Parameter(1, id);
	statement.execute();
    },

    set_node_next: function(id, next) {
	let statement = this.db.createStatement("UPDATE nodes SET next=?1 WHERE id=?2");
	statement.bindInt32Parameter(0, next);
	statement.bindInt32Parameter(1, id);
	statement.execute();
    },

    set_node_visited: function(id) {
	let statement = this.db.createStatement("UPDATE nodes SET visited=datetime('now') WHERE id=?1");
	statement.bindInt32Parameter(0, id);
	statement.execute();
    },

    remove_branch: function(branch) {
	//let statement = this.db.createStatement("");
    },

    add_child: function(parent, bookid) {
	let statement = this.db.createStatement("INSERT INTO tree (id, parent, book) VALUES (NULL, ?1, ?2)");
	statement.bindInt32Parameter(0, parent);
	statement.bindInt32Parameter(1, bookid);
	statement.execute();

	return this.db.lastInsertRowID;
    },

    set_preview: function(id, url) {
	let statement = this.db.createStatement("UPDATE books SET preview=?1 WHERE id=?2");
	statement.bindUTF8StringParameter(0, url);
	statement.bindUTF8StringParameter(1, id);
	statement.execute();
    },

    create_xid_index: function() {
	let statement = this.db.createStatement("CREATE INDEX xid ON books(xid)");
	statement.execute();
    },

    update_file: function(xid, location, type) {
	let statement = this.db.createStatement("UPDATE books SET location=?1, type=?2 WHERE xid=?3");
	statement.bindUTF8StringParameter(0, location);
	statement.bindUTF8StringParameter(1, type);
	statement.bindUTF8StringParameter(2, xid);
	statement.execute();
    },

    remove_book: function(id) {
	let statement = this.db.createStatement("DELETE FROM books WHERE id=?1");
	statement.bindUTF8StringParameter(0, id);
	statement.execute();
    },

    remove_book_xid: function(xid) {
	let statement = this.db.createStatement("DELETE FROM books WHERE xid=?1");
	statement.bindUTF8StringParameter(0, xid);
	statement.execute();
    },

    remove_children: function(id) {
	let statement = this.db.createStatement("DELETE FROM tree WHERE parent=?1");
	statement.bindUTF8StringParameter(0, id);
	statement.execute();
    },

    get_id: function(uid) {
	let statement = this.db.createStatement("SELECT id FROM books WHERE uid=?1");
	let id = -1;
	try {
	    statement.bindUTF8StringParameter(0, uid);
	    if (statement.executeStep())
		id = statement.getInt32(0);
	} finally {
	    statement.reset();
	}
	return id;
    },

    get_id_from_xid: function(xid) {
	let statement = this.db.createStatement("SELECT id FROM books WHERE xid=?1");
	let id = -1;
	try {
	    statement.bindUTF8StringParameter(0, xid);
	    if (statement.executeStep())
		id = statement.getInt32(0);
	} finally {
	    statement.reset();
	}
	return id;
    },

    get_id_from_location: function(location) {
	let id = -1;
	let statement = this.db.createStatement("SELECT id FROM books WHERE location=?1");
	try {
	    statement.bindUTF8StringParameter(0, location);
	    if (statement.executeStep())
		id = statement.getInt32(0);
	} finally {
	    statement.reset();
	}
	return id;
    },

    get_book_info: function(id) {
	debug("bookdb.get_book_info: " + id);

	let book;
	let statement = this.db.createStatement("SELECT id, uid, title, author, publisher, location, type, links, xid, fullid, content, preview FROM books WHERE id=?1");
	try {
	    statement.bindInt32Parameter(0, id);
	    if (statement.executeStep()) {
		book = {};
		book.id = statement.getInt32(0);
		//book.uid = statement.getUTF8String(1);
		book.title = statement.getUTF8String(2);
		book.author = statement.getUTF8String(3);
		book.publisher = statement.getUTF8String(4);
		book.location = statement.getUTF8String(5);
		book.type = statement.getUTF8String(6);
		let links = statement.getUTF8String(7);
		if (links) {
		    let JSON = Components.classes["@mozilla.org/dom/json;1"].
			createInstance(Components.interfaces.nsIJSON);
		    book.links = JSON.decode(links);
		}
		book.xid = statement.getUTF8String(8);
		book.fullid = statement.getUTF8String(9);
		book.content = statement.getUTF8String(10);
		book.preview = statement.getUTF8String(11);
	    }
	} finally {
	    statement.reset();
	}
	return book;
    },

    get_preview: function(id) {
	let url = null;
	let statement = this.db.createStatement("SELECT preview FROM books WHERE id=?1");
	try {
	    statement.bindInt32Parameter(0, id);
	    if (statement.executeStep())
		url = statement.getUTF8String(0);
	} finally {
	    statement.reset();
	}
	return url;
    },

    get_cover: function(id) {
	let url = null;
	let statement = this.db.createStatement("SELECT cover FROM books WHERE id=?1");
	try {
	    statement.bindInt32Parameter(0, id);
	    if (statement.executeStep())
		url = statement.getUTF8String(0);
	} finally {
	    statement.reset();
	}
	return url;
    },

    browse_path: function(path) {
	debug("browse_path: " + path);

	if (path == "" || path == "all") {
	    debug("NULL -> root");
	    path = this.root_node;
	}

	debug("FDAS");

	let folder = this.get_node(path);
	let link = { type: "folder",
		     path: path,
		     location: folder.path,
		     title: folder.title,
		     thisnode: true,
		     links: folder.links,
		     pageinfo: folder.pageinfo };
	if (folder.next > 0) {
	    let nfolder = this.get_node(folder.next);
	    link.nextpath = folder.next;
	    link.nexttitle = nfolder.title;
	}
	yield link;

	let statement = this.db.createStatement("SELECT node, book FROM tree WHERE parent=?1");
	try {
	    statement.bindInt32Parameter(0, path);	// FIXME Integer(path)
	    while (statement.executeStep()) {
		let node = statement.getInt32(0);
		let book = statement.getInt32(1);
		let link = {};
		debug("node: " + node + " book: " + book);
		if (node > 0) {
		    link = this.get_node(node);
		    link.type = "folder";
		    link.location = link.path;
		    link.path = node;
		    debug("FOLDER: " + link.path  + " " + link.location);
		} else {
		    link = this.get_book_info(book);
		    //link.type = "epub";
		}
		yield link;
	    }
	} finally {
	    statement.reset();
	}
    },

    get_node: function(id) {
	debug("get_node: " + id);
	let node = null;
	let statement = this.db.createStatement("SELECT path, title, content, links, next, pageinfo, visited FROM nodes WHERE id=?1");
	try {
	    statement.bindInt32Parameter(0, id);
	    if (statement.executeStep()) {
		node = {};
		node.id = id;
		node.path = statement.getUTF8String(0);
		node.title = statement.getUTF8String(1);
		node.content = statement.getUTF8String(2);
		let links = statement.getUTF8String(3);
		if (links) {
		    let JSON = Components.classes["@mozilla.org/dom/json;1"].
			createInstance(Components.interfaces.nsIJSON);
		    node.links = JSON.decode(links);
		}
		node.next = statement.getInt32(4);
		let pi = statement.getUTF8String(5);
		if (pi) {
		    let JSON = Components.classes["@mozilla.org/dom/json;1"].
			createInstance(Components.interfaces.nsIJSON);
		    node.pageinfo = JSON.decode(pi);
		} else {
		    node.pageinfo = null;
		}
		node.visited = statement.getUTF8String(6);
	    }
	} finally {
	    statement.reset();
	}
	return node;
    },

    get_node_by_path: function(path) {
	let id = -1;
	let statement = this.db.createStatement("SELECT id FROM nodes WHERE path=?1");
	try {
	    statement.bindUTF8StringParameter(0, path);
	    if (statement.executeStep())
		id = statement.getInt32(0);
	} finally {
	    statement.reset();
	}
	return id;
    },

    search: function(text, filter) {
	debug("Filter :'" + filter + "' Text: '" + text + "'");
	let stat = "SELECT id, author, title, location FROM books";
	let statement;
	if (filter == "where") {
	    stat += " " + text;
	    statement = this.db.createStatement(stat);
	} else if (filter == "untagged") {
	    stat += " WHERE tags = ''";	// FIXME
	    statement = this.db.createStatement(stat);
	} else if (text) {
	    if (filter == "author" || filter == "title") {
		stat += " WHERE LOWER(" + filter + ") LIKE LOWER(?1)";
		statement = this.db.createStatement(stat);
	    } else {	// meta
		stat += " WHERE LOWER(title) LIKE LOWER(?1) OR LOWER(author) LIKE LOWER(?1)";
		statement = this.db.createStatement(stat);
	    }
	    //text = text.replace("*", "%");
	    text = "%" + text + "%";
	    statement.bindUTF8StringParameter(0, text);
	} else {
	    statement = this.db.createStatement(stat);
	}
	try {
	    debug("sdf");
	    while (statement.executeStep()) {
		debug("SDF");
		let book = {};
		book.id = statement.getInt32(0);
		book.author = statement.getUTF8String(1);
		book.title = statement.getUTF8String(2);
		book.location = statement.getUTF8String(3);
		yield book;
	    }
	} finally {
	    statement.reset();
	}
    }
};

var components = [bookdb];

if (XPCOMUtils.generateNSGetFactory)
    var NSGetFactory = XPCOMUtils.generateNSGetFactory(components);
else
    var NSGetModule = XPCOMUtils.generateNSGetModule(components);
