/*
 *  Lucidor - e-book reader
 *  Copyright © 2008-2010  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

function debug(msg) {
    if (false) {
	try {
	    let cons = Components.classes["@mozilla.org/consoleservice;1"].
		getService(Components.interfaces.nsIConsoleService);
	    cons.logStringMessage(msg);
	} catch (e) {
	}
    }
}

function loop(thing) {
    if (!thing)
	return;
    for (let i = 0; i < thing.length; i++)
	yield thing[i];
}

function bookmanager() {
    this.wrappedJSObject = this;
    this.books = {};
}

bookmanager.prototype = {
    classDescription: "bookmanager",
    classID: Components.ID("{c9ff0f94-2b9a-43bb-96ec-b86ddef81c54}"),
    contractID: "@lucidor.org/bookmanagerService;1",
    _xpcom_factory: bookmanagerFactory,
    QueryInterface: XPCOMUtils.generateQI([Components.interfaces.nsISupports]),

    addBook: function(type, book) {
	let bookid;
	do {
	    bookid = Math.floor(Math.random() * 100000000).toString();
	} while (this.books[bookid] != undefined);

	debug("bookmanager.addBook: bookid=" + bookid);
	this.books[bookid] = { type: type, book: book };
        return bookid;
    },

    removeBook: function(bookid) {
	this.books[bookid] = undefined;	// FIXME
    },

    getChannel: function(uri) {
	debug("bookmanager.getChannel");
	debug("host: " + uri.spec);
	let bookid = uri.host;
	let entry = this.books[bookid];
	if (!entry)
	    throw "Error: No such book";

	let nu = Components.classes["@mozilla.org/network/util;1"].
		createInstance(Components.interfaces.nsINetUtil);
	let path = nu.unescapeString(uri.path, nu.ESCAPE_ALL);

	let i = 0;
	while (path.charAt(i) == "/")
	    i++;
	let j = path.indexOf("#", i);
	if (j < 0)
	    j = path.length;
	path = path.substring(i, j);

	let instream;
	let mediatype;

	if (this.books[bookid].type == "epub") {
	    let zipreader = entry.book.zipreader;

	    debug("PATH: " + path);
	    let item = entry.book.hrefToItem(path);
	    debug("IDREF: " + item.id);
	    let loader = entry.book.idrefToLoader(item.id);
	    debug("NEW PATH: " + loader);
	    path = loader[0];

	    if (!zipreader.hasEntry(path)) {
		debug("archive entry missing: " + path);
		entry.book.add_error("archive entry missing: " + path);
		if (path.substring(path.length - 4) == ".css") {
		    instream = Components.classes["@mozilla.org/io/string-input-stream;1"].
			createInstance(Components.interfaces.nsIStringInputStream);
		    instream.setData("", 0);
		    mediatype = "";
		} else
		    throw "Error: No such file entry";
	    } else {
		let item = entry.book.opf.manifest_href[path];
		if (!item) {
		    debug("manifest item missing: " + path);
		    entry.book.add_error("manifest item missing: " + path);
		    /* FIXME should not do this, only a workaround for
		       broken books (eg Death and the Senator) */
		    item = { mediatype: null };
		    //throw "Error: No such manifest item";
		} else if (item.encryption && !item.decrypt) {
		    throw "Error: The book creator dislikes freedom";
		}
		mediatype = item.mediatype;
		if (mediatype == "application/vnd.adobe-page-template+xml")
		    throw "Error: Adobe has fucked up the document";
		instream = zipreader.getInputStream(path);
		if (item.decrypt) {
		    instream = item.decrypt(entry.book, item, instream);

		    //let cs = Components.classes["@mozilla.org/streamConverters;1"].
		    //    getService(Components.interfaces.nsIStreamConverterService);
		    //instream = cs.convert(instream, "deflate", "uncompressed", null);
		    instream = getDeflateStream(instream);
		}
	    }
	} else
	    throw "Error: This will never happen";

        let channel = Components.classes["@mozilla.org/network/input-stream-channel;1"].createInstance(Components.interfaces.nsIInputStreamChannel);
	channel.contentStream = instream;
	channel.setURI(uri);
	if (mediatype) {
	    channel.QueryInterface(Components.interfaces.nsIChannel);
	    channel.contentType = mediatype;
	}

	return channel;
    }
};

var serviceObject = new bookmanager();

var bookmanagerFactory = {
    createInstance: function(outer, aIID) {
	if (outer != null)
	    throw Components.results.NS_ERROR_NO_AGGREGATION;
	return serviceObject.QueryInterface(aIID);
    }
};

var components = [bookmanager];

if (XPCOMUtils.generateNSGetFactory)
    var NSGetFactory = XPCOMUtils.generateNSGetFactory(components);
else
    var NSGetModule = XPCOMUtils.generateNSGetModule(components);

function getDeflateStream(instream) {
    let pipe = Components.classes["@mozilla.org/pipe;1"].
        createInstance(Components.interfaces.nsIPipe);
    pipe.init(false, false, 0, 0, null);
    let blocker = Components.classes["@mozilla.org/binaryinputstream;1"].
        createInstance(Components.interfaces.nsIBinaryInputStream);
    blocker.setInputStream(pipe.inputStream);

    let mylis = new listener(pipe);
    let cs = Components.classes["@mozilla.org/streamConverters;1"].
	getService(Components.interfaces.nsIStreamConverterService);
    let converter = cs.asyncConvertData("deflate", "uncompressed", mylis, null);
    converter.onDataAvailable(null, null, instream, 0, instream.available());
    try {
	blocker.read8();
    } catch (e) {
    }
    return mylis.stream;
};

function listener(pipe) {
    this.pipe = pipe;
};

listener.prototype = {
    onDataAvailable: function(request, context, inputStream, offset, count) {
	let main_control = Components.classes["@lucidor.org/mainControlService;1"].
	    getService().wrappedJSObject;
	let fil = main_control.get_tempdir();
	fil.append("lucidor.ttf");
	fil.createUnique(fil.NORMAL_FILE_TYPE, 0600);
 	let filout = Components.classes["@mozilla.org/network/file-output-stream;1"].
	    createInstance(Components.interfaces.nsIFileOutputStream);
	filout.init(fil, -1, -1, 0);

	let binin = Components.classes["@mozilla.org/binaryinputstream;1"].
            createInstance(Components.interfaces.nsIBinaryInputStream);
	binin.setInputStream(inputStream);
	let binout = Components.classes["@mozilla.org/binaryoutputstream;1"].
            createInstance(Components.interfaces.nsIBinaryOutputStream);
	binout.setOutputStream(filout);
	while (binin.available() > 0) {
	    let count = binin.available();
	    if (count > 1024)
		count = 1024;
	    let buf = binin.readByteArray(count);
	    binout.writeByteArray(buf, buf.length);
	}
	binout.close();
	filout.close();
	
 	let filin = Components.classes["@mozilla.org/network/file-input-stream;1"].
	    createInstance(Components.interfaces.nsIFileInputStream);
	filin.init(fil, -1, -1, 0);
	this.stream = filin;

	this.pipe.outputStream.close();
    }
};
