/*
 *  Lucidor - e-book reader
 *  Copyright © 2009-2011  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

function debug(msg) {
    if (false) {
	try {
	    let cons = Components.classes["@mozilla.org/consoleservice;1"].
		getService(Components.interfaces.nsIConsoleService);
	    cons.logStringMessage(msg);
	} catch (e) {
	}
    }
}

function loop(thing) {
    if (!thing)
	return;
    for (let i = 0; i < thing.length; i++)
	yield thing[i];
}

function collectionManager() {
    this.wrappedJSObject = this;
    this.listeners = [];
}

collectionManager.prototype = {
    classDescription: "collectionManager",
    classID: Components.ID("{b0e946c7-5d40-4e17-94ca-654d3b29d974}"),
    contractID: "@lucidor.org/collectionManagerService;1",
    _xpcom_factory: collectionManagerFactory,
    QueryInterface: XPCOMUtils.generateQI([Components.interfaces.nsISupports]),

    default_collections: [ { title: "Ebooks libres et gratuits",
			     type: "opds",
			     url: "http://www.ebooksgratuits.com/opds/" },
			   { title: "Feedbooks",
			     type: "opds",
			     url: "http://feedbooks.com/catalog.atom" },
		           /*{ title: "Lexcycle Online Catalog",
			     type: "stanza",
			     url: "http://catalog.lexcycle.com/" },*/
			   { title: "Project Gutenberg",
			     type: "opds",
			     url: "http://lucidor.org/opds/pg.php" },
			   { title: "Internet Archive Catalog",
			     type: "opds",
			     url: "http://bookserver.archive.org/catalog/" },
			   { title: "Smashwords",
			     type: "opds",
			     url: "http://www.smashwords.com/atom" },
			   /*{ title: "Google Books",
			     type: "googlebooks" },*/
			   /*{ title: "epubcatalog.org",
			     type: "epubcatalog_atom" },*/
		           /*{ title: "Meta Catalog",
			     type: "opds",
			     url: "http://lucidor.org/opds/meta.atom" },*/
			   { title: "Local Sample Books",
			     type: "folder",
			     folder: "resource://lucidor/samples/" }
			 ],

    type_config: { epubcatalog_atom: { url: false, urls: false, folder: false,
				       auth: false },
		   folder: { url: false, urls: false, folder: true,
			     auth: false },
		   gutenberg_rdf: { url: true, urls: false, folder: false,
				    auth: false },
		   opds: { url: true, urls: false, folder: false,
			   auth: false },
		   stanza: { url: true, urls: false, folder: false,
			     auth: false },
		   webpage: { url: true, urls: false, folder: false,
			      auth: false } },

    getBookcase: function(coll, tree) {
	let bcase;
	if (coll.type == "local")
	    bcase = Components.classes["@lucidor.org/localCase;1"].
		createInstance().wrappedJSObject;
	else if (coll.type == "epubcatalog_atom")
	    bcase = Components.classes["@lucidor.org/epubcatalogCase;1"].
		createInstance().wrappedJSObject;
	else if (coll.type == "folder")
	    bcase = Components.classes["@lucidor.org/folderCase;1"].
		createInstance().wrappedJSObject;
	else if (coll.type == "googlebooks")
	    bcase = Components.classes["@lucidor.org/googlebooksCase;1"].
		createInstance().wrappedJSObject;
	else if (coll.type == "gutenberg_rdf")
	    bcase = Components.classes["@lucidor.org/gutenbergRDFCase;1"].
		createInstance().wrappedJSObject;
	else if (coll.type == "opds")
	    bcase = Components.classes["@lucidor.org/opdsCase;1"].
		createInstance().wrappedJSObject;
	else if (coll.type == "stanza")
	    bcase = Components.classes["@lucidor.org/stanzaCase;1"].
		createInstance().wrappedJSObject;
	else if (coll.type == "webpage")
	    bcase = Components.classes["@lucidor.org/webpageCase;1"].
		createInstance().wrappedJSObject;
	else
	    return null;
	bcase.init(coll, tree);
	return bcase;
    },

    hasCollections: function() {
	try {
	    let prefs = Components.classes["@mozilla.org/preferences-service;1"].
		getService(Components.interfaces.nsIPrefService).
		getBranch("lucidor.");
	    prefs.getCharPref("collection.order");
	    return true;
	} catch (e) {
	    return false;
	}
    },

    getCollections: function() {
	debug("getCollections");
	let JSON = Components.classes["@mozilla.org/dom/json;1"].
		createInstance(Components.interfaces.nsIJSON);
	let prefs = Components.classes["@mozilla.org/preferences-service;1"].
		getService(Components.interfaces.nsIPrefService).
		getBranch("lucidor.");
	let value;
	try {
	    value = prefs.getCharPref("collection.order");
	} catch (e) {
	    debug("Default");
	    return this.default_collections;
	}
	debug("ORDER: " + value);
	let order = JSON.decode(value);
	debug("Order: " + order);
	let collections = [];
	for (let num in loop(order.order)) {
	    try {
		value = prefs.getCharPref("collection." + num);
		let val = JSON.decode(value);

		if (val.type == "feedbooks_xml") {
		    // Backward-compatible glue
		    val = { title: "Feedbooks",
			    type: "opds",
			    url: "http://feedbooks.com/catalog.atom" };
		}

		collections.push(val);
	    } catch (e) {
	    }
	}
	return collections;
    },

    getCollectionsJSON: function() {
	return this.collectionsToJSON(this.getCollections());
    },

    getCollectionsJSON2: function() {
	return this.collectionsToJSON2(this.getCollections());
    },

    getCollectionsJSON3: function() {
	return this.collectionsToJSON3(this.getCollections());
    },

    getDefaultCollectionsJSON2: function() {
	return this.collectionsToJSON2(this.default_collections);
    },

    collectionsToJSON: function(collections) {
	debug("collectionsToJSON " + collections);
	let JSON = Components.classes["@mozilla.org/dom/json;1"].
		createInstance(Components.interfaces.nsIJSON);
	let new_collections = [];
	for (let coll in loop(collections))
	    new_collections.push(JSON.encode(coll));
	return new_collections;
    },

    collectionsToJSON2: function(collections) {
	debug("collectionsToJSON2 " + collections);
	let JSON = Components.classes["@mozilla.org/dom/json;1"].
		createInstance(Components.interfaces.nsIJSON);
	let new_collections = [];
	for (let coll in loop(collections))
	    new_collections.push([coll.title, JSON.encode(coll)]);
	return new_collections;
    },

    collectionsToJSON3: function(collections) {
	debug("collectionsToJSON3 " + collections);
	let JSON = Components.classes["@mozilla.org/dom/json;1"].
		createInstance(Components.interfaces.nsIJSON);
	let new_collections = [];
	for (let coll in loop(collections)) {
	    if (coll.img)
		new_collections.push([coll.title, JSON.encode(coll), coll.img]);
	    else
		new_collections.push([coll.title, JSON.encode(coll), null]);
	}
	return new_collections;
    },

    saveOrder: function(order) {
	debug("saveOrder " + order);
	let prefs = Components.classes["@mozilla.org/preferences-service;1"].
		getService(Components.interfaces.nsIPrefService).
		getBranch("lucidor.");
	let JSON = Components.classes["@mozilla.org/dom/json;1"].
		createInstance(Components.interfaces.nsIJSON);
	let ord = { version: 1, order: order };      
	prefs.setCharPref("collection.order", JSON.encode(ord));
    },

    saveCollection: function(coll) {
	debug("saveCollection " + coll.id);
	let JSON = Components.classes["@mozilla.org/dom/json;1"].
		createInstance(Components.interfaces.nsIJSON);
	let prefs = Components.classes["@mozilla.org/preferences-service;1"].
		getService(Components.interfaces.nsIPrefService).
		getBranch("lucidor.");
	let old;
	try {
	    let pref = prefs.getCharPref("collection." + coll.id);
	    old = JSON.decode(pref);
	} catch (e) {
	}
	if (old) {
	    // FIXME
	}
	// FIXME check coll.type ...
	// delete coll.pass
	prefs.setCharPref("collection." + coll.id, JSON.encode(coll));
    },

    saveCollectionsJSON2: function(collections) {
	let new_collection = [];
	for (let coll in loop(collections))
	    new_collection.push(coll[1]);
	return this.saveCollectionsJSON(new_collection);
    },

    saveCollectionsJSON: function(collections) {
	debug("saveCollectionsJSON: " + collections);
	let JSON = Components.classes["@mozilla.org/dom/json;1"].
		createInstance(Components.interfaces.nsIJSON);
	let prefservice = Components.classes["@mozilla.org/preferences-service;1"].
		getService(Components.interfaces.nsIPrefService);
	let prefs = prefservice.getBranch("lucidor.");
	let old_order = [];
	try {
	    let pref = prefs.getCharPref("collection.order");
	    old_order = JSON.decode(pref).order;
	} catch (e) {
	}

	debug("1");

	let next_id = 1;
	for (let num in loop(old_order))
	    if (num >= next_id)
		next_id = num + 1;

	debug("2");

	let order = [];
	for (let coll in loop(collections)) {
	    let co = JSON.decode(coll);
	    debug("* " + co.id);
	    if (!co.id)
		co.id = next_id++;
	    else if (old_order)
		for (let i = 0; i < old_order.length; i++)
		    if (old_order[i] == co.id) {
			old_order[i] = -1;
			break;
		    }
	    debug("** " + order);
	    this.saveCollection(co);
	    order.push(co.id);
	}
	this.saveOrder(order);

	debug("3");

	prefservice.savePrefFile(null);

	for (let li in loop(this.listeners)) {
	    if (li.reread_collections)
		li.reread_collections();
	    // FIXME else remove listener?
	}

	debug("4");

	for (let id in loop(old_order))
	    if (id != -1)
		this.removeCollection(id);

	prefservice.savePrefFile(null);
    },

    changeCollectionDB: function(collid, db) {
	debug("changeCollectionDB");
	let JSON = Components.classes["@mozilla.org/dom/json;1"].
		createInstance(Components.interfaces.nsIJSON);
	let prefservice = Components.classes["@mozilla.org/preferences-service;1"].
		getService(Components.interfaces.nsIPrefService);
	let prefs = prefservice.getBranch("lucidor.");
	let coll;
	try {
	    let pref = prefs.getCharPref("collection." + collid);
	    coll = JSON.decode(pref);
	} catch (e) {
	}
	if (!coll)
	    return -1;
	coll.db = db.getFilename();
	prefs.setCharPref("collection." + coll.id, JSON.encode(coll));
	prefservice.savePrefFile(null);
	debug("changeCollectionDB finished");
	return 0;
    },

    // NOTE Does not change collection.order
    removeCollection: function(id) {
	let JSON = Components.classes["@mozilla.org/dom/json;1"].
		createInstance(Components.interfaces.nsIJSON);
	let prefs = Components.classes["@mozilla.org/preferences-service;1"].
		getService(Components.interfaces.nsIPrefService).
		getBranch("lucidor.");
	try {
	    let pref = prefs.getCharPref("collection." + id);
	    let old = JSON.decode(pref);
	    if (old.db) {
		let file = Components.classes["@mozilla.org/file/directory_service;1"].
		    getService(Components.interfaces.nsIProperties).
		    get("ProfD", Components.interfaces.nsIFile);
		file.append(old.db + ".sqlite");
		file.remove(false);
	    }
	} catch (e) {
	    debug("E: " +  e);
	}
	prefs.clearUserPref("collection." + id);
    },

    addCollection: function(coll) {
	let JSON = Components.classes["@mozilla.org/dom/json;1"].
		createInstance(Components.interfaces.nsIJSON);
	let prefs = Components.classes["@mozilla.org/preferences-service;1"].
		getService(Components.interfaces.nsIPrefService).
		getBranch("lucidor.");

	let pref = prefs.getCharPref("collection.order");
	let order = JSON.decode(pref);
	let no = 0;
	for (ord in loop(order.order))
	    if (ord >= no)
		no = ord + 1;
	coll.id = no;
	prefs.setCharPref("collection." + coll.id, JSON.encode(coll));
	order.order.push(no);
	prefs.setCharPref("collection.order", JSON.encode(order));
    },

    getConfig: function(type) {
	return this.type_config[type];
    },

    addListener: function(li) {
	this.listeners.push(li);
    },

    removeListener: function(li) {
	// FIXME
    }
};

var serviceObject = new collectionManager();

var collectionManagerFactory = {
    createInstance: function(outer, aIID) {
	if (outer != null)
	    throw Components.results.NS_ERROR_NO_AGGREGATION;
	return serviceObject.QueryInterface(aIID);
    }
};

var components = [collectionManager];

if (XPCOMUtils.generateNSGetFactory)
    var NSGetFactory = XPCOMUtils.generateNSGetFactory(components);
else
    var NSGetModule = XPCOMUtils.generateNSGetModule(components);
