/*
 *  Lucidor - e-book reader
 *  Copyright © 2008-2010  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

function debug(msg) {
    if (false) {
	try {
	    let cons = Components.classes["@mozilla.org/consoleservice;1"].
		getService(Components.interfaces.nsIConsoleService);
	    cons.logStringMessage(msg);
	} catch (e) {
	}
    }
}

function loop(thing) {
    if (!thing)
	return;
    for (let i = 0; i < thing.length; i++)
	yield thing[i];
}

function compareArray(a, b, start, max) {
    for (let i = 0; (i < a.length) && ((start + i) < max); i++)
	if (b[start + i] != a[i])
	    return false;
    return true;
}

function epubContentSniffer() {
}

epubContentSniffer.prototype = {
    classDescription: "epubContentSniffer",
    classID: Components.ID("{780d06f6-c257-47bb-9f39-1372a39075df}"),
    contractID: "@lucidor.org/epubContentSniffer;1",
    _xpcom_factory: epubContentSnifferFactory,
    _xpcom_categories: [{ category: "net-content-sniffers" },
			{ category: "content-sniffing-services" },
			{ category: "ext-to-type-mapping", entry: "epub",
			  value: "application/epub+zip" }] ,
    QueryInterface: XPCOMUtils.generateQI([Components.interfaces.nsISupports,
					   Components.interfaces.nsIContentSniffer]),

    pk_array: new Array(80, 75),	// "PK"
    // "mimetype"
    mimetype_array: new Array(109, 105, 109, 101, 116, 121, 112, 101),
    // "application/epub+zip"
    app_epub_array: new Array(97, 112, 112, 108, 105, 99, 97, 116, 105, 111,
			      110, 47, 101, 112, 117, 98, 43, 122, 105, 112),

    getMIMETypeFromContent: function(request, data, length) {
	if (length > 2 && compareArray(this.pk_array, data, 0, length)) {
	    let epub = false;
	    if (length > 58 &&
		compareArray(this.mimetype_array, data, 30, length) &&
		compareArray(this.app_epub_array, data, 38, length)) {
		//debug("Found application/epub+zip");
		epub = true;
	    }
	    if (!epub) {
		// In a perfect world this would not be necessary
		try {
		    let url = request.QueryInterface(Components.interfaces.nsIChannel).originalURI.QueryInterface(Components.interfaces.nsIURL);
		    //debug("File extention: " + url.fileExtension);
		    if (url.fileExtension.toLowerCase() == "epub") {
			//debug("Found .epub extension");
			epub = true;
		    }
		} catch (e) {
		}
	    }
	    if (epub) {
 		try {
		    // Documentation actually says that
		    // "Implementations should consider the request read-only."
		    // It is necessary to remove any Content-Disposition header
		    // otherwise the ExternalAppHandler will hijack the request
		    let chan = request.QueryInterface(Components.interfaces.nsIHttpChannel);
		    let head = chan.getResponseHeader("Content-Disposition");
		    // FIXME it maybe is enough to just remove "attachment;"
		    if (head)
			chan.setResponseHeader("Content-Disposition", "", false);
		} catch (e) {
		}
		return "application/epub+zip";
	    }
	}

	// FIXME what should be returned if you cannot determine the type?
	return null;
    },
};

var serviceObject = new epubContentSniffer();

var epubContentSnifferFactory = {
    createInstance: function(outer, aIID) {
	if (outer != null)
	    throw Components.results.NS_ERROR_NO_AGGREGATION;
	return serviceObject.QueryInterface(aIID);
    }
};

var components = [epubContentSniffer];

if (XPCOMUtils.generateNSGetFactory)
    var NSGetFactory = XPCOMUtils.generateNSGetFactory(components);
else
    var NSGetModule = XPCOMUtils.generateNSGetModule(components);
