/*
 *  Lucidor - e-book reader
 *  Copyright © 2009-2010  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

function debug(msg) {
    if (false) {
	try {
	    let cons = Components.classes["@mozilla.org/consoleservice;1"].
		getService(Components.interfaces.nsIConsoleService);
	    cons.logStringMessage(msg);
	} catch (e) {
	}
    }
}

function loop(thing) {
    if (!thing)
	return;
    for (let i = 0; i < thing.length; i++)
	yield thing[i];
}

const CNT_NS = "urn:oasis:names:tc:opendocument:xmlns:container";
const DC_NS = "http://purl.org/dc/elements/1.1/";
const NCX_NS = "http://www.daisy.org/z3986/2005/ncx/";
const OPF_NS = "http://www.idpf.org/2007/opf";
const XHTML_NS = "http://www.w3.org/1999/xhtml";

function epubCreator() {
    this.wrappedJSObject = this;
}

epubCreator.prototype = {
    classDescription: "epubCreator",
    classID: Components.ID("{73ce0ef7-2ae0-47eb-82ae-e63a256b499b}"),
    contractID: "@lucidor.org/epubCreator;1",
    QueryInterface: XPCOMUtils.generateQI([Components.interfaces.nsISupports]),

    init: function(d_meta, d_spine, d_nonspine, d_ncx) {
	debug("init: 0");
	let opf_ns = new Namespace(OPF_NS);
        default xml namespace = opf_ns;

	let pkg = <package version="2.0" unique-identifier="uid" xmlns="http://www.idpf.org/2007/opf">
		<metadata xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:opf="http://www.idpf.org/2007/opf"/>
		<manifest/>
		<spine toc="ncx"/>
		</package>;

	let item;

	debug("init: 1");
	let uid;
	let title;
	let authors = [];

	let dc_ns = new Namespace("dc", DC_NS);
	let opf_ns2 = new Namespace("opf", OPF_NS);
	for (let it in loop(d_meta)) {
	    if (it.name) {
		if (it.name == "unique-identifier") {
		    item = <identifier id="uid"/>;
		    uid = it.value;
		} else {
		    item = new XML("<" + it.name + "/>");
		    if (it.name == "title")
			title = it.value;
		    else if (it.name == "creator")
			authors.push(it.value);	// FIXME
		}
		if (it.namespace)
		    item.setNamespace(it.namespace);
		else
		    item.setNamespace(dc_ns);
		if (it.value)
		    item.appendChild(it.value);
		for (let att in loop(it.attrs)) {
		    item["@" + att.name] = att.value;
		    if (att.namespace)
			item["@" + att.name].setNamespace(att.namespace);
		    else
			item["@" + att.name].setNamespace(opf_ns2);
		}
	    } else {
		item = <meta name={it.meta} content={it.value}/>;
	    }
	    pkg.opf_ns::metadata.appendChild(item);
	}

	item = <item id="ncx" href="toc.ncx" media-type="application/x-dtbncx+xml"/>;
	pkg.opf_ns::manifest.appendChild(item);

	debug("init: 2");

	let id_counter = 0;
	function get_id() {
	    return "id" + id_counter++;
	}

	for (let it in loop(d_spine)) {
	    let sid = get_id();
	    item = <item id={sid} href={it.name} media-type={it.mediatype}/>;
	    pkg.opf_ns::manifest.appendChild(item);

	    item = <itemref idref={sid}/>;
	    if (it.nonlinear)
		item.@linear = "false";
	    pkg.opf_ns::spine.appendChild(item);
	}

	debug("init: 3");

	for (let it in loop(d_nonspine)) {
	    let sid = get_id();
	    item = <item id={sid} href={it.name} media-type={it.mediatype}/>;
	    pkg.opf_ns::manifest.appendChild(item);
	}

	// NCX
	let ncx_ns = new Namespace(NCX_NS);
        default xml namespace = ncx_ns;
	let ncx = <ncx version="2005-1" xmlns="http://www.daisy.org/z3986/2005/ncx/"/>;

	let max_depth = 1;

	if (d_ncx[0].level != undefined) {
	    for (let node in loop(d_ncx))
		if (node.level >= max_depth)
		    max_depth = node.level + 1;
	} else {
	    function get_depth(node, depth) {
		if (depth > max_depth)
		    max_depth = depth;
		for (let cld in loop(node.children))
		    get_depth(cld, depth + 1);
	    };
	    for (let node in loop(d_ncx))
		get_depth(node, 1);
	}

	item = <head><meta name="dtb:uid" content={uid}/>
		<meta name="dtb:depth" content={max_depth}/>
		<meta name="dtb:totalPageCount" content="0"/>
		<meta name="dtb:maxPageNumber" content="0"/></head>;
	ncx.appendChild(item);
	// FIXME docTitle, docAuthor
	item = <docTitle><text>{title}</text></docTitle>;
	ncx.appendChild(item);
	ncx.appendChild(<navMap/>);

	if (d_ncx[0].level != undefined) {
	    let levels = [ ncx.ncx_ns::navMap ];
	    let counter = 1;
	    for (let node in loop(d_ncx)) {
		let point = <navPoint><navLabel><text>{node.title}</text></navLabel>
			<content src={node.href}/></navPoint>;
		//point.content.@src = node.href;
		point.@id = "ncx" + counter;
		point.@playOrder = counter++;
		let parent = levels[node.level];
		parent.appendChild(point);
		levels.splice(node.level + 1, levels.length, point);
	    }
	} else {
	    let counter = 1;
	    function traverse_ncx(node, parent) {
		let point = <navPoint><navLabel><text>{node.title}</text></navLabel>
			<content src={node.href}/></navPoint>;
		//point.content.@src = node.href;
		point.@id = "ncx" + counter;
		point.@playOrder = counter++;
		parent.appendChild(point);
		for (let cld in loop(node.children))
		    traverse_ncx(cld, node);
	    };
	    for (let node in loop(d_ncx))
		traverse_ncx(node, ncx.ncx_ns::navMap);
	}

	debug("init: 4");

	let zip_file = null;
	let appinfo = Components.classes["@mozilla.org/xre/app-info;1"].
	    getService(Components.interfaces.nsIXULAppInfo);
	if (appinfo.ID == "Lucidor@lucidor.org") {
	    let main_control = Components.classes["@lucidor.org/mainControlService;1"].
		getService().wrappedJSObject;
	    zip_file = main_control.get_tempdir();
	} else {
	    let tmp = Components.classes["@mozilla.org/file/directory_service;1"].
		getService(Components.interfaces.nsIProperties).
		get("TmpD", Components.interfaces.nsIFile);
	    tmp.append("lucifox");
	    tmp.createUnique(Components.interfaces.nsIFile.DIRECTORY_TYPE, 0700);
	    zip_file = tmp;
	}
	debug(":" + zip_file.path);
	zip_file.append("lucidor.epub");
	debug(":" + zip_file.path);
	zip_file.createUnique(zip_file.NORMAL_FILE_TYPE, 0600);

	let zw = Components.classes["@mozilla.org/zipwriter;1"].
		createInstance(Components.interfaces.nsIZipWriter);
	zw.open(zip_file, 0x02 | 0x20);

	let modtime = 0;

	debug("init: 6");

	let converter = Components.classes["@mozilla.org/intl/scriptableunicodeconverter"].createInstance(Components.interfaces.nsIScriptableUnicodeConverter);
	converter.charset = "UTF-8";

	let mimetype = "application/epub+zip";
	let stream = converter.convertToInputStream(mimetype);
	zw.addEntryStream("mimetype", modtime, zw.COMPRESSION_NONE,
			  stream, false);

	let container = <container xmlns="urn:oasis:names:tc:opendocument:xmlns:container" version="1.0">
		<rootfiles>
		<rootfile full-path="OPS/package.opf" media-type="application/oebps-package+xml"/>
		</rootfiles>
		</container>;
	let xml_string = container.toXMLString();
	stream = converter.convertToInputStream(xml_string);
	zw.addEntryStream("META-INF/container.xml", modtime,
			  zw.COMPRESSION_DEFAULT, stream, false);

	xml_string = pkg.toXMLString();
	stream = converter.convertToInputStream(xml_string);
	zw.addEntryStream("OPS/package.opf", modtime,
			  zw.COMPRESSION_DEFAULT, stream, false);

	xml_string = ncx.toXMLString();
	stream = converter.convertToInputStream(xml_string);
	zw.addEntryStream("OPS/toc.ncx", modtime,
			  zw.COMPRESSION_DEFAULT, stream, false);

	debug("init: 9");

	for (let it in loop(d_spine)) {
	    let my_modtime = modtime;
	    if (it.string)
		stream = converter.convertToInputStream(it.string);
	    else	// href
		stream = hrefToStream(it.href);
	    zw.addEntryStream("OPS/" + it.name, my_modtime,
			      zw.COMPRESSION_DEFAULT, stream, false);
	}

	for (let it in loop(d_nonspine)) {
	    let my_modtime = modtime;
	    if (it.string)
		stream = converter.convertToInputStream(it.string);
	    else	// href
		stream = hrefToStream(it.href);
	    zw.addEntryStream("OPS/" + it.name, my_modtime,
			      zw.COMPRESSION_DEFAULT, stream, false);
	}

	debug("init: 10");

	zw.close();
	return zip_file;
    },

    generate_titlepage: function(title, subtitle, author, publisher) {
	let xhtml_ns = new Namespace(XHTML_NS);
        default xml namespace = xhtml_ns;

	let extra = null;
	// FIXME doctype
	let doc = <html xmlns="http://www.w3.org/1999/xhtml">
		<head>
		<title>{title}</title>
		</head>
		<body style="background-color: #ffffff; text-align: center; margin: 0px 0px 0px 0px;">
		<div style="font-family: Gentium, 'URW Palladio L', Georgia, serif; margin-bottom: 8em;">
		<div style="color: #7f0000; font-size: 2.4em; margin-top: 4em; margin-bottom: 0.3em; margin-left: 1.5em; margin-right: 1.5em;">{title}</div>
		</div>
		</body>
		</html>;

	let el;
	if (author) {
	    el = <div style="color: #00003f; font-size: 2.2em; font-variant: small-caps; margin-left: 1.5em; margin-right: 1.5em;">{author}</div>;
	    doc.xhtml_ns::body.xhtml_ns::div.appendChild(el);
	}

	if (publisher) {
	    el = <div style="color: #df0000; font-family: 'Doulos SIL', 'MgOpen Canonica', 'Times New Roman', serif; font-size: 1.7em; margin-bottom: 2.5em;">{publisher}</div>;
	    doc.xhtml_ns::body.appendChild(el);
	} else {
	    el = <div style="color: #df0000; font-family: 'Doulos SIL', 'MgOpen Canonica', 'Times New Roman', serif; font-size: 1.7em; margin-bottom: 2.5em;"><object data="lucidor.svg" type="image/svg+xml" style="height: 0.8em;">&#x2605; LUCIDOR &#x2605;</object></div>;
	    doc.xhtml_ns::body.appendChild(el);

	    extra = {name: "lucidor.svg", mediatype: "text/css",
		     href: "chrome://lucidor/content/lucidor.svg"};
	}

	let doctype = '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN"\n    "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">\n';

	return [doctype + doc.toXMLString(), extra];
    }
};

var components = [epubCreator];

if (XPCOMUtils.generateNSGetFactory)
    var NSGetFactory = XPCOMUtils.generateNSGetFactory(components);
else
    var NSGetModule = XPCOMUtils.generateNSGetModule(components);

function hrefToStream(href) {
    let curi = Components.classes["@mozilla.org/network/io-service;1"].
	getService(Components.interfaces.nsIIOService).
	newURI(href, null, null);
    let uri = Components.classes["@mozilla.org/chrome/chrome-registry;1"].
	getService(Components.interfaces.nsIChromeRegistry).
	convertChromeURL(curi);
    let file = Components.classes["@mozilla.org/network/protocol;1?name=file"].
	getService(Components.interfaces.nsIFileProtocolHandler).
	getFileFromURLSpec(uri.spec);
    let is = Components.classes["@mozilla.org/network/file-input-stream;1"].
	createInstance(Components.interfaces.nsIFileInputStream);
    is.init(file, -1, -1, 0);
    return is;
}
