/*
 *  Lucidor - e-book reader
 *  Copyright © 2008-2011  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

function debug(msg) {
    if (false) {
	try {
	    let cons = Components.classes["@mozilla.org/consoleservice;1"].
		getService(Components.interfaces.nsIConsoleService);
	    cons.logStringMessage(msg);
	} catch (e) {
	}
    }
}

function loop(thing) {
    if (!thing)
	return;
    for (let i = 0; i < thing.length; i++)
	yield thing[i];
}

const CNT_NS = "urn:oasis:names:tc:opendocument:xmlns:container";
const DC_NS = "http://purl.org/dc/elements/1.1/";
const ENC_NS = "http://www.w3.org/2001/04/xmlenc#";
const NCX_NS = "http://www.daisy.org/z3986/2005/ncx/";
const OPF_NS = "http://www.idpf.org/2007/opf";
const OPS_NS = "http://www.idpf.org/2007/ops";
const XOPF_NS = "http://lucidor/-/x-opf/";

var is_ws = [];
is_ws[" "] = true;
is_ws["\n"] = true;
is_ws["\t"] = true;
is_ws["\r"] = true;

function get_text(node) {
    let text = node.textContent;

    // Strip beginning and ending whitespace
    let i;
    for (i = 0; i < text.length; i++)
	if (!is_ws[text[i]])
	    break;
    let j;
    for (j = text.length - 1; j > i; j--)
	if (!is_ws[text[j]])
	    break;
    return text.substring(i, j + 1);
}

function epubDocument() {
    this.wrappedJSObject = this;
    this.has_errors = false;
    this.uses_encryption = false;
    this.errors = [];
}

epubDocument.prototype = {
    classDescription: "epubDocument",
    classID: Components.ID("{ccb876a9-eafe-4937-928b-8f54774779d5}"),
    contractID: "@lucidor.org/epubDocument;1",
    QueryInterface: XPCOMUtils.generateQI([Components.interfaces.nsISupports]),

    init: function(fil) {
	debug("epubDocument.init: " + fil.path);

	let fph = Components.classes["@mozilla.org/network/protocol;1?name=file"].getService(Components.interfaces.nsIFileProtocolHandler);
	this.type = "epub";
	this.location = fph.getURLSpecFromFile(fil);
	this.original_location = this.location;
	this.zipreader = Components.classes["@mozilla.org/libjar/zip-reader;1"].
		createInstance(Components.interfaces.nsIZipReader);
	this.zipreader.open(fil);

	//let entries = this.zipreader.findEntries(null);
	//while (entries.hasMore())
	//    debug(">" + entries.getNext());

	// Read OCF Container
	// FIXME should check namespace, but munseys.com do not include it
	let inpst;
	if (this.zipreader.hasEntry("META-INF/container.xml")) {
	    inpst = this.zipreader.getInputStream("META-INF/container.xml");
	} else if (this.zipreader.hasEntry("META_INF/container.xml")) {
	    // FIXME Only workaround for 'Dead Pan and Joe Progress'
	    inpst = this.zipreader.getInputStream("META_INF/container.xml");
	    this.add_error("'META_INF/container.xml' should be 'META-INF/container.xml'");
	} else {
	    throw "Error: no META-INF/container.xml";
	}
	let parser = Components.classes["@mozilla.org/xmlextras/domparser;1"].
		getService(Components.interfaces.nsIDOMParser);
	let ocf_doc = parser.parseFromStream(inpst, null,
					     inpst.available(), "text/xml");
	let doc_elem = ocf_doc.documentElement;
	if (doc_elem.tagName == "parsererror")
	    throw "Error: parsererror 1";

	if ((doc_elem.localName != "container") ||
	    //(doc_elem.namespaceURI != CNT_NS) ||
	    (doc_elem.getAttribute("version") != "1.0"))
	    throw "ERROR: container 1";

	if (doc_elem.namespaceURI != CNT_NS)
	    this.add_error("incorrect namespace in META-INF/container.xml");

	let rootfiles;
	for (let n in loop(doc_elem.childNodes))
	    if ((n.nodeType == doc_elem.ELEMENT_NODE) &&
		//(n.namespaceURI == CNT_NS) &&
		(n.localName == "rootfiles")) {
		rootfiles = n;
		break;
	    }
	if (rootfiles == undefined)
	    throw "ERROR: container 2";

	let full_path;
	for (let n in loop(rootfiles.childNodes))
	    if ((n.nodeType == doc_elem.ELEMENT_NODE) &&
		//(n.namespaceURI == CNT_NS) &&
		(n.localName == "rootfile")) {
		if (n.getAttribute("media-type") ==
		    "application/oebps-package+xml") {
		    full_path = n.getAttribute("full-path");
		    break;
		}
	    }
	if (full_path == undefined)
	    throw "ERROR: container 3";

	// Read OPF Package
	inpst = this.zipreader.getInputStream(full_path);
	let opf_doc = parser.parseFromStream(inpst, null,
					     inpst.available(), "text/xml");

	doc_elem = opf_doc.documentElement;
	if (doc_elem.tagName == "parsererror")
	    throw "Error: parsererror 2";
	if ((doc_elem.localName != "package") ||
	    (doc_elem.namespaceURI != OPF_NS))
	    throw "Error: not OPF or EPUB package";	// FIXME

	this.opf = {};
	this.opf.path = full_path.substring(0, full_path.lastIndexOf("/") + 1);

	if (doc_elem.getAttribute("version") == "2.0") {
	    this.opf.version = 2.0;
	    this.init_opf2(opf_doc);
	} else if (doc_elem.getAttribute("version") == "3.0") {
	    this.opf.version = 3.0;
	    this.init_epub3(opf_doc);
	} else {
	    throw "Error: not OPF 2.0 or EPUB 3.0 package";	// FIXME
	}

	// Calculate hash
	var is = Components.classes["@mozilla.org/network/file-input-stream;1"].
		createInstance(Components.interfaces.nsIFileInputStream);
	is.init(fil, -1, -1, is.CLOSE_ON_EOF);
	var hash = Components.classes["@mozilla.org/security/hash;1"].
		createInstance(Components.interfaces.nsICryptoHash);
	hash.initWithString("SHA1");
	hash.updateFromStream(is, is.available());
	let sha1 = hash.finish(true);
	this.sha1 = base64tohex(sha1);
    },

    init_epub3: function(opf_doc) {
	debug("epub3");
	let doc_elem = opf_doc.documentElement;

	// uid
	let uid_id = doc_elem.getAttribute("unique-identifier");
	if (!uid_id)
	    this.add_error("unique-identifier attribute missing");

	this.init_metadata(opf_doc, uid_id);

	// Read manifest - Create hash from id
	let manifest_elem = opf_doc.getElementsByTagNameNS(OPF_NS, "manifest").item(0);
	debug(manifest_elem);
	this.opf.manifest = {};
	this.opf.manifest_href = {};

	for (let n in loop(manifest_elem.childNodes)) {
	    if (!(n.nodeType == n.ELEMENT_NODE &&
		  n.localName == "item" &&
		  n.namespaceURI == OPF_NS))
		continue;
	    let item = {};
	    item.id = n.getAttribute("id");

	    item.href = n.getAttribute("href");
	    let nu = Components.classes["@mozilla.org/network/util;1"].
		createInstance(Components.interfaces.nsINetUtil);
	    item.href = nu.unescapeString(item.href, nu.ESCAPE_ALL);
	    //item.href = nu.escapeString(item.href, nu.ESCAPE_URL_PATH);

	    item.mediatype = n.getAttribute("media-type");
	    item.fallback = n.getAttribute("fallback");
	    item.properties = n.getAttribute("properties");
	    item.media_overlay = n.getAttribute("media-overlay");
	    item.encryption = null;

	    if (item.properties == "nav") {
		this.opf.navitem = item;
	    }
	    let docs = ["application/xhtml+xml", "image/svg+xml"];
	    if (docs.indexOf(item.mediatype) >= 0)
		item.is_content = true;
	    else
		item.is_content = false;
	    // FIXME Out-Of-Line XML Island (with required fallback) is also a content document

	    this.opf.manifest[item.id] = item;
	    debug("manifest item: " + this.opf.path + item.href);
	    this.opf.manifest_href[this.opf.path + item.href] = item;
	}

	let this_proxy = this;

	// Check chains and set .has_content and .parent
	// Maybe add check for circular chains too
	function check_content(item, parent) {
	    /* FIXME is this allowed?
	    if (item.parent && parent)
		this.add_error("item present in two fallback chains: " +
			       item.id);
	    */
	    if (!item.parent)
		item.parent = parent;
	    if (item.has_content != undefined)
		return;
	    if (item.fallback) {
		let fi = this_proxy.opf.manifest[item.fallback];
		if (fi) {
		    check_content(fi, item);
		    if (fi.has_content)
			item.has_content = true;
		} else {
		    this_proxy.add_error("manifest item referencing missing fallback: " + item.fallback);
		}
	    }
	    if (!item.has_content)
		item.has_content = item.is_content;
	}
	for (let id in this.opf.manifest)
	    check_content(this.opf.manifest[id], null);

	// Spine
	let spine_elem = opf_doc.getElementsByTagNameNS(OPF_NS, "spine").item(0);
	this.opf.spine = new Array();
	this.opf.spine_href = {};
	for (let n in loop(spine_elem.childNodes)) {
	    if (!(n.nodeType == n.ELEMENT_NODE &&
		  n.localName == "itemref" &&
		  n.namespaceURI == OPF_NS))
		continue;
	    let item = {};
	    item.idref = n.getAttribute("idref");
	    if (n.getAttribute("linear") == "no")
		item.linear = false;
	    else
		item.linear = true;
	    item.properties = n.getAttribute("properties");
	    debug("spine item: " + item.idref + " " + item.linear);
	    if (this.opf.manifest[item.idref] == undefined) {
		this.add_error("spine references missing item: " + item.idref);
	    } else {
		let mi = this.opf.manifest[item.idref];
		if (mi.has_content) {
		    this.opf.spine.push(item);
		    this.opf.spine_href[this.opf.path + mi.href] =
			this.opf.spine.length - 1;
		} else {
		    /*
		    this.add_error("spine includes non-content document: " +
				   item.idref);
		    */
		}
	    }
	}

	//this.init_guide(opf_doc);
	this.init_encryption(opf_doc);
	this.init_signatures(opf_doc);
	//this.init_ncx(opf_doc, spine_elem);
	this.init_nav(opf_doc, this.opf.navitem);
    },

    init_opf2: function(opf_doc) {
	let doc_elem = opf_doc.documentElement;

	// uid
	let uid_id = doc_elem.getAttribute("unique-identifier");
	if (!uid_id)
	    this.add_error("unique-identifier attribute missing");

	// // FIXME should not be needed, but xulrunner do not know the opf dtd
	// function get_element_by_id(doc, id) {
	//     let iter = doc.createTreeWalker(doc.documentElement, 0x01,
	// 				    null, false);
	//     let n;
	//     while (n = iter.nextNode())
	// 	if (n.getAttribute("id") == id)
	// 	    return n;
	// }
	// let elem = get_element_by_id(opf_doc, uid_id);
	// //let elem = opf_doc.getElementById(uid_id);
	// if (elem)
	//     this.opf.uid = get_text(elem);
	// debug("UID: " + this.opf.uid);

	this.init_metadata(opf_doc, uid_id);

	// Read manifest - Create hash from id
	let manifest_elem = opf_doc.getElementsByTagNameNS(OPF_NS, "manifest").item(0);
	debug(manifest_elem);
	this.opf.manifest = {};
	this.opf.manifest_href = {};

	for (let n in loop(manifest_elem.childNodes)) {
	    if (!(n.nodeType == n.ELEMENT_NODE &&
		  n.localName == "item" &&
		  n.namespaceURI == OPF_NS))
		continue;
	    let item = {};
	    item.id = n.getAttribute("id");

	    item.href = n.getAttribute("href");
	    let nu = Components.classes["@mozilla.org/network/util;1"].
		createInstance(Components.interfaces.nsINetUtil);
	    item.href = nu.unescapeString(item.href, nu.ESCAPE_ALL);
	    //item.href = nu.escapeString(item.href, nu.ESCAPE_URL_PATH);

	    item.mediatype = n.getAttribute("media-type");
	    item.fallback = n.getAttribute("fallback");
	    item.fallback_style = n.getAttribute("fallback-style");
	    item.required_modules = n.getAttribute("required-modules");
	    item.required_namespace = n.getAttribute("required-namespace");
	    item.encryption = null;

	    // FIXME should not really do this
	    if (item.mediatype == "text/x-oeb1-document")
		item.mediatype = "application/xhtml+xml";

	    let docs = ["application/xhtml+xml", "application/x-dtbook+xml",
			"text/x-oeb1-document"];
	    if (docs.indexOf(item.mediatype) >= 0)
		item.is_content = true;
	    else
		item.is_content = false;
	    // FIXME Out-Of-Line XML Island (with required fallback) is also a content document

	    this.opf.manifest[item.id] = item;
	    debug("manifest item: " + this.opf.path + item.href);
	    this.opf.manifest_href[this.opf.path + item.href] = item;
	}

	let this_proxy = this;

	// Check chains and set .has_content and .parent
	// Maybe add check for circular chains too
	function check_content(item, parent) {
	    /* FIXME is this allowed?
	    if (item.parent && parent)
		this.add_error("item present in two fallback chains: " +
			       item.id);
	    */
	    if (!item.parent)
		item.parent = parent;
	    if (item.has_content != undefined)
		return;
	    if (item.fallback) {
		let fi = this_proxy.opf.manifest[item.fallback];
		if (fi) {
		    check_content(fi, item);
		    if (fi.has_content)
			item.has_content = true;
		} else {
		    this_proxy.add_error("manifest item referencing missing fallback: " + item.fallback);
		}
	    }
	    if (!item.has_content)
		item.has_content = item.is_content;
	}
	for (let id in this.opf.manifest)
	    check_content(this.opf.manifest[id], null);

	// Spine
	let spine_elem = opf_doc.getElementsByTagNameNS(OPF_NS, "spine").item(0);
	this.opf.spine = new Array();
	this.opf.spine_href = {};
	for (let n in loop(spine_elem.childNodes)) {
	    if (!(n.nodeType == n.ELEMENT_NODE &&
		  n.localName == "itemref" &&
		  n.namespaceURI == OPF_NS))
		continue;
	    let item = {};
	    item.idref = n.getAttribute("idref");
	    if (n.getAttribute("linear") == "no")
		item.linear = false;
	    else
		item.linear = true;
	    debug("spine item: " + item.idref + " " + item.linear);
	    if (this.opf.manifest[item.idref] == undefined) {
		this.add_error("spine references missing item: " + item.idref);
	    } else {
		let mi = this.opf.manifest[item.idref];
		if (mi.has_content) {
		    this.opf.spine.push(item);
		    this.opf.spine_href[this.opf.path + mi.href] =
			this.opf.spine.length - 1;
		} else {
		    this.add_error("spine includes non-content document: " +
				   item.idref);
		}
	    }
	}

	// Populate spine_href with fallback items too
	// FIXME are there problems with doing this? check uses of spine_href
	for (let i = 0; i < this.opf.spine.length; i++) {
	    let it = this.opf.manifest[this.opf.spine[i].idref];
	    let fb = it.fallback;
	    let chain = [];
	    while (fb) {
		let item
		let it = this.opf.manifest[fb];
		if (!it)
		    break;
		if (chain.indexOf(it) >= 0)	// Check for circular chains
		    break;
		chain.push(it);
		this.opf.spine_href[this.opf.path + it.href] = i;
		fb = it.fallback;
	    }
	}

	// If there are OPF Content Documents in the manifest which are
	// not in the spine, and are referenced from other places, then
	// add them as nonlinear (see OPF 2.0 spec section 2.4.)
	// FIXME Implementation is not correct, all content documents
	// should not be added, but it is hard to do what the spec says.
	for (let id in this.opf.manifest) {
	    let item = this.opf.manifest[id];
	    if (item.is_content &&
		(this.opf.spine_href[this.opf.path + item.href] == undefined)) {
		let sitem = {};
		sitem.idref = id;
		sitem.linear = false;
		this.opf.spine.push(sitem);
		let href = this.opf.path + item.href;
		this.opf.spine_href[href] = this.opf.spine.length - 1;
	    }
	}

	for (let i in this.opf.spine) {
	    let item = this.opf.spine[i];
	    for (let it = item; it; it = it.fallback) {
		it.spine_chain = i;
	    }
	}

	this.init_guide(opf_doc);
	this.init_encryption(opf_doc);
	this.init_signatures(opf_doc);
	this.init_ncx(opf_doc, spine_elem);
    },

    init_metadata: function(opf_doc, uid_id) {
	// metadata
	this.opf.metadata = {};
	this.opf.metadata.meta = new Array();
	this.opf.metadata.titles = new Array();
	this.opf.metadata.creator = new Array();
	this.opf.metadata.identifier = new Array();

	let metadata_elem = opf_doc.getElementsByTagNameNS(OPF_NS,
							   "metadata").item(0);
	if (!metadata_elem)
	    throw "No metadata";
	let elem = metadata_elem.getElementsByTagNameNS(OPF_NS,
							"dc-metadata").item(0);
	let x_metadata_elem;
	if (elem) {
	    x_metadata_elem = metadata_elem.getElementsByTagNameNS(OPF_NS, "x-metadata").item(0);
	    metadata_elem = elem;
	}

	for (let n in loop(metadata_elem.childNodes)) {
	    if (n.nodeType != n.ELEMENT_NODE)
		continue;
	    debug("metadata: " + n.localName);
	    if (n.namespaceURI == DC_NS) {
		if (n.localName == "title") {
		    let title = {};
		    title.title = get_text(n);
		    title.language = n.getAttribute("xml:lang");	// FIXME is this correct?
		    this.opf.metadata.titles.push(title);
		    if (this.opf.metadata.title == undefined)
			this.opf.metadata.title = title.title;
		} else if (n.localName == "creator") {
		    let creator = {};
		    creator.name = get_text(n);
		    creator.role = n.getAttributeNS(OPF_NS, "role");
		    creator.fileas = n.getAttributeNS(OPF_NS, "file-as");
		    this.opf.metadata.creator.push(creator);
		} else if (n.localName == "subject") {
		    let subject = get_text(n);
		    if (this.opf.metadata.subject == undefined)
			this.opf.metadata.subject = new Array();
		    this.opf.metadata.subject.push(subject);
		} else if (n.localName == "description") {
		    let description = {};
		    description.description = get_text(n);
		    description.language = n.getAttribute("xml:lang");
		    if (this.opf.metadata.descriptions == undefined)
			this.opf.metadata.descriptions = new Array();
		    this.opf.metadata.descriptions.push(description);
		    if (this.opf.metadata.description == undefined)
			this.opf.metadata.description = description.description;
		} else if (n.localName == "publisher") {
		    let publisher = get_text(n);
		    if (this.opf.metadata.publishers == undefined)
			this.opf.metadata.publishers = new Array();
		    this.opf.metadata.publishers.push(publisher);
		    if (this.opf.metadata.publisher == undefined)
			this.opf.metadata.publisher = publisher;
		} else if (n.localName == "contributor") {
		    let contrib = {};
		    contrib.name = get_text(n);
		    contrib.role = n.getAttributeNS(OPF_NS, "role");
		    contrib.fileas = n.getAttributeNS(OPF_NS, "file-as");
		    if (this.opf.metadata.contributor == undefined)
			this.opf.metadata.contributor = new Array();
		    this.opf.metadata.contributor.push(contrib);
		} else if (n.localName == "date") {	// FIXME
		    let date = {};
		    date.date = get_text(n);
		    date.event = n.getAttributeNS(OPF_NS, "event");
		    if (this.opf.metadata.date == undefined)
			this.opf.metadata.date = new Array();
		    this.opf.metadata.date.push(date);
		} else if (n.localName == "type") {
		    let type = get_text(n);
		    if (this.opf.metadata.type == undefined)
			this.opf.metadata.type = new Array();
		    this.opf.metadata.type.push(type);
		} else if (n.localName == "format") {
		    let format = get_text(n);
		    if (this.opf.metadata.format == undefined)
			this.opf.metadata.format = new Array();
		    this.opf.metadata.format.push(format);
		} else if (n.localName == "identifier") {
		    let id = {};
		    id.id = get_text(n);
		    id.scheme = n.getAttributeNS(OPF_NS, "scheme");
		    this.opf.metadata.identifier.push(id);
		    if (n.getAttribute("id") == uid_id) {
			id.uid = true;
			this.opf.uid = id.id;
		    }
		} else if (n.localName == "source") {
		    let source = get_text(n);
		    if (this.opf.metadata.source == undefined)
			this.opf.metadata.source = new Array();
		    this.opf.metadata.source.push(source);
		} else if (n.localName == "language") {
		    let language = get_text(n);
		    if (this.opf.metadata.languages == undefined)
			this.opf.metadata.languages = new Array();
		    this.opf.metadata.languages.push(language);
		    if (this.opf.metadata.language == undefined)
			this.opf.metadata.language = language;
		} else if (n.localName == "relation") {
		    let relation = get_text(n);
		    if (this.opf.metadata.relation == undefined)
			this.opf.metadata.relation = new Array();
		    this.opf.metadata.relation.push(relation);
		} else if (n.localName == "coverage") {
		    let coverage = get_text(n);
		    if (this.opf.metadata.coverage == undefined)
			this.opf.metadata.coverage = new Array();
		    this.opf.metadata.coverage.push(coverage);
		} else if (n.localName == "rights") {
		    let rights = {};
		    rights.rights = get_text(n);
		    rights.language = n.getAttribute("xml:lang");
		    // FIXME use another name, not "rightss"
		    if (this.opf.metadata.rightss == undefined)
			this.opf.metadata.rightss = new Array();
		    this.opf.metadata.rightss.push(rights);
		    if (this.opf.metadata.rights == undefined)
			this.opf.metadata.rights = rights.rights;
		}
	    } else if ((n.localName == "meta") && (n.namespaceURI == DC_NS)) {
		if (n.hasAttribute("property")) {
		    let meta = {property: n.getAttribute("property")};
		    if (n.hasAttribute("refines")) {
			meta.refines = n.getAttribute("refines");
		    }
		    if (n.hasAttribute("scheme")) {
			meta.scheme = n.getAttribute("scheme");
		    }
		    if (meta.property = "modified") {

		    }
		}
	    } else if ((n.localName == "meta") && (n.namespaceURI == OPF_NS)) {
		let meta = {name: n.getAttribute("name"),
			    content: n.getAttribute("content")};
		this.opf.metadata.meta.push(meta);
	    } else if ((n.localName == "autopublication") &&
		       (n.namespaceURI == XOPF_NS)) {
		this.opf.metadata.autopublication = true;
	    }
	}
	if (x_metadata_elem) {
	    for (let n in loop(x_metadata_elem.childNodes)) {
		if ((n.nodeType == n.ELEMENT_NODE) &&
		    (n.localName == "meta") && (n.namespaceURI == OPF_NS)) {
		    let meta = {name: n.getAttribute("name"),
				content: n.getAttribute("content")};
		    this.opf.metadata.meta.push(meta);
		}
	    }
	}

	for (let m in loop(this.opf.metadata.meta))
	    if (m.name == "cover") {
		this.opf.metadata.cover = m.content;
		break;
	    }

	// Find real authors
	let authors = [];
	for (let creator in loop(this.opf.metadata.creator))
	    if (creator.role == "aut")
		authors.push(creator);
	if (authors.length == 0)
	    for (let creator in loop(this.opf.metadata.creator))
		if (!creator.role)
		    authors.push(creator);

	if (authors.length > 0)
	    this.opf.metadata.authors = authors[0].name;
	if (authors.length > 1) {
	    let i;
	    for (i = 1; i < authors.length - 1; i++)
		this.opf.metadata.authors += ", " + authors[i].name;
	    this.opf.metadata.authors += " & " + authors[i].name;
	}

	// Find first author
	if (authors.length > 0)
	    this.opf.metadata.author = authors[0];
	else if (this.opf.metadata.creator.length > 0)
	    this.opf.metadata.author = this.opf.metadata.creator[0];
	else
	    this.opf.metadata.author = {name: "", fileas: ""};	// FIXME

	// Metadata checks
	if (this.opf.metadata.title == undefined) {
	    this.add_error("title metadata missing");
	    this.opf.metadata.title = "[Unknown]";	// FIXME
	    let title = {title: this.opf.metadata.title,
			 lang: null};
	    this.opf.metadata.titles.push(title);
	}
	if (this.opf.metadata.language == undefined)
	    this.add_error("language metadata missing");
	if (this.opf.uid == undefined)
	    this.add_error("uid metadata missing");
	if (this.opf.metadata.date)
	    for (let date in loop(this.opf.metadata.date))
		try {
		    let d = parse_date(date.date);
		} catch (e) {
		    this.add_error("incorrect date string: " + date.date);
		}
	if (this.opf.metadata.languages)
	    for (let lang in loop(this.opf.metadata.languages))
		try {
		    let language_mapper = Components.classes["@lucidor.org/languageMapperService;1"].getService().wrappedJSObject;
		    //language_mapper.init();	// Not necessary for parse_lang
		    let lc = language_mapper.parse_lang(lang);
		    if (lc.future)
			this.add_error("probably incorrect language tag: " +
				       lang);
		    lang.code = lc;
		} catch (e) {
		    this.add_error("incorrect language tag: " + lang);
		}
    },

    init_guide: function(opf_doc) {
	// Guide
	let guide_elem = opf_doc.getElementsByTagNameNS(OPF_NS, "guide");
	if (guide_elem.length > 0) {
	    guide_elem = guide_elem.item(0);
	    this.opf.guide = new Array();
	    for (let n in loop(guide_elem.childNodes)) {
		if (!((n.localName == "reference") &&
		      (n.namespaceURI == OPF_NS)))
		    continue;
		let item = {};
		item.type = n.getAttribute("type");
		item.title = n.getAttribute("title");
		item.href = this.opf.path + n.getAttribute("href");
		let nu = Components.classes["@mozilla.org/network/util;1"].
		    createInstance(Components.interfaces.nsINetUtil);
		item.href = nu.unescapeString(item.href, nu.ESCAPE_ALL);

		this.opf.guide.push(item);
		if ((item.type == "cover") &&
		    (this.opf.guide_cover == undefined))
		    this.opf.guide_cover = item.href;

		if (this.hrefToPart(item.href) < 0) {
		    this.add_error("guide reference does not point to a spine item: " + item.title);
		    item.broken = true;
		}
	    }
	}
    },

    init_encryption: function(opf_doc) {
	let parser = Components.classes["@mozilla.org/xmlextras/domparser;1"].
		getService(Components.interfaces.nsIDOMParser);

	// Read META-INF/encryption.xml
	try {
	    let inpst = this.zipreader.getInputStream("META-INF/encryption.xml");
	    let enc_doc = parser.parseFromStream(inpst, null,
						 inpst.available(), "text/xml");
	    try {
		if (enc_doc.documentElement.tagName == "parsererror")
		    throw "Error: parsererror 5";
		this.parse_encryption(enc_doc);
	    } catch (e) {
		this.add_error("invalid encryption file");
	    }
	} catch (e) {
	}
    },

    init_signatures: function(opf_doc) {
	let parser = Components.classes["@mozilla.org/xmlextras/domparser;1"].
		getService(Components.interfaces.nsIDOMParser);

	// Read META-INF/signatures.xml
	try {
	    let inpst = this.zipreader.getInputStream("META-INF/signatures.xml");
	    let sig_doc = parser.parseFromStream(inpst, null,
						 inpst.available(), "text/xml");
	    try {
		if (sig_doc.documentElement.tagName == "parsererror")
		    throw "Error: parsererror 4";
		//this.parse_signature(sig_doc);
	    } catch (e) {
		this.add_error("invalid signatures file");
	    }
	} catch (e) {
	}
    },

    init_nav: function(opf_doc, item) {
	let parser = Components.classes["@mozilla.org/xmlextras/domparser;1"].
		getService(Components.interfaces.nsIDOMParser);

	debug("INIT_NAV: " + item.href);
	let path = this.opf.path + item.href;
	let nav_doc = null;
	if (item && item.encryption) {
	    return;
	} else if (path) {
	    let inpst = this.zipreader.getInputStream(path);
	    nav_doc = parser.parseFromStream(inpst, null,
					     inpst.available(),
					     "text/xml");
	    try {
		if (nav_doc.documentElement.tagName == "parsererror")
		    throw "Error: parsererror 3";
	    } catch (e) {
		this.add_error("invalid nav file");
	    }
	} else {
	    throw "Error: nav 0";
	}

	let path_prefix = "";
	let i = path.lastIndexOf("/");
	if (i > 0)
	    path_prefix = path.substring(0, i + 1);

	let doc_elem = nav_doc.documentElement;

	function get_list(node) {
	    let list = new Array();
	    for (let n in loop(node.childNodes)) {
		if (n.localName == "li") {
		    let item = {};
		    for (let ch in loop(n.childNodes)) {		    
			if (ch.localName == "span") {
			    item.head = get_text(ch);
			} else if (ch.localName == "a") {
			    item.href = ch.getAttribute("href");
			    if (item.href.substring(0, 1) == "#")
				item.href = path + item.href;
			    else
				item.href = path_prefix + item.href;
			    item.text = get_text(ch);
			    item.title = ch.getAttribute("title");
			    if (!item.text)
				item.text = item.title;
			    if (!item.text)
				item.text = "..."
			} else if (ch.localName == "ol") {
			    item.list = get_list(ch);
			}
		    }
		    list.push(item);
		}
	    }
	    return list;
	}

	function get_nav(node) {
	    let nav = {};
	    nav.type = node.getAttributeNS(OPS_NS, "type");
	    for (let n in loop(node.childNodes)) {
		if (n.localName == "h1") {
		    nav.head = get_text(n);
		} else if (n.localName == "h2") {
		    nav.head = get_text(n);
		} else if (n.localName == "h3") {
		    nav.head = get_text(n);
		} else if (n.localName == "h4") {
		    nav.head = get_text(n);
		} else if (n.localName == "h5") {
		    nav.head = get_text(n);
		} else if (n.localName == "h6") {
		    nav.head = get_text(n);
		} else if (n.localName == "ol") {
		    nav.list = get_list(n);
		}
	    }
	    return nav;
	}

	this.navs = new Array();
	for (let el in loop(doc_elem.getElementsByTagName("nav"))) {
	    let nav = get_nav(el);
	    this.navs.push(nav);
	}
    },

    init_ncx: function(opf_doc, spine_elem) {
	let parser = Components.classes["@mozilla.org/xmlextras/domparser;1"].
		getService(Components.interfaces.nsIDOMParser);

	// Read NCX
	let toc = spine_elem.getAttribute("toc");
	if (!toc) {
	    if (this.opf.version >= 3)
		return;
	    this.add_error("toc attribute missing");
	    // Try to find ncx anyway
	    for (let id in this.opf.manifest) {
		if (this.opf.manifest[id].mediatype ==
		    "application/x-dtbncx+xml") {
		    toc = id;
		    break;
		}
	    }
	    // FIXME Should not do this, only workaround for broken books
	    // like RosettaBooks' "The Father Thing" and "Death and the Senator"
	    if (!toc)
		for (let id in this.opf.manifest) {
		    let href = this.opf.manifest[id].href;
		    if (href.substring(href.length - 4) == ".ncx") {
			toc = id;
			break;
		    }
		}
	}

	if (toc) {
	    let item = this.opf.manifest[toc];
	    let path = this.idrefToHref(toc);
	    if (item && item.encryption) {
	    } else if (path) {
		let inpst = this.zipreader.getInputStream(path);
		let ncx_doc = parser.parseFromStream(inpst, null,
						     inpst.available(),
						     "text/xml");
		try {
		    if (ncx_doc.documentElement.tagName == "parsererror")
			throw "Error: parsererror 3";
		    let p_path = "";
		    let i = path.lastIndexOf("/");
		    if (i > 0)
			p_path = path.substring(0, i + 1);
		    this.parse_ncx(ncx_doc, p_path);
		} catch (e) {
		    this.add_error("invalid ncx file");
		}
	    } else
		this.add_error("incorrect toc attribute");
	}
    },

    parse_ncx: function(ncx_doc, path_prefix) {
	debug("PARSE_NCX");
	let doc_elem = ncx_doc.documentElement;

	// FIXME should check version too, but some documents are missing it
	if ((doc_elem.localName != "ncx") ||
	    (doc_elem.namespaceURI != NCX_NS))
	    throw "Error: ncx";

	if (doc_elem.getAttribute("version") != "2005-1")
	    this.add_error("incorrect ncx version");
	//    throw "Error: wrong ncx version";

	// head, docTitle, docAuthor

	let this_proxy = this;

	// <navMap>
	//let path_prefix = this.opf.path;	// FIXME should this be relative to the ncx file or the opf file?
	function get_points(node) {
	    let me = {};
	    me.playorder = node.getAttribute("playOrder");
	    for (let n in loop(node.childNodes)) {
		if (n.namespaceURI != NCX_NS)
		    continue;
		if (n.localName == "navLabel") {
		    // FIXME may exist several for different languages
		    let text = n.getElementsByTagNameNS(NCX_NS, "text").item(0);
		    if (text)
			me.label = get_text(text);
		    else
			me.label = "* * *";	// FIXME
		} else if (n.localName == "content") {
		    let src = n.getAttribute("src");
		    if (src) {
			me.content = path_prefix + src;
			if (this_proxy.hrefToPart(me.content) < 0) {
			    this_proxy.add_error("navigation point does not point to a spine item: " + me.label);
			    me.broken = true;
			}
		    }
		} else if (n.localName == "navPoint") {
		    let child = get_points(n);
		    if (!me.children)
			me.children = new Array();
		    me.children.push(child);
		}
	    }
	    return me;
	}

	this.ncx = {};
	let elem = doc_elem.getElementsByTagNameNS(NCX_NS, "navMap").item(0);
	if (elem) {
	    this.ncx.navmap = {};
	    for (let n in loop(elem.childNodes)) {
		if (n.namespaceURI != NCX_NS)
		    continue;
		if (n.localName == "navPoint") {
		    let child = get_points(n);
		    if (!this.ncx.navmap.children)
			this.ncx.navmap.children = new Array();
		    this.ncx.navmap.children.push(child);
		} else if (n.localName == "navLabel") {
		    // FIXME may exist several for different languages
		    let text = n.getElementsByTagNameNS(NCX_NS, "text").item(0);
		    if (text)
			this.ncx.navmap.label = get_text(text);
		    else
			this.ncx.navmap.label = "* * *";	// FIXME
		} else if (n.localName == "navInfo") {
		    let src = n.getAttribute("src");
		    if (src)
			this.ncx.navmap.content = path_prefix + src;
		}
	    }
	}

	// <pageList>
	elem = doc_elem.getElementsByTagNameNS(NCX_NS, "pageList").item(0);
	if (elem) {
	    this.ncx.pagelist = {};
	    let list = new Array();
	    this.ncx.pagelist.list = list;
	    for (let p in loop(elem.childNodes)) {
		if (p.namespaceURI != NCX_NS)
		    continue;
		if (p.localName == "pageTarget") {
		    let me = {};
		    for (let n in loop(p.childNodes)) {
			if (n.localName == "navLabel") {
			    // FIXME may exist several for different languages
			    let text = n.getElementsByTagNameNS(NCX_NS, "text").item(0);
			    if (text)
				me.label = get_text(text);
			    else
				me.label = "* * *";	// FIXME
			} else if (n.localName == "content") {
			    let src = n.getAttribute("src");
			    if (src) {
				me.content = path_prefix + src;
				if (this.hrefToPart(me.content) < 0) {
				    this.add_error("navigation point does not point to a spine item: " + me.label);
				    me.broken = true;
				}
			    }
			}
		    }
		    if (me.label && me.content)
			list.push(me);
		} else if (p.localName == "navLabel") {
		    // FIXME may exist several for different languages
		    let text = p.getElementsByTagNameNS(NCX_NS, "text").item(0);
		    if (text)
			this.ncx.pagelist.label = get_text(text);
		} else if (p.localName == "navInfo") {
		    // FIXME may exist several for different languages
		    let text = p.getElementsByTagNameNS(NCX_NS, "text").item(0);
		    if (text)
			this.ncx.pagelist.info = get_text(text);
		}
	    }
	}

	// <navList>
	let elems = doc_elem.getElementsByTagNameNS(NCX_NS, "navList");
	if (elems.length > 0)
	    this.ncx.navlists = new Array();
	for (elem in loop(elems)) {
	    let navlist = {};
	    this.ncx.navlists.push(navlist);
	    let list = new Array();
	    navlist.list = list;
	    for (let p in loop(elem.childNodes)) {
		if (p.namespaceURI != NCX_NS)
		    continue;
		if (p.localName == "navTarget") {
		    let me = {};
		    for (let n in loop(p.childNodes)) {
			if (n.localName == "navLabel") {
			    // FIXME may exist several for different languages
			    let text = n.getElementsByTagNameNS(NCX_NS, "text").item(0);
			    if (text)
				me.label = get_text(text);
			    else
				me.label = "* * *";	// FIXME
			} else if (n.localName == "content") {
			    let src = n.getAttribute("src");
			    if (src) {
				me.content = path_prefix + src;
				if (this.hrefToPart(me.content) < 0) {
				    this.add_error("navigation point does not point to a spine item: " + me.label);
				    me.broken = true;
				}
			    }
			}
		    }
		    if (me.label && me.content)
			list.push(me);
		} else if (p.localName == "navLabel") {
		    // FIXME may exist several for different languages
		    let text = p.getElementsByTagNameNS(NCX_NS, "text").item(0);
		    if (text)
			navlist.label = get_text(text);
		} else if (p.localName == "navInfo") {
		    // FIXME may exist several for different languages
		    let text = p.getElementsByTagNameNS(NCX_NS, "text").item(0);
		    if (text)
			navlist.info = get_text(text);
		}
	    }
	}
    },

    parse_encryption: function(enc_doc) {
	debug("PARSE_ENCRYPTION");
	let doc_elem = enc_doc.documentElement;

	if ((doc_elem.localName != "encryption") ||
	    (doc_elem.namespaceURI != CNT_NS))
	    throw "Error: encryption";

	for (let n in loop(doc_elem.childNodes)) {
	    if (n.namespaceURI != ENC_NS)
		continue;
	    if (n.localName == "EncryptedKey") {
		// Not much can be done with this
	    } else if (n.localName == "EncryptedData") {
		let em = n.getElementsByTagNameNS(ENC_NS, "EncryptionMethod").item(0);
		let alg = null;
		if (em && em.hasAttribute("Algorithm"))
		    alg = em.getAttribute("Algorithm");
		let cd = n.getElementsByTagNameNS(ENC_NS, "CipherReference").item(0);
		if (cd && cd.hasAttribute("URI")) {
		    let uri = cd.getAttribute("URI");
		    debug("encrypted: " + uri);
		    let nu = Components.classes["@mozilla.org/network/util;1"].
			createInstance(Components.interfaces.nsINetUtil);
		    uri = nu.unescapeString(uri, nu.ESCAPE_ALL);

		    let item = this.opf.manifest_href[uri];
		    if (item) {
			item.encryption = true;
			item.alg = alg;
			if (alg == "http://www.idpf.org/2008/embedding") {
			    if (this.make_idpf_key())
				item.decrypt = this.unmangle_idpf;
			} else if (alg == "http://ns.adobe.com/pdf/enc#RC") {
			    if (this.make_adobe_key())
				item.decrypt = this.unmangle_adobe;
			} else {
			    this.set_encryption_warning();
			}
		    }
		}
	    }
	}
    },

    parse_signature: function(sig_doc) {
	debug("PARSE_SIGNATURE");
	let doc_elem = sig_doc.documentElement;

	if ((doc_elem.localName != "signatures") ||
	    (doc_elem.namespaceURI != CNT_NS))
	    throw "Error: signatures";
    },

    add_error: function(msg) {
	this.has_errors = true;
	if (this.errors.indexOf(msg) < 0)
	    this.errors.push(msg);
	debug(msg);
    },

    set_encryption_warning: function() {
	this.uses_encryption = true;
    },

    close: function() {
	this.zipreader.close();
    },

    // FIXME turn into a generator?
    getSpine: function() {
	let arr = new Array();
	for (let i = 0; i < this.opf.spine.length; i++) {
	    let loader = this.idrefToLoader(this.opf.spine[i].idref);
	    debug(this.opf.spine[i].idref);
	    debug("loader: " + loader);
	    let item = { idref: this.opf.spine[i].idref,
			 href: loader[0],
			 stylesheet: loader[1],
			 linear: this.opf.spine[i].linear };
	    //yield item;
	    arr.push(item);
	}
	return arr;
    },

    hrefToPart: function(href) {
	let nhref = href.split("#");
	let part = this.opf.spine_href[nhref[0]];
	debug("hrefToPart: " + href + " : " + part);
	if (part == undefined)
	    part = -1;
	return part;
    },

    hrefToItem: function(href) {
	let nhref = href.split("#");
	let item = this.opf.manifest_href[nhref[0]];
	return item;
    },

    idrefToHref: function(idref) {
	let item = this.opf.manifest[idref];
	if (!item) {
	    debug("ERROR: no such manifest item [1]: " + idref);
	    return null;
	}
	return this.opf.path + item.href;
    },

    idrefToFallbackHref: function(idref) {
	let item = this.opf.manifest[idref];
	if (!item) {
	    debug("ERROR: no such manifest item [2]: " + idref);
	    return null;
	}
	if (item.real == undefined) {
	    let ok_array = ["application/xhtml+xml", "application/x-dtbook+xml",
			    "application/image/svg+xml",
			    "image/gif", "image/jpeg", "image/png",
			    "application/x-dtbncx+xml",
			    "text/plain"];
	    let chain = [];
	    let it = item;
	    while (it) {
		if (ok_array.indexOf(it.mediatype) >= 0)
		    break;
		chain.push(it);
		it = it.fallback;
		if (chain.indexOf(it) >= 0)	// Check for circular chains
		    it = null;
	    }
	    item.real = it;
	    if (it && (item != it)) {
		let href = this.opf.path + it.href;
		if (!this.opf.spine_href[href]) {
		    let part = this.hrefToPart(this.opf.path + item.href);
		    this.opf.spine_href[href] = part;
		}
	    }
	}
	if (!item.real)
	    return null;	// FIXME
	return this.opf.path + item.real.href;
    },

    idrefToLoader: function(idref) {
	let item = this.opf.manifest[idref];
	if (!item) {
	    debug("ERROR: no such manifest item [3]: " + idref);
	    return null;
	}
	if (item.loader == undefined) {
	    let type_ok = ["application/xhtml+xml", "application/x-dtbook+xml",
			   "application/image/svg+xml",
			   "image/gif", "image/jpeg", "image/png",
			   "application/x-dtbncx+xml",
			   "text/plain",
			   "text/html", "video/ogg", "video/mpeg4",
			   "text/css",
			   "application/x-font-ttf",
			   "application/vnd.ms-opentype"];
	    let module_ok = ["structure", "text", "hypertext", "list",
			     "object", "presentation", "edit",
			     "bidirectional-text", "forms", "table", "image",
			     "client-side-image-map", "metainformation",
			     "scripting", "stylesheet", "style-attribute",
			     "link", "base"];
	    //let module_ex = "ruby", "server-side-image-map", "intrinsic-events";
	    let chain = [];
	    let style_it;
	    let it = item;
	    while (it && it.mediatype) {
		debug("id: " + it.id);
		debug("href: " + it.href);
		debug("mediatype: " + it.mediatype);
		if (type_ok.indexOf(it.mediatype) >= 0) {
		    if (it.required_modules) {
			if (it.required_namespace == XHTML_NS) {
			    let mods = it.required_modules.split(",");
			    // FIXME
			}
		    } else if (it.required_namespace) {
			// FIXME
			let ns_ok = ["http://www.w3.org/1999/xhtml",
				     "http://www.daisy.org/z3986/2005/dtbook/",
				     "http://www.w3.org/2000/svg",
				     "http://www.w3.org/1998/Mateh/MathML"];
			if (ns_ok.indexOf(it.required_namespace) >= 0)
			    break;
		    } else
			break;
		}
		debug("FALLBACK");
		if (it.fallback) {
		    chain.push(it);
		    it = this.opf.manifest[it.fallback];
		    if (chain.indexOf(it) >= 0)	// Check for circular chains
			it = null;
		} else if (it.fallback_style) {
		    style_it = it;
		    break;
		} else {
		    it = null;
		}
	    }
	    if (it) {
		debug("1: " + it);
		item.loader = [this.opf.path + it.href, null];
	    } else if (style_it) {
		debug("2");
		let style_href = this.idrefToHref(style_it.fallback_style);
		item.loader = [this.opf.path + style_it.href, style_href];
	    } else {
		debug("3");
		item.loader = null;
	    }
	    if (item.loader) {
		if (!this.opf.spine_href[item.loader[0]]) {
		    let part = this.hrefToPart(this.opf.path + item.href);
		    this.opf.spine_href[item.loader[0]] = part;
		}
	    }
	}
	if (!item.loader)
	    return null;	// FIXME
	return item.loader;
    },

    get_cover: function() {
	debug("COVER");
	if (this.opf.metadata.cover)
	    return this.opf.metadata.cover;
	debug("NO COVER");
	if (this.opf.guide_cover) {
	    // FIXME is there a better way to handle fragment identifiers?
	    let loc = this.opf.guide_cover;
	    let j = loc.indexOf("#");
	    if (j >= 0)
		loc = loc.substring(0, j);
	    let item = this.opf.manifest_href[loc];
	    if (item)
		return item.id;
	}
	let part = this.get_first_part();
	if (part >= 0)
	    return this.opf.spine[part].idref;
	return null;
    },

    get_first_part: function() {
	for (let i = 0; i < this.opf.spine.length; i++)
	    if (this.opf.spine[i].linear)
		return i;
	if (this.opf.spine.length > 0)
	    return 0;
	return -1;
    },

    get_previous_part: function(ch) {
	if (!this.opf.spine[ch] || !this.opf.spine[ch].linear)
	    return -1;
	for (let i = ch - 1; i >= 0; i--)
	    if (this.opf.spine[i].linear)
		return i;
	return -1;
    },

    get_next_part: function(ch) {
	if (!this.opf.spine[ch] || !this.opf.spine[ch].linear)
	    return -1;
	for (let i = ch + 1; i < this.opf.spine.length; i++)
	    if (this.opf.spine[i].linear)
		return i;
	return -1;
    },

    is_encrypted: function(href) {
	let item = this.opf.manifest_href[href];
	if (item && item.encryption)
	    return true;
	return false;
    },

    /*
     * IDPF font obfuscation
     */
    make_idpf_key: function() {
	if (this.opf.idpfKeyData)
	    return true;

	let keyinput = this.opf.uid;
	keyinput = keyinput.replace(/[\x20\x09\x0d\x0a]+/g, "");
	if (keyinput.length == 0)
	    return false;
	let hash = Components.classes["@mozilla.org/security/hash;1"].
	    createInstance(Components.interfaces.nsICryptoHash);
	hash.initWithString("SHA1");
	let keyarray = utf8tobin(keyinput);
	hash.update(keyarray, keyarray.length);
	let sha1 = hash.finish(true);
	this.opf.idpfKeyData = base64tobin(sha1);
	return true;
    },

    make_adobe_key: function() {
	if (this.opf.adobeKeyData)
	    return true;

	let uid = null;
	for (let id in loop(this.opf.metadata.identifier)) {
	    if (id.id.substring(0, 9) == "urn:uuid:") {
		uid = id.id.substring(9).replace(/-/g, "");
		break;
	    }
	}
	if (!uid || uid.length != 32)
	    return false;

	let keyData = new Array(16);
	for (let i = 0; i < 16; i++) {
	    let val = Number("0x" + uid.substring(i * 2, i * 2 + 2));
	    if (!(val >= 0 && val <= 0xff))
		return false;
	    keyData[i] = val;
	}
	this.opf.adobeKeyData = keyData;
	return true;
    },

    unmangle_idpf: function(doc, item, instream) {
	debug("unmangle_idpf");
	let binis = Components.classes["@mozilla.org/binaryinputstream;1"].
            createInstance(Components.interfaces.nsIBinaryInputStream);
	binis.setInputStream(instream);
	let obfuscatedData = binis.readByteArray(1040);
	let data = item.decrypted_data;
	if (!item.decrypted_data) {
	    let keyData = doc.opf.idpfKeyData;
	    let total = obfuscatedData.length;
	    data = new Array(total);
	    let i = 0;
	    for (let j = 0; j < total; j++) {
		data[j] = obfuscatedData[j] ^ keyData[i];
		i++;
		if (i == 20)
		    i = 0;
	    }

	    item.decrypted_data = data;
	}

	let pipe = Components.classes["@mozilla.org/pipe;1"].
            createInstance(Components.interfaces.nsIPipe);
	pipe.init(false, false, 0, 0, null);
	let sis = pipe.inputStream;
	let binout = Components.classes["@mozilla.org/binaryoutputstream;1"].
            createInstance(Components.interfaces.nsIBinaryOutputStream);
	binout.setOutputStream(pipe.outputStream);
	binout.writeByteArray(data, data.length);
	binout.close();

	if (instream.available() > 0) {
	    let mis = Components.classes["@mozilla.org/io/multiplex-input-stream;1"].
		createInstance(Components.interfaces.nsIMultiplexInputStream);
	    mis.appendStream(sis);
	    mis.appendStream(instream);
	    return mis;
	} else
	    return sis;
    },

    unmangle_adobe: function(doc, item, instream) {
	let binis = Components.classes["@mozilla.org/binaryinputstream;1"].
            createInstance(Components.interfaces.nsIBinaryInputStream);
	binis.setInputStream(instream);
	let obfuscatedData = binis.readByteArray(1024);
	let data = item.decrypted_data;
	if (!item.decrypted_data) {
	    let keyData = doc.opf.adobeKeyData;
	    let total = obfuscatedData.length;
	    data = new Array(total);
	    let i = 0;
	    for (let j = 0; j < total; j++) {
		data[j] = obfuscatedData[j] ^ keyData[i];
		i++;
		if (i == 16)
		    i = 0;
	    }

	    item.decrypted_data = data;
	}

	let pipe = Components.classes["@mozilla.org/pipe;1"].
            createInstance(Components.interfaces.nsIPipe);
	pipe.init(false, false, 0, 0, null);
	let sis = pipe.inputStream;
	let binout = Components.classes["@mozilla.org/binaryoutputstream;1"].
            createInstance(Components.interfaces.nsIBinaryOutputStream);
	binout.setOutputStream(pipe.outputStream);
	binout.writeByteArray(data, data.length);
	binout.close();

	if (instream.available() > 0) {
	    let mis = Components.classes["@mozilla.org/io/multiplex-input-stream;1"].
		createInstance(Components.interfaces.nsIMultiplexInputStream);
	    mis.appendStream(sis);
	    mis.appendStream(instream);
	    return mis;
	} else
	    return sis;
    }
};

var components = [epubDocument];

if (XPCOMUtils.generateNSGetFactory)
    var NSGetFactory = XPCOMUtils.generateNSGetFactory(components);
else
    var NSGetModule = XPCOMUtils.generateNSGetModule(components);

function parse_date(dat) {
    let da = {};

    // Date
    let a = /^(\d{4})(?:-(\d{2})(?:-(\d{2}))?)?/.exec(dat);
    if (a) {
	da.year = a[1];
	if (a[2])
	    da.month = a[2];
	if (a[3])
	    da.day = a[3];
	if (a[0].length == dat.length)
	    return da;
	dat = dat.substring(a[0].length);

	// Time
	a = /^T(\d{2}):(\d{2})(?::(\d{2})(?:\.(\d+))?)?/.exec(dat);
	if (a) {
	    da.hours = a[1];
	    da.minutes = a[2];
	    if (a[3])
		da.seconds = a[3];
	    if (a[4])
		da.milliseconds = a[4];
	    if (a[0].length == dat.length)
		return da;
	    dat = dat.substring(a[0].length);

	    // Timezone
	    a = /^(?:Z|(?:[+-](\d{2}):(\d{2})))/.exec(dat);
	    if (a) {
		da.tz = a[0];
		da.tzsign = 0;
		if (dat[0] == "+")
		    da.tzsign = 1;
		else if (dat[0] == "-")
		    da.tzsign = -1;
		if (a[1])
		    da.tzhours = a[1];
		if (a[2])
		    da.tzminutes = a[2];
		if (a[0].length == dat.length)
		    return da;
	    }
	}
    }
    throw "Error";
}

// FIXME
function date_to_date(da) {
    let year = da.year;
    let month = 1;
    let day = 1;
    if (da.month)
	month = da.month;
    if (da.day)
	day = da.day;
    let date = new Date(da.year, month, day);
    date.setUTCHours(a[1]);
    date.setUTCMinutes(a[2]);
    date.setUTCSeconds(a[3]);
    date.setUTCMilliseconds(a[4]);
    let tzsign = 0;
    if (dat[0] == "+")
	tzsign = 1;
    else if (dat[0] == "-")
	tzsign = -1;
    let off = tzsign * (a[1] * 60 * 60 * 1000 + a[2] * 60 * 1000);
    if (off)
	date.setTime(date.getTime() + off);
    return date;
}

function utf8tobin(st) {
    let utf8 = unescape(encodeURIComponent(st));
    let len = utf8.length;
    let bin = new Array(len);
    for (let i = 0; i < len; i++)
	bin[i] = utf8.charCodeAt(i);
    return bin;
}

function base64tobin(input) {
    let alphabet = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";
    input = input.replace(/[^A-Za-z0-9\+\/\=]/g, "");
    let len = input.length;
    let bin = new Array();
    let j = 0;
    for (let i = 0; i < len - 3;) {
	let c1 = alphabet.indexOf(input[i++]);
	let c2 = alphabet.indexOf(input[i++]);
	let c3 = alphabet.indexOf(input[i++]);
	let c4 = alphabet.indexOf(input[i++]);

	bin[j++] = (c1 << 2) | (c2 >> 4);
	if (c3 == 64)
	    break;
	bin[j++] = ((c2 << 4) | (c3 >> 2)) & 0xff;
	if (c4 == 64)
	    break;
	bin[j++] = ((c3 << 6) | c4) & 0xff;
    }
    return bin;
}

function base64tohex(input) {
    let alphabet = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";
    input = input.replace(/[^A-Za-z0-9\+\/\=]/g, "");
    let len = input.length;
    let hex = "";
    for (let i = 0; i < len - 3;) {
	let c1 = alphabet.indexOf(input[i++]);
	let c2 = alphabet.indexOf(input[i++]);
	let c3 = alphabet.indexOf(input[i++]);
	let c4 = alphabet.indexOf(input[i++]);

	let out1 = (c1 << 2) | (c2 >> 4);
	let h = out1.toString(16);
	if (h.length == 1)
	    hex += "0";
	hex += h;

	if (c3 == 64)
	    break;

	let out2 = ((c2 << 4) | (c3 >> 2)) & 0xff;
	h = out2.toString(16);
	if (h.length == 1)
	    hex += "0";
	hex += h;

	if (c4 == 64)
	    break;

	let out3 = ((c3 << 6) | c4) & 0xff;
	h = out3.toString(16);
	if (h.length == 1)
	    hex += "0";
	hex += h;
    }
    return hex;
}
