/*
 *  Lucidor - e-book reader
 *  Copyright © 2009-2010  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

function debug(msg) {
    if (false) {
	try {
	    let cons = Components.classes["@mozilla.org/consoleservice;1"].
		getService(Components.interfaces.nsIConsoleService);
	    cons.logStringMessage(msg);
	} catch (e) {
	}
    }
}

function loop(thing) {
    if (!thing)
	return;
    for (let i = 0; i < thing.length; i++)
	yield thing[i];
}

function epubcatalogCase() {
    this.wrappedJSObject = this;
    this.init0();
}

epubcatalogCase.prototype = {
    classDescription: "epubcatalogCase",
    classID: Components.ID("{b470318f-b0e5-4a1e-87f2-c3bb5dfdc9af}"),
    contractID: "@lucidor.org/epubcatalogCase;1",
    QueryInterface: XPCOMUtils.generateQI([Components.interfaces.nsISupports]),

    catalog_url: "http://www.epubcatalog.org/epubcat/atom",
    db: null,

    info: { title: true,
	    author: true,
	    location: true },

    browsers: [ "new", "hosts" ],

    listers: [ "all" ],

    capabilities: { offline_db: true,
		    standard_search: true,
		    search_meta: true,
		    search_title: true,
		    search_author: true,
		    search_tags: false,
		    search_text: false },

    init0: function() {
	this.db = Components.classes["@lucidor.org/bookdb;1"].createInstance().
		wrappedJSObject;
    },

    init: function(coll, tree) {
	this.collid = coll.id;
	this.tree = tree;

	if (coll.db)
	    this.db.open(coll.db);
	else {
	    this.db.create("epubcatalog");
	    let cman = Components.classes["@lucidor.org/collectionManagerService;1"].
		getService().wrappedJSObject;
	    cman.changeCollectionDB(this.collid, this.db);
	}
	this.set_db();
    },

    set_db: function() {
	this.tree.datasources = this.db.file_uri;
	this.tree.builder.rebuild();
    },

    close: function() {
	if (this.db)
	    this.db.close();
	this.db = null;
    },

    add_book: null,

    remove_book: null,

    get_preview: function(id) {
	return null;
    },

    generate_preview: null,

    add_preview: null,

    get_book_info: function(id) {
	return this.db.get_book_info(id);
    },

    browse_cb: function(path, callback, sid) {
	debug("browse_cb: " + path);

	let nid = this.db.get_node_by_path(path);
	if (nid >= 0)
	    path = nid;

	let node = this.db.get_node(path);
	if (node) {
	    callback(this.db.browse_path(path), sid);
	} else {
	    callback(null, sid);
	}
    },

    sync_cache: function() {
	let this_proxy = this;
	let listener = {
	    handleEvent: function(event) {
		debug("EVENT: " + event);
		try {
		    this_proxy.step2(request.responseText);
		} catch (e) {
		    debug("E: " + e);
		}
	    }
	};
	let error_listener = {
	    handleEvent: function(event) {
		debug("ERROR EVENT: " + event);
	    }
	};

	/*
	let main_control = Components.classes["@lucidor.org/mainControlService;1"].
		getService().wrappedJSObject;
	debug("ONLINE: " + main_control.online());
	*/

	debug("B");
	let request = Components.classes["@mozilla.org/xmlextras/xmlhttprequest;1"].createInstance(Components.interfaces.nsIXMLHttpRequest);
	request.onload = listener;
	request.onerror = error_listener;
	try {
	    debug("URL: " + this.catalog_url);
	    request.open("GET", this.catalog_url, true);
	    request.send(null);
	} catch (e) {
	    debug("E3:" + e);
	    throw "error"; 
	}
	debug("C");
    },

    step2: function(doc) {
	debug("D");
	let this_proxy = this;

	var myFeedResultListener = {
	    handleResult: function(result) {
		debug("handle_result");
		try {
		    this_proxy.parse_results(result);
		} catch (e) {
		    debug("E1: " + e);
		}
	    }
	};

        try {
	    let parser = Components.classes["@mozilla.org/feed-processor;1"].
		createInstance(Components.interfaces.nsIFeedProcessor);
	    parser.listener = myFeedResultListener;
	    parser.parseFromString(doc, null);
        } catch (e) {
	    debug("Error parsing feed." + e);
        }
    },

    parse_results: function(result) {
	debug("parse_results");

	/*
	let new_db = Components.classes["@lucidor.org/bookdb;1"].
	    createInstance().wrappedJSObject;
	new_db.create("epubcatalog");
	 */
	let new_db = this.db;
	let new_entries = false;

	let host_node = new_db.add_or_get_node(null, "hosts", "Custom");
	let new_node = new_db.add_or_get_node(null, "new", "Custom");
	new_db.remove_children(new_node);

        let feed = result.doc.QueryInterface(Components.interfaces.nsIFeed);

	let ioService = Components.classes["@mozilla.org/network/io-service;1"].
		getService(Components.interfaces.nsIIOService);

	let nodelinks = [];
	let links;
	for (let i = 0; i < feed.items.length; i++) {
	    debug("ENTRY");
	    let entry = feed.items.queryElementAt(i, Components.interfaces.nsIFeedEntry);

	    if (!entry.fields.hasKey("links"))
		continue;
	    if (entry.id && new_db.get_id_from_xid(entry.id) >= 0)
		continue;

	    let ent = { };

	    links = entry.fields.getPropertyAsInterface("links", Components.interfaces.nsIArray);
	    for (let j = 0; j < links.length; j++) {
		let link = links.queryElementAt(j, Components.interfaces.nsIPropertyBag2);
		if (link.hasKey("rel") && link.hasKey("href")) {
		    let rel = link.getPropertyAsAUTF8String("rel");
		    if (rel != "alternate")
			continue;
		    let href = link.getPropertyAsAUTF8String("href");
		    try {
			ent.uri = ioService.newURI(href, null, feed.baseURI);
		    } catch (e) {
			continue;
		    }
		    ent.location = ent.uri.spec;
		    break;
		}
	    }

	    if (!ent.location)
		continue;

	    if (entry.id)
		ent.xid = entry.id;
	    if (!ent.xid)
		ent.xid = ent.location;
	    if (new_db.get_id_from_xid(ent.xid) >= 0)
		continue;

	    if (entry.summary)
		ent.author = entry.summary.plainText();
	    ent.type = "epub";
	    if (entry.title)
		ent.title = entry.title.plainText();
	    if (!ent.title)
		ent.title = "[Unknown]";	// FIXME

	    let bid = new_db.add_book_info(ent);
	    new_db.add_child(new_node, bid);
	    new_entries = true;

	    let domain = ent.uri.host;
	    let k = domain.lastIndexOf(".");
	    if (k > 0) {
		let l = domain.lastIndexOf(".", k - 1);
		if (l > 0)
		    domain = domain.substring(l + 1);
	    }
	    let anode = new_db.add_or_get_node(host_node, domain, domain);
	    new_db.add_child(anode, bid);
	}

	/*
	let old_db = this.db;
	this.db = new_db;
	let cman = Components.classes["@lucidor.org/collectionManagerService;1"].
		getService().wrappedJSObject;
	cman.changeCollectionDB(this.collid, this.db);
	this.set_db();
	old_db.delete();
	 */

	this.set_db();
    }
};

var components = [epubcatalogCase];

if (XPCOMUtils.generateNSGetFactory)
    var NSGetFactory = XPCOMUtils.generateNSGetFactory(components);
else
    var NSGetModule = XPCOMUtils.generateNSGetModule(components);
