/*
 *  Lucidor - e-book reader
 *  Copyright © 2009-2010  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

function debug(msg) {
    if (false) {
	try {
	    let cons = Components.classes["@mozilla.org/consoleservice;1"].
		getService(Components.interfaces.nsIConsoleService);
	    cons.logStringMessage(msg);
	} catch (e) {
	}
    }
}

function loop(thing) {
    if (!thing)
	return;
    for (let i = 0; i < thing.length; i++)
	yield thing[i];
}

function folderCase() {
    this.wrappedJSObject = this;
    this.init0();
}

folderCase.prototype = {
    classDescription: "folderCase",
    classID: Components.ID("{d6737d53-88b1-4c15-b25b-da71232f0c26}"),
    contractID: "@lucidor.org/folderCase;1",
    QueryInterface: XPCOMUtils.generateQI([Components.interfaces.nsISupports]),

    description: "Folder Offline Catalog",
    db: null,
    folder: null,
    collid: -1,

    info: { title: true,
	    author: true,
	    location: true },

    browsers: [ "all" ],

    listers: [ "all" ],

    capabilities: { offline_db: true,
		    standard_search: true,
		    search_meta: true,
		    search_title: true,
		    search_author: true,
		    search_tags: false,
		    search_text: false },

    init0: function() {
	this.db = Components.classes["@lucidor.org/bookdb;1"].createInstance().
		wrappedJSObject;
    },

    init: function(coll, tree) {
	this.collid = coll.id;
	this.folder = coll.folder;
	this.tree = tree;

	if (coll.db)
	    this.db.open(coll.db);
	else
	    this.db.create("folder");
	this.set_db();
    },

    set_db: function() {
	this.tree.datasources = this.db.file_uri;
	this.tree.builder.rebuild();
    },

    close: function() {
	if (this.db)
	    this.db.close();
	this.db = null;
    },

    add_book: null,

    remove_book: null,

    get_preview: function(id) {
	return this.db.get_preview(id);
    },

    generate_preview: null,

    add_preview: function(id, url) {
	return this.db.set_preview(id, url);
    },

    get_book_info: function(id) {
	return this.db.get_book_info(id);
    },

    browse_gen: function(path) {
	return this.db.browse_path(path);
    },

    sync_cache: function() {
	debug("sync_cache: " + new Date());

	let ioService = Components.classes["@mozilla.org/network/io-service;1"].
		getService(Components.interfaces.nsIIOService);
	let fph = Components.classes["@mozilla.org/network/protocol;1?name=file"].
		getService(Components.interfaces.nsIFileProtocolHandler);

	let dir;
	try {
	    let uri = ioService.newURI(this.folder, "UTF-8", null);
	    let location = uri.spec;
	    if (uri.schemeIs("resource"))
		location = Components.classes["@mozilla.org/network/protocol;1?name=resource"].
		    getService(Components.interfaces.nsIResProtocolHandler).
		    resolveURI(uri);
	    dir = fph.getFileFromURLSpec(location);
	} catch (e) {
	    debug("ERR: " + e);
	    return;
	}

	let new_db = Components.classes["@lucidor.org/bookdb;1"].
		createInstance().wrappedJSObject;
	new_db.create("folder");

	let root = new_db.add_node(null, dir.path, dir.leafName);
	this.addDirectory(new_db, dir, root);

	let old_db = this.db;
	this.db = new_db;
	let cman = Components.classes["@lucidor.org/collectionManagerService;1"].
		getService().wrappedJSObject;
	cman.changeCollectionDB(this.collid, this.db);
	this.set_db();
	old_db.delete();

	debug("sync_cache finished");
    },

    addDirectory: function(db, dir, node) {
	let empty = true;
	let en = dir.directoryEntries;
	while (en.hasMoreElements()) {
	    let fil = en.getNext();
	    fil.QueryInterface(Components.interfaces.nsIFile);
	    debug("file: " + fil.path);
	    if (fil.isDirectory()) {
		debug("Directory");
		let nnode = db.add_node(node, fil.path, fil.leafName);
		let ok = this.addDirectory(db, fil, nnode);
		if (!ok)
		    db.remove_branch(nnode);
		else
		    empty = false;
	    } else if (fil.isFile()) {
		let fext = fil.leafName.substring(fil.leafName.length - 5,
						  fil.leafName.length);
		if (fext.toLowerCase() == ".epub") {
		    let bid = db.add_book_file(fil);
		    if (bid) {
			empty = false;
			db.add_child(node, bid);
		    }
		}
	    }
	}
	return !empty;
    }
};

var components = [folderCase];

if (XPCOMUtils.generateNSGetFactory)
    var NSGetFactory = XPCOMUtils.generateNSGetFactory(components);
else
    var NSGetModule = XPCOMUtils.generateNSGetModule(components);
