/*
 *  Lucidor - e-book reader
 *  Copyright © 2009-2010  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

function debug(msg) {
    if (false) {
	try {
	    let cons = Components.classes["@mozilla.org/consoleservice;1"].
		getService(Components.interfaces.nsIConsoleService);
	    cons.logStringMessage(msg);
	} catch (e) {
	}
    }
}

function loop(thing) {
    if (!thing)
	return;
    for (let i = 0; i < thing.length; i++)
	yield thing[i];
}

function init_cache(this_proxy, name) {
    let ok = false;
    try {
	let prefs = Components.classes["@mozilla.org/preferences-service;1"].
	    getService(Components.interfaces.nsIPrefService).
	    getBranch("lucidor.");
	let value = prefs.getBoolPref("debug.cache");
	ok = value;
    } catch (e) {
    }

    if (!ok)
	return false;

    if (!this_proxy.debug_cache) {
	let file = Components.classes["@mozilla.org/file/directory_service;1"].
	    getService(Components.interfaces.nsIProperties).
	    get("ProfD", Components.interfaces.nsIFile);
	let fn = "debug_cache";
	if (name)
	    fn = name;
	file.append(fn + ".sqlite");
	let ss = Components.classes["@mozilla.org/storage/service;1"].
	    getService(Components.interfaces.mozIStorageService);
	let db = ss.openDatabase(file);
	if (!db.tableExists("cache")) {
	    db.createTable("cache", "id INTEGER NOT NULL PRIMARY KEY, uri TEXT, doc TEXT");
	}
	this_proxy.debug_cache = db;
    }
    return true;
}

function get_from_cache(this_proxy, url) {
    if (!this_proxy.debug_cache)
	return null;

    let doc = null;
    let statement = this_proxy.debug_cache.createStatement("SELECT doc FROM cache WHERE uri=?1");
    try {
	statement.bindUTF8StringParameter(0, url);
	if (statement.executeStep())
	    doc = statement.getUTF8String(0);
    } finally {
	statement.reset();
    }
    return doc;
}

function put_in_cache(this_proxy, url, doc) {
    if (!this_proxy.debug_cache)
	return;

    let statement = this_proxy.debug_cache.createStatement("INSERT INTO cache (id, uri, doc) VALUES (NULL, ?1, ?2)");
    statement.bindUTF8StringParameter(0, url);
    statement.bindUTF8StringParameter(1, doc);
    statement.execute();
}

const ATOM_NS = "http://www.w3.org/2005/Atom";
//const DC_NS = "http://purl.org/dc/elements/1.1/";
const DC_NS = "http://purl.org/dc/terms";
const GBS_NS = "http://schemas.google.com/books/2008";
const GD_NS = "http://schemas.google.com/g/2005"
const OS_NS = "http://a9.com/-/spec/opensearchrss/1.0/";

function googlebooksCase() {
    this.wrappedJSObject = this;
    this.init0();
}

googlebooksCase.prototype = {
    classDescription: "googlebooksCase",
    classID: Components.ID("{77713ca4-82d6-4a0a-8d4c-42b8b6511aa5}"),
    contractID: "@lucidor.org/googlebooksCase;1",
    QueryInterface: XPCOMUtils.generateQI([Components.interfaces.nsISupports]),

    db: null,
    enable_offline: false,
    debug_cache: null,

    info: { title: true,
	    author: true,
	    location: true },

    browsers: [],

    capabilities: { search_meta: false,
		    search_title: true,
		    search_author: true,
		    search_tags: false,
		    search_text: true },

    init0: function() {
	this.db = Components.classes["@lucidor.org/bookdb;1"].createInstance().
		wrappedJSObject;
    },

    init: function(coll, tree) {
	this.collid = coll.id;
	//this.coll_persistant = coll.persistant;
	this.tree = tree;

	if (coll.db) {
	    this.db.open(coll.db);
	    if (!coll.persistant) {
		let new_db = Components.classes["@lucidor.org/bookdb;1"].
		    createInstance().wrappedJSObject;
		new_db.create("googlebooks");
		this.db.delete();
		this.db = new_db;
		coll.db = null;
	    }
	} else
	    this.db.create("googlebooks");

	if (!coll.db) {
	    let cman = Components.classes["@lucidor.org/collectionManagerService;1"].
		getService().wrappedJSObject;
	    cman.changeCollectionDB(this.collid, this.db);
	}

	this.set_db();
	init_cache(this, "cache-gbs");
    },

    set_db: function() {
	this.tree.datasources = this.db.file_uri;
	this.tree.builder.rebuild();
    },

    close: function() {
	if (this.db)
	    this.db.close();
	this.db = null;
    },

    add_book: null,

    remove_book: null,

    get_preview_url: function(id) {
	return this.db.get_preview(id);
    },

    get_preview: null,

    generate_preview: null,

    add_preview: null,

    get_book_info: function(id) {
	return this.db.get_book_info(id);
    },

    search_cb: function(filter, text, text2, callback, sid) {
	let url = "http://books.google.com/books/feeds/volumes?q=";
	if (filter == "author") {
	    url += "inauthor:\"" + escape(text) + "\"";
	} else if (filter == "meta") {
	    url += escape(text);
	} else if (filter == "title") {
	    url += "intitle:\"" + escape(text) + "\"";
	}
	url += "&as_brr=1";	// full books
	//url += "&as_brr=4";	// public domain (this does not seem to work)
	//url += "&num=20";
	url += "&max-results=20";
	let nid = this.db.add_node(null, url, "Custom");	// FIXME inte Custom

	function gen() {
	    let link = { type: "folder",
			 path: nid,
			 title: "Results",
			 thisnode: true,
			 firstpath: true,
			 links: null };
	    yield link;
	}
	callback(gen(), sid);
	return;
    },

    browse_cb: function(nid, callback, sid) {
	let url;
	debug("NID: " + nid);
	if (nid.substring)
	    if (nid.substring(0, 4) == "url:") {
		url = nid.substring(4);
		nid = this.db.add_node(null, url, "Custom");	// FIXME inte Custom
	    } else if (nid.substring(0, 7) == "search:") {
		url = "http://books.google.com/books/feeds/volumes?q=";
		let tmp = nid.substring(7).split("=");
		if (tmp[0] == "author") {
		    url += "inauthor:";
		} else if (tmp[0] == "title") {
		    url += "intitle:";
		}
		url += "\"" + escape(tmp[1]) + "\"";
		url += "&as_brr=1";	// full books
		url += "&max-results=20";
		nid = this.db.add_node(null, url, "Custom");	// FIXME inte Custom
	    }

	let node = this.db.get_node(nid);
	if (node) {
	    if (node.visited) {
		callback(this.db.browse_path(nid), sid);
		return;
	    }
	    url = node.path;
	} else {
	    callback(null, sid);
	    return;
	}

	debug("URL: " + url);
	let doc = get_from_cache(this, url);
	if (doc && doc != "") {
	    this.step2(doc, url, nid, callback, sid);
	    return;
	}

 	let this_proxy = this;
	let listener = {
	    handleEvent: function(event) {
		debug("EVENT 1: " + event);
		try {
		    put_in_cache(this_proxy, url, request.responseText);
		    this_proxy.step2(request.responseText, url, nid,
				     callback, sid);
		} catch (e) {
		    debug("E: " + e);
		    callback(null, sid);
		}
	    }
	};
	let error_listener = {
	    handleEvent: function(event) {
		debug("ERROR EVENT: " + event);
		callback(null, sid);
	    }
	};

	/*
	let main_control = Components.classes["@lucidor.org/mainControlService;1"].
		getService().wrappedJSObject;
	debug("ONLINE: " + main_control.online());
	*/

	debug("B");
	let request = Components.classes["@mozilla.org/xmlextras/xmlhttprequest;1"].createInstance(Components.interfaces.nsIXMLHttpRequest);
	request.onload = listener;
	request.onerror = error_listener;
	try {
	    request.open("GET", url, true);
	    request.send(null);
	} catch (e) {
	    debug("E3:" + e);
	    callback(null, sid);
	}
    },

    step2: function(doc, url, nid, callback, sid) {
	debug("D");
	let this_proxy = this;

	var myFeedResultListener = {
	    handleResult: function(result) {
		debug("handle_result");
		try {
		    this_proxy.step3(result, this_proxy.db, nid, callback, sid);
		} catch (e) {
		    debug("E1: " + e);
		    callback(null, sid);
		}
	    }
	};

        try {
	    let parser = Components.classes["@mozilla.org/feed-processor;1"].
		createInstance(Components.interfaces.nsIFeedProcessor);
	    parser.listener = myFeedResultListener;
	    let ioService = Components.classes["@mozilla.org/network/io-service;1"].
		getService(Components.interfaces.nsIIOService);
	    let uri = ioService.newURI(url, null, null);
	    parser.parseFromString(doc, uri);
        } catch (e) {
	    debug("Error parsing feed." + e);
	    callback(null, sid);
        }
    },

    step3: function(result, db, node, callback, sid) {
	try {
	    this.parse_results(result, db, node);
	    callback(db.browse_path(node), sid);
	} catch (e) {
	    callback(null, sid);
	}
    },

    parse_results: function(result, db, node) {
	debug("parse_results");

	let next_node = null;
	//result.registerExtensionPrefix("openSearch", OS_NS);
        let feed = result.doc.QueryInterface(Components.interfaces.nsIFeedContainer);

	let ioService = Components.classes["@mozilla.org/network/io-service;1"].
		getService(Components.interfaces.nsIIOService);

	let nodelinks = [];
	let links;
	if (feed.fields.hasKey("links"))
	    links = feed.fields.getPropertyAsInterface("links", Components.interfaces.nsIArray);
	else
	    links = [];
	for (let j = 0; j < links.length; j++) {
	    let link = links.queryElementAt(j, Components.interfaces.nsIPropertyBag2);
	    if (link.hasKey("type") && link.hasKey("rel") &&
		link.hasKey("href")) {
		let type = link.getPropertyAsAUTF8String("type");
		let rel = link.getPropertyAsAUTF8String("rel");
		let ohref = link.getPropertyAsAUTF8String("href");
		let href = null;
		try {
		    href = ioService.newURI(ohref, null, feed.baseURI).spec;
		} catch (e) {
		    continue;
		}

		debug("TOP LINK:" + rel + ":" + href + ":" + type);
		// Not used: previous, first, last
		if (rel == "next" && type == "application/atom+xml") {
		    let title = "...";
		    if (link.hasKey("title"))
			title = link.getPropertyAsAUTF8String("title");
		    let nnode = db.add_node(-2, href, title);
		    db.set_node_next(node, nnode);
		    next_node = href;
		}
	    }
	}

	if (nodelinks.length > 0)
	    db.set_node_links(node, nodelinks);

	// FIXME
	//openSearch:totalResults
	//openSearch:startIndex
	//openSearch:itemsPerPage
	if (feed.fields.hasKey("nulltotalResults") &&
	    feed.fields.hasKey("nullstartIndex") &&
	    feed.fields.hasKey("nullitemsPerPage")) {
	    let tr = Number(feed.fields.get("nulltotalResults"));
	    let si = Number(feed.fields.get("nullstartIndex"));
	    let ipp = Number(feed.fields.get("nullitemsPerPage"));
	    db.set_node_pageinfo(node, tr, si, ipp);
	}

	feed.QueryInterface(Components.interfaces.nsIFeed);
	for (let i = 0; i < feed.items.length; i++) {
	    debug("ENTRY");
	    let entry = feed.items.queryElementAt(i, Components.interfaces.nsIFeedEntry);

	    let ent = this.parse_entry(entry);
	    if (!ent)
		continue;

	    if (ent.type == "epub" || ent.type == "link") {
		let bid = db.add_book_info(ent);
		db.add_child(node, bid);
	    }
	}

	debug("return" + node);
	db.set_node_visited(node);

	return next_node;
    },

    get_full_entry: function(nid, callback) {
	debug("get_full_entry: " + nid);
	let bi = this.db.get_book_info(nid);
	if (!bi || !bi.xid) {
	    callback(null);
	    return false;
	}
	let url = bi.xid;
	//let url = "http://www.google.com/books/feeds/volumes/" + bi.xid;

	debug("URL: " + url);
	let doc = get_from_cache(this, url);
	if (doc && doc != "") {
	    this.full_step2(doc, url, nid, callback);
	    return true;
	}

 	let this_proxy = this;
	let listener = {
	    handleEvent: function(event) {
		debug("EVENT: " + event);
		try {
		    put_in_cache(this_proxy, url, request.responseText);
		    this_proxy.full_step2(request.responseText, url, nid, callback);
		} catch (e) {
		    debug("E: " + e);
		    callback(null);
		}
	    }
	};
	let error_listener = {
	    handleEvent: function(event) {
		debug("ERROR EVENT: " + event);
		callback(null);
	    }
	};

	/*
	let main_control = Components.classes["@lucidor.org/mainControlService;1"].
		getService().wrappedJSObject;
	debug("ONLINE: " + main_control.online());
	*/

	debug("B");
	let request = Components.classes["@mozilla.org/xmlextras/xmlhttprequest;1"].createInstance(Components.interfaces.nsIXMLHttpRequest);
	request.onload = listener;
	request.onerror = error_listener;
	try {
	    request.open("GET", url, true);
	    request.send(null);
	} catch (e) {
	    debug("E3:" + e);
	    callback(null);
	    return false;
	}
	debug("C");
	return true;
    },

    full_step2: function(doc, url, nid, callback) {
	debug("D");
	let this_proxy = this;

	var myFeedResultListener = {
	    handleResult: function(result) {
		try {
		    this_proxy.full_step3(result, this_proxy.db, nid, callback);
		} catch (e) {
		    debug("E1: " + e);
		    callback(null);
		}
	    }
	};

        try {
	    let ioService = Components.classes["@mozilla.org/network/io-service;1"].
		getService(Components.interfaces.nsIIOService);
	    let uri = ioService.newURI(url, null, null);

	    // FIXME nsIFeedProcessor should be able to handle this, but doesn't
	    //let parser = Components.classes["@mozilla.org/feed-processor;1"].
	    //	createInstance(Components.interfaces.nsIFeedProcessor);
	    let parser = Components.classes["@lucidor.org/atomProcessor;1"].
		createInstance().wrappedJSObject;
	    parser.listener = myFeedResultListener;
	    parser.parseFromString(doc, uri);
        } catch (e) {
	    debug("Error parsing feed." + e);
	    callback(null);
        }
    },

    full_step3: function(result, db, node, callback) {
	//let item = result.doc.QueryInterface(Components.interfaces.nsIFeedEntry);
	let item = result.doc;
	let entry = this.parse_entry(item);
	if (entry) {
	    debug(":" + node);
	    db.update_book(node, entry);
	}
	callback(entry);
    },

    parse_entry: function(entry) {
	debug(entry.fields);
	entry.fields.QueryInterface(Components.interfaces.nsIPropertyBag2);
	if (!entry.fields.hasKey("links"))
	    return null;

	let ioService = Components.classes["@mozilla.org/network/io-service;1"].
		getService(Components.interfaces.nsIIOService);
	let ent = { links: [] };
	let link_type, location;
	let links = entry.fields.getPropertyAsInterface("links", Components.interfaces.nsIArray);
	for (let j = 0; j < links.length; j++) {
	    let link = links.queryElementAt(j, Components.interfaces.nsIPropertyBag2);
	    if (link.hasKey("type") && link.hasKey("href") &&
		link.hasKey("rel")) {
		let href = link.getPropertyAsAUTF8String("href");
		let nhref;
		try {
		    let ahref = ioService.newURI(href, null, entry.baseURI);
		    nhref = ahref.spec;
		} catch (e) {
		    nhref = href;
		}
		let typ = link.getPropertyAsAUTF8String("type");
		let rel = link.getPropertyAsAUTF8String("rel");

		let title = null;
		if (link.hasKey("title"))
		    title = link.getPropertyAsAUTF8String("title");

		if (rel == "http://schemas.google.com/books/2008/info") {
		    // FIXME
		    // type == "text/html"
		} else if (rel == "http://schemas.google.com/books/2008/preview") {
		    // FIXME
		    // type == "text/html"
		} else if (rel == "http://schemas.google.com/books/2008/thumbnail") {
		    // FIXME
		    // type == "image/jpeg" or "image/png" (or "image/x-unknown")
		    ent.thumbnail_link = nhref;
		} else if (rel == "http://schemas.google.com/books/2008/annotation") {
		    // FIXME
		} else if (rel == "http://schemas.google.com/books/2008/epubdownload" && (typ == "application/epub" || typ == "application/epub+zip")) {
		    // NOTE Google Books uses incorrect mimetype
		    ent.type = "epub";
		    ent.link = nhref;
		    debug("EPUB: " + ent.link);
		} else if (typ == "application/pdf") {
		    ent.links.push({ href: nhref, title: title,
				     type: typ });
		} else if (typ == "application/xhtml+xml" ||
			   typ == "text/html") {
		    ent.links.push({ href: nhref, title: title,
				     type: typ });
		}
	    }
	}

	if (!ent.type && ent.links.length > 0) {
	    ent.type = "link";
	    ent.link = ent.links[0].href;
	}

	debug("EPUB2: " + ent.link);

	debug("Entry type: " + ent.type);

	if (!ent.type)
	    return null;

        let location_uri;
        try {
	    ent.location_uri = ioService.newURI(ent.link, null, entry.baseURI);
        } catch (e) {
	    debug("E2: " + e);
            return null;
        }

	if (entry.title)
	    ent.title = entry.title.plainText();
	if (!ent.title)
	    ent.title = "[Unknown]";	// FIXME
	let catalog = true;

	debug("Title: " + ent.title);

	// dc:creator
	let creator = entry.fields.get("nullcreator");
	if (creator)
	    ent.author = creator;

	// dc:date
	// dc:description
	// dc:format
	// dc:identifier
	// dc:language
	let language = entry.fields.get("nulllanguage");
	if (language)
	    ent.publisher = language;

	// dc:publisher
	let publisher = entry.fields.get("nullpublisher");
	if (publisher)
	    ent.publisher = publisher;

	// dc:subject
	// dc:title
	// gbs:viewability
	// gbs:embeddability
	// gbs:openAccess
	// gbs:review
	// gd:rating

	/*
	if (entry.authors.length >= 1) {
	    let person = entry.authors.queryElementAt(0, Components.interfaces.nsIFeedPerson);
	    ent.author = person.name;
	}
	 */
	if (ent.location_uri)
	    ent.location = ent.location_uri.spec;
	debug("ID: " + entry.id + ":" + ent.location_uri);
	if (entry.id)
	    ent.xid = entry.id;
	else
	    ent.xid = ent.location;

	return ent;
    },

    sync_cache: null
};

var components = [googlebooksCase];

if (XPCOMUtils.generateNSGetFactory)
    var NSGetFactory = XPCOMUtils.generateNSGetFactory(components);
else
    var NSGetModule = XPCOMUtils.generateNSGetModule(components);
