/*
 *  Lucidor - e-book reader
 *  Copyright © 2009-2010  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

function debug(msg) {
    if (false) {
	try {
	    let cons = Components.classes["@mozilla.org/consoleservice;1"].
		getService(Components.interfaces.nsIConsoleService);
	    cons.logStringMessage(msg);
	} catch (e) {
	}
    }
}

function loop(thing) {
    if (!thing)
	return;
    for (let i = 0; i < thing.length; i++)
	yield thing[i];
}

function get_best_file(files) {
    let best = -1;
    let val = 0;
    for (let i = 0; i < files.length; i++) {
	//debug("Format: " + files[i].format);
	let format = files[i].format.split("; ");
	let v = 0;
	if (format[0] == "application/epub+zip") {	// 300
	    v = 300;
	} else if (format[0] == "text/plain") {		// 100-190
	    v = 100;
	    if (format[1] == "utf-8")
		v += 90;
	    else if (format[1] == "iso-8859-1")
		v += 80;
	    else if (format[1] == "ascii")
		v += 70;
	}
	if (v > val) {
	    best = i;
	    val = v;
	}
    }
    if (val == 0)
	return null;
    return files[best];
}

function saxHandler(db, caller) {
    this.db = db;
    this.caller = caller;
}

saxHandler.prototype = {
    db: null,
    etext: null,
    file: null,
    chars: "",
    value: null,
    last_xid: null,
    files: null,
    caller: null,

    guess: 30000,
    //midpoint: 0.44,
    midpoint: 1,
    texts: 0,

    characters: function(value) {
	//debug("characters: " + value);
	this.chars += value;
    },

    endDocument: function() {
	debug("endDocument");
    },

    endElement: function(uri, localName, qName) {
	//debug("endElement: " + qName);
	if (qName == "rdf:value") {
	    this.value = this.chars;
	//} else if (this.etext && this.texts < 600) {
	} else if (this.etext) {
	    if (qName == "pgterms:etext") {
		//debug("TITLE: " + this.etext.title);
		if (this.etext.friendlytitle)
		    this.etext.title = this.etext.friendlytitle;
		//this.etext.type = "text";
		try {
		    let id = this.etext.xid.substring(5);
		    this.etext.location = "http://www.gutenberg.org/cache/epub/" + id + "/pg" + id + ".epub";
		    this.etext.type = "epub";
		    this.db.add_book_info(this.etext);
		} catch (e) {
		    debug("err: " + e);
		}
		this.etext = null;
	    } else if (qName == "dc:publisher") {
		this.etext.publisher = this.chars;
	    } else if (qName == "dc:title") {
		this.etext.title = this.chars;
	    } else if (qName == "pgterms:friendlytitle") {
		this.etext.friendlytitle = this.chars;
	    } else if (qName == "dc:creator") {
		this.etext.author = this.chars;
	    } else if (qName == "dc:language") {
		this.etext.language = this.value;
		this.value = null;
	    }
	} else if (false && this.file) {	// FIXME
	    if (qName == "dc:format") {
		if (this.value == "application/zip")
		    this.file.zip = true;
		else
		    this.file.format = this.value;
		this.value = null;
	    } else if (qName == "pgterms:file") {
		try {
		if (this.file.xid != this.last_xid) {
		    if (this.texts % 500 == 0 && this.caller)
			this.caller.hello(1 - (1 - this.midpoint) * this.texts / this.guess);
		    this.texts--;

		    if (this.last_xid) {
			let fil = get_best_file(this.files);
			if (fil)
			    this.db.update_file(fil.xid.substring(1),
						fil.path, fil.format);
			else
			    this.db.remove_book_xid(this.last_xid);
		    }
		    this.files = [];
		    this.last_xid = this.file.xid;
		}
		this.files.push(this.file);
		this.file = null;
		} catch (e) {
		    debug("Error: " + e);
		}
	    }
	}
    },

    endPrefixMapping: function(prefix) {
    },

    ignorableWhitespace: function(whitespace) {
    },

    processingInstruction: function(target, data) {
    },

    startDocument: function() {
	debug("startDocument");
    },

    no_index: true,

    startElement: function(uri, localName, qName, attributes) {
	if (this.cancel) {
	    debug("CNANCL");
	    throw "cancel sync";
	}

	if (qName == "pgterms:etext") {
	    this.etext = {};
	    this.etext.xid = attributes.getValueFromQName("rdf:ID");

	    if (this.texts % 500 == 0 && this.caller)
		this.caller.hello(this.midpoint * this.texts /
				  (this.guess + 0.00001));
	    this.texts++;
	} else if (qName == "pgterms:file") {	// FIXME
	    throw "end parsing";

	    this.file = {};
	    this.file.path = attributes.getValueFromQName("rdf:about");

	    if (this.no_index) {
		this.guess = this.texts;
		debug("creating index: " + new Date());
		//this.db.create_xid_index();
		debug("index done: " + new Date());
		this.no_index = false;
	    }

	} else if (qName == "dcterms:isFormatOf") {
	    if (this.file)
		this.file.xid = attributes.getValueFromQName("rdf:resource");
	}
	this.chars = "";
	this.value = null;	// This is ok in this case, but not in general
    },

    startPrefixMapping: function(prefix, uri) {
    }
};

function gutenbergCase() {
    this.wrappedJSObject = this;
    this.init0();
}

gutenbergCase.prototype = {
    classDescription: "gutenbergCase",
    classID: Components.ID("{87f2f971-889d-40b2-a4b7-fa8cc6ad9289}"),
    contractID: "@lucidor.org/gutenbergRDFCase;1",
    QueryInterface: XPCOMUtils.generateQI([Components.interfaces.nsISupports]),

    description: "Project Gutenberg Offline Catalog",
    homepage: "http://gutenberg.org/",
    update_message: "The Project Gutenberg Catalog is very large and takes long time to update (maybe 10 minutes or more).",
    db: null,
    catalog_url: "http://www.gutenberg.org/feeds/catalog.rdf.zip",
    collid: -1,

    info: { title: true,
	    author: true,
	    location: true },

    listers: [ "all" ],

    capabilities: { offline_db: true,
		    standard_search: true,
		    search_meta: true,
		    search_title: true,
		    search_author: true,
		    search_tags: false,
		    search_text: false },

    init0: function() {
	this.db = Components.classes["@lucidor.org/bookdb;1"].createInstance().
		wrappedJSObject;
    },

    init: function(coll, tree) {
	this.collid = coll.id;
	if (coll.url)
	    this.catalog_url = coll.url;
	this.tree = tree;
	if (coll.db) {
	    this.db.open(coll.db);
	} else {
	    this.db.create("gutenberg");
	}
	this.set_db();
    },

    set_db: function() {
	this.tree.datasources = this.db.file_uri;
	this.tree.builder.rebuild();
    },

    close: function() {
	if (this.db)
	    this.db.close();
	this.db = null;
    },

    add_book: null,

    remove_book: null,

    get_preview: function(id) {
	return this.db.get_preview(id);
    },

    generate_preview: null,

    add_preview: function(id, url) {
	return this.db.set_preview(id, url);
    },

    get_book_info: function(id) {
	return this.db.get_book_info(id);
    },

    sync_cache: function(caller) {
	debug("sync_cache: " + new Date());
	this.caller = caller;
	this.cancel = false;

	let ioService = Components.classes["@mozilla.org/network/io-service;1"].
		getService(Components.interfaces.nsIIOService);
	let uri = ioService.newURI(this.catalog_url, null, null);

	let main_control = Components.classes["@lucidor.org/mainControlService;1"].
	    getService().wrappedJSObject;
	let fil = main_control.get_tempdir();
	fil.append("gutenberg");
	fil.createUnique(Components.interfaces.nsIFile.NORMAL_FILE_TYPE, 0600);

	let this_proxy = this;
	let wbp = Components.classes["@mozilla.org/embedding/browser/nsWebBrowserPersist;1"].createInstance(Components.interfaces.nsIWebBrowserPersist);

	wbp.progressListener = {
	    onLocationChange: function(webProgress, request, location) {
	    },

	    onProgressChange: function(webProgress, request, curSelfProgress,
				       maxSelfProgress, curTotalProgress,
				       maxTotalProgress) {
		if (this_proxy.cancel)
		    throw "cancel download";
	    },

	    onSecurityChange: function(webProgress, request, state) {
	    },

	    onStateChange: function(aWebProgress, aRequest, aStateFlags, aStatus) {
		debug("onStateChange");
		if (wbp.currentState == wbp.PERSIST_STATE_FINISHED) {
		    debug("Persister has finished downloading catalog.");
		    this_proxy.unzip_catalog(fil);
		}
	    },

	    onStatusChange: function(webProgress, request, status, message) {
	    },

	    QueryInterface: function(aIID) {
		if (aIID.equals(Components.interfaces.nsIWebProgressListener))
		    return this;
		throw Components.results.NS_NOINTERFACE;
	    }
	};

	debug("DOWNLOAD");
	wbp.saveURI(uri, null, null, null, null, fil);
    },

    unzip_catalog: function(zip_file) {
	debug("unzip_catalog: " + new Date());

	if (this.cancel) {
	    zip_file.remove(false);
	    return -1;
	}

	let zipreader = Components.classes["@mozilla.org/libjar/zip-reader;1"].
		createInstance(Components.interfaces.nsIZipReader);
	try {
	    zipreader.open(zip_file);
	} catch (e) {
	    zip_file.remove(false);
	    return -1;
	}

	if (!zipreader.hasEntry("catalog.rdf")) {
	    zipreader.close();
	    zip_file.remove(false);
	    return -1;
	}

	//let inpst = zipreader.getInputStream("catalog.rdf");

	let main_control = Components.classes["@lucidor.org/mainControlService;1"].
	    getService().wrappedJSObject;
	let cat_file = main_control.get_tempdir();
	cat_file.append("catalog");
	cat_file.createUnique(Components.interfaces.nsIFile.NORMAL_FILE_TYPE, 0600);
	zipreader.extract("catalog.rdf", cat_file);
	zipreader.close();
	zip_file.remove(false);

	let istream = Components.classes["@mozilla.org/network/file-input-stream;1"].createInstance(Components.interfaces.nsIFileInputStream);
	istream.init(cat_file, -1, -1, false);

	let new_db = Components.classes["@lucidor.org/bookdb;1"].
		createInstance().wrappedJSObject;
	new_db.create("gutenberg");

	debug("start sax");
	this.sax = new saxHandler(new_db, this.caller);
	let saxreader = Components.classes["@mozilla.org/saxparser/xmlreader;1"].
	    createInstance(Components.interfaces.nsISAXXMLReader);
	saxreader.contentHandler = this.sax;

	function do_parse() {
	    debug("do_parse");
	    try {
		saxreader.parseFromStream(istream, null, "application/xml");
	    } catch (e) {
		debug("E: " + e);
	    }
	}

	try {
	    if (!this.cancel)	// Prevent race
		this.run_background(do_parse, true);
	} catch (e) {
	    debug("0.3: " + e);
	}

	//istream.close();
	this.sax = null;

	debug("1");
	cat_file.remove(false);

	if (this.cancel) {
	    new_db.delete();
	    debug("Cancel sync");
	    return;
	}

	let old_db = this.db;
	this.db = new_db;
	debug("4");

	let cman = Components.classes["@lucidor.org/collectionManagerService;1"].
		getService().wrappedJSObject;
	cman.changeCollectionDB(this.collid, this.db);

	debug("5");

	this.set_db();
	old_db.delete();

	if (this.caller)
	    this.caller.finish();
	debug("finished unzip_catalog: " + new Date());
    },

    cancel_sync: function() {
	this.cancel = true;
	if (this.sax)
	    this.sax.cancel = true;
    }
};

var components = [gutenbergCase];

if (XPCOMUtils.generateNSGetFactory)
    var NSGetFactory = XPCOMUtils.generateNSGetFactory(components);
else
    var NSGetModule = XPCOMUtils.generateNSGetModule(components);
