/*
 *  Lucidor - e-book reader
 *  Copyright © 2008-2010  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

function debug(msg) {
    if (false) {
	try {
	    let cons = Components.classes["@mozilla.org/consoleservice;1"].
		getService(Components.interfaces.nsIConsoleService);
	    cons.logStringMessage(msg);
	} catch (e) {
	}
    }
}

function loop(thing) {
    if (!thing)
	return;
    for (let i = 0; i < thing.length; i++)
	yield thing[i];
}

function languageMapper() {
    this.wrappedJSObject = this;
    //this.init();
}

languageMapper.prototype = {
    classDescription: "languageMapper",
    classID: Components.ID("{825e64c9-cde4-43c8-b69f-ad3afc9f46bd}"),
    contractID: "@lucidor.org/languageMapperService;1",
    _xpcom_factory: languageMapperFactory,
    QueryInterface: XPCOMUtils.generateQI([Components.interfaces.nsISupports]),

    inited: false,
    quick_table: {},
    language_table: {},
    script_table: {},
    region_table: {},
    variant_table: {},
    grandfathered_table: {},
    redundant_table: {},

    debug: function() {
	debug_langtags();
    },

    init: function() {
	if (this.inited)
	    return;
	this.inited = true;
	let furis = [ "chrome://lucidor/content/language/language-subtags.en.xml" ];

	let chromeRegistry = Components.classes["@mozilla.org/chrome/chrome-registry;1"].getService(Components.interfaces.nsIXULChromeRegistry);
	let currentLocale = chromeRegistry.getSelectedLocale("lucidor");

	let locales = [ currentLocale ];
	let i = currentLocale.indexOf("-");
	if (i > 0)
	    locales.push(currentLocale.substring(0, i));

	for (let locale in loop(locales))
	    furis.push("chrome://lucidor/content/language/language-tags." +
		       locale + ".xml");

	this.init_tables(furis);

	// Workarounds for common incorrect language codes
	if (this.language_table["en"] != undefined) {
	    this.quick_table["english"] = this.language_table["en"];
	    // bookglutton.com
	    this.quick_table["en [dcterms:iso639-2]"] =
		this.language_table["en"];
	    this.quick_table["en [dcterms:iso639-2];"] =
		this.language_table["en"];
	}
	// smashwords.com
	if (this.quick_table["und"] != undefined)
	    this.quick_table["unknown"] = this.quick_table["und"];
    },

    init_tables: function(furis) {
	for (let furi in loop(furis)) {
	    debug(furi);
	    let request = Components.classes["@mozilla.org/xmlextras/xmlhttprequest;1"].createInstance(Components.interfaces.nsIXMLHttpRequest);
	    try {
		request.open("GET", furi, false);
		request.send(null);
	    } catch (e) {
		continue;
	    }
	    let doc = request.responseXML;
	    for (let item in loop(doc.documentElement.childNodes))
		if (item.nodeType == item.ELEMENT_NODE) {
		    let subtag = null;
		    let description = null;
		    for (let child in loop(item.childNodes)) {
			if (child.localName == "subtag")
			    subtag = child.textContent.toLowerCase();
			else if (child.localName == "description")
			    description = child.textContent;
		    }
		    if (subtag && description) {
			if (item.localName == "language")
			    this.language_table[subtag] = description;
			else if (item.localName == "region")
			    this.region_table[subtag] = description;
			else if (item.localName == "script")
			    this.script_table[subtag] = description;
			else if (item.localName == "tag")
			    this.quick_table[subtag] = description;
		    }
		}
	}
    },

    translate: function(code) {
	// Check if code is predefined
	let lcode = code.replace("_", "-");
	lcode = lcode.toLowerCase();
	if (this.quick_table[lcode] != undefined)
	    return this.quick_table[lcode];

	// Try to do real parsing
	try {
	    let lc = this.parse_lang(code);
	    return this.translate_code(lc);
	} catch (x) {
	}

	// Fall-back to just return it as it came
	return code;
    },

    translate_code: function(lc) {
	if (lc.priv) {
	    let str = "Private";
	    str += " (" + lc.toString() + ")";
	    return str;
	} else if (lc.gf) {
	    let str = "Grandfathered";
	    str += " (" + lc.toString() + ")";
	    return str;
	} else {
	    let str = "";
	    if (this.language_table[lc.language])
		str += this.language_table[lc.language];
	    else
		str += lc.language;
	    if (lc.region) {
		if (this.region_table[lc.region])
		    str += " (" + this.region_table[lc.region] + ")";
		else
		    str += " (" + lc.region.toUpperCase() + ")";
	    }
	    return str;
	}
    },

    parse_lang: function(str) {
	let lc = new LangCode();
	lc.init(str);
	return lc;
    }
};

var serviceObject = new languageMapper();

var languageMapperFactory = {
    createInstance: function(outer, aIID) {
	if (outer != null)
	    throw Components.results.NS_ERROR_NO_AGGREGATION;
	return serviceObject.QueryInterface(aIID);
    }
};

var components = [languageMapper];

if (XPCOMUtils.generateNSGetFactory)
    var NSGetFactory = XPCOMUtils.generateNSGetFactory(components);
else
    var NSGetModule = XPCOMUtils.generateNSGetModule(components);

// FIXME not finished
function canonize_tag(lc) {
    if (lc.region)
	while (serviceObject.region_table[lc.region] != undefined)
	    if (serviceObject.region_table[tag].preferred)
		lc.region = serviceObject.region_table[tag].preferred;
	    else
		break;
    if (lc.gf)
	if (serviceObject.grandfathered_table[lc.tag] != undefined)
	    if (serviceObject.grandfathered_table[lc.tag].preferred) {
		;	// FIXME
	    }
    if (!lc.gf && !lc.priv)
	if (serviceObject.redundant_table[lc.tag] != undefined)
	    if (serviceObject.redundant_table[lc.tag].preferred) {
		;	// FIXME
	    }

    lc.tag = lc.toString();
}

function is_grandfathered(tag) {
    if (serviceObject.grandfathered_table[tag] != undefined)
	return true;
    return false;
}

function get_language(ll, pos) {
    let re = /^[a-z]{2,8}$/;
    if (pos >= ll.length || !re.test(ll[pos]))
	return pos;
    pos++;
    if (ll[pos - 1].length < 4) {
	re = /^[a-z]{3}$/;
	let m = Math.min(pos + 3, ll.length);
	for (; pos < m && re.test(ll[pos]); pos++)
	    ;
    }
    return pos;
}

function get_script(str) {
    return /^[a-z]{4}$/.test(str);
}

function get_region(str) {
    return /^(?:[a-z]{2}|[0-9]{3})$/.test(str);
}

function get_variant(str) {
    return /^(?:[a-z0-9]{5,8}|(?:[0-9][a-z0-9]{3}))$/.test(str);
}

function get_extension(ll, pos) {
    let re = /^[a-wyz0-9]$/;
    if (pos >= ll.length || !re.test(ll[pos]))
	return pos;
    re = /^[a-z0-9]{2,8}$/;
    let j;
    for (j = pos + 1; j < ll.length && re.test(ll[j]); j++)
	;
    if (j < pos + 2)
	return pos;
    return j;
}

function get_privateuse(ll, pos) {
    if (pos >= ll.length || ll[pos] != "x")
	return pos;
    let re = /^[a-z0-9]{1,8}$/;
    let j;
    for (j = pos + 1; j < ll.length && re.test(ll[j]); j++)
	;
    if (j < pos + 2)
	return pos;
    return j;
}

function get_privateuse_only(ll, ret) {
    ret.reset();
    let j = get_privateuse(ll, 0);
    if (j == 0 || j != ll.length)
	return false;
    ret.priv = true;
    ret.privateuse = ll;
    return true;
}

function get_grandfathered(ll, ret) {
    ret.reset();
    let re = /^(?:[a-z]{2,3}|i)$/;
    if (!re.test(ll[0]))
	return false;
    re = /^[a-z0-9]{2,8}$/;
    let j;
    for (j = 1; j < 3 && re.test(ll[j]); j++)
	;
    if (j != ll.length || j < 2)
	return false;
    ret.grandfathered = ll;
    ret.gf = true;
    return true;
}

function get_langtag(ll, ret) {
    ret.reset();

    let j = get_language(ll, 0);
    if (j == 0)
	return false;
    ret.language = ll[0];
    if (ret.language.length >= 5)
	ret.future = true;	// There are currently no such language codes
    if (j > 1)
	ret.extlangs = splice(ll, 1, j);
    let i = j;

    if (i < ll.length && get_script(ll[i]))
	ret.script = ll[i++];

    if (i < ll.length && get_region(ll[i]))
	ret.region = ll[i++];

    while (i < ll.length && get_variant(ll[i])) {
	if (!ret.variants)
	    ret.variants = [];
	ret.variants.push(ll[i++]);
    }

    while (i < ll.length) {
	let j = get_extension(ll, i);
	if (j == i)
	    break;
	if (!ret.extensions)
	    ret.extensions = [];
	ret.extensions.push(splice(ll, i, j));
	i = j;
    }

    j = get_privateuse(ll, i);
    if (j > i)
	ret.privateuse = splice(ll, i, j);

    if (j == ll.length)
	return true;
    else
	return false;
}

function splice(a, i, j) {
    let b = new Array();
    for (let k = i; k < j; k++)
	b.push(a[k]);
    return b;
}

function check_langtag(ret) {
    if (ret.extensions)
	for (let i = 0; i < ret.extensions.length - 1; i++) {
	    let singleton = ret.extensions[i][0];
	    for (let j = i + 1; j < ret.extensions.length; j++)
		if (ret.extensions[j][0] == singleton)
		    return false;
	}
    return true;
}

function LangCode() {
}

LangCode.prototype = {
    init: function(str) {
	this.reset();
	this.tag = str.toLowerCase();
	let ll = this.tag.split("-");
	if (is_grandfathered(this.tag)) {
	    get_grandfathered(ll, this);
	    return;
	} else if (get_langtag(ll, this)) {
	    if (check_langtag(this))
		return;
	} else if (get_privateuse_only(ll, this))
	    return;
	throw "Incorrect string";
    },

    reset: function() {
	this.language = null;
	this.extlangs = null;
	this.script = null;
	this.region = null;
	this.variants = null;
	this.extensions = null;
	this.privateuse = null;
	this.grandfathered = null;
	this.priv = false;
	this.gf = false;
	this.future = false;
    },

    canonize: function() {
	canonize_tag(this);
    },

    toString: function() {
	let str;
	if (this.priv) {
	    str = this.format_list(this.privateuse);
	} else if (this.gf) {
	    str = this.format_list(this.grandfathered);
	} else {
	    str = this.language;
	    for (let xl in loop(this.extlangs))
		str += "-" + xl;
	    if (this.script)
		str += "-" + this.script[0].toUpperCase() +
		    this.script.substring(1, 4);
	    if (this.region)
		str += "-" + this.region.toUpperCase();
	    for (let va in loop(this.variants))
		str += "-" + va;
	    for (let xt in loop(this.extensions))
		str += "-" + this.format_list(xt);
	    if (this.privateuse)
		str += "-" + this.format_list(this.privateuse);
	}
	return str;
    },

    format_list: function(ll) {
	let str = "";
	let first = true;
	for (let l in loop(ll)) {
	    if (first) {
		str = l;
		first = false;
	    } else if (l.length == 2)
		str += "-" + l.toUpperCase();
	    else if (l.length == 4)
		str += "-" + l[0].toUpperCase() + l.substring(1);
	    else
		str += "-" + l;
	}
	return str;
    }
};

function debug_langtags() {
    let list = ["sv", "sv-SE", "English", "mN-cYrL-Mn", "zh-gan",
		"sr-Latn", "sr-Latn-CS", "es-419", "sl-IT-nedis", "zh-nedis",
		"de-CH-1996", "en-a-bbb-a-ccc", "en-a-bbb-x-a-ccc",
		"i-enochian", "sl-Latn-IT-nedis", "de-CH-x-phonebk",
		"az-Arab-x-AZE-derbend", "qaa-Qaaa-QM-x-southern",
		"de-Qaaa", "en-a-myExt-b-another", "de-419-DE", "a-DE",
		"ar-a-aaa-b-bbb-a-ccc"];
    list.push("unknown");
    list.push("UND");

    for (let lst in loop(list)) {
	debug("tag: " + lst);
	let lc = new LangCode();
	try {
	    lc.init(lst);
	    for (let a in lc) {
		if (a == "init")
		    break;
		if (a == "extensions") {
		    for (let e in loop(lc.extensions))
			debug("  " + a + ": " + e);
		} else
		    debug("  " + a + ": " + lc[a]);
	    }
	    debug("toString:" + lc.toString());
	} catch (e) {
	    debug("INCORR " + e);
	}
    }
}
