/*
 *  Lucidor - e-book reader
 *  Copyright © 2009-2010  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

function debug(msg) {
    if (false) {
	try {
	    let cons = Components.classes["@mozilla.org/consoleservice;1"].
		getService(Components.interfaces.nsIConsoleService);
	    cons.logStringMessage(msg);
	} catch (e) {
	}
    }
}

function loop(thing) {
    if (!thing)
	return;
    for (let i = 0; i < thing.length; i++)
	yield thing[i];
}

const SCHEME = "x-lucidor-opds";
const ATOM_NS = "http://www.w3.org/2005/Atom";
const DCTERMS_NS = "http://purl.org/dc/terms/";
const OS_NS = "http://a9.com/-/spec/opensearch/1.1/";

function lucidorOpdsProtocol() {
}

lucidorOpdsProtocol.prototype = {
    classDescription: "lucidorOpdsProtocol",
    classID: Components.ID("{5ee14210-a696-45be-92ee-37fc8a3568c7}"),
    contractID: "@mozilla.org/network/protocol;1?name=" + SCHEME,
    QueryInterface: XPCOMUtils.generateQI([Components.interfaces.nsISupports,
					   Components.interfaces.nsIProtocolHandler]),

    defaultPort: -1,
    protocolFlags: Components.interfaces.nsIProtocolHandler.URI_NOAUTH |
	Components.interfaces.nsIProtocolHandler.URI_IS_LOCAL_FILE,
    scheme: SCHEME,

    allowPort: function(port, scheme) {
        return false;
    },

    newURI: function(spec, originCharset, baseURI) {
        let uri = Components.classes["@mozilla.org/network/standard-url;1"].
		createInstance(Components.interfaces.nsIStandardURL);
	uri.init(uri.URLTYPE_STANDARD, this.defaultPort, spec,
		 originCharset, baseURI);

        return uri.QueryInterface(Components.interfaces.nsIURI);
    },

    newChannel: function(URI) {
	debug("My newChannel: " + URI.spec);
	// URL format: x-lucidor-opds://folder/?uri=...&si=...
	URI.QueryInterface(Components.interfaces.nsIURL);
	let params = URI.query.split("&");
	let diruri = null;
	let startIndex = 1;
	for (let par in loop(params)) {
	    let pair = par.split("=");
	    pair[1] = unescape(pair[1]);
	    if (pair[0] == "uri") {
		diruri = pair[1];
	    } else if (pair[0] == "si") {
		let a = parseInt(pair[1], 10);
		if (!isNaN(a))
		    startIndex = Math.floor(a);
	    }
	}

	let opds = null;
	if (URI.host == "folder") {
	    try {
		opds = this.create_opds(diruri, startIndex);
	    } catch (e) {
		debug("error: " + e);
	    }
	}

	let pipe = Components.classes["@mozilla.org/pipe;1"].
            createInstance(Components.interfaces.nsIPipe);
	pipe.init(true, false, 0, 0, null);

	let channel = Components.classes["@mozilla.org/network/input-stream-channel;1"].createInstance(Components.interfaces.nsIInputStreamChannel);
	channel.contentStream = pipe.inputStream;
	channel.setURI(URI);
	channel.QueryInterface(Components.interfaces.nsIChannel);
	channel.contentType = "application/atom+xml";
	channel.contentLength = -1;
	channel.contentCharset = "UTF-8";

	if (opds) {
	    let xs = Components.classes["@mozilla.org/xmlextras/xmlserializer;1"].
		createInstance(Components.interfaces.nsIDOMSerializer);
	    xs.serializeToStream(opds, pipe.outputStream, "UTF-8");
	}
	pipe.outputStream.close();

	return channel;
    },

    create_opds: function(location, startIndex) {
	debug("create_opds:" + location);

	let itemsPerPage = 20;
	let fph = Components.classes["@mozilla.org/network/protocol;1?name=file"].
	    getService(Components.interfaces.nsIFileProtocolHandler);

	let dir;
	try {
	    let ioService = Components.classes["@mozilla.org/network/io-service;1"].
		getService(Components.interfaces.nsIIOService);
	    let uri = ioService.newURI(location, "UTF-8", null);
	    if (uri.schemeIs("resource"))
		location = Components.classes["@mozilla.org/network/protocol;1?name=resource"].
		getService(Components.interfaces.nsIResProtocolHandler).
		resolveURI(uri);
	    dir = fph.getFileFromURLSpec(location);
	} catch (e) {
	    debug("error: " + e);
	    return null;
	}

	let document = Components.classes["@mozilla.org/xml/xml-document;1"].
            createInstance(Components.interfaces.nsIDOMDocument);
	let opds = document.implementation.createDocument(ATOM_NS, "", null);

	let pi = opds.createProcessingInstruction("xml", "version='1.0' encoding='UTF-8'");
	opds.appendChild(pi);

	let feed = opds.createElementNS(ATOM_NS, "feed");
	opds.appendChild(feed);
	feed.setAttribute("xmlns:dcterms", DCTERMS_NS);
	feed.setAttribute("xmlns:opensearch", OS_NS);

	let elem = opds.createElementNS(ATOM_NS, "id");
	let text = opds.createTextNode(location);
	elem.appendChild(text);
	feed.appendChild(elem);

	elem = opds.createElementNS(ATOM_NS, "title");
	text = opds.createTextNode(dir.leafName);
	elem.appendChild(text);
	feed.appendChild(elem);

	let date = new Date();
	function zfill(i) {
	    let st = i.toString();
	    if (st.length < 2)
		st = "0" + st;
	    return st;
	}
	let updated = date.getUTCFullYear() + "-" + zfill(date.getUTCMonth()) +
	    "-" + zfill(date.getUTCDate()) + "T" + zfill(date.getUTCHours()) +
	    ":" + zfill(date.getUTCMinutes()) +
	    ":" + zfill(date.getUTCSeconds()) + "Z";
	elem = opds.createElementNS(ATOM_NS, "updated");
	text = opds.createTextNode(updated);
	elem.appendChild(text);
	feed.appendChild(elem);

	let author = opds.createElementNS(ATOM_NS, "author");
	elem = opds.createElementNS(ATOM_NS, "name");
	text = opds.createTextNode("Lucidor");
	elem.appendChild(text);
	author.appendChild(elem);
	feed.appendChild(author);

	elem = opds.createElementNS(ATOM_NS, "generator");
	elem.setAttribute("uri", "http://lucidor.org/lucidor/");
	text = opds.createTextNode("Lucidor");
	elem.appendChild(text);
	feed.appendChild(elem);

	let num = 0;
	let dentries = dir.directoryEntries;
	while (dentries.hasMoreElements()) {
	    let dentry = dentries.getNext().QueryInterface(Components.interfaces.nsIFile);
	    if (dentry.isHidden() || !dentry.isReadable())
		continue;
	    let loc = fph.getURLSpecFromFile(dentry);
	    let type = null;
	    if (dentry.isFile() && loc.substring(loc.length - 5) == ".epub")
		type = "epub";
	    else if (dentry.isDirectory())
		type = "folder";
	    else
		continue;

	    num += 1;
	    if (num < startIndex)
		continue;
	    if (num >= (startIndex + itemsPerPage))
		continue;

	    let entry = opds.createElementNS(ATOM_NS, "entry");
	    elem = opds.createElementNS(ATOM_NS, "id");
	    text = opds.createTextNode(loc);
	    elem.appendChild(text);
	    entry.appendChild(elem);

	    elem = opds.createElementNS(ATOM_NS, "updated");
	    text = opds.createTextNode(updated);
	    elem.appendChild(text);
	    entry.appendChild(elem);

	    if (type == "folder") {
		elem = opds.createElementNS(ATOM_NS, "title");
		text = opds.createTextNode(dentry.leafName);
		elem.appendChild(text);
		entry.appendChild(elem);

		elem = opds.createElementNS(ATOM_NS, "link");
		elem.setAttribute("href", "x-lucidor-opds://folder/?uri=" +
				  escape(loc));
		elem.setAttribute("type", "application/atom+xml;profile=opds");
		elem.setAttribute("title", dentry.leafName);
		entry.appendChild(elem);
	    } else if (type == "epub") {
		let book = Components.classes["@lucidor.org/epubDocument;1"].
		    createInstance().wrappedJSObject;
		try {
		    book.init(dentry);
		} catch(e) {
		    continue;
		}

		for (let creator in loop(book.opf.metadata.creator)) {
		    if (!creator.role || creator.role == "aut") {
			author = opds.createElementNS(ATOM_NS, "author");
			elem = opds.createElementNS(ATOM_NS, "name");
			text = opds.createTextNode(creator.name);
			elem.appendChild(text);
			author.appendChild(elem);
			entry.appendChild(author);
		    }
		}

		elem = opds.createElementNS(ATOM_NS, "title");
		text = opds.createTextNode(book.opf.metadata.title);
		elem.appendChild(text);
		entry.appendChild(elem);

		if (book.opf.metadata.publisher != undefined) {
		    elem = opds.createElementNS(DCTERMS_NS, "publisher");
		    text = opds.createTextNode(book.opf.metadata.publisher);
		    elem.appendChild(text);
		    entry.appendChild(elem);
		}

		if (book.opf.metadata.languages != undefined) {
		    for (let lang in loop(book.opf.metadata.languages)) {
			elem = opds.createElementNS(DCTERMS_NS, "language");
			text = opds.createTextNode(book.opf.metadata.language);
			elem.appendChild(text);
			entry.appendChild(elem);
		    }
		}

		elem = opds.createElementNS(ATOM_NS, "link");
		elem.setAttribute("href", book.location);
		elem.setAttribute("rel", "http://opds-spec.org/acquisition");
		elem.setAttribute("type", "application/epub+zip");
		entry.appendChild(elem);
	    }
	    feed.appendChild(entry);
	}

	elem = opds.createElementNS(OS_NS, "itemsPerPage");
	text = opds.createTextNode(itemsPerPage);
	elem.appendChild(text);
	feed.appendChild(elem);

	elem = opds.createElementNS(OS_NS, "startIndex");
	text = opds.createTextNode(startIndex);
	elem.appendChild(text);
	feed.appendChild(elem);

	elem = opds.createElementNS(OS_NS, "totalResults");
	text = opds.createTextNode(num);
	elem.appendChild(text);
	feed.appendChild(elem);

	let base = "x-lucidor-opds://folder/?uri=" + escape(location);
	if (startIndex > 1) {
	    elem = opds.createElementNS(ATOM_NS, "link");
	    elem.setAttribute("rel", "first");
	    elem.setAttribute("type", "application/atom+xml");
	    elem.setAttribute("href", base + "&si=1");
	    feed.appendChild(elem);

	    elem = opds.createElementNS(ATOM_NS, "link");
	    elem.setAttribute("rel", "previous");
	    elem.setAttribute("type", "application/atom+xml");
	    elem.setAttribute("href", base + "&si=" +
			      Math.max(startIndex - itemsPerPage, 1));
	    feed.appendChild(elem);
	}
	if ((startIndex + itemsPerPage - 1) < num) {
	    elem = opds.createElementNS(ATOM_NS, "link");
	    elem.setAttribute("rel", "next");
	    elem.setAttribute("type", "application/atom+xml");
	    elem.setAttribute("href", base + "&si=" +
			      (startIndex + itemsPerPage));
	    feed.appendChild(elem);

	    elem = opds.createElementNS(ATOM_NS, "link");
	    elem.setAttribute("rel", "last");
	    elem.setAttribute("type", "application/atom+xml");
	    elem.setAttribute("href", base + "&si=" +
			      (Math.floor((num - 1) / itemsPerPage) * itemsPerPage + 1));
	    feed.appendChild(elem);
	}
	return opds;
    }
};

var components = [lucidorOpdsProtocol];

if (XPCOMUtils.generateNSGetFactory)
    var NSGetFactory = XPCOMUtils.generateNSGetFactory(components);
else
    var NSGetModule = XPCOMUtils.generateNSGetModule(components);
