/*
 *  Lucidor - e-book reader
 *  Copyright © 2008-2010  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

function debug(msg) {
    if (false) {
	try {
	    let cons = Components.classes["@mozilla.org/consoleservice;1"].
		getService(Components.interfaces.nsIConsoleService);
	    cons.logStringMessage(msg);
	} catch (e) {
	}
    }
}

function loop(thing) {
    if (!thing)
	return;
    for (let i = 0; i < thing.length; i++)
	yield thing[i];
}

const XHTML_NS = "http://www.w3.org/1999/xhtml";
const XUL_NS = "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul";

function mainControl() {
    this.wrappedJSObject = this;
    this.tempfiles = [];
    this.tempdir = null;
}

mainControl.prototype = {
    classDescription: "mainControl",
    classID: Components.ID("{df75f575-2d91-444f-adca-b6e01eaecf03}"),
    contractID: "@lucidor.org/mainControlService;1",
    _xpcom_factory: mainControlFactory,
    QueryInterface: XPCOMUtils.generateQI([Components.interfaces.nsISupports]),

    init: function(win) {
	this.main_window = win;
	let tmp = Components.classes["@mozilla.org/file/directory_service;1"].
	    getService(Components.interfaces.nsIProperties).
	    get("TmpD", Components.interfaces.nsIFile);
	tmp.append("lucidor");
	tmp.createUnique(Components.interfaces.nsIFile.DIRECTORY_TYPE, 0700);
	this.tempdir = tmp;
    },

    get_tempdir: function() {
	return this.tempdir.clone();
    },

    remove_tempfiles: function() {
	debug("remove_tempfiles");
	try {
	    if (this.tempdir && this.tempdir.isDirectory())
		this.tempdir.remove(true);
	} catch (e) {
	}
    },

    online: function() {
	return this.main_window.navigator.onLine;
    },

    add_tab: function(location) {
	return this.main_window.add_tab(location);
    },

    add_epub_tab: function(location) {
	return this.main_window.add_epub_tab(location);
    },

    add_web_tab: function(location) {
	return this.main_window.add_web_tab(location);
    },

    add_web_tab_post: function(location, post) {
	return this.main_window.add_web_tab_post(location, post);
    },

    add_web_tab_with_warning: function(location, texts, autohide) {
	let vbox = this.main_window.add_web_tab(location);
	try {
	    this.show_warning_vbox(null, vbox, texts, null, autohide);
	} catch (e) {
	    debug("EEE: " + e);
	}
    },

    open_epub_location: function(location) {
	return this.main_window.open_epub_location(location);
    },

    open_location: function(location) {
	return this.main_window.open_location(location);
    },

    update_bar: function() {
	return this.main_window.update_bar();
    },

    update_tab_title: function(win) {
	return this.main_window.update_tab_title(win);
    },

    toggle_tabbar: function() {
	return this.main_window.toggle_tabbar();
    },

    show_warning: function(win, texts, more_msg, autohide) {
	let document = this.main_window.document;
	let tabpanels = document.getElementById("tabpanels");
	for (let panel in loop(tabpanels.childNodes)) {
	    if (panel.firstChild.firstChild.nextSibling.contentWindow == win) {
		let msg_box = panel.firstChild.firstChild;
		this.show_warning_vbox(win, msg_box, texts, more_msg, autohide);
	    }
	}
    },

    show_warning_vbox: function(win, msg_box, texts, more_msg, autohide, type) {
	let document = msg_box.ownerDocument;

	let new_box = document.createElementNS(XUL_NS, "notification");
	if (type)	// "info", "warning", "critical"
	    new_box.setAttribute("type", type);
	let label = "";
	for (let text_msg in loop(texts)) {
	    if (label)
		label += "\n\n";
	    label += text_msg;
	}
	new_box.setAttribute("label", label);

	if (more_msg) {
	    let button = document.createElementNS(XUL_NS, "button");
	    button.setAttribute("label", more_msg);
	    button.setAttribute("oncommand", "Components.classes['@lucidor.org/mainControlService;1'].getService().wrappedJSObject.show_more(event);");
	    new_box.setUserData("more", win, null);
	    new_box.appendChild(button);
	}

	msg_box.appendChild(new_box);

	if (autohide) {
	    let this_proxy = this;
	    var tim = {
		notify: function(timer) {
		    try {
			new_box.close();
		    } catch (e) {
		    }
		    tim.timer = null;	// break circular reference
		}
	    };

	    let timer = Components.classes["@mozilla.org/timer;1"].
		createInstance(Components.interfaces.nsITimer);
	    tim.timer = timer;	// NOTE circular reference
	    timer.initWithCallback(tim, 15000, timer.TYPE_ONE_SHOT);
	}
    },

    show_more: function(event) {
	event.target.parentNode.getUserData("more").show_details();
    }
};

var serviceObject = new mainControl();

var mainControlFactory = {
    createInstance: function(outer, aIID) {
	if (outer != null)
	    throw Components.results.NS_ERROR_NO_AGGREGATION;
	return serviceObject.QueryInterface(aIID);
    }
};

var components = [mainControl];

if (XPCOMUtils.generateNSGetFactory)
    var NSGetFactory = XPCOMUtils.generateNSGetFactory(components);
else
    var NSGetModule = XPCOMUtils.generateNSGetModule(components);
