/*
 *  Lucidor - e-book reader
 *  Copyright © 2009-2011  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

function debug(msg) {
    if (false) {
	try {
	    let cons = Components.classes["@mozilla.org/consoleservice;1"].
		getService(Components.interfaces.nsIConsoleService);
	    cons.logStringMessage(msg);
	} catch (e) {
	}
    }
}

function loop(thing) {
    if (!thing)
	return;
    for (let i = 0; i < thing.length; i++)
	yield thing[i];
}

function init_cache(this_proxy, name) {
    let ok = false;
    try {
	let prefs = Components.classes["@mozilla.org/preferences-service;1"].
	    getService(Components.interfaces.nsIPrefService).
	    getBranch("lucidor.");
	let value = prefs.getBoolPref("debug.cache");
	ok = value;
    } catch (e) {
    }

    if (!ok)
	return false;

    if (!this_proxy.debug_cache) {
	let file = Components.classes["@mozilla.org/file/directory_service;1"].
	    getService(Components.interfaces.nsIProperties).
	    get("ProfD", Components.interfaces.nsIFile);
	let fn = "debug_cache";
	if (name)
	    fn = name;
	file.append(fn + ".sqlite");
	let ss = Components.classes["@mozilla.org/storage/service;1"].
	    getService(Components.interfaces.mozIStorageService);
	let db = ss.openDatabase(file);
	if (!db.tableExists("cache")) {
	    db.createTable("cache", "id INTEGER NOT NULL PRIMARY KEY, uri TEXT, doc TEXT");
	}
	this_proxy.debug_cache = db;
    }
    return true;
}

function get_from_cache(this_proxy, url) {
    if (!this_proxy.debug_cache)
	return null;

    let doc = null;
    let statement = this_proxy.debug_cache.createStatement("SELECT doc FROM cache WHERE uri=?1");
    try {
	statement.bindUTF8StringParameter(0, url);
	if (statement.executeStep())
	    doc = statement.getUTF8String(0);
    } finally {
	statement.reset();
    }
    return doc;
}

function put_in_cache(this_proxy, url, doc) {
    if (!this_proxy.debug_cache)
	return;

    let statement = this_proxy.debug_cache.createStatement("INSERT INTO cache (id, uri, doc) VALUES (NULL, ?1, ?2)");
    statement.bindUTF8StringParameter(0, url);
    statement.bindUTF8StringParameter(1, doc);
    statement.execute();
}

const OS_NS = "http://a9.com/-/spec/opensearch/1.1/";
const LUCI_NS = "http://lucidor.org/-/x-opds/";

function opensearch() {
    this.type = "opds";
    this.syntax = {};
    this.template = null;
    this.wrappedJSObject = this;
}

opensearch.prototype = {
    get_search_uri: function(phrase) {
	let url = this.template;
	for (let par in this.template_params) {	// FIXME
	    // FIXME hantera { } och ?
	    if (par == "searchTerms") {
		url = url.replace("{searchTerms}", escape(phrase));
	    } else if (this.template_params[par] == "opt") {
		url = url.replace("{" + par + "?}", "");
	    } else if (par == "count") {
		url = url.replace(par, "25");	// FIXME
	    } else if (par == "startIndex") {
		url = url.replace(par, "1");	// FIXME
	    } else if (par == "startPage") {
		url = url.replace(par, "1");	// FIXME
	    } else if (par == "language") {
		url = url.replace(par, "*");
	    } else if (par == "inputEncoding") {
		url = url.replace(par, "UTF-8");	// FIXME
	    } else if (par == "outputEncoding") {
		url = url.replace(par, "UTF-8");	// FIXME
	    } else {	// This should never happen
		url = url.replace(par, "Barbie");
	    }
	}
	return url;
    }
};

function opdsCase() {
    this.wrappedJSObject = this;
    this.init0();
}

opdsCase.prototype = {
    classDescription: "opdsCase",
    classID: Components.ID("{4981452d-396a-4274-9e29-932a3bb99f99}"),
    contractID: "@lucidor.org/opdsCase;1",
    QueryInterface: XPCOMUtils.generateQI([Components.interfaces.nsISupports]),

    catalog_url: null,
    db: null,
    offline_db: null,
    enable_offline: false,
    allow_persistant: true,
    in_memory: true,
    debug_cache: null,

    info: { title: true,
	    author: true,
	    location: true },

    browsers: [ "all" ],

    capabilities: { search_meta: false,
		    search_title: false,
		    search_author: false,
		    search_tags: false,
		    search_text: false },

    init0: function() {
	this.db = Components.classes["@lucidor.org/bookdb;1"].createInstance().
	    wrappedJSObject;
	//this.search_uris = {};
    },

    set_in_memory: function(in_memory) {
	this.in_memory = in_memory;
    },

    init: function(coll, tree) {
	this.collid = coll.id;
	//this.coll_persistant = coll.persistant;
	this.tree = tree;
	this.catalog_url = coll.url;

	if (this.in_memory) {
	    this.db.create_memory();
	} else if (coll.db) {
	    this.db.open(coll.db);
	    if (!coll.persistant) {
		let new_db = Components.classes["@lucidor.org/bookdb;1"].
		    createInstance().wrappedJSObject;
		new_db.create("opds");
		this.db.delete();
		this.db = new_db;
		coll.db = null;
	    }
	} else
	    this.db.create("opds");

	if (!coll.db) {
	    let cman = Components.classes["@lucidor.org/collectionManagerService;1"].
		getService().wrappedJSObject;
	    cman.changeCollectionDB(this.collid, this.db);
	}

	this.set_db();
	init_cache(this, "cache-opds");
    },

    set_db: function() {
	if (this.db.file_uri && !this.in_memory) {
	    this.tree.datasources = this.db.file_uri;
	    this.tree.builder.rebuild();
	}
    },

    close: function() {
	//if (!this.coll_persistant)
	//    this.db.delete();
	if (this.db)
	    this.db.close();
	this.db = null;
    },

    add_book: null,

    remove_book: null,

    get_preview_url: function(id) {
	return this.db.get_preview(id);
    },

    get_preview: null,

    generate_preview: null,

    add_preview: null,

    get_book_info: function(id) {
	return this.db.get_book_info(id);
    },

    browse_cb: function(nid, callback, sid) {
	let url;
	debug("NID: " + nid);
	if (nid.substring)
	    if (nid == "all") {
		url = this.catalog_url;
		nid = this.db.add_or_get_node(null, url, "All");
	    } else if (nid.substring(0, 4) == "uri:") {
		url = nid.substring(4);
		nid = this.db.add_node(null, url, "...");	// FIXME not ...
	    }

	let node = this.db.get_node(nid);
	if (node) {
	    if (node.visited) {
		callback(this.db.browse_path(nid), sid);
		return;
	    }
	    url = node.path;
	} else {
	    callback(null, sid);
	    return;
	}

	debug("URL: " + url);
	let doc = get_from_cache(this, url);
	if (doc && doc != "") {
	    this.step2(doc, url, nid, callback, sid);
	    return;
	}

 	let this_proxy = this;
	let listener = {
	    handleEvent: function(event) {
		debug("EVENT 1: " + event);
		try {
		    put_in_cache(this_proxy, url, request.responseText);
		    this_proxy.step2(request.responseText, url, nid,
				     callback, sid);
		} catch (e) {
		    debug("E: " + e);
		    callback(null, sid);
		}
	    }
	};
	let error_listener = {
	    handleEvent: function(event) {
		debug("ERROR EVENT: " + event);
		callback(null, sid);
	    }
	};

	/*
	let main_control = Components.classes["@lucidor.org/mainControlService;1"].
		getService().wrappedJSObject;
	debug("ONLINE: " + main_control.online());
	*/

	let request = Components.classes["@mozilla.org/xmlextras/xmlhttprequest;1"].createInstance(Components.interfaces.nsIXMLHttpRequest);
	request.onload = listener;
	request.onerror = error_listener;
	try {
	    request.open("GET", url, true);
	    request.send(null);
	} catch (e) {
	    debug("E3:" + e);
	    callback(null, sid);
	}
    },

    step2: function(doc, url, nid, callback, sid) {
	debug("D");
	let this_proxy = this;

	var myFeedResultListener = {
	    handleResult: function(result) {
		debug("handle_result");
		try {
		    this_proxy.step3(result, this_proxy.db, nid, callback, sid);
		    return;
		} catch (e) {
		    debug("E1: " + e);
		    callback(null, sid);
		}
	    }
	};

        try {
	    let parser = Components.classes["@mozilla.org/feed-processor;1"].
		createInstance(Components.interfaces.nsIFeedProcessor);
	    parser.listener = myFeedResultListener;
	    let ioService = Components.classes["@mozilla.org/network/io-service;1"].
		getService(Components.interfaces.nsIIOService);
	    let uri = ioService.newURI(url, null, null);
	    parser.parseFromString(doc, uri);
        } catch (e) {
	    debug("Error parsing feed." + e);
	    callback(null, sid);
        }
    },

    step3: function(result, db, node, callback, sid) {
	try {
	    this.parse_results(result, db, node, false);
	    callback(db.browse_path(node), sid);
	} catch (e) {
	    callback(null, sid);
	}
    },

    parse_results: function(result, db, node, update_mode) {
	debug("parse_results");

	let next_node = null;
        let feed = result.doc.QueryInterface(Components.interfaces.nsIFeedContainer);

	/*
	let catalog_title, catalog_author, updated;
	if (feed.title)
	    catalog_title = feed.title.plainText();
	if (feed.subtitle)
	    catalog_subtitle = feed.subtitle.plainText();
	if (feed.authors.length >= 1) {
	    let person = entry.authors.queryElementAt(0, Components.interfaces.nsIFeedPerson);
	    catalog_author = person.name;
	}
	if (feed.updated)
	    catalog_updated = feed.updated;
	*/

	let ioService = Components.classes["@mozilla.org/network/io-service;1"].
		getService(Components.interfaces.nsIIOService);

	let nodelinks = [];
	let links;
	if (feed.fields.hasKey("links"))
	    links = feed.fields.getPropertyAsInterface("links", Components.interfaces.nsIArray);
	else
	    links = [];
	for (let j = 0; j < links.length; j++) {
	    let link = links.queryElementAt(j, Components.interfaces.nsIPropertyBag2);
	    if (!link.hasKey("type"))
		continue;
	    let type = link.getPropertyAsAUTF8String("type");
	    type = type.replace(";profile=opds-catalog", "")
	    type = type.replace(";type=feed", "")
	    if (!link.hasKey("rel"))
		continue;
	    let rel = link.getPropertyAsAUTF8String("rel");
	    let title = null;
	    if (link.hasKey("title"))
		title = link.getPropertyAsAUTF8String("title");
	    let href = null;
	    if (link.hasKey("href")) {
		let ohref = link.getPropertyAsAUTF8String("href");
		try {
		    href = ioService.newURI(ohref, null, feed.baseURI).spec;
		} catch (e) {
		}

		// index / start
		// previous, first, last
		if (rel == "http://opds-spec.org/bookshelf") {
		} else if (rel == "http://opds-spec.org/subscriptions") {
		} else if (rel == "http://opds-spec.org/crawlable") {
		    // Ignore
		} else if (rel == "http://opds-spec.org/sort/new" &&
			   type == "application/atom+xml") {
		    if (!title)
			title = "New";
		    nodelinks.push({ href: href, title: title,
				     type: "toplink" });
		} else if (rel == "http://opds-spec.org/sort/popular" &&
			   type == "application/atom+xml") {
		    if (!title)
			title = "Popular";
		    nodelinks.push({ href: href, title: title,
				     type: "toplink" });
		} else if (rel == "http://opds-spec.org/sort/featured" &&
			   type == "application/atom+xml") {
		    if (!title)
			title = "Featured";
		    nodelinks.push({ href: href, title: title,
				     type: "toplink" });
		} else if (rel == "related" && type == "application/atom+xml") {
		    if (!title)
			title = "Related";
		    nodelinks.push({ href: href, title: title,
				     type: "toplink" });
		} else if (rel == "next" && type == "application/atom+xml") {
		    let title = "...";
		    if (link.hasKey("title"))
			title = link.getPropertyAsAUTF8String("title");
		    if (!update_mode) {
			let nnode = db.add_node(-2, href, title);
			db.set_node_next(node, nnode);
		    }
		    next_node = href;
		} else if (rel == "search") {
		    if (type == "application/opensearchdescription+xml") {
			nodelinks.push({ href: href, type: "opensearch" });
		    } else if (type == "application/atom+xml") {
			// FIXME old style search, should not be used
			let title = "Search";
			if (link.hasKey("title"))
			    title = link.getPropertyAsAUTF8String("title");
			nodelinks.push({ href: href, title: title,
					 type: "search" });
		    }
		}
	    }
	}

	if (nodelinks.length > 0 && !update_mode)
	    db.set_node_links(node, nodelinks);

	// FIXME
	//openSearch:totalResults
	let tr = -1;
	if (feed.fields.hasKey("nulltotalResults"))
	    tr = Number(feed.fields.get("nulltotalResults"));
	//openSearch:startIndex
	let si = -1;
	if (feed.fields.hasKey("nullstartIndex"))
	    si = Number(feed.fields.get("nullstartIndex"));
	//openSearch:itemsPerPage
	let ipp = -1;
	if (feed.fields.hasKey("nullitemsPerPage"))
	    ipp = Number(feed.fields.get("nullitemsPerPage"));
	if (tr != -1 && si != -1 && ipp != -1) {
	    db.set_node_pageinfo(node, tr, si, ipp);
	}

	// FIXME ändra ev node->treerow->treecell enligt föregående
	// t ex catalog_subtitle som tooltip

	feed.QueryInterface(Components.interfaces.nsIFeed);
	for (let i = 0; i < feed.items.length; i++) {
	    debug("ENTRY");
	    let entry = feed.items.queryElementAt(i, Components.interfaces.nsIFeedEntry);

	    let ent = this.parse_entry(entry);
	    if (!ent)
		continue;

	    debug("Title: " + ent.title);

	    if (ent.type == "folder") {
		if (!update_mode)
		    db.add_node(node, ent.location_uri.spec, ent.title);
	    } else if (ent.type == "epub" || ent.type == "link") {
		if (update_mode) {
		    db.add_or_update_book(ent);
		} else {
		    let bid = db.add_book_info(ent);
		    db.add_child(node, bid);
		}
	    }
	}

	debug("return" + node);
	if (!update_mode)
	    db.set_node_visited(node);

	return next_node;
    },

    parse_entry: function(entry) {
	if (!entry.fields.hasKey("links"))
	    return null;

	let ioService = Components.classes["@mozilla.org/network/io-service;1"].
	    getService(Components.interfaces.nsIIOService);

	let ent = { links: [] };
	let link_type, location;
	let links = entry.fields.getPropertyAsInterface("links", Components.interfaces.nsIArray);
	for (let j = 0; j < links.length; j++) {
	    let link = links.queryElementAt(j, Components.interfaces.nsIPropertyBag2);
	    if (link.hasKey("type") && link.hasKey("href")) {
		let href = link.getPropertyAsAUTF8String("href");
		let nhref;
		try {
		    let ahref = ioService.newURI(href, null, entry.baseURI);
		    nhref = ahref.spec;
		} catch (e) {
		    nhref = href;
		}
		let type = link.getPropertyAsAUTF8String("type");
		//let typo = type.split(";");
		type = type.replace(";profile=http://opds-spec.org/2010/catalog", "")
		type = type.replace(";profile=opds", "")
		type = type.replace(";type=feed", "")
		let title = null;
		if (link.hasKey("title"))
		    title = link.getPropertyAsAUTF8String("title");
		let rel = "alternate";
		if (link.hasKey("rel"))
		    rel = link.getPropertyAsAUTF8String("rel");

		// Workaround for Smashwords
		if (rel == "alternate" && type =="application/epub+zip")
		    rel = "http://opds-spec.org/acquisition";

		if (rel == "http://opds-spec.org/acquisition" ||
		    rel == "http://opds-spec.org/acquisition/open-access") {
		    if (type == "application/epub+zip") {
			ent.type = "epub";
			if (!ent.epub_link) {
			    ent.epub_link = nhref;
			} else {
			    if (!title)
				title = "Alternative version";
			    ent.links.push({ href: nhref, title: title,
					     type: type });
			}
		    } else if (type == "application/pdf") {
			ent.links.push({ href: nhref, title: title,
					 type: type });
		    }
		} else if (rel == "http://opds-spec.org/borrow") {
		    // FIXME
		} else if (rel == "http://opds-spec.org/buy") {
		    // FIXME
		} else if (rel == "http://opds-spec.org/sample") {
		    // FIXME
		} else if (rel == "http://opds-spec.org/subscribe") {
		    // FIXME
		} else if (rel == "replies") {	// Feedbooks extension
		    // FIXME
		    /*
		    if (type == "application/atom+xml") {
			ent.links.push({ href: nhref, title: title,
					 type: "replies" });

		    }
		     */
		} else if (rel == "subsection" || rel == "alternate") {
		    if (type == "application/atom+xml;type=entry") {
			ent.fullid = nhref;
		    } else if (type == "application/atom+xml") {
			// Should be a hierarchical link
			if (ent.type) {
			    ent.links.push({ href: nhref, title: title,
					     type: type });
			} else {
			    ent.type = "folder";
			    ent.folder_link = nhref;
			    // This will be ignored if an EPUB link comes later
			}
		    } else if (type == "application/xhtml+xml" ||
			       type == "text/html") {
			ent.links.push({ href: nhref, title: title, type: type });
		    }
		} else if (rel == "http://opds-spec.org/image" ||
			   // old variants
			   rel == "http://opds-spec.org/cover" ||
			   rel == "http://opds-spec.org/opds-cover-image" ||
			   rel == "x-stanza-cover-image") {
		    ent.cover_link = nhref;
		} else if (rel == "http://opds-spec.org/image/thumbnail" ||
			   // old variants
			   rel == "http://opds-spec.org/thumbnail" ||
			   rel == "http://opds-spec.org/opds-cover-image-thumbnail" ||
			   rel == "x-stanza-cover-image-thumbnail") {
		    ent.thumbnail_link = nhref;
		} else if (rel == "related") {
		    ent.links.push({ href: nhref, title: title, type: type });
		}
	    }
	}

	// FIXME Get "summary" and "content" from entry

	if (ent.type == "epub") {
	    ent.link = ent.epub_link;
	} else if (ent.type == "folder") {
	    ent.link = ent.folder_link;
	} else if (!ent.type && ent.links.length > 0) {
	    ent.type = "link";
	    ent.link = ent.links[0].href;
	}

	debug("Entry type: " + ent.type);

	/*
	if (!ent.type)
	    return null;
	 */

        let location_uri;
        try {
	    ent.location_uri = ioService.newURI(ent.link, null, entry.baseURI);
        } catch (e) {
	    debug("E2: " + e);
            return null;
        }

	if (entry.title)
	    ent.title = entry.title.plainText();
	if (!ent.title)
	    ent.title = "[Unknown]";	// FIXME

	if (ent.type == "epub" || ent.type == "link") {
	    if (entry.authors.length >= 1) {
		let person = entry.authors.queryElementAt(0, Components.interfaces.nsIFeedPerson);
		ent.author = person.name;
	    }

	    ent.location = ent.location_uri.spec;
	    if (entry.id)
		ent.xid = entry.id;
	    else
		ent.xid = ent.location;
	}

	for (a in ent)
	    debug("ent:" + ent + "=" + ent[a]);

	return ent;
    },

    get_full_entry: function(nid, callback) {
	debug("get_full_entry: " + nid);
	let bi = this.db.get_book_info(nid);
	if (!bi || !bi.fullid) {
	    callback(null);
	    return false;
	}
	let url = bi.fullid;

	debug("URL: " + url);
	let doc = get_from_cache(this, url);
	if (doc && doc != "") {
	    this.full_step2(doc, url, nid, callback);
	    return true;
	}

 	let this_proxy = this;
	let listener = {
	    handleEvent: function(event) {
		debug("EVENT: " + event);
		try {
		    put_in_cache(this_proxy, url, request.responseText);
		    this_proxy.full_step2(request.responseText, url, nid, callback);
		} catch (e) {
		    debug("E: " + e);
		    callback(null);
		}
	    }
	};
	let error_listener = {
	    handleEvent: function(event) {
		debug("ERROR EVENT: " + event);
		callback(null);
	    }
	};

	debug("B");
	let request = Components.classes["@mozilla.org/xmlextras/xmlhttprequest;1"].createInstance(Components.interfaces.nsIXMLHttpRequest);
	request.onload = listener;
	request.onerror = error_listener;
	try {
	    request.open("GET", url, true);
	    request.send(null);
	} catch (e) {
	    debug("E3:" + e);
	    callback(null);
	    return false;
	}
	debug("C");

	return true;
    },

    full_step2: function(doc, url, nid, callback) {
	debug("D");
	let this_proxy = this;

	var myFeedResultListener = {
	    handleResult: function(result) {
		try {
		    this_proxy.full_step3(result, this_proxy.db, nid, callback);
		} catch (e) {
		    debug("E: " + e);
		    callback(null);
		}
	    }
	};

        try {
	    let ioService = Components.classes["@mozilla.org/network/io-service;1"].
		getService(Components.interfaces.nsIIOService);
	    let uri = ioService.newURI(url, null, null);

	    // FIXME nsIFeedProcessor should be able to handle this, but doesn't
	    //let parser = Components.classes["@mozilla.org/feed-processor;1"].
	    //	createInstance(Components.interfaces.nsIFeedProcessor);
	    let parser = Components.classes["@lucidor.org/atomProcessor;1"].
		createInstance().wrappedJSObject;
	    parser.listener = myFeedResultListener;
	    parser.parseFromString(doc, uri);
	} catch (e) {
	    debug("Error parsing feed." + e);
	    callback(null);
	}
    },

    full_step3: function(result, db, node, callback) {
	debug("full_step3");
	//let item = result.doc.QueryInterface(Components.interfaces.nsIFeedEntry);
	let item = result.doc;
	let entry = this.parse_entry(item);
	if (entry) {
	    debug(":" + node);
	    db.update_book(node, entry);
	}
	callback(entry);
    },

    get_opensearch: function(uri, callback) {
	debug("get_opensearch");
	let this_proxy = this;
	let listener = {
	    handleEvent: function(event) {
		if (request.responseXML) {
		    let se = this_proxy.parse_opensearch(request.responseXML);
		    callback(se, uri);
		} else
		    callback(null, uri);
	    }
	};
	let error_listener = {
	    handleEvent: function(event) {
		callback(null, uri);
	    }
	};

	let request = Components.classes["@mozilla.org/xmlextras/xmlhttprequest;1"].createInstance(Components.interfaces.nsIXMLHttpRequest);
	request.onload = listener;
	request.onerror = error_listener;
	try {
	    request.open("GET", uri, true);
	    request.send(null);
	} catch (e) {
	    debug("E:" + e);
	    callback(null, uri);
	}
    },

    parse_opensearch: function(doc) {
	debug("parse_opensearch");
	if (!(doc.documentElement.localName == "OpenSearchDescription" &&
	      doc.documentElement.namespaceURI == OS_NS))
	    return null;
	let se = new opensearch();
	for (let el in loop(doc.documentElement.childNodes)) {
	    if (!(el.nodeType == el.ELEMENT_NODE && el.namespaceURI == OS_NS))
	    	continue;
	    if (el.localName == "ShortName") {
		se.shortname = el.textContent;
	    } else if (el.localName == "Description") {
		se.description = el.textContent;
	    } else if (el.localName == "InputEncoding") {
		se.encoding = el.textContent;
	    //} else if (el.localName == "OutputEncoding") {
	    } else if (el.localName == "Image") {
		se.image = el.textContent;
	    } else if (el.localName == "Url" && el.hasAttribute("type") &&
		       el.hasAttribute("template")) {
		if (el.hasAttributeNS(LUCI_NS, "querysyntax")) {
		    let syntax = el.getAttributeNS(LUCI_NS, "querysyntax");
		    debug("SYN: " + syntax);
		    for (let word in loop(syntax.split(" "))) {
			debug("syn: " + word);
			se.syntax[word] = true;
		    }
		}

		let type = el.getAttribute("type");
		let temp = el.getAttribute("template");
		let params = {};
		for (let pa in loop(temp.split("{"))) {
		    let pat = pa.split("}");
		    if (pat.length == 2) {
			if (pat[0].substring(pat[0].length - 1) == "?") {
			    params[pat[0].substring(0, pat[0].length - 1)] = "opt";
			} else {
			    params[pat[0]] = "req";
			}
		    }
		}
		if (params["searchTerms"]) {
		    if (type == "application/atom+xml") {
			se.template = temp;
			se.template_params = params;
		    } else if (type == "application/x-suggestions+json" &&
			       el.hasAttribute("rel") &&
			       el.getAttribute("rel") == "suggestions") {
			se.suggestions = temp;
			se.suggestions_params = params;
		    }
		}
	    }
	}
	if (se.template)
	    return se;
	return null;
    },

    sync_cache: function() {
	// FIXME
	let db = Components.classes["@lucidor.org/bookdb;1"].createInstance().
	    wrappedJSObject;
	db.create("opds");
	this.sync1(db, url);
    },

    sync1: function(db, url) {
	let this_proxy = this;
	let listener = {
	    handleEvent: function(event) {
		debug("EVENT 1: " + event);
		try {
		    //put_in_cache(this_proxy, url, request.responseText);
		    this_proxy.sync2(request.responseText, db, url);
		} catch (e) {
		    debug("E: " + e);
		}
	    }
	};
	let error_listener = {
	    handleEvent: function(event) {
		debug("ERROR EVENT: " + event);
	    }
	};

	let request = Components.classes["@mozilla.org/xmlextras/xmlhttprequest;1"].createInstance(Components.interfaces.nsIXMLHttpRequest);
	request.onload = listener;
	request.onerror = error_listener;
	try {
	    request.open("GET", url, true);
	    request.send(null);
	} catch (e) {
	    debug("E3:" + e);
	    throw "error"; 
	}
    },

    sync2: function(doc, db, url) {
	debug("D");
	let this_proxy = this;

	var myFeedResultListener = {
	    handleResult: function(result) {
		debug("handle_result");
		try {
		    let nn = this_proxy.parse_results(result, db, null, true);
		    if (nn) {
			this_proxy.sync1(db, nn);
		    } else {
			this_proxy.sync3(db, url);
		    }
		} catch (e) {
		    debug("E1: " + e);
		}
	    }
	};

        try {
	    let parser = Components.classes["@mozilla.org/feed-processor;1"].
		createInstance(Components.interfaces.nsIFeedProcessor);
	    parser.listener = myFeedResultListener;
	    let ioService = Components.classes["@mozilla.org/network/io-service;1"].
		getService(Components.interfaces.nsIIOService);
	    let uri = ioService.newURI(url, null, null);
	    parser.parseFromString(doc, uri);
        } catch (e) {
	    debug("Error parsing feed." + e);
        }
    },

    sync3: function(db, url) {
	
    }
};

var components = [opdsCase];

if (XPCOMUtils.generateNSGetFactory)
    var NSGetFactory = XPCOMUtils.generateNSGetFactory(components);
else
    var NSGetModule = XPCOMUtils.generateNSGetModule(components);
