/*
 *  Lucifox - e-book reader
 *  Copyright © 2009-2011  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

function debug(msg) {
    if (true) {
	try {
	    let cons = Components.classes["@mozilla.org/consoleservice;1"].
		getService(Components.interfaces.nsIConsoleService);
	    cons.logStringMessage(msg);
	} catch (e) {
	}
    }
}

function loop(thing) {
    if (!thing)
	return;
    for (let i = 0; i < thing.length; i++)
	yield thing[i];
}

const OS_NS = "http://a9.com/-/spec/opensearch/1.1/";
const LUCI_NS = "http://lucidor.org/-/x-opds/";

function opensearch() {
    this.type = "opds";
    this.syntax = {};
    this.template = null;
    this.wrappedJSObject = this;
}

opensearch.prototype = {
    get_search_uri: function(phrase) {
	let url = this.template;
	for (let par in this.template_params) {	// FIXME
	    // FIXME hantera { } och ?
	    if (par == "searchTerms") {
		url = url.replace("{searchTerms}", escape(phrase));
	    } else if (this.template_params[par] == "opt") {
		url = url.replace("{" + par + "?}", "");
	    } else if (par == "count") {
		url = url.replace(par, "25");	// FIXME
	    } else if (par == "startIndex") {
		url = url.replace(par, "1");	// FIXME
	    } else if (par == "startPage") {
		url = url.replace(par, "1");	// FIXME
	    } else if (par == "language") {
		url = url.replace(par, "*");
	    } else if (par == "inputEncoding") {
		url = url.replace(par, "UTF-8");	// FIXME
	    } else if (par == "outputEncoding") {
		url = url.replace(par, "UTF-8");	// FIXME
	    } else {	// This should never happen
		url = url.replace(par, "Barbie");
	    }
	}
	return url;
    }
};

function opdsParser() {
    this.wrappedJSObject = this;
}

opdsParser.prototype = {
    classDescription: "opdsParser",
    classID: Components.ID("{5bb54c1c-5f7f-49b4-ad8b-bbf2bc8932f3}"),
    contractID: "@lucidor.org/opdsParser;1",
    QueryInterface: XPCOMUtils.generateQI([Components.interfaces.nsISupports]),

    catalog_url: null,

    info: { title: true,
	    author: true,
	    location: true },

    get_feed: function(url, callback, sid) {
	debug("URL: " + url);

 	let this_proxy = this;
	let listener = {
	    handleEvent: function(event) {
		debug("EVENT 1: " + event);
		try {
		    this_proxy.step2(request.responseText, url, callback, sid);
		} catch (e) {
		    debug("E: " + e);
		    callback(null, sid);
		}
	    }
	};
	let error_listener = {
	    handleEvent: function(event) {
		debug("ERROR EVENT: " + event);
		callback(null, sid);
	    }
	};

	/*
	let main_control = Components.classes["@lucidor.org/mainControlService;1"].
		getService().wrappedJSObject;
	debug("ONLINE: " + main_control.online());
	*/

	let request = Components.classes["@mozilla.org/xmlextras/xmlhttprequest;1"].createInstance(Components.interfaces.nsIXMLHttpRequest);
	request.onload = listener;
	request.onerror = error_listener;
	try {
	    request.open("GET", url, true);
	    request.send(null);
	} catch (e) {
	    debug("E3:" + e);
	    callback(null, sid);
	}
    },

    step2: function(doc, url, callback, sid) {
	debug("D");
	let this_proxy = this;

	var myFeedResultListener = {
	    handleResult: function(result) {
		debug("handle_result");
		try {
		    this_proxy.step3(result, url, callback, sid);
		} catch (e) {
		    debug("E1: " + e);
		    callback(null, sid);
		}
	    }
	};

        try {
	    let parser = Components.classes["@mozilla.org/feed-processor;1"].
		createInstance(Components.interfaces.nsIFeedProcessor);
	    parser.listener = myFeedResultListener;
	    let ioService = Components.classes["@mozilla.org/network/io-service;1"].
		getService(Components.interfaces.nsIIOService);
	    let uri = ioService.newURI(url, null, null);
	    parser.parseFromString(doc, uri);
        } catch (e) {
	    debug("Error parsing feed: " + e);
	    callback(null, sid);
        }
    },

    step3: function(result, url, callback, sid) {
	try {
	    this.parse_results(result, url, callback, sid);
	} catch (e) {
	    debug("Error parsing step3: " + e);
	    callback(null, sid);
	}
    },

    parse_results: function(res, uri, callback, sid) {
	debug("parse_results");

	let result = { uri: uri,
		       title: null,
		       subtitle: null,
		       entries: [],		// Feed entries
		       links: [],		// Feed-level links
		       previous: null,		// Previous link
		       next: null,		// Next link
		       parent: null,		// Parent link
		       start: null,		// Start link
		       opensearch: null,	// OpenSearch link
		       totalResults: -1,
		       startIndex: -1,
		       itemsPerPage: -1 };

        let feed = res.doc.QueryInterface(Components.interfaces.nsIFeedContainer);

	if (feed.title)
	    result.title = feed.title.plainText();
	if (feed.subtitle)
	    result.subtitle = feed.subtitle.plainText();
	/*
	if (feed.authors.length >= 1) {
	    let person = entry.authors.queryElementAt(0, Components.interfaces.nsIFeedPerson);
	    catalog_author = person.name;
	}
	if (feed.updated)
	    catalog_updated = feed.updated;
	*/

	let ioService = Components.classes["@mozilla.org/network/io-service;1"]
	    .getService(Components.interfaces.nsIIOService);

	let links;
	if (feed.fields.hasKey("links"))
	    links = feed.fields.getPropertyAsInterface("links", Components.interfaces.nsIArray);
	else
	    links = [];
	for (let j = 0; j < links.length; j++) {
	    let link = links.queryElementAt(j, Components.interfaces.nsIPropertyBag2);
	    if (!link.hasKey("type"))
		continue;
	    let type = link.getPropertyAsAUTF8String("type");
	    type = type.replace(";profile=http://opds-spec.org/2010/catalog", "")
	    type = type.replace(";profile=opds", "")
	    type = type.replace(";type=feed", "")
	    if (!link.hasKey("rel"))
		continue;
	    let rel = link.getPropertyAsAUTF8String("rel");
	    let title = null;
	    if (link.hasKey("title"))
		title = link.getPropertyAsAUTF8String("title");
	    let href = null;
	    if (link.hasKey("href")) {
		let ohref = link.getPropertyAsAUTF8String("href");
		try {
		    href = ioService.newURI(ohref, null, feed.baseURI).spec;
		} catch (e) {
		}

		// index / start
		// previous, first, last
		if (rel == "http://opds-spec.org/bookshelf") {
		} else if (rel == "http://opds-spec.org/subscriptions") {
		} else if (rel == "http://opds-spec.org/crawlable") {
		    // Ignore
		} else if (rel == "http://opds-spec.org/sort/new" &&
			   type == "application/atom+xml") {
		    if (!title)
			title = "New";
		    result.links.push({ href: href, title: title,
					type: "toplink" });
		} else if (rel == "http://opds-spec.org/sort/popular" &&
			   type == "application/atom+xml") {
		    if (!title)
			title = "Popular";
		    result.links.push({ href: href, title: title,
					type: "toplink" });
		} else if (rel == "http://opds-spec.org/sort/featured" &&
			   type == "application/atom+xml") {
		    if (!title)
			title = "Featured";
		    result.links.push({ href: href, title: title,
					type: "toplink" });
		} else if (rel == "related" && type == "application/atom+xml") {
		    if (!title)
			title = "Related";
		    result.links.push({ href: href, title: title,
					type: "toplink" });
		} else if (rel == "previous" &&
			   type == "application/atom+xml") {
		    if (!title)
			title = "Previous";
		    result.previous = href;
		} else if (rel == "next" && type == "application/atom+xml") {
		    if (!title)
			title = "Next";
		    result.next = href;
		} else if (rel == "parent" && type == "application/atom+xml") {
		    if (!title)
			title = "Parent";
		    result.parent = href;
		} else if (rel == "start" && type == "application/atom+xml") {
		    if (!title)
			title = "Start";
		    result.start = href;
		} else if (rel == "search") {
		    if (type == "application/opensearchdescription+xml") {
			result.opensearch = href;
		    }
		}
	    }
	}

	//openSearch:totalResults
	if (feed.fields.hasKey("nulltotalResults"))
	    result.totalResults = Number(feed.fields.get("nulltotalResults"));
	//openSearch:startIndex
	if (feed.fields.hasKey("nullstartIndex"))
	    result.startIndex = Number(feed.fields.get("nullstartIndex"));
	//openSearch:itemsPerPage
	if (feed.fields.hasKey("nullitemsPerPage"))
	    result.itemsPerPage = Number(feed.fields.get("nullitemsPerPage"));

	feed.QueryInterface(Components.interfaces.nsIFeed);
	for (let i = 0; i < feed.items.length; i++) {
	    debug("ENTRY");
	    let entry = feed.items.queryElementAt(i, Components.interfaces.nsIFeedEntry);

	    let ent = this.parse_entry(entry);
	    if (!ent)
		continue;

	    debug("Title: " + ent.title);
	    ent.idx = result.entries.length;
	    result.entries.push(ent);
	}

	callback(result, sid);
    },

    parse_entry: function(entry) {
	if (!entry.fields.hasKey("links"))
	    return null;

	let ioService = Components.classes["@mozilla.org/network/io-service;1"].
	    getService(Components.interfaces.nsIIOService);

	let ent = { links: [] };
	let link_type, location;
	let links = entry.fields.getPropertyAsInterface("links", Components.interfaces.nsIArray);
	for (let j = 0; j < links.length; j++) {
	    let link = links.queryElementAt(j, Components.interfaces.nsIPropertyBag2);
	    if (link.hasKey("type") && link.hasKey("href")) {
		let href = link.getPropertyAsAUTF8String("href");
		let nhref;
		try {
		    let ahref = ioService.newURI(href, null, entry.baseURI);
		    nhref = ahref.spec;
		} catch (e) {
		    nhref = href;
		}
		let type = link.getPropertyAsAUTF8String("type");
		//let typo = type.split(";");
		type = type.replace(";profile=http://opds-spec.org/2010/catalog", "")
		type = type.replace(";profile=opds", "")
		type = type.replace(";type=feed", "")
		let title = null;
		if (link.hasKey("title"))
		    title = link.getPropertyAsAUTF8String("title");
		let rel = "alternate";
		if (link.hasKey("rel"))
		    rel = link.getPropertyAsAUTF8String("rel");

		// Workaround for Smashwords
		if (rel == "alternate" && type =="application/epub+zip")
		    rel = "http://opds-spec.org/acquisition";

		if (rel == "http://opds-spec.org/acquisition" ||
		    rel == "http://opds-spec.org/acquisition/open-access") {
		    if (type == "application/epub+zip") {
			ent.type = "epub";
			if (!ent.epub_link) {
			    ent.epub_link = nhref;
			} else {
			    if (!title)
				title = "Alternative version";
			    ent.links.push({ href: nhref, title: title,
					     type: type });
			}
		    } else if (type == "application/pdf") {
			ent.links.push({ href: nhref, title: title,
					 type: type });
		    }
		} else if (rel == "http://opds-spec.org/buying") {
		    // FIXME
		} else if (rel == "http://opds-spec.org/lending") {
		    // FIXME
		} else if (rel == "http://opds-spec.org/subscription") {
		    // FIXME
		} else if (rel == "http://opds-spec.org/sample") {
		    // FIXME
		} else if (rel == "replies") {	// Feedbooks extension
		    // FIXME
		    /*
		    if (type == "application/atom+xml") {
			ent.links.push({ href: nhref, title: title,
					 type: "replies" });

		    }
		     */
		} else if (rel == "alternate") {
		    if (type == "application/atom+xml;type=entry") {
			ent.fullid = nhref;
		    } else if (type == "application/atom+xml") {
			// Should be a hierarchical link
			if (ent.type) {
			    ent.links.push({ href: nhref, title: title,
					     type: type });
			} else {
			    ent.type = "folder";
			    ent.folder_link = nhref;
			    // This will be ignored if an EPUB link comes later
			}
		    } else if (type == "application/xhtml+xml" ||
			       type == "text/html") {
			ent.links.push({ href: nhref, title: title, type: type });
		    }
		} else if (rel == "http://opds-spec.org/cover") {
		    ent.cover_link = nhref;
		} else if (rel == "http://opds-spec.org/thumbnail") {
		    ent.thumbnail_link = nhref;
		} else if (rel == "related") {
		    ent.links.push({ href: nhref, title: title, type: type });
		}
	    }
	}

	// FIXME Get "summary" from entry

	if (ent.type == "epub") {
	    ent.link = ent.epub_link;
	} else if (ent.type == "folder") {
	    ent.link = ent.folder_link;
	} else if (!ent.type && ent.links.length > 0) {
	    ent.type = "link";
	    ent.link = ent.links[0].href;
	}

	debug("Entry type: " + ent.type);

	/*
	if (!ent.type)
	    return null;
	 */

        let location_uri;
        try {
	    location_uri = ioService.newURI(ent.link, null, entry.baseURI);
        } catch (e) {
	    debug("E2: " + e);
            return null;
        }

	ent.location = location_uri.spec;

	if (entry.title)
	    ent.title = entry.title.plainText();
	if (!ent.title)
	    ent.title = "[Unknown]";	// FIXME

	if (ent.type == "epub" || ent.type == "link") {
	    if (entry.authors.length >= 1) {
		let person = entry.authors.queryElementAt(0, Components.interfaces.nsIFeedPerson);
		ent.author = person.name;
	    }

	    if (entry.id)
		ent.id = entry.id;
	}

	for (a in ent)
	    debug("ent:" + a + "=" + ent[a]);

	return ent;
    },

    get_entry: function(url, callback, sid) {
	debug("get_entry: " + url);

 	let this_proxy = this;
	let listener = {
	    handleEvent: function(event) {
		debug("EVENT: " + event);
		try {
		    this_proxy.full_step2(request.responseText, url, callback,
					  sid);
		} catch (e) {
		    debug("E: " + e);
		    callback(null);
		}
	    }
	};
	let error_listener = {
	    handleEvent: function(event) {
		debug("ERROR EVENT: " + event);
		callback(null);
	    }
	};

	debug("B");
	let request = Components.classes["@mozilla.org/xmlextras/xmlhttprequest;1"].createInstance(Components.interfaces.nsIXMLHttpRequest);
	request.onload = listener;
	request.onerror = error_listener;
	try {
	    request.open("GET", url, true);
	    request.send(null);
	} catch (e) {
	    debug("E3:" + e);
	    callback(null);
	    return false;
	}
	debug("C");

	return true;
    },

    full_step2: function(doc, url, callback, sid) {
	debug("D");
	let this_proxy = this;

	var myFeedResultListener = {
	    handleResult: function(result) {
		try {
		    this_proxy.full_step3(result, callback, sid);
		} catch (e) {
		    debug("E: " + e);
		    callback(null);
		}
	    }
	};

        try {
	    let ioService = Components.classes["@mozilla.org/network/io-service;1"].
		getService(Components.interfaces.nsIIOService);
	    let uri = ioService.newURI(url, null, null);

	    // FIXME nsIFeedProcessor should be able to handle this, but doesn't
	    //let parser = Components.classes["@mozilla.org/feed-processor;1"].
	    //	createInstance(Components.interfaces.nsIFeedProcessor);
	    let parser = Components.classes["@lucidor.org/atomProcessor;1"].
		createInstance().wrappedJSObject;
	    parser.listener = myFeedResultListener;
	    parser.parseFromString(doc, uri);
	} catch (e) {
	    debug("Error parsing feed." + e);
	    callback(null);
	}
    },

    full_step3: function(result, callback, sid) {
	debug("full_step3");
	//let item = result.doc.QueryInterface(Components.interfaces.nsIFeedEntry);
	let item = result.doc;
	let entry = this.parse_entry(item);
	callback(entry, sid);
    },

    get_opensearch: function(uri, callback) {
	debug("get_opensearch");
	let this_proxy = this;
	let listener = {
	    handleEvent: function(event) {
		if (request.responseXML) {
		    let se = this_proxy.parse_opensearch(request.responseXML);
		    callback(se, uri);
		} else
		    callback(null, uri);
	    }
	};
	let error_listener = {
	    handleEvent: function(event) {
		callback(null, uri);
	    }
	};

	let request = Components.classes["@mozilla.org/xmlextras/xmlhttprequest;1"].createInstance(Components.interfaces.nsIXMLHttpRequest);
	request.onload = listener;
	request.onerror = error_listener;
	try {
	    request.open("GET", uri, true);
	    request.send(null);
	} catch (e) {
	    debug("E:" + e);
	    callback(null, uri);
	}
    },

    parse_opensearch: function(doc) {
	debug("parse_opensearch");
	if (!(doc.documentElement.localName == "OpenSearchDescription" &&
	      doc.documentElement.namespaceURI == OS_NS))
	    return null;
	let se = new opensearch();
	for (let el in loop(doc.documentElement.childNodes)) {
	    if (!(el.nodeType == el.ELEMENT_NODE && el.namespaceURI == OS_NS))
	    	continue;
	    if (el.localName == "ShortName") {
		se.shortname = el.textContent;
	    } else if (el.localName == "Description") {
		se.description = el.textContent;
	    } else if (el.localName == "InputEncoding") {
		se.encoding = el.textContent;
	    //} else if (el.localName == "OutputEncoding") {
	    } else if (el.localName == "Image") {
		se.image = el.textContent;
	    } else if (el.localName == "Url" && el.hasAttribute("type") &&
		       el.hasAttribute("template")) {
		if (el.hasAttributeNS(LUCI_NS, "querysyntax")) {
		    let syntax = el.getAttributeNS(LUCI_NS, "querysyntax");
		    debug("SYN: " + syntax);
		    for (let word in loop(syntax.split(" "))) {
			debug("syn: " + word);
			se.syntax[word] = true;
		    }
		}

		let type = el.getAttribute("type");
		let temp = el.getAttribute("template");
		let params = {};
		for (let pa in loop(temp.split("{"))) {
		    let pat = pa.split("}");
		    if (pat.length == 2) {
			if (pat[0].substring(pat[0].length - 1) == "?") {
			    params[pat[0].substring(0, pat[0].length - 1)] = "opt";
			} else {
			    params[pat[0]] = "req";
			}
		    }
		}
		if (params["searchTerms"]) {
		    if (type == "application/atom+xml") {
			se.template = temp;
			se.template_params = params;
		    } else if (type == "application/x-suggestions+json" &&
			       el.hasAttribute("rel") &&
			       el.getAttribute("rel") == "suggestions") {
			se.suggestions = temp;
			se.suggestions_params = params;
		    }
		}
	    }
	}
	if (se.template)
	    return se;
	return null;
    }
};

var components = [opdsParser];

if (XPCOMUtils.generateNSGetFactory)
    var NSGetFactory = XPCOMUtils.generateNSGetFactory(components);
else
    var NSGetModule = XPCOMUtils.generateNSGetModule(components);
