/*
 *  Lucidor - e-book reader
 *  Copyright © 2008-2009  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

function debug(msg) {
    if (false) {
	try {
	    let cons = Components.classes["@mozilla.org/consoleservice;1"].
		getService(Components.interfaces.nsIConsoleService);
	    cons.logStringMessage(msg);
	} catch (e) {
	}
    }
}

function loop(thing) {
    if (!thing)
	return;
    for (let i = 0; i < thing.length; i++)
	yield thing[i];
}

function previewCreator() {
    this.wrappedJSObject = this;
    this.pstack = [];
    this.running = false;
}

previewCreator.prototype = {
    classDescription: "previewCreator",
    classID: Components.ID("{919e26ea-0cbb-4545-a7f6-f19c19984f5d}"),
    contractID: "@lucidor.org/previewCreatorService;1",
    _xpcom_factory: previewCreatorFactory,
    QueryInterface: XPCOMUtils.generateQI([Components.interfaces.nsISupports]),

    init: function(browser) {
	this.browser = browser;
	this.previewProgressListener.this_proxy = this;
	if (this.browser.has_listener == undefined) {
	    this.browser.addProgressListener(this.previewProgressListener);
	    this.browser.has_listener = true;

	    this.browser.docShell.allowAuth = false;
	    this.browser.docShell.allowJavascript = false;
	    this.browser.docShell.allowMetaRedirects = false;
	    this.browser.docShell.allowPlugins = false;
	    this.browser.docShell.useErrorPages = true;

	    //let necp = Components.classes["@lucidor.org/noExternalContentPolicy;1"].
	    //	getService().wrappedJSObject;
	    //necp.addBrowser(this.browser);
	}
    },

    generate_preview: function(location, canvas, add_preview, id) {
	debug("generate_preview");
	canvas.width = 100;
	canvas.height = 120;
	this.pstack.push({ location: location, canvas: canvas, id: id,
		    add: add_preview });
	if (!this.running)
	    this.run_generator();
    },

    run_generator: function() {
	debug("run_generator");
	this.running = true;
	let entry = this.pstack.pop();
	if (!entry) {
	    this.running = false;
	    this.browser.loadURIWithFlags("about:blank", null, null, null);
	    debug("stop_generator");
	    return;
	}
	this.preview_epub(entry);
    },

    preview_epub: function(entry) {
	debug("preview_epub");
	let nsuri;
	let file = null;
	try {
	    let ioService = Components.classes["@mozilla.org/network/io-service;1"].
		getService(Components.interfaces.nsIIOService);
	    nsuri = ioService.newURI(entry.location, "UTF-8", null);
	    if (nsuri.schemeIs("file"))
		file = Components.classes["@mozilla.org/network/protocol;1?name=file"].
		    getService(Components.interfaces.nsIFileProtocolHandler).
		    getFileFromURLSpec(entry.location);
	} catch (e) {
	    debug("PW Error 1: " + e);
	    return this.run_generator();
	}

	if (file)
	    return this.preview_epub2(file, entry.canvas, entry.id, entry.add);

	// FIXME preview_epub should only be used with local files
	return this.run_generator();

	/*
	let main_control = Components.classes["@lucidor.org/mainControlService;1"].
	    getService().wrappedJSObject;
	file = main_control.get_tempdir();
	file.append("preview.epub");
	try {
	    file.createUnique(Components.interfaces.nsIFile.NORMAL_FILE_TYPE,
			      0600);
	} catch (e) {
	    // FIXME
	    debug("PW Error 2: " + e);
	    return this.run_generator();
	}

	let this_proxy = this;
	let wbp = Components.classes["@mozilla.org/embedding/browser/nsWebBrowserPersist;1"].createInstance(Components.interfaces.nsIWebBrowserPersist);
	wbp.persistFlags = wbp.PERSIST_FLAGS_SERIALIZE_OUTPUT;
	wbp.progressListener = {
	    onLocationChange: function(webProgress, request, location) {
	    },

	    onProgressChange: function(webProgress, request, curSelfProgress,
				       maxSelfProgress, curTotalProgress,
				       maxTotalProgress) {
	    },

	    onSecurityChange: function(webProgress, request, state) {
	    },

	    onStateChange: function(aWebProgress, aRequest,
				    aStateFlags, aStatus) {
		if (wbp.currentState == wbp.PERSIST_STATE_FINISHED)
		    this_proxy.preview_epub2(file, entry.canvas, entry.id,
					     entry.add);
	    },

	    onStatusChange: function(webProgress, request, status, message) {
	    },

	    QueryInterface: function(aIID) {
		if (aIID.equals(Components.interfaces.nsIWebProgressListener))
		    return this;
		throw Components.results.NS_NOINTERFACE;
	    }
	};

	try {
	    wbp.saveURI(nsuri, null, null, null, null, file);
	} catch (e) {
	    // FIXME
	    debug("PW Error 3: " + e);
	    return this.run_generator();
	}
	 */
    },

    draw_it: function() {
	debug("draw_it");
	this.draw_book(this.browser.data.canvas);
	if (this.browser.data.add)
	    this.browser.data.add.add_preview(this.browser.data.id,
					      this.browser.data.canvas.toDataURL());
	let bookman = Components.classes["@lucidor.org/bookmanagerService;1"].
		getService().wrappedJSObject;
	bookman.removeBook(this.browser.data.bookid);
	this.browser.data = null;
	return this.run_generator();
    },

    // FIXME return this.run_generator() on errors
    previewProgressListener: {
	this_proxy: null,

	// This implements nsITimerCallback
	notify: function(timer) {
	    this.this_proxy.draw_it();
	},

	onLocationChange: function(webProgress, request, location) {
	},

	onProgressChange: function(webProgress, request, curSelfProgress,
				   maxSelfProgress, curTotalProgress,
				   maxTotalProgress) {
	},

	onSecurityChange: function(webProgress, request, state) {
	},

	onStateChange: function(webProgress, request, stateFlags, status) {
	    if ((stateFlags & 16) && (stateFlags & 131072)) {
		if (this.this_proxy.browser.data != null) {
		    if (webProgress.DOMWindow !=
			this.this_proxy.browser.contentWindow)
			return;
		    // FIXME Why is the timer necessary?
		    let timer = Components.classes["@mozilla.org/timer;1"].
			createInstance(Components.interfaces.nsITimer);
		    timer.initWithCallback(this, 100, 0);
		}
	    }
	},

	onStatusChange: function(webProgress, request, status, message) {
	},

	QueryInterface: function(aIID) {
	    if (aIID.equals(Components.interfaces.nsIWebProgressListener) ||
		aIID.equals(Components.interfaces.nsISupportsWeakReference) ||
		aIID.equals(Components.interfaces.nsITimerCallback))
		return this;
	    throw Components.results.NS_NOINTERFACE;
	}
    },

    preview_epub2: function(file, canvas, id, add_preview) {
	let book = Components.classes["@lucidor.org/epubDocument;1"].
		createInstance().wrappedJSObject;
	try {
	    book.init(file);
	} catch (e) {
	    debug("PW Error 4: " + e);
	    return this.run_generator();
	}

	let bookman = Components.classes["@lucidor.org/bookmanagerService;1"].
		getService().wrappedJSObject;
	let bookid = bookman.addBook("epub", book);

	let cover = book.get_cover();
	let href = book.idrefToFallbackHref(cover);
	if (!href)
	    return this.run_generator();
	let location = "x-lucidor://" + bookid + "/" + href;
	this.browser.data = { canvas: canvas, id: id, add: add_preview,
			      bookid: bookid };
	try {
	    this.browser.loadURI(location, null, null);
	} catch (e) {
	    return this.run_generator();
	}
    },

    draw_book: function(canvas) {
	let content = this.browser.contentWindow;
	let w = content.innerWidth + content.scrollMaxX - 2;
	let h = content.innerHeight + content.scrollMaxY - 2;

	// Limit height
	h = Math.max(h, w);
	//h = Math.min(h, w * 1.3);
	if (h > 1.5 * w)
	    h = w * 1.3;

	let cw = 128;
	let ch = 128;
	if (h > w)
	    cw = ch * w / h;
	else
	    ch = cw * h / w;

	debug("W: " + w + " H: " + h + " CW: " + cw + " CH: " + ch);
	//canvas.style.width = cw + "px";
	//canvas.style.height = ch + "px";
	canvas.width = cw;
	canvas.height = ch;

	let ctx = canvas.getContext("2d");
	ctx.scale(cw / (w - 2), ch / (h - 2));
	ctx.drawWindow(content, 2, 2, w, h, "rgb(255,255,255)");
	//ctx.drawWindow(content, 2, 2, w, h, "rgba(0,0,0,0)");
    }
};

var serviceObject = new previewCreator();

var previewCreatorFactory = {
    createInstance: function(outer, aIID) {
	if (outer != null)
	    throw Components.results.NS_ERROR_NO_AGGREGATION;
	return serviceObject.QueryInterface(aIID);
    }
};

var components = [previewCreator];

if (XPCOMUtils.generateNSGetFactory)
    var NSGetFactory = XPCOMUtils.generateNSGetFactory(components);
else
    var NSGetModule = XPCOMUtils.generateNSGetModule(components);
