/*
 *  Lucidor - e-book reader
 *  Copyright © 2009  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

function debug(msg) {
    if (false) {
	try {
	    let cons = Components.classes["@mozilla.org/consoleservice;1"].
		getService(Components.interfaces.nsIConsoleService);
	    cons.logStringMessage(msg);
	} catch (e) {
	}
    }
}

function loop(thing) {
    if (!thing)
	return;
    for (let i = 0; i < thing.length; i++)
	yield thing[i];
}

function init_cache(this_proxy, name) {
    let ok = false;
    try {
	let prefs = Components.classes["@mozilla.org/preferences-service;1"].
	    getService(Components.interfaces.nsIPrefService).
	    getBranch("lucidor.");
	let value = prefs.getBoolPref("debug.cache");
	ok = value;
    } catch (e) {
    }

    if (!ok)
	return false;

    if (!this_proxy.debug_cache) {
	let file = Components.classes["@mozilla.org/file/directory_service;1"].
	    getService(Components.interfaces.nsIProperties).
	    get("ProfD", Components.interfaces.nsIFile);
	let fn = "debug_cache";
	if (name)
	    fn = name;
	file.append(fn + ".sqlite");
	let ss = Components.classes["@mozilla.org/storage/service;1"].
	    getService(Components.interfaces.mozIStorageService);
	let db = ss.openDatabase(file);
	if (!db.tableExists("cache")) {
	    db.createTable("cache", "id INTEGER NOT NULL PRIMARY KEY, uri TEXT, doc TEXT");
	}
	this_proxy.debug_cache = db;
    }
    return true;
}

function get_from_cache(this_proxy, url) {
    if (!this_proxy.debug_cache)
	return null;

    let doc = null;
    let statement = this_proxy.debug_cache.createStatement("SELECT doc FROM cache WHERE uri=?1");
    try {
	statement.bindUTF8StringParameter(0, url);
	if (statement.executeStep())
	    doc = statement.getUTF8String(0);
    } finally {
	statement.reset();
    }
    return doc;
}

function put_in_cache(this_proxy, url, doc) {
    if (!this_proxy.debug_cache)
	return;

    let statement = this_proxy.debug_cache.createStatement("INSERT INTO cache (id, uri, doc) VALUES (NULL, ?1, ?2)");
    statement.bindUTF8StringParameter(0, url);
    statement.bindUTF8StringParameter(1, doc);
    statement.execute();
}

function stanzaCase() {
    this.wrappedJSObject = this;
    this.init0();
}

stanzaCase.prototype = {
    classDescription: "stanzaCase",
    classID: Components.ID("{c1c8e938-6941-4f94-a773-5415eef49676}"),
    contractID: "@lucidor.org/stanzaCase;1",
    QueryInterface: XPCOMUtils.generateQI([Components.interfaces.nsISupports]),

    catalog_url: null,
    db: null,
    allow_persistant: true,
    warn_about_drm: true,
    in_memory: true,
    debug_cache: null,

    info: { title: true,
	    author: true,
	    location: true },

    browsers: [ "all" ],

    capabilities: { search_meta: false,
		    search_title: false,
		    search_author: false,
		    search_tags: false,
		    search_text: false },

    init0: function() {
	this.db = Components.classes["@lucidor.org/bookdb;1"].createInstance().
		wrappedJSObject;
    },

    set_in_memory: function(in_memory) {
	this.in_memory = in_memory;
    },

    init: function(coll, tree) {
	this.collid = coll.id;
	this.tree = tree;
	this.catalog_url = coll.url;

	if (this.in_memory) {
	    this.db.create_memory();
	} else if (coll.db) {
	    this.db.open(coll.db);
	    if (!coll.persistant) {
		let new_db = Components.classes["@lucidor.org/bookdb;1"].
		    createInstance().wrappedJSObject;
		new_db.create("atom");
		this.db.delete();
		this.db = new_db;
		coll.db = null;
	    }
	} else
	    this.db.create("atom");

	if (!coll.db) {
	    let cman = Components.classes["@lucidor.org/collectionManagerService;1"].
		getService().wrappedJSObject;
	    cman.changeCollectionDB(this.collid, this.db);
	}

	this.set_db();
	init_cache(this, "cache-stanza");
    },

    set_db: function() {
	this.tree.datasources = this.db.file_uri;
	this.tree.builder.rebuild();
    },

    close: function() {
	if (this.db)
	    this.db.close();
	this.db = null;
    },

    add_book: null,

    remove_book: null,

    get_preview_url: function(id) {
	return this.db.get_preview(id);
    },

    get_preview: null,

    generate_preview: null,

    add_preview: null,

    get_book_info: function(id) {
	return this.db.get_book_info(id);
    },

    browse_cb: function(nid, callback, sid) {
	let url;
	debug("NID: " + nid);
	if (nid.substring)
	    if (nid == "all") {
		url = this.catalog_url;
		nid = this.db.add_or_get_node(null, url, "All");
	    } else if (nid.substring(0, 4) == "url:") {
		url = nid.substring(4);
		nid = this.db.add_node(null, url, "Custom");	// FIXME inte Custom
	    } else if (nid.substring(0, 7) == "search:") {
		let tmp = nid.substring(7).split("=");
		url = unescape(tmp[0]).replace("{searchTerms}", escape(tmp[1]));
		nid = this.db.add_node(null, url, "Custom");	// FIXME inte Custom
	    }

	let node = this.db.get_node(nid);
	if (node) {
	    if (node.visited) {
		callback(this.db.browse_path(nid), sid);
		return;
	    }
	    url = node.path;
	} else {
	    callback(null, sid);
	    return;
	}

	debug("URL: " + url);
	let doc = get_from_cache(this, url);
	if (doc && doc != "") {
	    this.step2(doc, url, nid, callback, sid);
	    return;
	}

	let this_proxy = this;
	let listener = {
	    handleEvent: function(event) {
		debug("EVENT 1: " + event);
		try {
		    put_in_cache(this_proxy, url, request.responseText);
		    this_proxy.step2(request.responseText, url, nid,
				     callback, sid);
		} catch (e) {
		    debug("E: " + e);
		    callback(null, sid);
		}
	    }
	};
	let error_listener = {
	    handleEvent: function(event) {
		callback(null, sid);
	    }
	};

	/*
	let main_control = Components.classes["@lucidor.org/mainControlService;1"].
		getService().wrappedJSObject;
	debug("ONLINE: " + main_control.online());
	*/

	let request = Components.classes["@mozilla.org/xmlextras/xmlhttprequest;1"].createInstance(Components.interfaces.nsIXMLHttpRequest);
	request.onload = listener;
	request.onerror = error_listener;
	try {
	    request.open("GET", url, true);
	    request.send(null);
	} catch (e) {
	    debug("E3:" + e);
	    callback(null, sid);
	}
    },

    step2: function(doc, url, nid, callback, sid) {
	let this_proxy = this;

	var myFeedResultListener = {
	    handleResult: function(result) {
		debug("handle_result");
		try {
		    this_proxy.parse_results(result, nid, url, callback, sid);
		    return;
		} catch (e) {
		    debug("E1: " + e);
		    callback(null, sid);
		}
	    }
	};

        try {
	    let parser = Components.classes["@mozilla.org/feed-processor;1"].
		createInstance(Components.interfaces.nsIFeedProcessor);
	    parser.listener = myFeedResultListener;
	    let ioService = Components.classes["@mozilla.org/network/io-service;1"].
		getService(Components.interfaces.nsIIOService);
	    let uri = ioService.newURI(url, null, null);
	    parser.parseFromString(doc, uri);
        } catch (e) {
	    debug("Error parsing feed:" + e);
	    callback(null, sid);
        }
    },

    parse_results: function(result, node, baseurl, callback, sid) {
	debug("parse_results");

	let feed = result.doc.QueryInterface(Components.interfaces.nsIFeedContainer);
	/*
	let catalog_title, catalog_author, updated;
	if (feed.title)
	    catalog_title = feed.title.plainText();
	if (feed.subtitle)
	    catalog_subtitle = feed.subtitle.plainText();
	if (feed.authors.length >= 1) {
	    let person = entry.authors.queryElementAt(0, Components.interfaces.nsIFeedPerson);
	    catalog_author = person.name;
	}
	if (feed.updated)
	    catalog_updated = feed.updated;
	*/

	// FIXME ändra ev node->treerow->treecell enligt föregående
	// t ex catalog_subtitle som tooltip

	let ioService = Components.classes["@mozilla.org/network/io-service;1"].
		getService(Components.interfaces.nsIIOService);

	let nodelinks = [];
	let links;
	if (feed.fields.hasKey("links"))
	    links = feed.fields.getPropertyAsInterface("links", Components.interfaces.nsIArray);
	else
	    links = [];
	for (let j = 0; j < links.length; j++) {
	    let link = links.queryElementAt(j, Components.interfaces.nsIPropertyBag2);
	    if (!link.hasKey("type"))
		continue;
	    let type = link.getPropertyAsAUTF8String("type");
	    if (!link.hasKey("rel"))
		continue;
	    let rel = link.getPropertyAsAUTF8String("rel");
	    let href = null;
	    if (link.hasKey("href")) {
		let ohref = link.getPropertyAsAUTF8String("href");
		try {
		    href = ioService.newURI(ohref, null, feed.baseURI).spec;
		} catch (e) {
		}

		debug("TOP LINK:" + rel + ":" + href + ":" + type);
		// previous, first, last
		if (rel == "next" && type == "application/atom+xml") {
		    let title = "...";
		    if (link.hasKey("title"))
			title = link.getPropertyAsAUTF8String("title");
		    let nnode = this.db.add_node(-2, href, title);
		    this.db.set_node_next(node, nnode);
		} else if (rel == "search" && type == "application/atom+xml") {
		    let title = "Search";
		    if (link.hasKey("title"))
			title = link.getPropertyAsAUTF8String("title");
		    nodelinks.push({ href: href, title: title, type: "search" });
		}
	    }
	}

	if (nodelinks.length > 0)
	    this.db.set_node_links(node, nodelinks);

	feed.QueryInterface(Components.interfaces.nsIFeed);
	for (let i = 0; i < feed.items.length; i++) {
	    debug("ENTRY");
	    let entry = feed.items.queryElementAt(i, Components.interfaces.nsIFeedEntry);

	    if (!entry.fields.hasKey("links"))
		continue;
	    let ent = { links: [] };
	    let link_type, location;
	    let links = entry.fields.getPropertyAsInterface("links", Components.interfaces.nsIArray);
	    for (let j = 0; j < links.length; j++) {
		let link = links.queryElementAt(j, Components.interfaces.nsIPropertyBag2);
		if (link.hasKey("type") && link.hasKey("href")) {
		    let href = link.getPropertyAsAUTF8String("href");
		    let nhref;
		    try {
			let ahref = ioService.newURI(href, null, feed.baseURI);
			nhref = ahref.spec;
			// this is necessary for Fictionwise: rel="goodreads"
			if (ahref.schemeIs("stanza"))
			    nhref = "http://" + nhref.substring(9);
		    } catch (e) {
			nhref = href;
		    }
		    let typ = link.getPropertyAsAUTF8String("type");
		    let title = null;
		    if (link.hasKey("title"))
			title = link.getPropertyAsAUTF8String("title");

		    if (link.hasKey("rel")) {
			let rel = link.getPropertyAsAUTF8String("rel");
			if (rel == "x-stanza-cover-image") {
			    debug("COVER");
			    ent.cover_link = nhref;
			} else if (rel == "x-stanza-cover-image-thumbnail") {
			    debug("THUMBNAIL");
			    ent.thumbnail_link = nhref;
			} else if (rel == "buynow") {
			    // O'Reilly, BooksOnBoard" 
			    ent.links.push({ href: nhref, title: title });
			    if (!ent.type) {
				ent.type = "link";
				ent.link = nhref;
			    }
			} else if (rel == "All Romance Ebooks" &&
				   typ == "application/xhtml+xml") {
			    // All Romance Ebooks
			    ent.links.push({ href: nhref, title: title,
					type: "application/epub+zip" });
			    if (!ent.type) {
				ent.type = "link";
				ent.link = nhref;
			    }
			} else if (typ == "application/atom+xml") {
			    // Fictionwise, Munseys, Gutenberg, Smashwords:
			    //	rel="author", rel="series"
			    // Fictionwise:
			    //  rel="goodreads"
			    // All Romance Ebooks: rel="Author Search"
			    // Harlequin: rel="fictionwise_harlequin"
			    ent.links.push({ href: nhref, title: title,
					type: "application/atom+xml" });
			    if (!ent.type) {
				ent.type = "link";
				ent.link = nhref;
			    }
			} else if (typ == "application/xhtml+xml") {
			    // Fictionwise, Munseys
			    ent.links.push({ href: nhref, title: title });
			    if (!ent.type) {
				ent.type = "link";
				ent.link = nhref;
			    }
			} else if (typ == "text/html") {
			    // All Romance Ebooks
			    // Smashwords: rel="bookinfo", rel="authorprofile"
			    ent.links.push({ href: nhref, title: title });
			    if (!ent.type) {
				ent.type = "link";
				ent.link = nhref;
			    }
			}
		    } else {
			if (typ == "application/atom+xml") {
			    ent.type = "folder";
			    ent.link = nhref;
			} else if (typ == "application/epub+zip") {
			    ent.type = "epub";
			    ent.link = nhref;
			} else if (typ == "text/html") {
			    ent.links.push({ href: nhref, title: title });
			    if (!ent.type) {
				ent.type = "link";
				ent.link = nhref;
			    }
			}
		    }
		}
	    }

	    // FIXME Get "content" from entry

	    debug("Entry type: " + ent.type);

	    /*
	    if (!ent.type)
		continue;
	    */

            let location_uri;
            try {
		ent.location_uri = ioService.newURI(ent.link, null, feed.baseURI);
            } catch (e) {
		debug("E2: " + e);
                continue;
            }

	    if (entry.title)
		ent.title = entry.title.plainText();
	    if (!ent.title)
		ent.title = "[Unknown]";	// FIXME
	    let catalog = true;

	    debug("Title: " + ent.title);

	    if (ent.type == "folder") {
		let nnode = this.db.add_node(node, ent.location_uri.spec, ent.title);
	    } else if (ent.type == "epub" || ent.type == "link") {
		if (entry.authors.length >= 1) {
		    let person = entry.authors.queryElementAt(0, Components.interfaces.nsIFeedPerson);
		    ent.author = person.name;
		}
		/*
		let bi = {};
		bi.id = 0;	// FIXME
		//bi.uid = id;	// FIXME
		bi.title = ent.title;
		bi.author = ent.author;
		//bi.publisher = null;
		*/
		ent.location = ent.location_uri.spec;
		if (ent.links)
		    debug("LINKS!!\n");
		let bid = this.db.add_book_info(ent);
		this.db.add_child(node, bid);
	    }
	}

	debug("return" + node);
	this.db.set_node_visited(node);
	callback(this.db.browse_path(node), sid);
    },

    sync_cache: function() {
	// FIXME
    }
};

var components = [stanzaCase];

if (XPCOMUtils.generateNSGetFactory)
    var NSGetFactory = XPCOMUtils.generateNSGetFactory(components);
else
    var NSGetModule = XPCOMUtils.generateNSGetModule(components);
