/*
 *  Lucidor - e-book reader
 *  Copyright © 2009-2010  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

function debug(msg) {
    if (false) {
	try {
	    let cons = Components.classes["@mozilla.org/consoleservice;1"].
		getService(Components.interfaces.nsIConsoleService);
	    cons.logStringMessage(msg);
	} catch (e) {
	}
    }
}

function loop(thing) {
    if (!thing)
	return;
    for (let i = 0; i < thing.length; i++)
	yield thing[i];
}

function webpageCase() {
    this.wrappedJSObject = this;
    this.init0();
}

webpageCase.prototype = {
    classDescription: "webpageCase",
    classID: Components.ID("{c0b2ec78-02ff-4ddc-9e42-e8abd0446024}"),
    contractID: "@lucidor.org/webpageCase;1",
    QueryInterface: XPCOMUtils.generateQI([Components.interfaces.nsISupports]),

    description: "Webpage Offline Catalog",
    //update_message: "",
    db: null,
    catalog_url: null,
    collid: -1,

    info: { title: true,
	    author: false,
	    location: true },

    listers: [ "all" ],

    capabilities: { offline_db: true,
		    standard_search: true,
		    search_meta: true,
		    search_title: true,
		    search_author: false,
		    search_tags: false,
		    search_text: false },

    init0: function() {
	this.db = Components.classes["@lucidor.org/bookdb;1"].createInstance().
		wrappedJSObject;
    },

    init: function(coll, tree) {
	this.collid = coll.id;
	this.catalog_url = coll.url;
	this.tree = tree;
	if (coll.db) {
	    this.db.open(coll.db);
	} else {
	    this.db.create("webpage");
	}
	this.set_db();
    },

    set_db: function() {
	this.tree.datasources = this.db.file_uri;
	this.tree.builder.rebuild();
    },

    close: function() {
	if (this.db)
	    this.db.close();
	this.db = null;
    },

    add_book: null,

    remove_book: null,

    get_preview: function(id) {
	return this.db.get_preview(id);
    },

    generate_preview: null,

    add_preview: function(id, url) {
	return this.db.set_preview(id, url);
    },

    get_book_info: function(id) {
	return this.db.get_book_info(id);
    },

    sync_cache: function() {
	debug("sync_cache: " + new Date());

	let ioService = Components.classes["@mozilla.org/network/io-service;1"].
		getService(Components.interfaces.nsIIOService);
	let uri = ioService.newURI(this.catalog_url, null, null);

	let main_control = Components.classes["@lucidor.org/mainControlService;1"].
	    getService().wrappedJSObject;
	let fil = main_control.get_tempdir();
	fil.append("webpage.html");
	fil.createUnique(Components.interfaces.nsIFile.NORMAL_FILE_TYPE, 0600);

	let this_proxy = this;

	let wbp = Components.classes["@mozilla.org/embedding/browser/nsWebBrowserPersist;1"].createInstance(Components.interfaces.nsIWebBrowserPersist);
	wbp.progressListener = {
	    onLocationChange: function(webProgress, request, location) {
	    },

	    onProgressChange: function(webProgress, request, curSelfProgress,
				       maxSelfProgress, curTotalProgress,
				       maxTotalProgress) {
	    },

	    onSecurityChange: function(webProgress, request, state) {
	    },

	    onStateChange: function(aWebProgress, aRequest, aStateFlags, aStatus) {
		debug("onStateChange");
		if (wbp.currentState == wbp.PERSIST_STATE_FINISHED) {
		    debug("Persister has finished saving data.");
		    try {
			this_proxy.finish_sync(fil, aRequest.name);
		    } catch (e) {
			debug("ERW: " + e);
		    }
		    debug("H 6");
		}
	    },

	    onStatusChange: function(webProgress, request, status, message) {
	    },

	    QueryInterface: function(aIID) {
		if (aIID.equals(Components.interfaces.nsIWebProgressListener))
		    return this;
		throw Components.results.NS_NOINTERFACE;
	    }
	};

	wbp.saveURI(uri, null, null, null, null, fil);
    },

    finish_sync: function(fil, furi) {
	debug("HA");
	let fis = Components.classes["@mozilla.org/network/file-input-stream;1"].
		createInstance(Components.interfaces.nsIFileInputStream);
	fis.init(fil, -1, -1, 0);

	let charset = "UTF-8";
	const replacementChar = Components.interfaces.nsIConverterInputStream.DEFAULT_REPLACEMENT_CHARACTER;
	let is = Components.classes["@mozilla.org/intl/converter-input-stream;1"].
		createInstance(Components.interfaces.nsIConverterInputStream);
	is.init(fis, charset, 1024, replacementChar);

	debug("11");
	let txt = "";
	let str = {};
	while (is.readString(4096, str) != 0)
	    txt += str.value;
	is.close();
	fis.close();

	let document = Components.classes["@mozilla.org/xml/xml-document;1"].
		createInstance(Components.interfaces.nsIDOMDocument);
	let implementation = document.implementation;
	//let dt = implementation.createDocumentType("html", "-//W3C//DTD XHTML 1.1//EN", "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd");
	let doc = implementation.createDocument("", "", null);
	debug("P " + doc.documentElement + " " + doc.ownerDocument);

	//let root = doc.createElement("root");
	//doc.appendChild(root);

	try {
	    debug("Start");
	    let parser = Components.classes["@lucidor.org/htmlParser;1"].
		createInstance().wrappedJSObject;
	    parser.myHTMLtoDOM(txt, doc);
	    debug("End");
	} catch (e) {
	    debug("ERROR: " + e);
	}

	/*
	let xser = Components.classes["@mozilla.org/xmlextras/xmlserializer;1"].
		createInstance();
	debug(";;;" + xser.serializeToString(root));
	*/

	let new_db = Components.classes["@lucidor.org/bookdb;1"].
		createInstance().wrappedJSObject;
	new_db.create("webpage");

	let ioService = Components.classes["@mozilla.org/network/io-service;1"].
		getService(Components.interfaces.nsIIOService);
	let baseuri = null;
	try {
	    // FIXME baseuri is not correct
	    baseuri = ioService.newURI(furi, "UTF-8", null).
		QueryInterface(Components.interfaces.nsIURL);
	} catch (e) {
	}

	for (let link in loop(doc.getElementsByTagName("a"))) {
	    let href = link.getAttribute("href");
	    if (!href)
		continue;

	    let linkuri;
	    let ok = false;
	    try {
		linkuri = ioService.newURI(href, "UTF-8", baseuri);
		if (linkuri.schemeIs("epub")) {	// Non-standard Stanza thing
		    linkuri = ioService.newURI("http" + linkuri.spec.substring(4), "UTF-8", null).QueryInterface(Components.interfaces.nsIURL);
		    ok = true;
		} else {
		    linkuri.QueryInterface(Components.interfaces.nsIURL);
		    if (linkuri.schemeIs("http") || linkuri.schemeIs("https") ||
			linkuri.schemeIs("ftp"))
			if (linkuri.fileExtension.toLowerCase() == "epub")
			    ok = true;
		}
	    } catch (e) {
		continue;
	    }
	    if (ok) {
		debug("LOCATION: " + linkuri.spec + "\n");
		if (new_db.test_location(linkuri.spec) < 0) {
		    let title = unescapeEntities(link.textContent);
		    if (!title)
			title = linkuri.fileBaseName;
		    if (!title)
			title = "[Unknown]";
		    let book = { title: title,
				 location: linkuri.spec,
				 type: "epub" };
		    new_db.add_book_info(book);
		}
	    }
	}

	this.db.delete();
	this.db = new_db;
	let cman = Components.classes["@lucidor.org/collectionManagerService;1"].
		getService().wrappedJSObject;
	cman.changeCollectionDB(this.collid, this.db);
	this.set_db();
    }
};

var components = [webpageCase];

if (XPCOMUtils.generateNSGetFactory)
    var NSGetFactory = XPCOMUtils.generateNSGetFactory(components);
else
    var NSGetModule = XPCOMUtils.generateNSGetModule(components);

function unescapeEntities(inp) {
    let re0 = new RegExp("&#([0-9]+);");
    let re1 = new RegExp("&#[xX]([a-fA-F0-9]+);");
    let st = inp;
    let res;
    while ((res = re0.exec(st)) != null) {
	let chr = String.fromCharCode(parseInt(res[1], 10));
	st = st.substring(0, res.index) + chr +
	    st.substring(res.index + res[0].length);
    }
    while ((res = re1.exec(st)) != null) {
	let chr = String.fromCharCode(parseInt(res[1], 16));
	st = st.substring(0, res.index) + chr +
	    st.substring(res.index + res[0].length);
    }
    return st;
}
